#!/usr/local/bin/python3
# coding=utf-8

"""
INFO
*** Paketaufpbau

- gui           GUI-Komponenten (MVC)
- logik         Die zentrale Logik
- helper        Ausgelagerte etwas allgemeinere Verwaltungslogik (befindet sich bezueglich Komplexitaet/Spezialisierung zwischen Logik und Tools)
- tools         Allgemeine Tools
critterbase.py
start.pyw

"""
import random

try:
    import argparse, sys, os, textwrap, getpass, configparser, datetime, requests, re, psutil, subprocess, urllib, ssl, webbrowser, folium, zipfile, pathlib
    from collections import OrderedDict
    from PyQt5.QtWidgets import *
    from gui.controller import Controller
    from gui.model import Model
    from tools.printing import SmartPrinter
    from logik.batchbot import BatchBot
    from tools.text import TextEncoding
    from tools.fileio import FileTools
    from helper.database import Database
    from tools.text import TextEncoding, PrettyText
    from tools.coding import Result, MultiSheet, Package
    from tools.taxon import Taxon
    from tools.taxa import Taxa
    from tools.ui import ColorGradient, Colors
    from logik.ingestor import Ingestor
    from tools.geocruncher import GeoCruncher
    from tools.taxacruncher import TaxaCruncher
    from distutils.dir_util import copy_tree
    from tools.excel import ExcelTools, ExcelNavigator
    from helper.thread_container import Worker_CheckDois
    import openpyxl, warnings

except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")
    exit(1)

# Main
if __name__ == "__main__":

    # Hilfsmethoden
    def testFileExistsRead(sFile):
        if not os.path.isfile(sFile):
            print("File '"+sFile+"' does not exist ... Abort")
            exit(1)

    def testFileExistsWrite(sFile):
        if os.path.isfile(sFile):
            print("File '"+sFile+"' does exist")
            ein = input("Overwrite the file [yN]? ")
            if ein == "" or ein.lower()[0] not in ('y','j'):
                print("Abort ...")
                exit(1)

    def testEncoding(sFile):
        sFileEndung = os.path.splitext(sFile)[1]
        if sFileEndung == ".csv":
            print("* Checking type of text encoding")
            (sEncoding, fConfidence) = TextEncoding.checkout(sFile)
            print("Determined type of encoding: " + sEncoding.upper() + " (Probability: " + str(
                round(fConfidence * 100)) + "%)")
            if sEncoding.lower() not in ("ascii", "utf-8", "ibm852", "iso-8859-1"):
                print("Found a not allowed type of encoding. Abort ...")
                exit(1)

    # Model erzeugen

    m = Model()
    m.ui_getSmartprinter().setCliNoPorgress(False)
    sImportTypes = PrettyText.lst2strC(m.subsystem_getMapper().workflow_getAll())

    # Workflowliste
    # Normal
    lstWfInfo = []
    for sWfShort in m.subsystem_getMapper().workflow_getAll():
        sWfLong = m.subsystem_getMapper().workflow_short2long(sWfShort)
        lstWfInfo.append(sWfShort + " (" + sWfLong+")")
    sWf_nativ_liste = PrettyText.lst2strC(lstWfInfo)
    # Converter: Import
    lstWf_Cnv_import = []
    for sWfCnfShort in m.subsystem_getMapper().workflow_converter_import_getAll():
        sWfCnfLong = m.subsystem_getMapper().workflow_converter_import_short2long(sWfCnfShort)
        lstWf_Cnv_import.append(sWfCnfShort + " (" + sWfCnfLong + ")")
    sWf_Cnv_import_liste = PrettyText.lst2strC(lstWf_Cnv_import)
    # Converter: Export
    lstWf_Cnv_export = []
    for sWfCnfShort in m.subsystem_getMapper().workflow_converter_export_getAll():
        sWfCnfLong = m.subsystem_getMapper().workflow_converter_export_short2long(sWfCnfShort)
        lstWf_Cnv_export.append(sWfCnfShort + " (" + sWfCnfLong + ")")
    sWf_Cnv_export_liste = PrettyText.lst2strC(lstWf_Cnv_export)
    # BatchBot Configs
    sBatchBotConfigs = PrettyText.lst2strC(m.subsystem_getBatchBot().getConfig_all())

    # CLI-Parser
        # Kurzinfo zum Programm
    sInfo = m.conf_getInfo("prgname") + " " + m.conf_getInfo("version") + " (" + m.conf_getInfo("dbschema") + ")\n" \
            + m.conf_getInfo("comment").replace("<br>", "\n").replace("<i>", "").replace("</i>", "") + "\n" \
            "Contact: " + m.conf_getInfo("www") + "\n" + \
            m.conf_getInfo("date")
        # Importer/Exporter: Arten des Importes
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(sInfo),
        add_help=False)

    # Import Ingest
    gruppeImportIngest = parser.add_argument_group("Import: Ingest")
    gruppeImportIngest.add_argument("-imp-ing", metavar=("<workflow>", "<file.xlsx>"), help="Import selected ingest tables from Excel file (Workflows: direct: " + sWf_nativ_liste + ", with conversion: " + sWf_Cnv_import_liste, type=str, nargs=2)
    gruppeImportIngest.add_argument("-imp-ing-ext", help="Allow extension of records in database (yes/no)", type=str)
    gruppeImportIngest.add_argument("-imp-ing-ext-react", help="Allow extension of records in database with reaction level if not matching 100 percent", type=str)
    gruppeImportIngest.add_argument("-imp-ing-lup-single", help="Import each lookup on the fly", action="store_true")

    # Import Lookup
    gruppeImportLookup = parser.add_argument_group("Import: Lookup")
    gruppeImportLookup.add_argument("-imp-lup", metavar=("<file.xlsx>", "<table>"), help="Add selected lookup tables from Excel file (Sheets: " + ", ".join(m.subsystem_getMapper().lookup_getExcelSheets()), type=str, nargs='+')
    gruppeImportLookup.add_argument("-imp-lup-in", metavar=("<type>"), help="Type of records to add (types: all,gold,not-gold)", type=str)
    gruppeImportLookup.add_argument("-imp-lup-mode", metavar=("<mode>"), help="Mode of adding data records: (modes: new,update,both)", type=str)
    gruppeImportLookup.add_argument("-imp-lup-update", metavar=("<type>"), help="Type of matching records to update: (types: all,gold,not-gold)", type=str)
    gruppeImportLookup.add_argument("-imp-lup-reuse-react", metavar=("<type>"), help="When update is dewactivated, reaction if existing record is not 100 percent matching with import: (types: error,waring,info)", type=str)
    gruppeImportLookup.add_argument("-imp-lup-goldrush", help="Only allow gold records while import", action="store_true")

    # Import TaxonWizard Data
    gruppeImportTaxon = parser.add_argument_group("Import: TaxonWizard")
    gruppeImportTaxon.add_argument("-imp-tax", metavar=("<file.xlsx>", "<table>"),
                                   help="Add selected lookup tables from Excel file (Sheets: " + ", ".join(
                                        m.subsystem_getMapper().taxon_getExcelSheets()), type=str, nargs='+')
    gruppeImportTaxon.add_argument("-imp-tax-update", help="update existing db records", action="store_true")

    # Esport Subset
    gruppeExportSubset = parser.add_argument_group("Export: Subset")
    gruppeExportSubset.add_argument("-exp-sub-file", metavar=("<file.xlsx>"), help="Export to Excel file - if not set console output is active", type=str)
    gruppeExportSubset.add_argument("-exp-sub-converter", metavar=("<converter>"), help="Use converter for ingest data export: " + sWf_Cnv_export_liste, type=str)
    gruppeExportSubset.add_argument("-exp-sub-selector", help="Show selector ranges", action="store_true")
    gruppeExportSubset.add_argument("-exp-sub-filter-date-start", metavar=("<date start>"), type=str)
    gruppeExportSubset.add_argument("-exp-sub-filter-date-end", metavar=("<date end>"), type=str)
    gruppeExportSubset.add_argument("-exp-sub-filter-ing", metavar=("<ingest id>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-filter-gear", metavar=("<gear>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-filter-cov", metavar=("<taxonomic coverage>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-filter-dat", metavar=("<dataset>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-filter-region", metavar=("<region>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-filter-wf", metavar=("<workflows>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-id", help="Export ids of each record", action="store_true")
    gruppeExportSubset.add_argument("-exp-sub-ing-id", help="Export ingest id of each record", action="store_true")
    gruppeExportSubset.add_argument("-exp-sub-rfriendly", help="Export R-friendly", action="store_true")
    gruppeExportSubset.add_argument("-exp-sub-sheets", metavar=("<sheet>"), nargs="+")
    gruppeExportSubset.add_argument("-exp-sub-stat", help="Only show information an no actual records in console output mode", action="store_true")

    # Export Template
    gruppeExportTemplate = parser.add_argument_group("Export: Template")
    gruppeExportTemplate.add_argument("-exp-template", metavar=("<workflow>"), help="Available workflows: native: " + sWf_nativ_liste + "; for converter import: " + sWf_Cnv_import_liste, type=str)

    # Export Lookup
    gruppeExportLookup = parser.add_argument_group("Export: Lookup")
    gruppeExportLookup.add_argument("-exp-lup-all", help="Export all lookups", action="store_true")
    gruppeExportLookup.add_argument("-exp-lup-ing", metavar=("<ingest id>"), nargs="+", help="Export used lookups for given ingests")
    gruppeExportLookup.add_argument("-exp-lup-id", help="Export ids of each record", action="store_true")
    gruppeExportLookup.add_argument("-exp-lup-sheets", metavar=("<sheet>"), nargs="+", help="Export only given lookups")
    gruppeExportLookup.add_argument("-exp-lup-gold", metavar=("<all/gold/not-gold>"), type=str, help="Export only lookups with the following gold status: all, gold, not-gold lookups")
    gruppeExportLookup.add_argument("-exp-lup-file", metavar=("<file.xlsx>"), help="Export to Excel file - if not set console output is active", type=str)

    # TaxonWizard: Update Taxon List
    gruppeTaxWiz = parser.add_argument_group("TaxonWizard: TaxonList")
    gruppeTaxWiz.add_argument("-tax-update", help="Update the taxon list", action="store_true")

    # Export TaxonWizard
    gruppeExportTaxon = parser.add_argument_group("Export Data: TaxonWizard")
    gruppeExportTaxon.add_argument("-exp-tax", help="Export all taxon system admin tables", action="store_true")
    gruppeExportTaxon.add_argument("-exp-tax-sheets", metavar=("<sheet>"), nargs="+", help="Export only given taxon admin tables")
    gruppeExportTaxon.add_argument("-exp-tax-file", metavar=("<file.xlsx>"),
                                    help="Export to Excel file - if not set console output is active", type=str)
    # Database
    gruppeDB = parser.add_argument_group("Database")
    gruppeDB.add_argument("-db", metavar="<number of account>", help="Display/configure database connection for account number (0-4)", type=str, nargs=1)

    # BatchBot
    gruppeBatch = parser.add_argument_group("Batch")
    gruppeBatch.add_argument("-bot", metavar="<configname>", help="Start the BatchBot (Configs: "+ sBatchBotConfigs+")", type=str, nargs=1)

    # FAO
    gruppeBatch = parser.add_argument_group("FAO")
    gruppeBatch.add_argument("-fao-new", metavar=("<level>"), help="Create new FAO regions (all: all available level, std: major, division, subarea)", type=str, nargs="+")
    gruppeBatch.add_argument("-fao-del", help="Delete mapping (ingest_id: delete mapping for ingest id only, or when nocthing is given delete all)",  metavar="<ingest id>", nargs="?", type=int, const=-1)
    gruppeBatch.add_argument("-fao-update", help="Update mapping", action="store_true")
    gruppeBatch.add_argument("-fao-stat", metavar=("<size>"), help="Show info on mapping (size: short, long <default>)", nargs="?", type=str, const="full")

    # Etc
    gruppeEtc = parser.add_argument_group("Etc")
    gruppeEtc.add_argument("-gui", metavar="Style", help="Start in GUI mode, all other parameters are ignored. (Styles: "+", ".join(QStyleFactory.keys())+")", type=str, nargs='?', const="_system_")
    gruppeEtc.add_argument("-i", metavar=("<sub system>"), help="Shows information on sub systems: all, " + PrettyText.lst2strC(m.subsystem_getSubLst()), type=str, nargs='+')
    gruppeEtc.add_argument("-dry", help="writing database actions are only simulated", action="store_true")
    gruppeEtc.add_argument("-cnp", help="no progress bar in CLI mode", action="store_true")
    gruppeEtc.add_argument("-v", help="verbose", action="store_true")
    gruppeEtc.add_argument("-d", help="debugging", action="store_true")
    gruppeEtc.add_argument("-vs", "--version", help="Version", action="store_true")
    gruppeEtc.add_argument("-update", help="Check for available pdates from server", action="store_true")
    gruppeEtc.add_argument("-h", "--help", help="Shows this help", action="store_true")

    # Testing
    gruppeTesting = parser.add_argument_group("TESTING")
    gruppeTesting.add_argument("-test", help="Generic test area", action="store_true")

    # Parse CLI-Argumente
    try:
        args = parser.parse_args()
    except:
        print("Faulty parameters ... abort")
        exit(1)

    # --- Keine Serververbindung nötig ---

    if len(sys.argv) <= 1:
        # Nichts übergeben -> GUI starten
        del m  # Sauberkeit: loesche das Model, welches hier im CLI-Modus benutzt wird
        Controller()
        exit(1)
    elif args.gui:
        # GUI
        del m # Sauberkeit: loesche das Model, welches hier im CLI-Modus benutzt wird
        sGuiStyle = args.gui
        if args.gui == "_system_":
            sGuiStyle = None
        Controller(sGuiStyle)
        exit(1)
    elif args.help:
        # Hilfe
        parser.print_help()
        exit(1)
    elif args.version:
        # Version
        print(m.conf_getInfo("version") + " (" + m.conf_getInfo("dbschema") + ")")
        exit(1)

    # Veraenderungen (zur eingelesenen Konfigurationsdatei) am Model vornehmen (CLI-Parameter)
        # Geschwaetzigkeit
    m.conf_setVerbosity(args.v, args.d)
        # Dry Run
    m.conf_setDry(args.dry)
        # Logging
    m.conf_setLogging(False) # Std: Logging ausschalten
        # CLI: Kein Progressbar anzeigen
    m.ui_getSmartprinter().setCliNoPorgress(args.cnp)

    # Server-Config veraendern
    if args.db:
        iAccount = int(args.db[0])
        print("* Settings of the database connection")
        print("Account number: " + str(iAccount))
        if iAccount < 0 or iAccount >= m.subsystem_getDb().account_getNumberOfAccounts():
            print("number not valid")
            exit(1)
        m.subsystem_getDb().account_setIndexOfActiveAccount(iAccount)
        if m.subsystem_getDb().account_single_getPart("pw") != "":
            print("Test of the server connection: ", end="")
            resCheck = m.subsystem_getDb().check_isSchemaExisting()
            bOk = resCheck.getData("exist") if resCheck.isOk() else False
            print("ok" if bOk == True else "")
        print("---")
        print("Preset: " + m.subsystem_getDb().account_single_getPart("preset"))
        print("Host:   " + m.subsystem_getDb().account_single_getPart("host"))
        print("Port:   " + m.subsystem_getDb().account_single_getPart("port"))
        print("Name:   " + m.subsystem_getDb().account_single_getPart("name"))
        print("Schema: " + m.subsystem_getDb().account_single_getPart("schema"))
        print("User:   " + m.subsystem_getDb().account_single_getPart("user"))
        if m.subsystem_getDb().account_single_getPart("pw") == "":
            print("Pw:     <not saved>")
        else:
            print("Pw:     <saved>")
        print("---")
        ein = input("Should the data be changed [yN]? ")
        if ein == "" or ein.lower()[0] not in ('y', 'j'):
            exit(1)
        print("* Change of the database connection")
        ein = input("Preset (" + m.subsystem_getDb().account_single_getPart("preset") + "): ")
        if ein != "": m.subsystem_getDb().account_single_setPart("preset", ein)
        ein = input("Server (" + m.subsystem_getDb().account_single_getPart("host") + "): ")
        if ein != "": m.subsystem_getDb().account_single_setPart("host", ein)
        ein = input("Port (" + m.subsystem_getDb().account_single_getPart("port") + "): ")
        if ein != "": m.subsystem_getDb().account_single_setPart("port", ein)
        ein = input("DB-Name (" + m.subsystem_getDb().account_single_getPart("name") + "): ")
        if ein != "": m.subsystem_getDb().account_single_setPart("name", ein)
        ein = input("DB-Schema (" + m.subsystem_getDb().account_single_getPart("schema") + "): ")
        if ein != "": # Check auf Alphanumerik
            if m.subsystem_getDb().account_single_setPart("schema", ein) == False:
                exit(1)
        ein = input("User(" + m.subsystem_getDb().account_single_getPart("user") + "): ")
        if ein != "": m.subsystem_getDb().account_single_setPart("user", ein)
        if m.subsystem_getDb().account_single_getPart("pw") == "":
            sPrompt = "Password (<not saved>): "
        else:
            sPrompt = "Password (<saved>): "
        m.subsystem_getDb().account_single_setPart("pw", getpass.getpass(prompt=sPrompt))
        m.conf_cfgSave()
        exit(1)

    # Teste ob Account leer ist
    if m.subsystem_getDb().account_single_isEmpty():
        print("Configuration of account is empty -> Please configure first")
        exit(1)

    # Initialisiere Workflows (Serververbindung und Passwort)
    if m.subsystem_getDb().account_single_getPart("pw") == "":
        m.subsystem_getDb().account_single_setPart("pw", getpass.getpass(prompt="Password: ")) # Pw holen
    # Programmkonfiguration zeigen
    if args.i:
        if args.i == ["all"]:
            lst = m.subsystem_getSubLst()
        else:
            lst = args.i
        for sSub in lst:
            sInfo = m.subsystem_getSubInfo(sSub)
            if sInfo is not None:
                print(sSub.upper())
                print(sInfo)
                print()
        exit(1)
    # Teste auf Update
    if args.update:
        print("* Checking for available updates")
        res = m.help_updateCheck()
        if res.isErr():
            print("Error:", res.getResult())
        else:
            bUpdate = res.getData("update")
            print(res.getData("update-info"))
            if bUpdate:
                print("Update Level:", res.getData("update-level"))
                ein = input("Do you want to update [yN]? ")
                if ein == "" or ein.lower()[0] not in ('y', 'j'):
                    print("abort ...")
                    exit(0)
                print("* Starting update")
                res = m.help_updateDownload()
                if res.isErr():
                    print("Error:", res.getResult())
                    exit(1)
                print("* Installation")
                sSys = res.getData("sys")
                sSrc = res.getData("src")
                sTarget = res.getData("target")
                print("System: " + sSys)
                print("Source: " + sSrc)
                print("Target: " + sTarget)
                ein = input("Do you want to install the update [yN]? ")
                if ein == "" or ein.lower()[0] not in ('y', 'j'):
                    print("Installation aborted ...")
                    exit(0)
                # Installation ---
                print("Copying files:")
                files = copy_tree(sSrc, sTarget)
                print(PrettyText.lst2strR(files))
                print("Installation is finished")
    elif args.tax_update:
        # Update Taxon list
        res = m.subsystem_getTaxonWizard().update()
        if res.isErr():
            print("\n* Taxon system update failed - nothing is wrtten to database")
            print(res.getData("text"))
            exit(1)
        else:
            print("\n* Taxon system update sucessfully finished")
            print(res.getData("text"))
            exit(0)

    # --- Serververbindung noetig ---
    resDir = m.subsystem_getConnector().reconnect()
    if resDir.isErr():
        print(resDir.getErrWarnInfoString())
        exit(0)

    if args.imp_ing:
        # Import: Ingest
        sWf = args.imp_ing[0]
        sFile = args.imp_ing[1]
        bLupSingle = args.imp_ing_lup_single
        sExcel = args.imp_lup_in
        bGoldrush = args.imp_lup_goldrush
        sMode = args.imp_lup_mode
        sUpdate = args.imp_lup_update
        sExtend = args.imp_ing_ext
        sExtendReact = args.imp_ing_ext_react
        sReuse = args.imp_lup_reuse_react
        if sExtend == "yes":
            bExtend = True
        elif sExtend == "no":
            bExtend = False
            sExtendReact = None
        else:
            bExtend = None
            sExtendReact = None
        bGoldrush = None
        res = m.subsystem_getImporter().startImport(sFile, ingest=True, lookup=True, ingest_wf=sWf, lookup_import_excel=sExcel, lookup_import_mode=sMode, lookup_import_db=sUpdate, lookup_import_goldrush=bGoldrush, ingest_lup_search_single=bLupSingle, tamer=True, extend=bExtend, extend_react=sExtendReact, lookup_reuse=sReuse)
        print(res.getData("log"))
        if res.isAnyMessage():
            print(res.getErrWarnInfoString())
        exit(1)
    elif args.imp_lup is not None:
        # Import: Lookup
        sFile = args.imp_lup.pop(0)
        lstTabs = args.imp_lup
        sExcel = args.imp_lup_in
        bGoldrush = args.imp_lup_goldrush
        sMode = args.imp_lup_mode
        sUpdate = args.imp_lup_update
        sReuse = args.imp_lup_reuse_react
        res = m.subsystem_getImporter().startImport(sFile, lookup=True, lookup_sheets=lstTabs, lookup_import_excel=sExcel, lookup_import_mode=sMode, lookup_import_db=sUpdate, lookup_import_goldrush=bGoldrush, tamer=True, lookup_reuse=sReuse)
        print(res.getData("log"))
        if res.isAnyMessage():
            print(res.getErrWarnInfoString())
            exit(0)
    elif args.imp_tax is not None:
        # Import: Taxon System
        sFile = args.imp_tax.pop(0)
        lstTabs = args.imp_tax
        bUpdate = args.imp_tax_update
        res = m.subsystem_getImporter().startImport(sFile, taxon=True, taxon_sheets=lstTabs, taxon_update=bUpdate, tamer=True)
        print()
        if res.isErr():
            print(res.getResult())
            exit(0)
        print("Occurred errors:", res.getData("err"))
        print("Log:")
        print(res.getData("log"))

    elif args.exp_tax:
        # Export: Taxon System
        sExportFile = args.exp_tax_file
        lstSheets = args.exp_tax_sheets
        res = m.subsystem_getExporter().subsystem_getTaxonWizard(sheets=lstSheets)
        if res.isErr():
            print("Error:", res.getResult())
            exit(1)
        odicTax = res.getData("taxon")
        if sExportFile is None:
            # Console
            if m.conf_isDebug():
                print("* Log")
                print(res.getData("log"))
                print("* Tables")
            for sSheet, dicSheet in odicTax.items():
                lstHeaders = dicSheet["header"]
                lstTab = dicSheet["content"]
                print("--- "+sSheet)
                print(PrettyText.table([lstHeaders] + lstTab))
        else:
            # Excel
            print()
            print("Export to file: "+sExportFile)
            res = m.subsystem_getExporter().writeExcel(taxon=True, taxon_sheets=lstSheets, taxon_data=odicTax, file=sExportFile)
            print()
            if res.isErr():
                print("Error:", res.getResult())
                exit(1)
    elif args.exp_lup_ing or args.exp_lup_all:
        # Export: Lookup
        sExportFile = args.exp_lup_file
        bShowId = args.exp_lup_id
        lstIng = args.exp_lup_ing
        bLupAll = args.exp_lup_all
        lstSheets = args.exp_lup_sheets
        sGold = args.exp_lup_gold
        if sGold not in [None, "all","gold","not-gold"]:
            print("Error: Unknown gold status: ", sGold)
            exit(1)
        dicSelect = {"ingest": lstIng}
        if bLupAll:
            dicSelect = None
        res = m.subsystem_getExporter().getLookup(selector=dicSelect, sheets=lstSheets, showid=bShowId, gold=sGold)
        if res.isErr():
            print("Error:", res.getResult())
            exit(1)
        odicLup = res.getData("lookup")
        if sExportFile is None:
            # Console
            if m.conf_isDebug():
                print("* Log")
                print(res.getData("log"))
                print("* SQL")
                for sSheet, dicSheet in odicLup.items():
                    sSql = dicSheet["sql"]
                    print("--- " + sSheet)
                    print(sSql)
                print("* Tables")
            for sSheet, dicSheet in odicLup.items():
                lstHeaders = dicSheet["header"]
                lstTab = dicSheet["content"]
                print("--- "+sSheet)
                print(PrettyText.table([lstHeaders] + lstTab))
        else:
            # Excel
            print()
            print("Export to file: "+sExportFile)
            res = m.subsystem_getExporter().writeExcel(lookup_sheets=lstSheets, showid=bShowId, rfriendly=b, lookup_data=odicLup, lookup=True, file=sExportFile)
            if res.isErr():
                print("Error:", res.getResult())
                exit(1)
    elif args.exp_sub_selector:
        # Export: Subset - Zeige Selector Ranges
        sel = m.ui_selector_getSelector()
        sel.reset()
        print("* Selector ranges")
        print(sel.infoConfigShort2())
        exit(0)
    elif any((args.exp_sub_filter_date_start, args.exp_sub_filter_date_end, args.exp_sub_filter_ing, args.exp_sub_filter_gear, args.exp_sub_filter_cov, args.exp_sub_filter_dat, args.exp_sub_filter_region, args.exp_sub_filter_wf)) :
        # Export: Subset
        # Parameter
        lstExportSheets = args.exp_sub_sheets
        sExportFile = args.exp_sub_file
        bExportId = args.exp_sub_id
        bExportIngestId = args.exp_sub_ing_id
        bRfriendly = args.exp_sub_rfriendly
        bInfoOnly = args.exp_sub_stat
        sConverter = args.exp_sub_converter
        sDateStart = args.exp_sub_filter_date_start
        sDateEnd = args.exp_sub_filter_date_end
        lstIng = args.exp_sub_filter_ing
        lstGea = args.exp_sub_filter_gear
        lstCov = args.exp_sub_filter_cov
        lstDat = args.exp_sub_filter_dat
        lstReg = args.exp_sub_filter_region
        lstWf = args.exp_sub_filter_wf
        # Auswerten
        dicSelect = dict()
        if lstIng is not None:
            dicSelect["ingest"] = lstIng
        if sDateStart is not None:
            dicSelect["sample_date_start"] = sDateStart
        if sDateEnd is not None:
            dicSelect["sample_date_end"] = sDateEnd
        if lstGea is not None:
            dicSelect["gear"] = lstGea
        if lstCov is not None:
            dicSelect["cover"] = lstCov
        if lstDat is not None:
            dicSelect["dataset"] = lstDat
        if lstReg is not None:
            dicSelect["region"] = lstReg
        if lstWf is not None:
            dicSelect["workflow"] = lstWf

        # Ingest
        res = m.subsystem_getExporter().getSubset(selector=dicSelect, showid=bExportId, showingestid=bExportIngestId, sheets=lstExportSheets, converter=sConverter, rfriendly=bRfriendly)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        odicIng = res.getData("data")
        sLogIng = res.getData("log")
        sWf = res.getData("wf")

        # Lookup
        res = m.subsystem_getExporter().getLookup(selector=dicSelect)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        odicLup = res.getData("lookup")
        sLogLookup = res.getData("log")

        if sExportFile is None:
            print()
            # --- Console Output
            if m.conf_isVerbose():
                print("* Console output")
            # Ingest
            print(">>> Ingest")
            sLogIngest = res.getData("log")
            if m.conf_isVerbose():
                print("* Overview")
                for sSheet, dicSheet in odicIng.items():
                    if sSheet != "Selector":
                        print(sSheet, ":", dicSheet["rows"],"rows")
                print("* Tables")
            for sSheet, dicSheet in odicIng.items():
                if sSheet != "Selector":
                    if bInfoOnly:
                        print(sSheet + ":", len(dicSheet["content"]))
                    else:
                        print("---", sSheet)
                        print(PrettyText.table([dicSheet["header"]] + dicSheet["content"], True))
            print(">>> Lookup")
            if m.conf_isVerbose():
                print("* Tables")
            for sSheet, dicSheet in odicLup.items():
                lstHeaders = dicSheet["header"]
                lstTab = dicSheet["content"]
                iRows = dicSheet["rows"]
                if bInfoOnly:
                    print(sSheet + ":", iRows)
                else:
                    print("--- " + sSheet)
                    print(PrettyText.table([lstHeaders] + lstTab))
            # Log
            sLog = "*** LOG\n>>> Ingest\n" + sLogIngest + "\n>>> Lookup\n" + sLogLookup
            if m.conf_isVerbose():
                print(sLog)
        else:
            # Excel
            print("Export to file: " + sExportFile)
            res = m.subsystem_getExporter().writeExcel(ingest=True, ingest_data=odicIng, ingest_wf=sWf, rfriendly=bRfriendly, ingest_sheets=lstExportSheets, lookup_data=odicLup, lookup=True, file=sExportFile)
            if res.isErr():
                print("Error:", res.getErrWarnInfoString(short=True))
                exit(1)
    elif args.exp_template:
        # Export: Template
        sWf = args.exp_template
        sFile = None
        if m.subsystem_getMapper().workflow_converter_import_isExisting(sWf):
            # Converter
            sWf_long = m.subsystem_getMapper().workflow_converter_import_short2long(sWf)
            sFile = PrettyText.clean_2filename(sWf_long) + "_template.xlsx" if sWf_long is not None else None
            res = m.subsystem_getConverter().write_converter_import_template(sWf, sFile)
            if res.isErr():
                print("Error:", res.getErrWarnInfoString(short=True))
                exit(1)
            print("Converter template written:", sFile)
        elif m.subsystem_getMapper().workflow_isExisting(sWf):
            # Normal
            sWf_long = m.subsystem_getMapper().workflow_short2long(sWf)
            sFile = PrettyText.clean_2filename(sWf_long) + "_template.xlsx" if sWf_long is not None else None
            res = m.subsystem_getExporter().writeExcel(ingest_wf=sWf, file=sFile, ingest=True, lookup=True)
            if res.isErr():
                print("Error:", res.getErrWarnInfoString(short=True))
                exit(1)
            print("Template written:", sFile)
        else:
            print("Unknown workflow:", sWf)
    elif args.bot:
        # Export: Template
        sConf = args.bot[0]
        bat = BatchBot(m)
        res = bat.startBatch(sConf)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        else:
            pass
            print(res.getData("log"))
    elif args.fao_new:
        # FAO: New
        lstLevel = args.fao_new
        level = None
        if len(lstLevel) ==1 and lstLevel[0] == "std":
            level = ["major", "division", "subarea"]
            sInfoLevel = PrettyText.lst2strC(level)
        elif len(lstLevel) ==1 and lstLevel[0] == "all":
            sInfoLevel = "<all available>"
            level = "*"
        else:
            level = lstLevel
            sInfoLevel = PrettyText.lst2strC(level)
        ein = input("New FAO regions: "+sInfoLevel+" [yN]? ")
        if ein == "" or ein.lower()[0] not in ('y', 'j'):
            print("Abort ...")
            exit(1)
        res = m.subsystem_getDb().fao_regions_create(level=level)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        else:
            print(res.getData("log"))
    elif args.fao_del:
        # FAO: Del mapping
        iIngId = args.fao_del
        if iIngId == -1:
            iIngId = None
            ein = input("Del whole FAO mapping [yN]? ")
        else:
            ein = input("Del FAO mapping for ingest "+str(iIngId) + " [yN]? ")
        if ein == "" or ein.lower()[0] not in ('y', 'j'):
            print("Abort ...")
            exit(1)
        res = m.subsystem_getDb().fao_map_clean(ing=iIngId)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        else:
            iDel = res.getData("del")
            if iDel is not None:
                print("Removed samples from mapping: "+ str(iDel))
    elif args.fao_stat:
        bShort = True if args.fao_stat == "short" else False
        # FAO: Info
        res = m.subsystem_getDb().fao_map_getData(short=bShort)
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        else:
            print(res.getData("log"))
    elif args.fao_update:
        # FAO: Update
        res = m.subsystem_getDb().fao_map_update()
        if res.isErr():
            print("Error:", res.getErrWarnInfoString(short=True))
            exit(1)
        else:
            print(res.getData("log"))
    elif args.test:
        # Generic test
        pass
    else:
        # Nichts ausgewaehlt
        print("Nothing to do ...")
