"""
Modul mit Controller
"""

try:
    from gui.view import View
    from gui.model import Model
    from gui.observer import Observer
    from PyQt5 import QtCore, QtGui, QtGui, uic
    from PyQt5.QtWidgets import *
    from PyQt5.QtGui import *
    import sys, os, platform, ctypes
except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")

    quit()



class Controller():
    """
    CONTROLLER

    - Der Controller ist "dumm" und besitzt keine Geschäftslogik
    """

    # Konstruktor
    def __init__(self, sGuiStyle=None):
        """
        Args:
            sGuiStyle (str): String der das Thema der QT-GUI beschreibt
        """

        # Bildschirmauflösung unter Windows holen
        sOs = platform.system().lower()
        if sOs == "windows":
            user32 = ctypes.windll.user32
            user32.SetProcessDPIAware()

        # Icon/Text-Scaling aktivieren
        QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)  # enable highdpi scaling
        QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)  # use highdpi icons

        # Keine Icons in Menus
        QApplication.setAttribute(QtCore.Qt.AA_DontShowIconsInMenus, True)
        QApplication.setAttribute(QtCore.Qt.AA_DontShowShortcutsInContextMenus, True)

        # Keine Nativen Menus
        #QApplication.setAttribute(QtCore.Qt.AA_DontUseNativeDialogs, True)
        QApplication.setAttribute(QtCore.Qt.AA_DontUseNativeMenuBar, True)

        # Icon/Text-Scaling deaktivieren
        #QApplication.setAttribute(QtCore.Qt.AA_DisableHighDpiScaling)
        #QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, False)  # enable highdpi scaling
        #QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, False)  # use highdpi icons

        app = QApplication(sys.argv)

        if sGuiStyle is not None:
            app.setStyle(QStyleFactory.create(sGuiStyle))

        # Instanzvariablen
        self.__model = Model() # Model
        self.__view = View(self.__model) # View
        self.__model.thread_ui_setObserver(Observer(self.__view)) # View-Observer bei Model anmelden
        self.__view.show()

        sys.exit(app.exec_())


