"""
Modul mit Model
"""

try:
    import sys, os, getpass, psycopg2, configparser, hashlib, base64, requests, re, platform, zipfile, shutil, subprocess, site
    from os.path import expanduser
    from tools.printing import SmartPrinter
    from tools.fileio import FileTools
    from helper.database import Database
    from helper.selector import Selector
    from logik.dbschema import DbSchema
    from logik.batchbot import BatchBot
    from tools.coding import Result
    from logik.taxon_wizard import TaxonWizard
    from logik.virtual_assistant import VirtualAssistant
    from logik.myworms import MyWorms
    from logik.ingestor import Ingestor
    from logik.exporter import Exporter
    from logik.importer import Importer
    from logik.tamer import Tamer
    from logik.validator import ExcelValidator
    from logik.converter import Converter
    from logik.feeder import Feeder
    from logik.lookup import Lookup
    from logik.mapper import Mapper
    from logik.regions import Regions
    from logik.connector import Connector
    from logik.datapatrol import DataPatrol
    from tools.coding import PrettyText

    from collections import Counter, OrderedDict
except ImportError as ex:
    print("Module missing ("+str(__file__)+"): " + ex.name + "\nPlease install.")

    quit()

class Model:
    """
    MODEL

    - Diese Klasse besitzt die Geschaeftslogik und -daten
    """

    def __init__(self):

        # Member

        self.__prnt = SmartPrinter(model=self)
        pn = self.__prnt.normal

        # App und Datenmodell: Version, Changelog und Dateien
        self.__sAppVersion = "?"  # Backup falls Datei nicht lesbar ist
        self.__sAppVersionFile = "version_app.txt"
        self.__sAppChangelogFile = "changelog_app.txt"
        self.__sDatenmodellVersion = "?"  # Backup falls Datei nicht lesbar ist
        self.__sDatenmodellVersionFile = "version_dm.txt"
        self.__sDatenmodellChangelogFile = "changelog_dm.txt"

        # Storage (Flüchtig im Speicher)
        self.__dicStorage = {}

        # Worker
        self.__worker = None

        # Observer der View
        self.__observer = None

        # Geschwaetzigkeit
        self.__bVerbose = False
        self.__bDebug = False

        # Dry Run
        self.__bDry = False

        # App Base Verzeichnis
        tmp_path_model = os.path.dirname(os.path.realpath(__file__))
        tmp_path_base = os.path.split(tmp_path_model)[0]
        self.__sAppBase = tmp_path_base

        # Pfad fuer SQL Scripte
        self.__sPathSqlScripts = self.__sAppBase + "/postgres/"

        # SQL: Leeres Datenmodell
        self.__sPathEmptyDatamodel = self.__sPathSqlScripts  + "cb_empty.sql"

        # User Data: Verzeichnis
        self.__sAppUserData_path = os.path.expanduser("~")
        self.__sAppUserData_dirName = "critterbase_user_data"
        self.__sAppUserData = self.__sAppUserData_path + "/" + self.__sAppUserData_dirName

        # ---

        # ABWÄRTSKOMPATIBILITÄT: Altes Marlin Science User Data verzeichnig gefunden -> Rename
        sOldMarlinScienceUserData = self.__sAppUserData_path + "/marlin_science_user_data"
        if os.path.exists(sOldMarlinScienceUserData):
            print("Found old Marlin Science User data folder: " + sOldMarlinScienceUserData)
            print("Rename to CRITTERBASE data folder: " + str(self.__sAppUserData))
            try:
                os.rename(sOldMarlinScienceUserData, self.__sAppUserData)
                print("... now its made.")
            except:
                print("Error: could not rename user data folder")
        sOldCritterUserData = self.__sAppUserData_path + "/critter_user_data"
        if os.path.exists(sOldCritterUserData):
            print("Found old CRITTER User data folder: " + sOldCritterUserData)
            print("Rename to CRITTERBASE data folder: " + str(self.__sAppUserData))
            try:
                os.rename(sOldCritterUserData, self.__sAppUserData)
                print("... now its made.")
            except:
                print("Error: could not rename user data folder")
        # -----

        # Suche Verzeichnis
        if not os.path.exists(self.__sAppUserData):
            try:
                print("No user data folder found: " + str(self.__sAppUserData))
                os.mkdir(self.__sAppUserData)
                print("... now its made.")
            except:
                print("Error: could not generate user data folder: " + str(self.__sAppUserData))

        # User Data: Tmp
        self.__sAppUserData_tmpDir = self.__sAppUserData + "/tmp"
        if not os.path.exists(self.__sAppUserData_tmpDir):
            try:
                print("No user data tmp folder found: " + str(self.__sAppUserData_tmpDir))
                os.mkdir(self.__sAppUserData_tmpDir)
                print("... now its made.")
            except:
                print("Error: could not generate user data tmp folder: " + str(self.__sAppUserData_tmpDir))

        # User Data: Batch
        self.__sAppUserData_batchDir = self.__sAppUserData + "/batch"
        if not os.path.exists(self.__sAppUserData_batchDir):
            try:
                print("No user data batch folder found: " + str(self.__sAppUserData_batchDir))
                os.mkdir(self.__sAppUserData_batchDir)
                print("... now its made.")
            except:
                print("Error: could not generate user data tmp folder: " + str(self.__sAppUserData_batchDir))

        # User Data: Updates
        self.__sAppUserData_updates = self.__sAppUserData + "/updates"
        if not os.path.exists(self.__sAppUserData_updates):
            try:
                print("No user data updates folder found: " + str(self.__sAppUserData_updates))
                os.mkdir(self.__sAppUserData_updates)
                print("... now its made.")
            except:
                print("Error: could not generate user data tmp folder: " + str(self.__sAppUserData_updates))

        # Generic WorkingDir
        self.__sWorkingDir = ""

        # User Data: Logging
        self.__bLogging = False
        self.__sAppUserData_loggingFile = self.__sAppUserData + "/log.txt"
        # User Data: Map
        self.__sAppUserData_htmlMapFile = self.__sAppUserData_tmpDir + "/map.html"
        # User Data: Snapshots
        self.__sAppUserData_snapshotDir = self.__sAppUserData + "/snapshots"
        self.__sAppUserData_snapshotDirSchema = self.__sAppUserData_snapshotDir + "/cb"
        self.__sAppUserData_snapshotDirWorms = self.__sAppUserData_snapshotDir + "/worms"
        if not os.path.exists(self.conf_getAppSnapshotDir()):
            try:
                print("No user data snapshot folder found: " + str(self.__sAppUserData_snapshotDir))
                os.mkdir(self.__sAppUserData_snapshotDir)
                print("... now its made.")
            except:
                print("Error: could not generate user data snapshot folder: " + str(self.__sAppUserData_snapshotDir))
        if not os.path.exists(self.conf_getAppSnapshotDirSchema()):
            try:
                print("No user data snapshot folder for schema data found: " + str(self.conf_getAppSnapshotDirSchema()))
                os.mkdir(self.conf_getAppSnapshotDirSchema())
                print("... now its made.")
            except:
                print("Error: could not generate snapshot folder for schema data: " + str(self.conf_getAppSnapshotDirSchema()))
        if not os.path.exists(self.conf_getAppSnapshotDirMyWorms()):
            try:
                print("No user data snapshot folder for WoRMS data found: " + str(self.conf_getAppSnapshotDirMyWorms()))
                os.mkdir(self.conf_getAppSnapshotDirMyWorms())
                print("... now its made.")
            except:
                print("Error: could not generate snapshot folder for WoRMSdata: " + str(self.conf_getAppSnapshotDirMyWorms()))

        # FAO
        self.__sFAO_server = 'http://www.fao.org/fishery/geoserver/fifao/ows?service=WFS&request=GetFeature&version=1.0.0&typeName=fifao:FAO_AREAS_CWP&outputFormat=json'
        self.__sFAO_server_mirror_file_zipped = "fao.zip"
        self.__sFAO_server_mirror_file_unzipped = "fao.txt"

        # Passwort
        self.__bSavePw = True

        # Defaults: Extension of ingest records in database
        self.__bRecordExtension = True
        self.__sRecordExtensionProblemReactionlevel = "error"

        # Defaults: Lookup import
        self.__bLookupImport_excel_goldrush = False
        self.__sLookupImport_excel = "all"
        self.__sLookupImport_db = "new"
        self.__sLookupImport_db_update = "all"
        self.__sLookupReuseDiffersReactionlevel = "error"

        # Map Block
        self.__bMapHandbreak = True

        # Observer der View
        self.__observer = None

        # Infos zum Programm
        self.__sInfoPrgName = "CRITTERBASE Collector\'s App"
        self.__sInfoDatum = u"2021" # u"Copyright \N{COPYRIGHT SIGN}"
        self.__sInfoKommentar = """
        <i>CRITTERBASE Collector's App</i>
        <br>
        <br>
        Copyright 2021 - Alfred Wegener Institute, Helmholtz Centre for Polar and Marine Research
        <small>
        <br>
        <br>
        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are met:
        <br>
        1. Redistributions of source code must retain the above copyright notice, this
           list of conditions and the following disclaimer.
        <br>
        2. Redistributions in binary form must reproduce the above copyright notice,
           this list of conditions and the following disclaimer in the documentation
           and/or other materials provided with the distribution.
        <br>
        3. Neither the name of the copyright holder nor the names of its
           contributors may be used to endorse or promote products derived from
           this software without specific prior written permission.
        <br>
        <br>
        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
        AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
        IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
        FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
        DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
        SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
        CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
        OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
        """
        self.__sInfoWeb = "http://www.awi.de/en/science/biosciences/functional-ecology/main-research-focus/ecosystem-functions.html"

        # Update-Server (neu2)
        self.__sUpdateServer = "https://critterbase.awi.de/update"
        self.__sUpdateAppLinux = "critterbase_linux.zip"
        self.__sUpdateAppWin = "critterbase_win.zip"
        self.__sUpdateAppMac = "critterbase_mac.zip"

        # Admin
        self.__sAdminPsqlPath = None
        self.__sAdminPgdumpPath = None
        self.__sAdminPsqlVersion = None
        self.__sAdminPgdumpVersion = None
        # Hole Programmversion aus Datei (darf nicht mit ins Git gepushed werden -> wird durch git-push-skript erzeugt)
        try:
            sVersionDatei = self.__sAppBase + "/" + self.__sAppVersionFile
            with open(sVersionDatei, 'r') as f:
                self.__sAppVersion = f.read().strip()
        except:
            print("File with app version not found: " + self.__sAppVersionFile + " -> using backup value:" + self.__sAppVersion)
        # Hole Datenmodellversion aus Datei
        try:
            sVersionDatei = self.__sAppBase + "/" + self.__sDatenmodellVersionFile
            with open(sVersionDatei, 'r') as f:
                self.__sDatenmodellVersion = f.read().strip()
        except:
            print(
                "File with datamodel version not found: " + self.__sDatenmodellVersionFile + " -> using backup value:" + self.__sDatenmodellVersion)

        # Datenbank
        self.__db = Database(self)

        # Timeout für Datenbankanfragen
        self.__iDbTimeOut = 3  # Sekunden

        # Ist Adminbereich offen
        self.__bAdminLock = False

        # Selector
        self.__selector = Selector(self)
        self.__dicStatistics = None

        # Hole die Konfiguration
        self.__sConfigFilename = self.__sAppUserData + "/cfg.txt"
        self.conf_cfgRead()

        # Evaluiere die PG Admin Tools
        self.core_evalPgTools()

        # --- Sub Systeme

        # Mapper
        self.__mapper = Mapper(self)
        if not self.__mapper.isWorking():
            pn("Error initializing mapper module -> quit")
            quit()

        # Lookup
        self.__lookup = Lookup(self)
        if not self.__lookup.isWorking():
            pn("Error initializing lookup module -> quit")
            quit()

        # Regions
        self.__regions = Regions(self)
        if not self.__regions.isWorking():
            pn("Error initializing regionsl module -> quit")
            quit()

        # DbSchema
        self.__dbschema = DbSchema(self)
        if not self.__dbschema.isWorking():
            pn("Error initializing comments module -> quit")
            quit()

        # Converter
        self.__converter = Converter(self)
        if not self.__converter.isWorking():
            pn("Error initializing converter module -> quit")
            quit()

        # Tamer
        self.__tamer = Tamer(self)
        if not self.__tamer.isWorking():
            pn("Error initializing Tamer module -> quit")
            quit()

        # Validator
        self.__validator = ExcelValidator(self)
        if not self.__validator.isWorking():
            pn("Error initializing cecking module -> quit")
            quit()

        # Ingestor
        self.__ingestor = Ingestor(self)
        if not self.__ingestor.isWorking():
            pn("Error initializing ingestor module -> quit")
            quit()

        # Feeder
        self.__dbfeeder = Feeder(self)
        if not self.__dbfeeder.isWorking():
            pn("Error initializing feeder module -> quit")
            quit()

        # Exporter
        self.__exporter = Exporter(self)
        if not self.__exporter.isWorking():
            pn("Error initializing exporter module -> quit")
            quit()

        # MyWoRMS
        self.__myworms = MyWorms(self)
        if not self.__myworms.isWorking():
            pn("Error initializing My WoRMS module -> quit")
            quit()
        self.__myworms.puffer_load()
        self.__myworms.puffer_createStats()

        # TaxonWizard
        self.__wizard = TaxonWizard(self)
        if not self.__wizard.isWorking():
            pn("Error initializing TaxonWizard module -> quit")
            quit()

        # TaxonWizard
        self.__virtualAssistant = VirtualAssistant(self)
        if not self.__virtualAssistant.isWorking():
            pn("Error initializing VirtualAssistant module -> quit")
            quit()

        # DataPatrol
        self.__dataPatrol = DataPatrol(self)
        if not self.__dataPatrol.isWorking():
            pn("Error initializing DataPatrol module -> quit")
            quit()

        # Importer
        self.__importer = Importer(self)
        if not self.__importer.isWorking():
            pn("Error initializing importer module -> quit")
            quit()

        # BatchBot
        self.__batchbot = BatchBot(self)
        if not self.__batchbot.isWorking():
            pn("Error initializing BatchBot -> quit")
            quit()

        # Director
        self.__connector = Connector(self)
        if not self.__connector.isWorking():
            pn("Error initializing Connector -> quit")
            quit()

        # Subsysteme
        self.__odicSubSys = OrderedDict([
            ("connector", self.__connector),
            ("db", self.__db),
            ("mapper", self.__mapper),
            ("dbschema", self.__dbschema),
            ("regions", self.__regions),
            ("checker", self.__validator),
            ("tamer", self.__tamer),
            ("lookup", self.__lookup),
            ("converter", self.__converter),
            ("importer", self.__importer),
            ("ingestor", self.__ingestor),
            ("feeder", self.__dbfeeder),
            ("exporter", self.__exporter),
            ("myworms", self.__myworms),
            ("wizard", self.__wizard),
            ("virtualassistant", self.__virtualAssistant),
            ("datapatrol", self.__dataPatrol),
            ("batchbot", self.__batchbot),
            ("base", self)
        ])

    # -- Info

    def info(self):
        """
        Informationen über die Konfiguration wird ausgedruckt (Smartprinter wird benutzt)
        """

        pn = self.__prnt.normal
        pv = self.__prnt.verbose
        try:
            sPyVersion = sys.version.split()[0]
        except:
            sPyVersion = sys.version
        try:
            sPyMods = PrettyText.lst2strC(os.listdir(site.getusersitepackages()))
        except:
            sPyMods = "?"

        sInfo = "*** Basic Information\n\n"
        sInfo += "PyVersion:  " + str(sPyVersion) + "\n"
        sInfo += "PyDir :     " + str(site.getuserbase()) + "\n"
        sInfo += "PyModDir:   " + str(site.getusersitepackages()) + "\n"
        sInfo += "PyModules:  " + str(sPyMods) + "\n"
        sInfo += "AppVer:     " + self.__sAppVersion + "\n"
        sInfo += "SchemaVer:  " + self.__sDatenmodellVersion + "\n"
        sInfo += "Accounts:   " + str(self.__db.account_all_getNr()) + "\n"
        sInfo += "AppBaseDir: " + self.__sAppBase + "\n"
        sInfo += "UserDatDir: " + self.__sAppUserData + " " + ("<exists>" if os.path.exists(self.__sAppUserData) else "<missing>") + "\n"
        sInfo += "Config:     " + self.__sConfigFilename + " " + ("<exists>" if os.path.isfile(self.__sConfigFilename) else "<missing>") + "\n"
        sInfo += "Logging:    " + ("on" if self.__bLogging == True else "off") + "\n"
        sInfo += "Log file:   " + (self.__sAppUserData_loggingFile + " " if self.__sAppUserData_loggingFile != "" else "") + ("<exists>" if os.path.isfile(self.__sAppUserData_loggingFile) else "<missing>") + "\n"
        sInfo += "TmpDir:     " + self.__sAppUserData_tmpDir + " " + ("<exists>" if os.path.exists(self.__sAppUserData_tmpDir) else "<missing>") + "\n"
        sInfo += "WorkDir:    " + self.__sWorkingDir + " " + ("<exists>" if os.path.exists(self.__sWorkingDir) else "<missing>") + "\n"
        sInfo += "SnapsDir:   " + self.__sAppUserData_snapshotDir + " " + ("<exists>" if os.path.exists(self.__sAppUserData_snapshotDir) else "<missing>") + "\n"
        sInfo += "Verbose:    " + ("on" if self.__bVerbose == True else "off") + "\n"
        sInfo += "Debug:      " + ("on" if self.__bDebug == True else "off") + "\n"
        sInfo += "Dry:        " + ("on" if self.__bDry == True else "off") + "\n"
        sInfo += "RecExt:     " + ("on" if self.__bRecordExtension == True else "off") + "\n"
        sInfo += "RecExtLvl:  " + str(self.__sRecordExtensionProblemReactionlevel) + "\n"
        sInfo += "UpdateSrv:  " + self.conf_getUpdateServer() + "\n"
        sInfo += "psql:       " + ("<working>" if self.core_isAdminPsqlWorking() else "<not working>") + "\n"
        sInfo += "psql-bin:   " + (self.__sAdminPsqlPath if self.__sAdminPsqlPath is not None and self.__sAdminPsqlPath != "" else "<not explicitly set - using psql in PATH>") + "\n"
        sInfo += "psql-ver:   " + (self.__sAdminPsqlVersion if self.__sAdminPsqlVersion is not None else "-") + "\n"
        sInfo += "pgdump:     " + ("<working>" if self.core_isAdminPgdumpWorking() else "<not working>") + "\n"
        sInfo += "pgdump-bin: " + (self.__sAdminPgdumpPath if self.__sAdminPgdumpPath is not None and self.__sAdminPgdumpPath != "" else "<not explicitly set - using pg_dump in PATH>") + "\n"
        sInfo += "pgdump-ver: " + (self.__sAdminPgdumpVersion if self.__sAdminPgdumpVersion is not None else "-") + "\n"
        sInfo += "db-model:   " + self.__sPathEmptyDatamodel +"\n"
        sInfo += "\n"
        return sInfo

    # -- Info Subsysteme
    def subsystem_getSubLst(self):
        return list(self.__odicSubSys.keys())

    def subsystem_getSubInfo(self, sSys):
        if sSys in self.__odicSubSys:
            return self.__odicSubSys[sSys].info()
        else:
            return None

    def subsystem_getMapper(self):
        return self.__mapper

    def subsystem_getConverter(self):
        return self.__converter

    def subsystem_getTamer(self):
        return self.__tamer

    def subsystem_getValidator(self):
        return self.__validator

    def subsystem_getImporter(self):
        return self.__importer

    def subsystem_getExporter(self):
        return self.__exporter

    def subsystem_getFeeder(self):
        return self.__dbfeeder

    def subsystem_getIngestor(self):
        return self.__ingestor

    def subsystem_getLookup(self):
        return self.__lookup

    def subsystem_getTaxonWizard(self):
        return self.__wizard

    def subsystem_getMyWorms(self):
        return self.__myworms

    def subsystem_getVirtualAssistant(self):
        return self.__virtualAssistant

    def subsystem_getDataPatrol(self):
        return self.__dataPatrol

    def subsystem_getBatchBot(self):
        return self.__batchbot

    def subsystem_getRegions(self):
        return self.__regions

    def subsystem_getDb(self):
        return self.__db

    def subsystem_getSchema(self):
        return self.__dbschema

    def subsystem_getConnector(self):
        return self.__connector

    # --- Conf

    def conf_getDbTimeout(self):
        """
        Hole Timeout für Datenbankzugriffe
        Returns: (int) timeout : Timeout in Sekunden
        """
        return self.__iDbTimeOut

    def conf_getDbSchema(self):
        return self.__dbschema

    def conf_getMap(self):
        """
        Hole die Datei für die Map
        """
        return self.__sAppUserData_htmlMapFile

    def conf_getDatenmodellChangelogFile(self):
        """
        Hole Datei zu Changelog Datenmodell
        """
        return self.__sDatenmodellChangelogFile

    def conf_getAppChangelogFile(self):
        """
        Hole Datei zu Changelog App
        """
        return self.__sAppChangelogFile

    def conf_getEmptyDatenmodellFile(self):
        """
        Hole leeres Datenmodell
        """
        return self.__sPathEmptyDatamodel

    def conf_getUserDataDir(self):
        """
        Hole das App-Userdata-Verzeichnis
        """
        return self.__sAppUserData

    def conf_getUserHomeDir(self):
        """
        Hole das User Homeverzeichnis
        """
        return expanduser("~")

    def conf_getAppBaseDir(self):
        """
        Hole App Base Dir
        """
        return self.__sAppBase

    def conf_getAppSqlScriptsDir(self):
        """
        Hole SQL Scripts Dir
        """
        return self.__sPathSqlScripts

    def conf_getAppTmpDir(self):
        """
        Hole App Temp Dir
        """
        return self.__sAppUserData_tmpDir

    def conf_getAppBatchDir(self):
        """
        Hole App Batch Dir
        """
        return self.__sAppUserData_batchDir

    def conf_getAppUpdateDir(self):
        """
        Hole Update Temp Dir
        """
        return self.__sAppUserData_updates

    def conf_getAppSnapshotDir(self):
        """
        Hole App Snapshot Dir
        """
        return self.__sAppUserData_snapshotDir

    def conf_getAppSnapshotDirSchema(self):
        """
        Hole App Snapshot Dir
        """
        return self.__sAppUserData_snapshotDirSchema

    def conf_getAppSnapshotDirMyWorms(self):
        """
        Hole App Snapshot Dir
        """
        return self.__sAppUserData_snapshotDirWorms

    def conf_getUpdateServer(self):
        """
        Hole die Adresse des Update Servers
        """
        return self.__sUpdateServer

    # FAO: Dateiname der FAO-Regioen auf Updateserver (zipped und unzipped)
    def conf_getFAO_server(self):
        return self.__sFAO_server
    def conf_getFAO_mirror_zipped(self):
        return self.__sFAO_server_mirror_file_zipped
    def conf_getFAO_mirror_unzipped(self):
        return self.__sFAO_server_mirror_file_unzipped

    """
    Hole die Defaultwerte zum Lookup Import
    """

    def conf_getLookupReuseDiffersReactionlevel(self):
        return self.__sLookupReuseDiffersReactionlevel
    def conf_setLookupReuseDiffersReactionlevel(self, sLevel):
        if sLevel not in ["error", "warning", "info"]:
            sLevel = "error"
        self.__sLookupReuseDiffersReactionlevel = sLevel

    def conf_getLookupImport_excel(self):
        return self.__sLookupImport_excel
    def conf_setLookupImport_excel(self, sMode):
        if sMode not in ["all", "gold", "not-gold"]:
            sMode = "all"
        self.__sLookupImport_excel = sMode

    def conf_getLookupImport_excel_goldrush(self):
        return self.__bLookupImport_excel_goldrush
    def conf_setLookupImport_excel_goldrush(self, bBool):
        self.__bLookupImport_excel_goldrush = bBool

    def conf_getLookupImport_db(self):
        return self.__sLookupImport_db
    def conf_setLookupImport_db(self, sMode):
        if sMode not in ["both", "new", "update"]:
            sMode = "both"
        self.__sLookupImport_db = sMode

    def conf_getLookupImport_db_update(self):
        return self.__sLookupImport_db_update
    def conf_setLookupImport_db_update(self, sMode):
        if sMode not in ["all", "gold", "not-gold"]:
            sMode = "all"
        self.__sLookupImport_db_update = sMode

    def conf_getRecordExtensionProblemReactionlevel(self):
        return self.__sRecordExtensionProblemReactionlevel
    def conf_setRecordExtensionProblemReactionlevel(self, sLevel):
        if sLevel not in ["error", "warning", "info"]:
            sLevel = "error"
        self.__sRecordExtensionProblemReactionlevel = sLevel

    def conf_isRecordExtensionAllowed(self):
        """
        Ist die Record Extension aktiviert?
        """
        return self.__bRecordExtension

    def conf_setRecordExtensionAllowed(self, bBool):
        """
        Setze die Record Extension
        """
        self.__bRecordExtension = bBool

    def conf_getWorkingDir(self):
        return self.__sWorkingDir

    def conf_setWorkingDir(self, sDir):
        self.__sWorkingDir = sDir

    def conf_getAdminPsqlPath(self):
        """
        Hole Path zu psql Kommando
        """
        return self.__sAdminPsqlPath

    def conf_getAdminPsqlVersion(self):
        """
        Hole Version von psql Kommando
        """
        return self.__sAdminPsqlVersion

    def conf_setAdminPsqlPath(self, sPath):
        """
        Setze Path psql Kommando
        """
        self.__sAdminPsqlPath = sPath

    def conf_getAdminPgdumpPath(self):
        """
        Hole Path zu pg_dump Kommando
        """
        return self.__sAdminPgdumpPath

    def conf_getAdminPgdumpVersion(self):
        """
        Hole Version von pg_dump Kommando
        """
        return self.__sAdminPgdumpVersion

    def conf_setAdminPgdumpPath(self, sPath):
        """
        Setze Path zu pg_dump Kommando
        """
        self.__sAdminPgdumpPath = sPath

    def conf_getOs(self):
        """
        Hole den Betriebssystem-Typ

        Returns: win,linux,mac
        """

        sOs = platform.system().lower()
        if sOs == "windows":
            return "win"
        elif sOs == "linux":
            return "lin"
        elif sOs == "darwin":
            return "mac"
        else:
            return None

    def conf_setVerbosity(self, bVerbose, bDebug):
        """
        Die Geschwaetzigkeit setzten

        Args:
            bVerbose (bool): Verbose an/aus?
            bDebug (bool): Debug an/aus?
        """

        self.conf_setVerbose(bVerbose)
        self.conf_setDebug(bDebug)

    def conf_setVerbose(self, bVerbose):
        self.__bVerbose = bVerbose

    def conf_isVerbose(self):
        return self.__bVerbose

    def conf_setDebug(self, bDebug):
        self.__bDebug = bDebug

    def conf_isDebug(self):
        return self.__bDebug

    def conf_isNormal(self):
        return not self.__bVerbose and not self.__bDebug

    def conf_setDry(self, bDry):
        self.__bDry = bDry

    def conf_isDry(self):
        return self.__bDry

    def conf_setAdminLock(self, bLock):
        self.__bAdminLock = bLock

    def conf_getAdminLock(self):
        return self.__bAdminLock

    def conf_setPwSave(self, bSave):
        self.__bSavePw = bSave

    def conf_isPwSave(self):
        return self.__bSavePw

    def conf_getInfo(self, sTyp):
        if sTyp == "prgname":
            return self.__sInfoPrgName
        elif sTyp == "version":
            return self.__sAppVersion
        elif sTyp == "dbschema":
            return self.__sDatenmodellVersion
        elif sTyp == "date":
            return self.__sInfoDatum
        elif sTyp == "comment":
            return self.__sInfoKommentar
        elif sTyp == "www":
            return self.__sInfoWeb
        else:
            return None


    def conf_getDateiLogging(self):
        return self.__sAppUserData_loggingFile

    def conf_isLogging(self):
        return self.__bLogging

    def conf_setLogging(self, logging):
        self.__bLogging = logging

    def conf_cfgRead(self):
        """
        Konfigurationsdatei für das gesamte Programm einlesen

        Returns:
            (bool): Vorgang war erfolgreich?
        """

        try:
            cfg = configparser.ConfigParser()
            cfg.read(self.__sConfigFilename)
            # DB (Erzeuge Accounts und fülle diese mit gefunenenen Daten aus Konfigurationsdatei)
            i = 0
            bFund = False
            while True:
                sPreset = sHost = sPort = sName = sSchema = sUser = sPw = ""
                sIdx = "_" + str(i)
                if not cfg.has_option("DB", "db_preset" + sIdx):
                    break
                if cfg.has_option("DB", "db_preset" + sIdx): sPreset = str(cfg["DB"]["db_preset" + sIdx])
                if cfg.has_option("DB", "db_host" + sIdx): sHost = str(cfg["DB"]["db_host" + sIdx])
                if cfg.has_option("DB", "db_port" + sIdx): sPort = str(cfg["DB"]["db_port" + sIdx])
                if cfg.has_option("DB", "db_name" + sIdx): sName = str(cfg["DB"]["db_name" + sIdx])
                if cfg.has_option("DB", "db_schema" + sIdx): sSchema = (str(cfg["DB"]["db_schema" + sIdx]))
                if cfg.has_option("DB", "db_user" + sIdx): sUser = str(cfg["DB"]["db_user" + sIdx])
                if cfg.has_option("DB", "db_pw" + sIdx): sPw = self.tools_decrypt(str(cfg["DB"]["db_pw" + sIdx]))
                self.__db.account_single_add({"account": sPreset, "host": sHost, "port": sPort, "db": sName, "schema": sSchema, "user": sUser, "pw": sPw})
                bFund = True
                i += 1
            # Dummyserver
            if not bFund:
                self.__db.account_single_add({"account": "new", "host": "", "port": "", "db": "", "schema": "", "user": "", "pw": ""})
                # Default
            if cfg.has_option("DEFAULT", "record_extension"): self.__bRecordExtension = (True if str(cfg["DEFAULT"]["record_extension"]) == "True" else False)
            if cfg.has_option("DEFAULT","record_extension_problem_reaction_level"): self.conf_setRecordExtensionProblemReactionlevel(str(cfg["DEFAULT"]["record_extension_problem_reaction_level"]))
            if cfg.has_option("DEFAULT","lookup_reuse_problem_reaction_level"): self.conf_setLookupReuseDiffersReactionlevel(str(cfg["DEFAULT"]["lookup_reuse_problem_reaction_level"]))
            if cfg.has_option("DEFAULT","lookup_import_excel"): self.conf_setLookupImport_excel(str(cfg["DEFAULT"]["lookup_import_excel"]))
            if cfg.has_option("DEFAULT","lookup_import_excel_goldrush"): self.__bLookupImport_excel_goldrush = (True if str(cfg["DEFAULT"]["lookup_import_excel_goldrush"]) == "True" else False)
            if cfg.has_option("DEFAULT","lookup_import_db"): self.conf_setLookupImport_db(str(cfg["DEFAULT"]["lookup_import_db"]))
            if cfg.has_option("DEFAULT","lookup_import_db_update"): self.conf_setLookupImport_db_update(str(cfg["DEFAULT"]["lookup_import_db_update"]))
            if cfg.has_option("DEFAULT", "batchbot_logdir"): self.conf_setWorkingDir(str(cfg["DEFAULT"]["batchbot_logdir"]))
                # Etc
            if cfg.has_option("ETC", "verbose"): self.__bVerbose = (
                True if str(cfg["ETC"]["verbose"]) == "True" else False)
            if cfg.has_option("ETC", "debug"): self.__bDebug = (True if str(cfg["ETC"]["debug"]) == "True" else False)
            if cfg.has_option("ETC", "dry"): self.__bDry = (True if str(cfg["ETC"]["dry"]) == "True" else False)
            if cfg.has_option("ETC", "save_pw"): self.__bSavePw = (
                True if str(cfg["ETC"]["save_pw"]) == "True" else False)
            if cfg.has_option("ETC", "logging"): self.__bLogging = (
                True if str(cfg["ETC"]["logging"]) == "True" else False)
            if cfg.has_option("ETC", "path_psql"): self.__sAdminPsqlPath = str(cfg["ETC"]["path_psql"])
            if cfg.has_option("ETC", "path_pgdump"): self.__sAdminPgdumpPath = str(cfg["ETC"]["path_pgdump"])
            self.__db.account_active_setIndex(0)  # Default Account Nr
            if cfg.has_option("ETC", "aktiver_db_account"): self.__db.account_active_setIndex(
                int(cfg["ETC"]["aktiver_db_account"]))  # Hole aus Config
        except (configparser.Error, ValueError) as ex:
            self.__prnt.normal("Error while reading the config file (" + self.__sConfigFilename, "): " + str(ex))
            return False
        return True

    def conf_cfgSave(self, **kwargs):
        """
        Konfigurationsdatei für das gesamte Programm abspeichern

        Kwargs:
            savePw (bool): Sollen die Userpassworte mit abgespeichert werden?

        Returns:
            (bool): Vorgang war erfolgreich?
        """

        bSavePw = True  # Passwort speichern
        # Hole Argumente
        for arg in kwargs:
            if "savePw" in kwargs:
                bSavePw = kwargs['savePw']
        try:
            cfg = configparser.ConfigParser()
            cfg.read(self.__sConfigFilename)
            # Accounts
            cfg["DB"] = {}
            for iIdx, sAccount in enumerate(self.__db.account_all_getNames()):
                sIdx = "_" + str(iIdx)
                cfg["DB"]["db_preset" + sIdx] = sAccount
                cfg["DB"]["db_host" + sIdx] = self.__db.account_single_getPart("host", account=sAccount)
                cfg["DB"]["db_port" + sIdx] = self.__db.account_single_getPart("port", account=sAccount)
                cfg["DB"]["db_name" + sIdx] = self.__db.account_single_getPart("db", account=sAccount)
                cfg["DB"]["db_schema" + sIdx] = self.__db.account_single_getPart("schema", account=sAccount)
                cfg["DB"]["db_user" + sIdx] = self.__db.account_single_getPart("user", account=sAccount)
                if bSavePw == True:
                    cfg["DB"]["db_pw" + sIdx] = self.tools_encrypt(self.__db.account_single_getPart("pw", account_index=iIdx))
                else:
                    cfg["DB"]["db_pw" + sIdx] = ""
                # Default
            cfg["DEFAULT"] = {}
            cfg["DEFAULT"]["record_extension"] = "True" if self.__bRecordExtension else "False"
            cfg["DEFAULT"]["record_extension_problem_reaction_level"] = self.__sRecordExtensionProblemReactionlevel
            cfg["DEFAULT"]["lookup_reuse_problem_reaction_level"] = self.__sLookupReuseDiffersReactionlevel
            cfg["DEFAULT"]["lookup_import_excel"] = self.__sLookupImport_excel
            cfg["DEFAULT"]["lookup_import_excel_goldrush"] = "True" if self.__bLookupImport_excel_goldrush else "False"
            cfg["DEFAULT"]["lookup_import_db"] = self.__sLookupImport_db
            cfg["DEFAULT"]["lookup_import_db_update"] = self.__sLookupImport_db_update
            cfg["DEFAULT"]["batchbot_logdir"] = self.conf_getWorkingDir()
                # Etc
            cfg["ETC"] = {}
            cfg["ETC"]["verbose"] = "True" if self.__bVerbose else "False"
            cfg["ETC"]["debug"] = "True" if self.__bDebug else "False"
            cfg["ETC"]["logging"] = "True" if self.__bLogging else "False"
            cfg["ETC"]["dry"] = "True" if self.__bDry else "False"
            cfg["ETC"]["save_pw"] = "True" if self.__bSavePw else "False"
            cfg["ETC"]["path_psql"] = self.__sAdminPsqlPath
            cfg["ETC"]["path_pgdump"] = self.__sAdminPgdumpPath
            cfg["ETC"]["aktiver_db_account"] = str(self.__db.account_active_getIndex())
            with open(self.__sConfigFilename, 'w') as cfgdat:
                cfg.write(cfgdat)
            self.__prnt.verbose("Configuration is saved in file: " + self.__sConfigFilename)
        except (configparser.Error, ValueError) as ex:
            self.__prnt.normal("Error while saving the configuration file (" + self.__sConfigFilename, "): " + str(ex))
            return False
        return True

    # --- Store things in Memory

    def memory_save(self, sName, data):
        if sName not in self.__dicStorage:
            self.__dicStorage[sName] = data

    def memory_del(self, sName=None):
        if sName == None:
            self.__dicStorage = {}
        else:
            if sName in self.__dicStorage:
                del self.__dicStorage[sName]

    def memory_load(self, sName):
        if sName in self.__dicStorage:
            return self.__dicStorage[sName]
        else:
            return None

    # --- Thread

    def thread_getWorker(self):
        """

        Worker holen

        Returns:
            tools.parallel.Worker: Worker
        """
        return self.__worker

    def thread_setWorker(self, w):
        """
        Worker setzen

        Args:
            tools.threadtools.Worker: Worker

        """
        self.__worker = w

    def thread_ui_setObserver(self, observer):
        """
        Den Observer an der View anmelden

        Args:
            observer (gui.observer.Observer): Observer
        """

        self.__observer = observer

    def thread_ui_getObserver(self):
        """
        Den an die View angemeldeten Observer zurückliefern

        Returns:
            gui.observer.Observer: Observer
        """

        return self.__observer

    # --- UI

    def ui_selector_getSelector(self):
        """
        Hole den Selector
        Returns:
            helper.selector.Selector: selector

        """
        return self.__selector

    def ui_selector_getStatistics(self):
        """
        Hole die Statistik der Selection
        Returns:
            dict: statistik

        """
        return self.__dicStatistics

    def ui_selector_setStatistics(self, dicStat):
        """
        Setze die Statistik der Selection
        Args:
            (dict): Statisitk

        """
        self.__dicStatistics = dicStat

    def ui_getSmartprinter(self):
        """
        Hole Smartprinter

        """
        return self.__prnt

        # --- Core

    def core_isAdminPsqlWorking(self):
        """
        Funktioniert das psql Kommando
        """
        return True if self.__sAdminPsqlVersion is not None else False

    def core_isAdminPgdumpWorking(self):
        """
        Funktioniert das pg_dump Kommando
        """
        return True if self.__sAdminPgdumpVersion is not None else False

    def core_evalPgTools(self):
        """
        Ermittle die Postgres Command line tools
        """
        pd = self.ui_getSmartprinter().debug

        # Helper: teste ob Prg wirklich am Ende von Path steht korrekt ist
        def testPath(sPath, sName):

            # Existiert die Datei
            if not os.path.isfile(sPath):
                pd("Error evaluate PostgreSQL command line tools: The path [" + str(
                    sPath) + "] is no file")
                return False

            # Ist das Binary erlaubt (psql, pg_dump)
            sFront, sTail = FileTools.cutAtPathTail(sPath)
            if sTail == sName or sTail == sName + ".exe":
                return True
            else:
                pd("Error evaluate PostgreSQL command line tools: The executable of path [" + str(
                    sPath) + "] is not [" + str(sName) + "]")
                return False

        # Helper: Extrahiere Versionsnummer aus Text
        def extractVersionNumber(sText):
            try:
                sText = sText.replace("\n", "")
                lstText = sText.split()
                sVersion = lstText[2]  # Zweiter Textteil = Versionsnummer
                # Version enthält mehr als Zahlen und Punkte -> Irgendwas ist schiefgegangen
                sNotAllowedChars = re.sub(r'[0-9.]', "", sVersion)
                if sNotAllowedChars != "":
                    pd("Error evaluate PostgreSQL command line tools: Could not extract version number of  [" + str(
                        sText) + "]")
                    return None
                else:
                    return sVersion
            except:
                return None

        # Helper: Teste Version von einem Tool
        def testVersion(sBin):
            try:
                # Windows-only: Startup-Info -> verändern, damit die Konsole verborgen bleibt (würde auch über shell=True funktionieren)
                if self.conf_getOs() == "win":
                    startup_info = subprocess.STARTUPINFO()
                    startup_info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                else:
                    startup_info = None

                # Kommando testen
                sCmd = sBin + " --version"
                lstCmd = sCmd.split()

                p = subprocess.Popen(lstCmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                     startupinfo=startup_info)
                p.wait()
                lstStdOut = p.stdout.readlines()
                sStdOut = ""
                if lstStdOut is not None and len(lstStdOut) > 0:
                    for sStdOutLine in lstStdOut:
                        sStdOut += str(sStdOutLine.decode("utf-8"))
                sOut = extractVersionNumber(sStdOut)
            except Exception as ex:
                sOut = None
            return sOut

        # Teste versionen
        # Psql
        if self.conf_getAdminPsqlPath() is None or self.conf_getAdminPsqlPath() == "":
            self.__sAdminPsqlVersion = testVersion("psql")
        else:
            if testPath(self.conf_getAdminPsqlPath(), "psql"):
                self.__sAdminPsqlVersion = testVersion(self.conf_getAdminPsqlPath())
            else:
                self.__sAdminPsqlVersion = None
        # PgDump
        if self.conf_getAdminPgdumpPath() is None or self.conf_getAdminPgdumpPath() == "":
            self.__sAdminPgdumpVersion = testVersion("pg_dump")
        else:
            if testPath(self.conf_getAdminPgdumpPath(), "pg_dump"):
                self.__sAdminPgdumpVersion = testVersion(self.conf_getAdminPgdumpPath())
            else:
                self.__sAdminPgdumpVersion = None

    # --- Tools

    def tools_encrypt(self, txt, mode="base64"):
        """
        Text verschlüsseln

        - Momentan wird lediglich eine Verschleierung durch den base64-Algorithmus angeboten

        Args:
            txt (str): unverschlüsselter Text
            mode="base64" (str): Verschlüsselungsart

        Returns:
            (str): verschlüsselter Text
        """

        if mode == "base64":
            return base64.encodebytes(txt.encode()).decode()
        return None

    def tools_decrypt(self, txt, mode="base64"):
        """
        Text entschlüsseln

        Args:
            txt (str): verschlüsselter Text
            mode="base64" (str): Verschlüsselungsart

        Returns:
            (str): entschlüsselter Text
        """

        if mode == "base64":
            return base64.decodebytes(txt.encode()).decode()
        return None

    def help_updateCheck(self):
        """
        Teste ob die installierte Version geupdatet werden kann

        Versionsnummern:
            build           Kompilation
            patch level     Bugfixes
            minor release   Funktionale Erweiterung
            major release   Signifikante Veränderung am Programm
        """

        # Helper
        # Whitespaces aus String entfernen
        def cleanWhiteSpaces(sEin):
            if sEin is None:
                return None
            else:
                return re.sub(r'[\ \t\n]', "", sEin)

        # Evaluiere Versionsstring
        def evalVersion(sVersion):

            if sVersion is None:
                return None, None, None, None

            # Whitespaces entfernen
            sVersion = cleanWhiteSpaces(sVersion)

            # Teste auf unerlaubte Zeichen
            sNotAllowedChars = re.sub(r'[0-9.-]', "", sVersion)
            if sNotAllowedChars != "":
                return None, None, None, None

            iBuild = None
            iPatch = None
            iMin = None
            iMaj = None

            try:
                # Check String-Format (major.minor.patch-build)
                erg = re.fullmatch(r'^[0-9]+\.[0-9]+\.[0-9]+-[0-9]+', sVersion)
                if erg is not None:
                    # Extrahiere Einzelwerte
                    lst1 = sVersion.split("-")
                    lst2 = lst1[0].split(".")
                    iBuild = int(lst1[1])
                    iMaj = int(lst2[0])
                    iMin = int(lst2[1])
                    iPatch = int(lst2[2])
            except:
                pass

            return iMaj, iMin, iPatch, iBuild

        # Hole Information von Updateserver
        def getInfoFromServer(sUrl):
            try:
                r = requests.get(sUrl, allow_redirects=True, timeout=3, auth=('critter', 'P0lych4eta')) # TWEAK für Preview Version
                if r.status_code == 200 and ('text/plain' in r.headers['Content-Type']):
                    return r.text
                else:
                    return None
            except Exception as ex:
                return None

        # Helper
        def isVersion1Greater2(sV1, sV2):
            # Vergleich V1 mit V2
            lst1 = sV1.split(".")
            i1Maj = int(lst1[0])
            i1Min = int(lst1[1])
            i1Patch = int(lst1[2])
            lst2 = sV2.split(".")
            i2Maj = int(lst2[0])
            i2Min = int(lst2[1])
            i2Patch = int(lst2[2])
            if i1Maj > i2Maj:
                return True
            elif i1Maj == i2Maj:
                if i1Min > i2Min:
                    return True
                elif i1Min == i2Min:
                    if i1Patch > i2Patch:
                        return True
                    elif i1Patch == i2Patch:
                        return False
            return False

        # ---

        pn = self.ui_getSmartprinter().normal
        progress = self.ui_getSmartprinter().progress
        workflow = self.ui_getSmartprinter().workflow

        progress("on-pulse")
        workflow(top="Update", sub="checking online")

        pn("Connecting to server: "+self.__sUpdateServer)

        res = Result(name="Checking for available update")

        # Daten vom Server holen: Version, Changelog
        sVersionServer = cleanWhiteSpaces(getInfoFromServer(self.__sUpdateServer +"/" + self.__sAppVersionFile))
        sChangeLogServer = getInfoFromServer(self.__sUpdateServer +"/" + self.__sAppChangelogFile)

        # Server nicht erreichbar
        if None in [sVersionServer, sChangeLogServer]:
            res.addErr("Server is unreachable. Are you connected to the AWI intranet?")
            return res

        # Installierte Version holen
        #sVersionInstalled = cleanWhiteSpaces("3.2.0-1") # TODO: Tweak: Immer updaten
        sVersionInstalled = cleanWhiteSpaces(self.__sAppVersion)
        # TODO: DEBUG
        #sVersionInstalled = "2.4.1-12"

        # Ergebnis: Versionen
        res.addData("version-installed", sVersionInstalled)
        res.addData("version-server", sVersionServer)

        # Teste ob Kompatabilität gebrochen wird
        bIncompatibleWarning = False
        sIncompatibleWarningVersion= None
        try:
            sVersionInstalled_nobuild = sVersionInstalled.split("-")[0]
            lstRows = sChangeLogServer.split(sep="\n")
            lstVersions = []
            for sRow in lstRows:
                hit = re.search('^\[\d*\.\d*\.\d*]!', sRow)
                if hit is not None:
                    sVersion = re.sub(r'[\[\]!]', "", hit.group())
                    lstVersions.append(sVersion)
            for sV in lstVersions:
                if isVersion1Greater2(sV, sVersionInstalled_nobuild):
                    bIncompatibleWarning = True
                    sIncompatibleWarningVersion = sV
                    break
        except:
            pass
        res.addData("update-incomp", bIncompatibleWarning)
        res.addData("update-incomp-version", sIncompatibleWarningVersion)

        # Zerlege Versionsstrings in Einzelteile
        iMajInst, iMinInst, iPatchInst, iBuildInst = evalVersion(sVersionInstalled)
        iMajServer, iMinServer, iPatchServer, iBuildServer = evalVersion(sVersionServer)

        # Check: Fehler in Version Strings
        if None in [iMajInst, iMinInst, iPatchInst, iBuildInst]:
            res.addErr("Internal error: installed version string is corrupt")
            return res
        if None in [iMajServer, iMinServer, iPatchServer, iBuildServer]:
            res.addErr("Internal error: server version string is corrupt")
            return res

        # Check Update Level (Lokal ist neuer, Update möglich, Gleich)
        bNewer = False
        bUpdate = False
        sUpdatelevel = None
            # Major
        if iMajServer < iMajInst:
            bNewer = True
        elif iMajServer > iMajInst:
            bUpdate = True
            sUpdatelevel = "major"
            # Minor
        elif iMinServer < iMinInst:
            bNewer = True
        elif iMinServer > iMinInst:
            bUpdate = True
            sUpdatelevel = "minor"
            # Patch
        elif iPatchServer < iPatchInst:
            bNewer = True
        elif iPatchServer > iPatchInst:
            bUpdate = True
            sUpdatelevel = "patch"
            # Build
        elif iBuildServer < iBuildInst:
            bNewer = True
        elif iBuildServer > iBuildInst:
            bUpdate = True
            sUpdatelevel = "build"
        if bNewer:
            # Lokal > Server
            res.addData("update-info", "That's strange. The installed version is newer than the one on the server. Are you a developer?\n\nVersion installed:  "+sVersionInstalled+"\nVersion on server: "+sVersionServer)
        elif bUpdate:
            # Server > Lokal
            res.addData("update-info", "There is a newer version available on the server.\n\nVersion installed:  "+sVersionInstalled+"\nVersion on server: "+sVersionServer)
        else:
            # Server == Lokal
            res.addData("update-info", "Version installed:  "+sVersionInstalled+"\nVersion on server: "+sVersionServer)

        # Update Level
        res.addData("update", bUpdate)
        if bUpdate:
            res.addData("update-level", sUpdatelevel)

        # Changelog
        if bUpdate:
            res.addData("changelog-server", sChangeLogServer)

        return res

    def help_updateDownload(self):
        """
        Starte App download
        """

        # Helper
        # Helper: Hole AppZip von Updateserver
        def getAppZip(sUrl, sFilename):
            res = Result(name="Download App")
            try:
                r = requests.get(sUrl, allow_redirects=True, timeout=1, auth=('critter', 'P0lych4eta')) # TWEAK für Preview-Version
                if r.status_code == 200 and r.headers['Content-Type'] == 'application/zip':
                    open(sFilename, 'wb').write(r.content)
                    res.addData("length",int(r.headers['Content-Length']))
                else:
                    res.addErr("Download failed: Server reachable but data not available", sys="Update", kat="Download")
            except Exception as ex:
                res.addErr("Download failed: " + str(ex))
            return res

        # ---

        res = Result(name="Downloading update")

        pn = self.ui_getSmartprinter().normal
        pv = self.ui_getSmartprinter().verbose
        pn = self.ui_getSmartprinter().normal
        progress = self.ui_getSmartprinter().progress
        workflow = self.ui_getSmartprinter().workflow

        progress("on-pulse")
        workflow(top="Update", sub="downloading update")

        # App directory neu erzeugen
        sUpdateFolder = self.conf_getAppUpdateDir() + "/update_" + PrettyText.createTimestamp()
        sUpdateFolder_app = sUpdateFolder+"/app"
        try:
            os.mkdir(sUpdateFolder)
            os.mkdir(sUpdateFolder_app)
        except Exception as ex:
            res.addErr("Internal error: Could not create update folder: " + str(ex))
            return res

        # *** Platform bestimmen
        sOs = self.conf_getOs()
        if sOs is None:
            res.addErr("Internal error: Unknown platform" + str(sOs))
            return res

        # *** Download
        pn("--- Downloading app from server")
        if sOs == "win":
            sSystem = "Windows"
        elif sOs == "lin":
            sSystem = "Linux"
        elif sOs == "max":
            sSystem = "MacOs"
        else:
            sSystem = "OS could not be determined"

        pn("Currently used system: " + sSystem)
        if sOs == "win":
            sUrl = self.__sUpdateAppWin
            sFile = self.__sUpdateAppWin
            pn("Downloading Windows version from server: " + sUrl)
        elif sOs == "mac":
            sUrl = self.__sUpdateAppMac
            sFile = self.__sUpdateAppMac
            pn("Downloading MacOs version from server: " + sUrl)
        elif sOs == "lin":
            sUrl = self.__sUpdateAppLinux
            sFile = self.__sUpdateAppLinux
            pn("Downloading Linux version from server: " + sUrl)
        else:
            res.addErr("Internal error: Unknown platform" + str(sOs))
            return res

        # Download Pfad
        sUrl = self.__sUpdateServer + "/" + sUrl
        sFile = sUpdateFolder + "/" + sFile
        pn("Saving to folder: " + sFile)
        pn("Downloading ...")
        # Download App
        resDL = getAppZip(sUrl, sFile)
        if resDL.isErr():
            res.addResult(resDL, msg_only=True)
            return res
        pn("Downloaded " + str(resDL.getData("length")) + " bytes")

        # *** Testing Zip
        workflow(top="Update", sub="prepare installation")
        pn("--- Checking integrity of downloaded zip file")
        try:
            with zipfile.ZipFile(sFile, 'r') as zip_ref:
                zip_ref.testzip()
        except Exception as ex:
            res.addErr("Error while testing zip integrity: "+str(ex))
            return res
        pn("Zip file is ok")

        # *** Unpacking zip
        pn("--- Unpacking zip file")
        pn("Unpacking to folder: " + sUpdateFolder_app)
        # Enpacke zip file mit Ausführungsrechten (Tweak, da die Standardfunktion das nicht kann)
        try:
            with zipfile.ZipFile(sFile, 'r') as unZipFile:
                for info in unZipFile.infolist():
                    real_path = unZipFile.extract(info, sUpdateFolder_app)
                    # Rechte setzen
                    unix_attributes = info.external_attr >> 16
                    target = os.path.join(sUpdateFolder_app, info.filename)
                    pv(str(target))
                    if unix_attributes:
                        os.chmod(target, unix_attributes)
                    if not real_path:
                        raise NameError(info.filename)
        except Exception as ex:
            res.addErr("Error while unpacking zip file: "+str(ex))
            return res
        pn("Zip file has been successfully unpacked")

        # *** Preparing for installation

        sInstallSystem = None
        sInstallSource = None
        sInstallTarget = None
        sRestart = None

        pn("* Preparing for installation")
        pn("--- Testing local installation path")
        pn("Appbase: " + str(self.__sAppBase))

        # Teste auf gefährlichen Installationspfad
        lstDangerPathParts = ["/git/", "git/", "git/", "~git", "~git"]
        for sDangerPart in lstDangerPathParts:
            if sDangerPart in self.__sAppBase:
                pass
                res.addErr("App base directory is not suitable for an update: "+ self.__sAppBase+" - Dangerours part found: " + sDangerPart)
                return res
        pn("ok")
        if sOs == "win":
            # ------------- Windows
            pn("--- Testing downloaded Windows app")
            sTestPath = sUpdateFolder_app+"/critterbase.py"
            pn("Testing file: " + sTestPath)
            bTest = FileTools.isPathFile(sTestPath)
            if not bTest:
                res.addErr("The downloaded app seems not to be a valid")
                return res
            pn("ok")
            sInstallSystem = "Windows"
            sInstallSource = sUpdateFolder_app
            sInstallTarget = self.__sAppBase
            sRestart = self.__sAppBase + "/start.pyw"
        elif sOs == "mac":
            # -------------- Mac
            pn("- Testing if local installation is a valid Mac Application")
            try:
                pn("Appbase: " + str(self.__sAppBase))
                pn("Testing for parent directory structure: ../Contents/MacOS/")
                sFront, sParent = FileTools.cutAtPathTail(self.__sAppBase)
                pn("Parent: " + str(sParent))
                sMacContainerBase, sGrandparent = FileTools.cutAtPathTail(sFront)
                pn("Grandparent: " +str(sGrandparent))
                if sParent != "MacOS" or sGrandparent != "Contents":
                    raise NameError("Uups")
                pn("ok")
                sRestart = sMacContainerBase
            except:
                res.addErr("The running app seems not to be a valid MacOs Application bundle (.app).")
                return res
            pn("--- Testing if downloaded Mac app is a valid Mac Application")
            sTestPath = sUpdateFolder_app+"/CRITTERBASE.app/Contents/MacOS/critterbase.py"
            pn("Testing file: "+sTestPath)
            bTest = FileTools.isPathFile(sTestPath)
            if not bTest:
                res.addErr("The downloaded app seems not to be a valid MacOs Application bundle (.app).")
                return res
            pn("ok")
            sInstallSystem = "MacOs"
            sInstallSource = sUpdateFolder_app+"/CRITTERBASE.app/Contents/MacOS"
            sInstallTarget = self.__sAppBase

        elif sOs == "lin":
            # -------------- Linux
            pn("--- Testing downloaded Linux app")
            sTestPath = sUpdateFolder_app+"/critterbase.py"
            pn("Testing file: " + sTestPath)
            bTest = FileTools.isPathFile(sTestPath)
            if not bTest:
                res.addErr("The downloaded app seems not to be a valid Linux program")
                return res
            pn("ok")
            sInstallSystem = "Linux"
            sInstallSource = sUpdateFolder_app
            sInstallTarget = self.__sAppBase
            sRestart = self.__sAppBase + "/critterbase.py"
        else:
            # --------------- Fehler
            res.addErr("Error: Wrong system found: " + str(sOs))
            return res
        # --- Installation kann beginnen

        # Quelle und Ziel sind korrekt ermittelt
        res.addData("sys", sInstallSystem)
        res.addData("src", sInstallSource)
        res.addData("target", sInstallTarget)
        res.addData("restart", sRestart)

        return res


