"""
Modul mit speziellen Containern für Thread Funktionalitäten
"""

# coding=utf-8
import psycopg2, csv, os, folium, traceback, subprocess, stat, zipfile
from pathlib import Path
from shutil import copyfile
from folium.plugins import MarkerCluster
from tools.coding import Result, MultiSheet
from tools.parallel import Worker
from tools.csvfiles import CsvTools
from tools.text import PrettyText
from tools.fileio import FileTools
from tools.ui import ColorGradient, Colors
from logik.batchbot import BatchBot
from logik.version_control import VersionControl
from collections import OrderedDict


# ---------------------------------------------------------
# View
# ---------------------------------------------------------

class Worker_AdminCmd(Worker):
    """
    Starte Admin Kommando
    """

    def __init__(self, model, dicCmd):
        super().__init__()
        self.__model = model
        self.__dicCmd = dicCmd
        self.__smartprinter = self.__model.ui_getSmartprinter()

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal

        # Helper: Eval Dictionary
        def getData(dic, sVal):
            if sVal in dic:
                return dic[sVal]
            else:
                return None

        # Werte Werte des Kommando aus
        sTitle = getData(self.__dicCmd, "title")
        sInfo = getData(self.__dicCmd,"info")
        sCmd = getData(self.__dicCmd,"cmd")
        sSqlScript = getData(self.__dicCmd,"sqlscript")
        sFile = getData(self.__dicCmd,"file")
        bPopup = getData(self.__dicCmd,"popup")
        sHost = getData(self.__dicCmd,"db_host")
        sPort = getData(self.__dicCmd,"db_port")
        sUser = getData(self.__dicCmd,"db_user")
        sPw = getData(self.__dicCmd,"db_pw")
        sDb = getData(self.__dicCmd,"db_name")
        sSchema = getData(self.__dicCmd,"db_schema")
        sSchemaBackupFile = getData(self.__dicCmd,"db_schema_backup_file")
        sSchemaRestoreFile = getData(self.__dicCmd,"db_schema_restore_file")
        sSchemaRenameName = getData(self.__dicCmd,"db_schema_rename_name")
        sSchemaNew = getData(self.__dicCmd,"db_schema_new")

        # Teste auf Existenz der Basisdaten
        if None in [sTitle,sInfo,sCmd,bPopup,sHost,sPort,sUser,sPw, sDb]:
            self.sendResult(Result(name="Internal error: Db command admin tool", err="Missing basic parameter"))
            return
        # Result
        result = Result(name=sTitle)

        # Teste ob Dateien ent-/gezippt werden sollen?
        # Backup
        if sCmd == "backup_schema" and FileTools.getFileType(sSchemaBackupFile) == "zip":
            bSchemaBackupFileZip = True
            sSchemaBackupFileZip = sSchemaBackupFile
            sSchemaBackupFile = FileTools.getFileWithoutType(sSchemaBackupFileZip)
            resCheckFileExists = FileTools.prepareSaveFile(sSchemaBackupFile)
            if resCheckFileExists.getData("bak") is not None:
                pn("* Backup existing file")
                pn(str(sSchemaBackupFile) +" -> " + str(resCheckFileExists.getData("bak")))
            if sSchemaBackupFile in ["", None] or sSchemaBackupFileZip in ["", None]:
                result.addErr("Problems zipping file", add=str(sSchemaBackupFile) + " -> "+ str(sSchemaBackupFileZip), kat="Zip", sys="Admin")
        else:
            bSchemaBackupFileZip = False
            sSchemaBackupFileZip = None
        # Restore Schema
        if sCmd == "restore_schema" and FileTools.getFileType(sSchemaRestoreFile) == "zip":
            bSchemaRestoreFileZip = True
            sSchemaRestoreFileZip = sSchemaRestoreFile
            sSchemaRestoreFile = FileTools.getFileWithoutType(sSchemaRestoreFileZip)
            sSchemaRestoreFileName = FileTools.getFileWithoutPath(sSchemaRestoreFile)
            sSchemaRestoreFilePath = FileTools.getFilePath(sSchemaRestoreFile)
            resCheckFileExists = FileTools.prepareSaveFile(sSchemaRestoreFile)
            if resCheckFileExists.getData("bak") is not None:
                pn("* Backup existing file")
                pn(str(sSchemaBackupFile) + " -> " + str(resCheckFileExists.getData("bak")))
            if sSchemaRestoreFile in ["", None] or sSchemaRestoreFileZip in ["", None]:
                result.addErr("Problems unzipping file", add = str(sSchemaRestoreFileZip), kat="Zip", sys="Admin")
            # Unzip
            workflow(top="Unzipping file", sub="Working", fmt=False)
            pn("* Unzipping file")
            pn(str(sSchemaRestoreFileZip))
            pn("->")
            pn(str(sSchemaRestoreFile))
            lstZipContent = []
            try:
                lstZipContent = zipfile.ZipFile(sSchemaRestoreFileZip).namelist()
                zipfile.ZipFile(sSchemaRestoreFileZip).extract(sSchemaRestoreFileName, path=sSchemaRestoreFilePath)
                pn("done")
            except Exception as ex:
                result.addErr("File not found in zip package", add= str(sSchemaRestoreFileName) + " -> found only this files: " + PrettyText.lst2strC(lstZipContent), kat="Zip", sys="Admin")
                pn("error")
        else:
            bSchemaRestoreFileZip = False
            sSchemaRestoreFileZip = None
            pass

        # Info
        progress("on-pulse-abort")
        workflow(top="Command", sub=str(sTitle), fmt=False)

        # Consolenausgabe
        sOut = sInfo + "\n"

        # Path zu Tools
        sPsqlBin = "psql" if self.__model.conf_getAdminPsqlPath() in [None, ""] else self.__model.conf_getAdminPsqlPath()
        sPgdumpBin = "pg_dump" if self.__model.conf_getAdminPgdumpPath() in [None, ""] else self.__model.conf_getAdminPgdumpPath()

        # OS
        sOs = self.__model.conf_getOs()

        # Environment -> aktuelles Postgres-User-Passwort hinzufuegen
        dicEnv = os.environ
        dicEnv["PGPASSWORD"] = sPw

        # --- Arbeit

        # Login String
        sPgConsoleLogin = "-U " + sUser + " -d " + sDb + " -h " + sHost + " -p " + sPort

        if sCmd == "start_psql":

            # *** Psql Console losgelöst starten
            try:
                if sOs == "win":
                    # Win
                    sPgConsoleCmd = 'start cmd /c "'+sPsqlBin+'" '+sPgConsoleLogin
                    sOut += sPgConsoleCmd + "\n"
                    p = subprocess.Popen(sPgConsoleCmd, shell=True, env=dicEnv) # TODO: Shell True entfernen
                elif sOs == "lin":
                    # Linux
                    sPgConsoleCmd = 'gnome-terminal -e "' + sPsqlBin + ' ' + sPgConsoleLogin + '"'
                    sOut += sPgConsoleCmd + "\n"
                    p = subprocess.Popen(sPgConsoleCmd, shell=True, env=dicEnv) # TODO: Shell True entfernen
                elif sOs == "mac":
                    # Mac
                    # Starter file
                    sFilePsqlOpenMac = self.__model.conf_getAppTmpDir() + "/start_psql_mac.sh"
                    sFilePsqlOpenMacContent = 'PGPASSWORD="'+sPw+'" '+sPsqlBin+' -U '+sUser+' -h '+sHost+' -p '+sPort+' -d '+sDb
                    with open(sFilePsqlOpenMac, 'w') as out:
                        out.write(sFilePsqlOpenMacContent + '\n')
                    os.chmod(sFilePsqlOpenMac, os.stat(sFilePsqlOpenMac).st_mode | stat.S_IEXEC)  # Datei ausführbar machen
                    # Starte
                    sPgConsoleCmd = 'open -a Terminal ' + sFilePsqlOpenMac
                    sOut += sPgConsoleCmd + "\n"
                    p = subprocess.Popen(sPgConsoleCmd, shell=True, env=dicEnv) # TODO: Shell True entfernen
                else:
                    raise NameError("Could not determine OS")
            except Exception as ex:
                result.addErr("Problems starting psql terminal", add=str(ex), kat="Terminal", sys="Core")
        else:
            # *** Andere Befehle
            sPgConsoleCmd = sPsqlBin + " " + sPgConsoleLogin + " -c"
            lstPgConsoleCmd = sPgConsoleCmd.split()

            if sCmd == "list_databases":
                # List Database ---
                lstPgConsoleCmd.append("\l")
            elif sCmd == "list_schemas":
                # List Schema ---
                lstPgConsoleCmd.append("\dn")
            elif sCmd == "list_users":
                # List Users ---
                lstPgConsoleCmd.append("\du")
            elif sCmd == "list_users_marlin":
                # List Users ---
                lstPgConsoleCmd.append("SELECT * FROM critterbase_users ORDER BY login_name")
            elif sCmd == "sql_script":
                # Run external SQL Script ---
                # Kommando
                sCmdTmp = sPsqlBin + " " + sPgConsoleLogin + " -f " + sSqlScript
                lstPgConsoleCmd = sCmdTmp.split()
            elif sCmd == "restore_schema":
                # Restore Schema ---
                if None in [sSchema, sSchemaRestoreFile]:
                    self.sendResult(Result(name="Internal error: "+str(sTitle), err="Missing parameter"))
                    return
                # Kommando
                sCmdTmp = sPsqlBin + " " + sPgConsoleLogin + " -f " + sSchemaRestoreFile
                lstPgConsoleCmd = sCmdTmp.split()
            elif sCmd == "backup_schema":
                # Backup Schema ---
                if None in [sSchemaBackupFile]:
                    self.sendResult(Result(name="Internal error: "+str(sTitle), err="Missing parameter"))
                    return
                sCmdTmp = sPgdumpBin+" -U "+sUser+" -h "+sHost+ " -d "+sDb+" -n "+sSchema +" -O -v -f "+sSchemaBackupFile
                lstPgConsoleCmd = sCmdTmp.split()
            elif sCmd == "rename_schema":
                # Rename Schema ---
                if None in [sSchema, sSchemaRenameName]:
                    self.sendResult(Result(name="Internal error: "+str(sTitle), err="Missing parameter"))
                    return
                lstPgConsoleCmd.append("ALTER SCHEMA "+sSchema + " RENAME TO "+sSchemaRenameName)
            elif sCmd == "del_schema":
                # Delete Schema ---
                if None in [sSchema]:
                    self.sendResult(Result(name="Internal error: "+str(sTitle), err="Missing parameter"))
                    return
                lstPgConsoleCmd.append("DROP SCHEMA IF EXISTS " + sSchema + " CASCADE")
            elif sCmd == "new_schema":
                # New Schema ---
                sCmdTmp = sPsqlBin + " " + sPgConsoleLogin+" -f " + str(Path(self.__model.conf_getEmptyDatenmodellFile()))
                lstPgConsoleCmd = sCmdTmp.split()
            elif sCmd == "snapshot":
                # Snapshot ---
                if None in [sSchema]:
                    self.sendResult(Result(name="Internal error: "+str(sTitle), err="Missing parameter"))
                    return
                sHostFilename = sHost.replace(".", "_")
                sSchemaVersion = self.__model.subsystem_getDb().version_getVersionConnectedDatamodel().getData("version")
                if sSchemaVersion is not None:
                    sSchemaVersion = sSchemaVersion.replace(".", "-")
                    sSchemaVersion = "_v"+sSchemaVersion
                else:
                    sSchemaVersion = ""
                sSnapshotFileName = sHostFilename+"_"+sDb+"_"+sSchema+sSchemaVersion+"_"+PrettyText.createTimestamp()+".sql"
                sSnapshotFile = self.__model.conf_getAppSnapshotDirSchema() + "/" + sSnapshotFileName
                result.addData("snapshot_file", sSnapshotFileName+".zip")
                sCmdTmp = sPgdumpBin+" -U "+sUser+" -h "+sHost+ " -d "+sDb+" -n "+sSchema +" -O -v -f "+ sSnapshotFile
                lstPgConsoleCmd = sCmdTmp.split()
            else:
                sPgConsoleCmd = None

            # --- Kommando ausfuehren

            if sPgConsoleCmd not in [None, ""]:
                try:

                    # Windows-only: Startup-Info -> verändern, damit die Konsole verborgen bleibt (würde auch über shell=True funktionieren)
                    if self.__model.conf_getOs() == "win":
                        startup_info = subprocess.STARTUPINFO()
                        startup_info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                    else:
                        startup_info = None

                    sPgConsoleCmd = " ".join(lstPgConsoleCmd)
                    p = subprocess.Popen(lstPgConsoleCmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, startupinfo=startup_info, env=dicEnv)

                    # Warten mit Prozess: StdOut Realtime Ausgabe
                    sStdOut = ""
                    sStdErr = ""
                    iStdErr = 0
                    for byteLine in p.stdout:
                        sLine = str(byteLine.decode("utf-8", errors="ignore"))
                        pn(sLine, "")
                        if "error" in sLine.lower() or "fail" in sLine.lower():
                            sStdErr += sLine
                            iStdErr += 1
                        sStdOut += sLine
                        if self.isAbort():
                            result.addErr("Aborted by user", kat="Abort", sys="Core")
                            sOut +="Aborted by user ..."
                            p.terminate()
                    sOut += sPgConsoleCmd + "\n\n"
                    if sStdErr == "":
                        sOut += sStdOut
                    else:
                        sOut += "*** "+str(iStdErr)+" lines with possible errors detected (please check) ***\n\n"
                        sOut += sStdErr+"\n"
                        sOut += "*** Full output ***\n\n"
                        sOut += sStdOut

                    # Command error
                    if p.returncode not in [0, None]:
                        raise NameError("Error executing SQL command")

                    # New empty CRITTERBASE: Trage die aktuelle Version in Admintabelle ein
                    if sCmd == "new_schema":
                        try:
                            sConnStr = self.__model.subsystem_getDb().conn_makeConnStr(user=sUser, pw=sPw, host=sHost, port=sPort, db=sDb)
                            conn = psycopg2.connect(sConnStr)
                            cursor = conn.cursor()
                            sSqlCmd = "DELETE FROM cb_empty.admin WHERE key='version'"
                            cursor.execute(sSqlCmd)
                            sSqlCmd = "INSERT INTO cb_empty.admin (key, value) VALUES ('version', '" + self.__model.conf_getInfo("dbschema") + "')"
                            cursor.execute(sSqlCmd)
                            sSqlCmd = "ALTER SCHEMA cb_empty RENAME TO " + str(sSchemaNew)
                            cursor.execute(sSqlCmd)
                            conn.commit()
                            cursor.close()
                            conn.close()
                            sOut += "\nSetting version number of new empty critterbase to: " + str(self.__model.conf_getInfo("dbschema") + "\n")
                            sOut += "\nRenaming new empty critterbase to: " + str(sSchemaNew) + "\n"

                        except Exception as ex:
                            sErr = "\nThe new schema cb_empty is not usable, cause I could not set the version number.\nPlease delete it and try again.\nError message: " + str(ex)
                            raise NameError(sErr)
                            conn.rollback()
                            cursor.close()
                            conn.close()
                    # Snapshot: Snapshot zippen und Original löschen
                    if sCmd == "snapshot":
                        zipfile.ZipFile(sSnapshotFile+".zip", mode='w', compression=zipfile.ZIP_DEFLATED).write(
                            sSnapshotFile, sSnapshotFileName)
                        os.remove(sSnapshotFile)
                    # Backup: Backup zippen und Original löschen
                    if sCmd == "backup_schema" and bSchemaBackupFileZip:
                        workflow(top="Zipping file", sub="Working", fmt=False)
                        pn("* Zipping file")
                        pn(str(FileTools.getFileWithoutPath(sSchemaBackupFile)))
                        pn("->")
                        pn(str(FileTools.getFileWithoutPath(sSchemaBackupFileZip)))
                        zipfile.ZipFile(sSchemaBackupFileZip, mode='w', compression=zipfile.ZIP_DEFLATED).write(
                            sSchemaBackupFile, FileTools.getFileWithoutPath(sSchemaBackupFile))
                        pn("* Removing unzipped file")
                        pn(str(FileTools.getFileWithoutPath(sSchemaBackupFile)))
                        os.remove(sSchemaBackupFile)
                        pn("done")
                    # Restore: Die enpackte Datei loeschen
                    if sCmd == "restore_schema" and bSchemaRestoreFileZip:
                        pn("* Removing unzipped file")
                        pn(str(sSchemaRestoreFile))
                        os.remove(sSchemaRestoreFile)
                        pn("done")
                    # Alles ist fertig
                    sOut += "\nDone."

                except Exception as ex:
                    # Command Errors
                    result.addErr("Problems starting postgres admin command", add=str(ex), kat="Command", sys="Core")
        # Ergebnis
        result.addData("cmd", sCmd)
        result.addData("out", sOut)
        result.addData("title", sTitle)
        result.addData("info", sInfo)
        result.addData("popup", bPopup)
        self.sendResult(result)

class Worker_CheckUpdateApp(Worker):
    """
    Starte Update Check
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model

    def abort(self):
        pass

    def run(self):
        result = self.__model.help_updateCheck()
        self.sendResult(result)


class Worker_DownloadUpdate(Worker):
    """
    Download Update
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model

    def abort(self):
        pass

    def run(self):
        result = self.__model.help_updateDownload()
        self.sendResult(result)

class Worker_Ingest(Worker):
    """
    Import
    Args:
            model: Modell
            sFile (str): Excel-File
            lstIng (lst): Ingest Sheets
            lstLup (list(: Lookups
            sExcel (str): Excel-Import Gold/NoGold
            sMode (str): Db-Update-Mode
            sUpdate (str): Db-Update Gold/NoGold
    Returns:
            Result
    """

    def __init__(self, model):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

    def run(self):

        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        res = Result("Ingest")
        progress("on-pulse")

        sFile = self.getInit("file")
        sWf = self.getInit("wf")
        lstIng = self.getInit("ing")
        lstLup = self.getInit("lup")
        sExcel = self.getInit("lupExcel")
        sMode = self.getInit("lupMode")
        sUpdate = self.getInit("lupDb")
        sLupReuseReact = self.getInit("lupReuseReact")
        bTamer = self.getInit("tamer")
        bLupSingle = self.getInit("ingest_lup_single")
        bLupGoldRush= self.getInit("lupExcelGoldrush")
        bExt = self.getInit("extend")
        sExtReact = self.getInit("extend_react")

        res = self.__model.subsystem_getImporter().startImport(sFile, ingest=True, lookup=True, ingest_sheets=lstIng, lookup_sheets=lstLup, ingest_wf=sWf, lookup_import_excel=sExcel, lookup_import_goldrush=bLupGoldRush, lookup_import_mode=sMode, lookup_import_db=sUpdate, lookup_reuse_react=sLupReuseReact, ingest_lup_search_single=bLupSingle, tamer=bTamer, extend=bExt, extend_react=sExtReact)

        self.sendResult(res)

class Worker_ImportLookup(Worker):
    """
    Import Lookup
    Args:
            model: Modell
            sFile (str): Excel-File
            bTamer (bool): Excel Tamer
            lstTabs (list(: Lookups
            sExcel (str): Excel-Import Gold/NoGold
            sMode (str): Db-Update-Mode
            sUpdate (str): Db-Update Gold/NoGold
    Returns:
            Result
    """

    def __init__(self, model, sFile, bTamer=False, lstTabs=None, sExcel=None, sMode=None, sUpdate=None, sReuseReact=None, bGoldRush=False):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

        self.__sFile = sFile
        self.__lstTabs = lstTabs
        self.__sExcel = sExcel
        self.__sMode = sMode
        self.__sUpdate = sUpdate
        self.__bTamer = bTamer
        self.__sReuseReact = sReuseReact
        self.__bGoldrush = bGoldRush

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        # Import
        res = self.__model.subsystem_getImporter().startImport(self.__sFile, lookup=True, lookup_sheets=self.__lstTabs, lookup_import_excel=self.__sExcel, lookup_import_goldrush=self.__bGoldrush, lookup_import_mode=self.__sMode, lookup_import_db=self.__sUpdate, lookup_reuse=self.__sReuseReact, tamer=self.__bTamer)

        self.sendResult(res)

class Worker_ExportLookup(Worker):
    """
    Exportiere Lookup aus System in Excel-Tabelle
    Args:
            model: Modell
            sFile: Excel-File
            lstIng: Nur diese Lookups (wenn leer -> Alle)
            lstSheets: Ingest Ids (wenn leer -> Alle )
    """

    def __init__(self, model, sFile, lstIng=None, lstSheets=None, sGold=None):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

        self.__lstIngests = lstIng
        self.__lstSheets = lstSheets
        self.__sGold = sGold
        self.__sFile = sFile

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        res = Result(name="Export lookups")

        # Export Lookup
        sLogLookup = ""
        try:
            progress("on-pulse")
            sFile = self.__sFile
            if sFile is None:
                NameError("No export file given")
            workflow(top="Export lookup", sub="ask db", steps=(1,2), fmt=False)
            if self.__lstIngests is not None:
                selector = {"ingest": self.__lstIngests}
            else:
                selector =  None
            res = self.__model.subsystem_getExporter().getLookup(sheets=self.__lstSheets, showid=False, gold=self.__sGold, selector=selector)
            if res.isErr():
                workflow()
                raise NameError(res.getResult())
            odicLup = res.getData("lookup")
            sLogLookup = res.getData("log")
            workflow(top="Export lookup", sub="write Excel", steps=(2,2))
            res = self.__model.subsystem_getExporter().writeExcel(lookup=True, lookup_data=odicLup, file=self.__sFile)
        except Exception as ex:
            res.addErr("Unexcepted Problem", add=str(ex), kat="Export", sys="Lookup")

        # Stats
        res.addData("log", sLogLookup)
        self.sendResult(res)

class Worker_ImportTaxon(Worker):
    """
    Import Taxon Admin Tables
    Args:
            model: Modell
            bTamer (bool): Excel Tamer
            sFile (str): Excel-File
            lstTabs (list(: Lookups

    Returns:
            Result
    """

    def __init__(self, model, sFile, bTamer=False, lstTabs=None, bUpdate=False):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

        self.__sFile = sFile
        self.__lstTabs = lstTabs
        self.__bUpdate = bUpdate
        self.__bTamer = bTamer

    def run(self):
        # Import
        res = self.__model.subsystem_getImporter().startImport(self.__sFile, taxon=True, taxon_sheets=self.__lstTabs, taxon_update=self.__bUpdate, tamer=self.__bTamer)
        self.sendResult(res)

class Worker_ExportTaxon(Worker):
    """
    Exportiere Taxon System
    Args:
            model: Modell
            sFile: Excel-File
            lstSheets: nur diese Sheets (wenn leer -> Alle )
    """

    def __init__(self, model, sFile, lstSheets=None):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

        self.__lstSheets = lstSheets
        self.__sFile = sFile

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        res = Result(name="Export TaxonWizard Data")

        # Export Taxon
        sLogTaxon = ""
        try:
            progress("on-pulse")
            sFile = self.__sFile
            if sFile is None:
                NameError("No export file given")
            workflow(top="Export TaxonWizard", sub="ask db", steps=(1,2))
            res = self.__model.subsystem_getExporter().getTaxon(sheets=self.__lstSheets, showid=False)
            if res.isErr():
                workflow()
                raise NameError(res.getResult())
            odicTaxon = res.getData("taxon")
            sLogTaxon = res.getData("log")
            workflow(top="Export TaxonWizard", sub="write Excel", steps=(2,2))
            res = self.__model.subsystem_getExporter().writeExcel(taxon=True, taxon_sheets=self.__lstSheets, taxon_data=odicTaxon, file=self.__sFile)
        except Exception as ex:
            res.addErr("Unexcepted Problem", add=str(ex), kat="Export", sys="Taxon")

        # Stats
        res.addData("log", sLogTaxon)
        self.sendResult(res)

class Worker_ExportTemplate(Worker):
    """
    Exportiere Template
    Args:
            model: Modell
            sFile: Excel-File
    kwargs:
            ingest (bool): Ingest
            lookup (bool): Lookup
            taxon (bool): Taxon
            wf (bool: Workflow

    Returns:
            Result:
    """

    def __init__(self, model, sFile, **kwargs):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__bIngest = kwargs["ingest"] if ("ingest" in kwargs and kwargs["ingest"] in [True,False]) else False
        self.__bLookup = kwargs["lookup"] if ("lookup" in kwargs and kwargs["lookup"] in [True, False]) else False
        self.__bTaxon = kwargs["taxon"] if ("taxon" in kwargs and kwargs["taxon"] in [True, False]) else False
        self.__sWf = kwargs["wf"] if "wf" in kwargs else None
        self.__sFile = sFile

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        res = Result(name="Export template", sub="Working", fmt=False)
        # Export Template
        try:
            progress("on-pulse")
            workflow(top="Export Template")
            if self.__sFile is None:
                NameError("No export file given")
            res = self.__model.subsystem_getExporter().writeExcel(file=self.__sFile, lookup=self.__bLookup, taxon=self.__bTaxon, ingest=self.__bIngest, ingest_wf=self.__sWf)
        except Exception as ex:
            res.addErr("Unexcepted Problem", add=str(ex), kat="Export", sys="Template")
        self.sendResult(res)

class Worker_BatchBot(Worker):
    """
    BatchBot
    Args:
            model: Modell
            sFile: Excel-File
            sWf: Workflow
    Returns:
            Result:
    """

    def __init__(self, model, sName, sLogDir):
        super().__init__()
        self.__model = model
        self.__smartprinter = self.__model.ui_getSmartprinter()
        self.__map = self.__model.subsystem_getMapper()
        self.__sName = sName
        self.__sLogDir = sLogDir
        self.__batchbot = self.__model.subsystem_getBatchBot()

    def abort(self, bBool=True):
        if self.__batchbot is not None:
            self.__batchbot.abort(bBool)

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        res = Result(name="BatchBot", sub="Working", fmt=False)
        progress("on-pulse")
        workflow(top="BatchBot")
        if self.__sName is None:
            NameError("No Batch Config given")
        if self.__sLogDir is None:
            NameError("No Log directory given")
        resBat = self.__batchbot.startBatch(self.__sName, self.__sLogDir)
        sLog = resBat.getData("log")
        lstFiles = resBat.getData("files")
        sDirLog = resBat.getData("dirlog")
        sTimeStamp = resBat.getData("timestamp")

        if resBat.isErr():
            res.addResult(resBat, msg_only=True)
        res.addData("log", sLog)
        res.addData("dirlog", sDirLog)
        res.addData("timestamp", sTimeStamp)
        res.addData("files", lstFiles)
        self.sendResult(res)

class Worker_ExportConverterTemplate(Worker):
    """
    Exportiere Template
    Args:
            model: Modell
            sFile: Excel-File
            sWf: Workflow
    Returns:
            Result:
    """

    def __init__(self, model, sFile, sWf):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__sWf = sWf
        self.__sFile = sFile

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        res = Result(name="Export COnverter Template")
        # Export Template
        try:
            progress("on-pulse")
            workflow(top="Converter", sub ="Export Template")
            if self.__sFile is None:
                NameError("No export file given")
            res = self.__model.subsystem_getConverter().write_converter_import_template(self.__sWf, self.__sFile)
        except Exception as ex:
            res.addErr("Unexcepted Problem", add=str(ex), kat="Export", sys="Converter Template")
        self.sendResult(res)

class Worker_ShowSubset(Worker):
    """
    Hole ein Subset und liefere es an die GUT zurück um es dort anzuzeigen in Tabellenform
    Args:
        model: Modell
        selector (selector): Selector
    Returns:
        Result:

    """

    def __init__(self, model, dicSelector, **kwargs):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__dicSelector = dicSelector
        self.__sConverter = kwargs.get("converter")
        self.__bRfriendly = kwargs.get("rfriendly")

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        progress("on-pulse")
        result = Result(name="Show Subset")
        try:
            workflow(top="Getting subset form database", sub="Working", fmt=False)
            res = self.__model.subsystem_getExporter().getSubset(selector=self.__dicSelector, converter=self.__sConverter, rfriendly=self.__bRfriendly)
            sWf = res.getData("wf")
            odicData = res.getData("data")
            ms = MultiSheet(name="Subset")
            for sSheet, dicSheet in odicData.items():
                if sSheet != "Selector":
                    lstHeader = dicSheet["header"]
                    lstContent = dicSheet["content"]
                    ms.setSheet(sSheet, lstHeader, lstContent)
        except Exception as ex:
            result.addErr("Error during getting data from database", add=str(ex), kat="Show Ingest Table", sys="Admin")
            progress("off")
            workflow()
            self.sendResult(result)

        # Antwort
        result.addData("diaTitle", "Subset")
        result.addData("multisheet", ms)
        progress("off")
        workflow()
        self.sendResult(result)


class Worker_ExportIngest(Worker):
    """
    Exportiere ein Ingest mit Lookups
    Args:
        model
        sFile (str)
        selector (dic)
    Kwargs:
        ingest_sheets (lst)
        lookup_sheets (lst)
        pangaea (bool)
        converter (str)
    Returns:
        Result:

    """

    def __init__(self, model, sFile, dicSelector, **kwargs):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()

        self.__sFile = sFile
        self.__dicSelector = dicSelector
        self.__lstSheetsIngest = kwargs.get("ingest_sheets")
        self.__lstSheetsLookup = kwargs.get("lookup_sheets")
        self.__sConverter = kwargs.get("converter")
        self.__bPangaea = kwargs.get("pangaea") if "pangaea" in kwargs else False
        self.__bRfriendly = kwargs.get("rfriendly") if "rfriendly" in kwargs else False

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        res = Result(name="Export lookups")

        progress("on-pulse")
        if self.__sFile is None:
            NameError("No export file given")
        bIngest = True if (self.__lstSheetsIngest is None or len(self.__lstSheetsIngest)>0) else False
        bLookup = True if (self.__lstSheetsLookup is None or len(self.__lstSheetsLookup) > 0) else False

        # Lookup
        workflow(top="Export subset", sub="getting lookup", steps=(1,3))
        if bLookup:
            res = self.__model.subsystem_getExporter().getLookup(selector=self.__dicSelector, sheets=self.__lstSheetsLookup)
            if res.isErr():
                self.sendResult(res)
            odicLup = res.getData("lookup")
            sLogLookup = res.getData("log")

        # Ingest
        workflow(top="Export subset", sub="getting ingest", steps=(2,3))
        if bIngest:
            res = self.__model.subsystem_getExporter().getSubset(selector=self.__dicSelector, sheets=self.__lstSheetsIngest, converter=self.__sConverter, rfriendly=self.__bRfriendly)
            if res.isErr():
                self.sendResult(res)
            odicIng = res.getData("data")
            sLogIng = res.getData("log")
            sWf = res.getData("wf")

        # Excel
        workflow(top="Export subset", sub="writing Excel", steps=(3,3))
        res = self.__model.subsystem_getExporter().writeExcel(ingest=bIngest, ingest_sheets=self.__lstSheetsIngest, ingest_data=odicIng, ingest_wf=sWf, pangaea=self.__bPangaea, rfriendly=self.__bRfriendly, lookup_data=odicLup, lookup=bLookup, lookup_sheets=self.__lstSheetsLookup, file=self.__sFile, converter=self.__sConverter)
        # Log
        sLog = sLogIng + sLogLookup
        res.addData("log", sLog)
        self.sendResult(res)

class Worker_VersionControl(Worker):
    """
    Starte Version control
    """

    def __init__(self, smartprinter, sIn1, sIn2, sOut):
        super().__init__()

        self.__smartprinter = smartprinter
        self.__sIn1 = sIn1
        self.__sIn2 = sIn2
        self.__sOut = sOut
        self.__vc = VersionControl(self.__smartprinter)

    def abort(self):
        self.__vc.abort()

    def run(self):
        result = self.__vc.diff(self.__sIn1, self.__sIn2, self.__sOut)
        self.sendResult(result)

class Worker_VirtualAssistant_taxaReport(Worker):
    """
    Virtual Assistant: TaxonSpotter
    """
    def __init__(self, model, dicParam):
        super().__init__()
        self.__taxaReport = model.subsystem_getVirtualAssistant().getTaxonSpotter()
        self.__dicParam = dicParam

    def abort(self):
        self.__taxaReport.abort()

    def run(self):
        result = self.__taxaReport.start_report(**self.__dicParam)
        self.sendResult(result)

class Worker_VirtualAssistant_polishTaxa(Worker):
    """
    Virtual Assistant: Polishing Taxa
    """
    def __init__(self, model, sFileOriginal, sFilePolished, sSheetBiota, sColAid, sColSn, bLike, bFuzzy, bMarineOnly, bBuffer):
        super().__init__()
        self.__taxaPolishing = model.subsystem_getVirtualAssistant().getTaxaPolishing()
        self.__sFileOriginal = sFileOriginal
        self.__sFilePolished = sFilePolished
        self.__sSheetBiota = sSheetBiota
        self.__sColAid = sColAid
        self.__sColSn = sColSn
        self.__bLike = bLike
        self.__bFuzzy = bFuzzy
        self.__bMarineOnly = bMarineOnly
        self.__bBuffer = bBuffer

    def abort(self):
        self.__taxaPolishing.abort()

    def run(self):
        result = self.__taxaPolishing.start_polishing(self.__sFileOriginal, file_polished=self.__sFilePolished, sheet_biota=self.__sSheetBiota, col_aid=self.__sColAid, col_scientificname=self.__sColSn, marine_only=self.__bMarineOnly, like=self.__bLike, fuzzy=self.__bFuzzy, puffer=self.__bBuffer)
        self.sendResult(result)

class Worker_Fao(Worker):
    """
    FAO: Update
    """
    def __init__(self, model, sCmd, **kwargs):
        super().__init__()
        self.__model = model
        self.__sCmd = sCmd
        self.__kwargs = kwargs

    def run(self):
        res = Result()
        reg = self.__model.subsystem_getRegions()
        if self.__sCmd == "reg_update":
            # Map Update
            res = reg.map_update()
        elif self.__sCmd == "reg_clean":
            # Clean Mapoing
            res = reg.map_clean()
        elif self.__sCmd == "reg_stat":
            # Map Stat
            iSample_all = reg.map_getSamples_all_nr()
            sInfo = reg.info_map()
            if sInfo in [None, ""]:
                sInfo = "Region mapping has not been executed yet ..."
            res.addData("title", "Info on Regions")
            res.addData("text", sInfo)
        elif self.__sCmd == "reg_create":
            # (Re)Create FAO regions
            res = reg.regions_create(**self.__kwargs)
        elif self.__sCmd == "import":
            # Import FAO regions from file
            sFile = self.__kwargs.get("file")
            res = reg.regions_import(sFile)
        elif self.__sCmd == "export":
            # Export FAO regions from file
            sFile = self.__kwargs.get("file")
            res = reg.regions_export(sFile)
        else:
            res.addErr("Unknown command:", add=self.__sCmd, kat="Internal", sys="FAO")
        self.sendResult(res)

class Worker_DataPatrol_biotaControl(Worker):
    """
    DataPatrol
    """
    def __init__(self, dataPatrol):
        super().__init__()
        self.__dataPatrol = dataPatrol

    def abort(self):
        self.__dataPatrol.abort()

    def run(self):
        result = self.__dataPatrol.start_biotaControl()
        self.sendResult(result)

class Worker_MyWorms_searchManualByAphiaId(Worker):
    """
    MyWoRMS: Search Taxa
    """
    def __init__(self, model, iAid, bPuffer):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__iAid = iAid
        self.__bPuffer = bPuffer

    def abort(self):
        self.__model.subsystem_getMyWorms().abort()

    def run(self):
        res = Result("Search manually in MyWorms by AphiaId")
        status = self.__smartprinter.status
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        progress("on-pulse-abort")
        workflow(top="Asking MyWoRMS by AphiaId", sub="Working", fmt=False)
        resMyworms = self.__model.subsystem_getMyWorms().getTaxon(self.__iAid, worms=True, puffer=self.__bPuffer)
        if resMyworms.isErr():
            # Fehler/Abort
            sStatus = "Search aborted"
            status(sStatus, time=5000, fmt=False)
            res.addData("dont_show", True)
        elif resMyworms.getData("found") is True:
            # Gefunden
            t = resMyworms.getData("taxon")
            bAaskedPuffer = resMyworms.getData("asked_puffer")
            bPufferToOld = resMyworms.getData("puffer_age_too_old")
            bPufferWasDeaged = resMyworms.getData("puffer_de_aged")
            bAskedWorms = resMyworms.getData("asked_worms")
            sTaxon = t.info()
            sTaxon += "\n"
            if bAskedWorms:
                sStatus = "Taxon not found in MyWoRMS -> asked WoRMS"
            else:
                sStatus = "Taxon found in MyWorMS"
            status(sStatus, time=10000, fmt=False)
            # Antwort
            sInfo = sTaxon
            res = Result()
            res.addData("title", "Search by AphiaId: " + str(self.__iAid))
            res.addData("text", sInfo)
            res.addData("title_capitalize", False)
        else:
            # Nicht gefunden
            sStatus = "Taxon not found"
            status(sStatus, time=10000, fmt=False)
            res.addData("dont_show", True)
        self.sendResult(res)

class Worker_MyWorms_searchManualByName(Worker):
    """
    MyWoRMS: Search Taxa
    """
    def __init__(self, model, sName, bFuzzy, bLike, bPuffer):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__sName = sName
        self.__bPuffer = bPuffer
        self.__bLike = bLike
        self.__bFuzzy = bFuzzy

    def abort(self):
        self.__model.subsystem_getMyWorms().abort()

    def run(self):
        res = Result("Search manually in MyWorms by Name")
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        status = self.__smartprinter.status
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        progress("on-pulse-abort")
        workflow(top="MyWoRMS ", sub="Asking by Name", fmt=False)

        resMyworms = self.__model.subsystem_getMyWorms().queryByName(self.__sName, like=self.__bLike, fuzzy=self.__bFuzzy, puffer=self.__bPuffer)
        if resMyworms.isErr():
            # Fehler/Abort
            sStatus = "Query aborted"
            status(sStatus, time=5000, fmt=False)
            res.addData("dont_show", True)
        elif resMyworms.getData("found"):
            # Gefunden
            lstTaxa = resMyworms.getData("taxa")
            # Protokol
            bFound = resMyworms.getData("found")
            bFallbackUsed = resMyworms.getData("fallback_used")
            bFoundInPuffer = resMyworms.getData("found_in_puffer")
            lstTaxa_discardedUnaccepted = resMyworms.getData("taxa_discarded_unaccepted")
            lstTaxa_addedAcceptedTraced = resMyworms.getData("taxa_added_accepted_traced")
            lstTaxa_wormsOriginal = resMyworms.getData("taxa_worms_original")
            sProtocol = ""
            if bFallbackUsed:
                sProtocol += "--- Fallback (original unaccepted taxa by WoRMS) used to prevent empty result\n"
            if len(lstTaxa_discardedUnaccepted) > 0 or len(lstTaxa_addedAcceptedTraced) > 0:
                sProtocol += "--- Taxa found by WoRMS: " + str(len(lstTaxa_wormsOriginal)) + "x\n"
                for t in lstTaxa_wormsOriginal:
                    sProtocol += t.signature() +"\n"
            if len(lstTaxa_discardedUnaccepted) > 0:
                sProtocol += "--- Discarded unaccepted taxa: " + str(len(lstTaxa_discardedUnaccepted)) + "x\n"
                for t in lstTaxa_discardedUnaccepted:
                    sProtocol += t.signature() +"\n"
            if len(lstTaxa_addedAcceptedTraced) > 0:
                sProtocol += "--- Added traced accepted taxa: " + str(len(lstTaxa_addedAcceptedTraced)) + "x\n"
                for t in lstTaxa_addedAcceptedTraced:
                    sProtocol += t.signature() +"\n"
            if sProtocol != "":
                pv("* Query MyWoRMS by name: "+str(self.__sName), fmt=False)
                pv(sProtocol, fmt=False)
            # Info
            sTaxa = ""
            for t in lstTaxa:
                sTaxa += t.signature() + "\n"
            sInfo = sTaxa
            res = Result()
            res.addData("title", "Search by name: " + str(self.__sName)+ " ("+str(len(lstTaxa))+"x)")
            res.addData("text", sInfo)
            res.addData("title_capitalize", False)
            bFoundInBuffer = resMyworms.getData("found_in_puffer")
            if not bFoundInBuffer:
                sStatus = "Query not found in MyWoRMS -> asked WoRMS"
            else:
                sStatus = "Query found in MyWorMS"
            status(sStatus, time=10000, fmt=False)
        else:
            # Nicht gefunden
            sStatus = "Query not found"
            status(sStatus, time=10000, fmt=False)
            res.addData("dont_show", True)
        # Antwort
        self.sendResult(res)

class Worker_TaxonWizard_update(Worker):
    """
    Starte den Taxon Wizzard
    """

    def __init__(self, taxonSystem):
        super().__init__()
        self.__taxonSystem = taxonSystem

    def abort(self):
        self.__taxonSystem.abort()

    def run(self):
        result = self.__taxonSystem.update()
        self.sendResult(result)

class Worker_MyWorms_antiage(Worker):
    """
    Starte den Taxon Wizzard
    """

    def __init__(self, taxonSystem):
        super().__init__()
        self.__taxonSystem = taxonSystem

    def abort(self):
        self.__taxonSystem.abort()

    def run(self):
        result = self.__taxonSystem.puffer_update()
        self.sendResult(result)

class Worker_ShowingCsvFile(Worker):
    """
    Lies eine Csv-Datei ein und sende sie an die View (um sie dort anzuzeigen)
    """

    def __init__(self, smartprinter, sFile):
        super().__init__()

        self.__smartprinter = smartprinter
        self.__sFile = sFile

    def run(self):

        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        sFileOnly = os.path.basename(self.__sFile)

        result = Result(name="Showing file: "+sFileOnly)

        if self.__sFile is None:
            result.addErr("No Filename given", kat="Show Csv File", sys="Toolbox")
            self.sendResult(result)

        result.addData("tablename", "File: "+sFileOnly)
        progress("on-abort")
        workflow(sop="Reading csv file")

        # Datei einlesen
        lstHeader = list()
        lstContent = list()

        try:
            iZeilenGesamt = sum(1 for line in open(self.__sFile, newline='', encoding='utf-8'))
            with open(self.__sFile, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstHeader = csvReader.__next__()
                iZeile = 1
                for lstZeile in csvReader:
                    lstContent.append(lstZeile)
                    self.showProgress()
                    workflow(top="Building table", sub="assembling rows")
                    if self.isAbort():
                        result.addInfo("Aborted by user ...", kat="Show Csv File", sys="Toolbox")
                        self.sendResult(result)
                    iZeile += 1
        except Exception as ex:

            result.addErr("Error while reading file", add = str(ex), kat = "Show Csv File", sys = "Toolbox")

        result.addData("header", lstHeader)
        result.addData("content", lstContent)

        workflow()
        progress("off")

        self.sendResult(result)

class Worker_ExctractNewCsvData(Worker):
    """
    Extraktion von neuen Einträgen in eine Csv-Datei
    """

    def __init__(self, smartprinter, sHead, sIn1, sIn2, sOut):
        super().__init__()

        self.__smartprinter = smartprinter
        self.__sHead = sHead
        self.__sIn1 = sIn1
        self.__sIn2 = sIn2
        self.__sOut = sOut
        self.__csvTool = CsvTools()

    def abort(self):
        self.__csvTool.abort()

    def run(self):
        result = self.__csvTool.extractNewData(self.__smartprinter, self.__sHead, self.__sIn1, self.__sIn2, self.__sOut)
        self.sendResult(result)


class Worker_SqlCmd(Worker):
    """
    Stelle dem SQL-Server eine Anfrage und sende das Ergebnis an die View ()
    """

    def __init__(self, smartprinter, sConnection, sCmd, sTableName=None):
        super().__init__()
        self.__smartprinter = smartprinter
        self.__sTableName = sTableName
        self.__sCmd = sCmd
        self.__sConnection = sConnection


    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result()
        progress("on-pulse")
        workflow(top="Cmd", sub="Waiting for db")

        try:
            # Db connect
            conn = psycopg2.connect(self.__sConnection)
            cursor = conn.cursor()

            # Cmd
            cursor.execute(self.__sCmd)
            lstHeader = [head[0] for head in cursor.description]
            lstContent = cursor.fetchall()

            # Antwort
            sName = self.__sTableName if self.__sTableName is not None else self.__sCmd
            ms = MultiSheet(name=sName)
            ms.setSheet(sName, lstHeader, lstContent)
            result.addData("multisheet", ms)
            result.addData("save", True)
            result.addData("diaTitle", self.__sCmd)

            # Db disconnect
            conn.rollback()
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Sql error", add=str(ex), kat="Execute SQL command", sys="Core")

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_SaveDataToFile(Worker):
    """
    KLasse speichert Daten in eine Datei ab
    """

    def __init__(self, smartprinter, data, sFilename):
        super().__init__()
        self.__smartprinter = smartprinter
        self.__data = data
        self.__sFilename = sFilename

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        progress("on-pulse")
        result = Result()

        if type(self.__data) == list:
            # Csv (list)
            workflow(top="Saving table to file")
            result.setName("Saving all table data to a csv file")
            result.addData("file", str(self.__sFilename))
            try:
                with open(self.__sFilename, 'w', newline='', encoding='utf-8') as csvWriter:
                    writer = csv.writer(csvWriter, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
                    writer.writerows(self.__data)
                    sLen = "-"
                    try:
                        sLen = str(len(self.__data))
                    except:
                        pass
                    result.addInfo("Table saved to file", add=self.__sFilename, kat="I/O", sys="Admin")
            except Exception as ex:
                result.addErr("Error saving to file", add= self.__sFilename + ": " + str(ex), kat="I/O", sys="Admin")
        elif type(self.__data) == str:
            # Text (str)
            workflow(top="Saving text to file")
            result.setName("Saving text data to file")
            result.addData("file", str(self.__sFilename))
            try:
                with open(self.__sFilename, "w", newline="\r\n") as file:
                    file.write(self.__data)
            except Exception as ex:
                result.addErr("Error saving to file", add= str(self.__sFilename) + ": " + str(ex), kat="I/O", sys="Admin")

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_ShowIngest(Worker):
    """
    Laed eine Tabelle und sende den Inhalt an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model, iId):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__iId = iId

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        progress("on-pulse")
        result = Result(name="Show ingest " + str(self.__iId))
        sFilename = None
        try:
            workflow(top="Show Ingest", sub="Loading: "+str(self.__iId))
            dicSelect={"ingest": [self.__iId]}
            res = self.__model.subsystem_getExporter().getSubset(selector=dicSelect, showid=True, showingestid=True)
            sWf = res.getData("wf")
            odicData = res.getData("data")
            ms = MultiSheet(name="Ingest Table: "+str(self.__iId))
            for sSheet, dicSheet in odicData.items():
                if sSheet != "Selector":
                    lstHeader = dicSheet["header"]
                    lstContent = dicSheet["content"]
                    ms.setSheet(sSheet, lstHeader, lstContent)

            resFilename = self.__model.subsystem_getDb().ingest_getIngestFilenames(self.__iId)
            sFilename = resFilename.getData("file")
            # Antwort
            sTitle = str(sFilename) + " (" + str(self.__iId) + ")"
            result.addData("ingFilename", sFilename)
            result.addData("ingId", self.__iId)
            result.addData("diaTitle", sTitle)
            result.addData("multisheet", ms)
            result.addData("wf", sWf)
        except Exception as ex:
            result.addErr("Error during getting data from database" , add=str(ex), kat="Show Ingest Table", sys="Admin")
            progress("off")
            workflow()
            self.sendResult(result)

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_StatisticOnTable(Worker):
    """
    Zeige Statistiken zu Tabelle
    """

    def __init__(self, model, sTab):
        super().__init__()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__sTab = sTab
        if self.__model.subsystem_getMapper().all_isTableLookup(sTab):
            self.__sTitle = self.__map.lookup_table2sheet(self.__sTab)
        elif self.__model.subsystem_getMapper().all_isTableTaxon(sTab):
            self.__sTitle = self.__map.taxon_table2sheet(self.__sTab)
        else:
            self.__sTitle  = str(self.__sTab).title()

    def run(self):
        # --- Helper
        # Get AccessLevel of Dataset
        def getAccessLevel(iId):
            # Data
            sData = "?"
            if iId in setLookup_dataset_useddatafree_ids:
                sData = "free"
            elif iId in setLookup_dataset_useddataRequest_ids:
                sData = "request"
            elif iId in setLookup_dataset_useddataConfidential_ids:
                sData = "confidential"
            # Product
            sProduct = "?"
            if iId in setLookup_dataset_usedproductfree_ids:
                sProduct = "free"
            elif iId in setLookup_dataset_usedproductRequest_ids:
                sProduct = "request"
            elif iId in setLookup_dataset_usedproductConfidential_ids:
                sProduct = "confidential"
            return (sData, sProduct)

        # ---
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Retrieving statistics for table: "+str(self.__sTab))
        progress("on-pulse")
        workflow(top="Database", sub="Waiting for answer")
        bHtml = False
        bBrowser = False
        try:
            # DB-Verbindung
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # Status zu Lookup-Tabellen (ausser Taxon) holen
            if self.__model.subsystem_getMapper().all_isTableLookup(self.__sTab) and self.__sTab != "taxon":
                resDatasetStatus = self.__model.subsystem_getDb().lookup_getStatus(self.__sTab)
                result.addResult(resDatasetStatus, msg_only=True)
                iLookup_all = resDatasetStatus.getData("all")
                dicLookup_all = resDatasetStatus.getData("all-names")

                iLookup_gold = resDatasetStatus.getData("gold")
                setLookup_gold_ids = resDatasetStatus.getData("gold-ids")
                iLookup_gold_proz = resDatasetStatus.getData("gold-proz")

                iLookup_used = resDatasetStatus.getData("used")
                setLookup_used_id = resDatasetStatus.getData("used-ids")
                iLookup_used_proz = resDatasetStatus.getData("used-proz")

                iLookup_usedNot = resDatasetStatus.getData("not-used")
                setLookup_usedNot_ids = resDatasetStatus.getData("not-used-ids")
                iLookup_usedNot_proz = resDatasetStatus.getData("not-used-proz")

                iLookup_used_gold = resDatasetStatus.getData("used-gold")
                setLookup_used_gold_ids = resDatasetStatus.getData("used-gold-ids")
                iLookup_used_gold_proz = resDatasetStatus.getData("used-gold-proz")

                iLookup_used_goldNot = resDatasetStatus.getData("used-not-gold")
                setLookup_used_goldNot_ids = resDatasetStatus.getData("used-not-gold-ids")
                iLookup_used_goldNot_proz = resDatasetStatus.getData("used-not-gold-proz")

                iLookup_usedNot_goldNot = resDatasetStatus.getData("not-used-not-gold")
                setLookup_usedNot_goldNot_ids = resDatasetStatus.getData("not-used-not-gold-ids")
                iLookup_usedNot_goldNot_proz = resDatasetStatus.getData("not-used-not-gold-proz")

                iUse_ing_files = resDatasetStatus.getData("used-ing-sourcefiles")
                dicUse_ing_files = resDatasetStatus.getData("used-ing-sourcefiles-dic")
                iUse_lup_names = resDatasetStatus.getData("used-lup-sourcelup")
                dicUse_lup_names = resDatasetStatus.getData("used-lup-sourcelup-dic")

                # Special: Dataset
                dicLookup_dataset_files = resDatasetStatus.getData("dataset-files")
                iLookup_dataset_accessPossible = resDatasetStatus.getData("dataset-access-possible")
                iLookup_dataset_accessPossible_proz = resDatasetStatus.getData("dataset-access-possible-proz")
                iLookup_dataset_usedAccessPossible = resDatasetStatus.getData("dataset-used-access-possible")
                setLookup_dataset_usedAccessPossible_ids = resDatasetStatus.getData("dataset-used-access-possible-ids")
                iLookup_dataset_usedAccessPossible_proz = resDatasetStatus.getData("dataset-used-access-possible-proz")
                iLookup_dataset_usedNotGoldAccessPossible = resDatasetStatus.getData("dataset-used-not-gold-access-possible")
                iLookup_dataset_datafree = resDatasetStatus.getData("dataset-access-data-free")
                iLookup_dataset_datafree_proz = resDatasetStatus.getData("dataset-access-data-free-proz")
                iLookup_dataset_dataRequest = resDatasetStatus.getData("dataset-access-data-request")
                iLookup_dataset_dataRequest_proz = resDatasetStatus.getData("dataset-access-data-request-proz")
                iLookup_dataset_dataConfidential = resDatasetStatus.getData("dataset-access-data-confidential")
                iLookup_dataset_dataConfidential_proz = resDatasetStatus.getData("dataset-access-data-confidential-proz")
                iLookup_dataset_productfree = resDatasetStatus.getData("dataset-access-product-free")
                iLookup_dataset_productfree_proz = resDatasetStatus.getData("dataset-access-product-free-proz")
                iLookup_dataset_productRequest = resDatasetStatus.getData("dataset-access-product-request")
                iLookup_dataset_productRequest_proz = resDatasetStatus.getData("dataset-access-product-request-proz")
                iLookup_dataset_productConfidential = resDatasetStatus.getData("dataset-access-product-confidential")
                iLookup_dataset_productConfidential_proz = resDatasetStatus.getData("dataset-access-product-confidential-proz")
                
                iLookup_dataset_useddatafree = resDatasetStatus.getData("dataset-used-access-data-free")
                setLookup_dataset_useddatafree_ids = resDatasetStatus.getData("dataset-used-access-data-free-ids")
                iLookup_dataset_useddatafree_proz = resDatasetStatus.getData("dataset-used-access-data-free-proz")
                iLookup_dataset_useddataRequest = resDatasetStatus.getData("dataset-used-access-data-request")
                setLookup_dataset_useddataRequest_ids = resDatasetStatus.getData("dataset-used-access-data-request-ids")
                iLookup_dataset_useddataRequest_proz = resDatasetStatus.getData("dataset-used-access-data-request-proz")
                iLookup_dataset_useddataConfidential = resDatasetStatus.getData("dataset-used-access-data-confidential")
                setLookup_dataset_useddataConfidential_ids = resDatasetStatus.getData("dataset-used-access-data-confidential-ids")
                iLookup_dataset_useddataConfidential_proz = resDatasetStatus.getData("dataset-used-access-data-confidential-proz")
                iLookup_dataset_usedproductfree = resDatasetStatus.getData("dataset-used-access-product-free")
                setLookup_dataset_usedproductfree_ids = resDatasetStatus.getData("dataset-used-access-product-free-ids")
                iLookup_dataset_usedproductfree_proz = resDatasetStatus.getData("dataset-used-access-product-free-proz")
                iLookup_dataset_usedproductRequest = resDatasetStatus.getData("dataset-used-access-product-request")
                setLookup_dataset_usedproductRequest_ids = resDatasetStatus.getData("dataset-used-access-product-request-ids")
                iLookup_dataset_usedproductRequest_proz = resDatasetStatus.getData("dataset-used-access-product-request-proz")
                iLookup_dataset_usedproductConfidential = resDatasetStatus.getData("dataset-used-access-product-confidential")
                setLookup_dataset_usedproductConfidential_ids = resDatasetStatus.getData("dataset-used-access-product-confidential-ids")
                iLookup_dataset_usedproductConfidential_proz = resDatasetStatus.getData("dataset-used-access-product-confidential-proz")
                
                iLookup_dataset_usedAccessPossible = resDatasetStatus.getData("dataset-used-access-possible")
                setLookup_dataset_usedAccessPossible_ids = resDatasetStatus.getData("dataset-used-access-possible-ids")
                iLookup_dataset_usedAccessPossible_proz = resDatasetStatus.getData("dataset-used-access-possible-proz")
                iLookup_dataset_usedAccessRestricted = resDatasetStatus.getData("dataset-used-access-restricted")
                setLookup_dataset_usedAccessRestricted_ids = resDatasetStatus.getData("dataset-used-access-restricted-ids")
                iLookup_dataset_usedAccessRestricted_proz = resDatasetStatus.getData("dataset-used-access-restricted-proz")

                iLookup_dataset_usedAccessPossible_notGold = resDatasetStatus.getData("dataset-used-access-possible")
                iLookup_dataset_usedAccessPossible_notGold_proz = resDatasetStatus.getData("dataset-used-access-possible-proz")

                # Special: Gear
                lstLookup_gear_doubleHeader = resDatasetStatus.getData("gear-double-header")
                dicLookup_gear_double = resDatasetStatus.getData("gear-double-dic")
                iLookup_gear_double = resDatasetStatus.getData("gear-double")
                lstLookup_gear_doubleIds = resDatasetStatus.getData("gear-double-ids")
                iLookup_gear_double_groups = resDatasetStatus.getData("gear-double-groups")
                setLookup_gear_double_groups = resDatasetStatus.getData("gear-double-groups-set")

            sTitle = ""
            sStat = ""
            if self.__sTab == "taxlistupdate":
                # *** Special: Taxon List Update
                sTitle = "Last Taxon List Update"
                resLastTaxListUpdate = self.__model.subsystem_getConnector().status_getData("wizard", "taxonlist_lastupdate")
                if resLastTaxListUpdate is not None and resLastTaxListUpdate.isOk():
                    sLog = resLastTaxListUpdate.getData("log")
                    if sLog is None:
                        bHtml = False
                        bBrowser = False
                        sStat += "Taxon List update has not been executed yet ..."
                    else:
                        bHtml = True
                        bBrowser = True
                        sStat += sLog
                else:
                    sStat += "Problems getting TaxonWizard update log ..."
            elif self.__sTab == "DOI":
                # *** Special: DOI Status in Datasets
                sTitle = "Report on: DOI status"
                packDoi = self.__model.memory_load("doi")
                if packDoi is not None:
                    sLog = packDoi.get("log")
                    dicDoiOk = packDoi.get("ok")
                    dicDoiNo = packDoi.get("no")
                    dicDoiErr = packDoi.get("err")
                    sStat += sLog
                    bBrowser = True
                    bHtml = True
                else:
                    sStat += "Please start the automatic DOI check first ..."
            elif self.__model.subsystem_getMapper().all_isTableTaxon(self.__sTab):
                # *** Taxon Wizard
                # --- Alle Admin Listen ausser Base
                sSheet = str(self.__model.subsystem_getMapper().taxon_table2sheet(self.__sTab))
                sTitle = "Report on: "+sSheet+"\n\n"
                sStat += "*** Report on TaxonWizard List: "+sSheet+"\n\n"
                sSqlCmd = "select count(*) from " + self.__model.subsystem_getDb().account_single_getPart(
                    "schema") + "." + self.__sTab + ";"
                cursor.execute(sSqlCmd)
                lstContent = cursor.fetchall()
                try:
                    iNumberEntries = lstContent[0][0]
                except Exception as ex:
                    iNumberEntries = 0
                # sStat += "<h3>Number of entries</h3>" + str(iNumberEntries) +"<p>"
                sStat += "Number of entries: " + str(iNumberEntries) + "\n\n"
                if self.__sTab == "wormsbase":
                    # --- WormsBase
                    sSqlCmd = "select * from " + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + "." + self.__sTab + " where gold is false"
                    cursor.execute(sSqlCmd)
                    lstContent = cursor.fetchall()
                    lstHeader = [head[0] for head in cursor.description]
                    lstOut = [lstHeader]
                    lstOut += lstContent
                    if len(lstContent) > 0:
                        sStat += "--- Gold status not reached yet (" + str(len(lstContent)) + ") ---\n\n"
                        sStat += PrettyText.table(lstOut, True, False) + "\n\n"
                    else:
                        sStat += "All entries reached gold status\n\n"
                        # --- Taxon Admin
                        if self.__model.subsystem_getMapper().all_isTableTaxon(self.__sTab):
                            if self.__sTab == "wormsbase":
                                # --- TaxBase
                                sSqlCmd = "select * from " + self.__model.subsystem_getDb().account_single_getPart(
                                    "schema") + "." + self.__sTab + " where gold is false"
                                cursor.execute(sSqlCmd)
                                lstContent = cursor.fetchall()
                                lstHeader = [head[0] for head in cursor.description]
                                lstOut = [lstHeader]
                                lstOut += lstContent
                                if len(lstContent) > 0:
                                    sStat += "--- Gold status not reached yet (" + str(
                                        len(lstContent)) + ") ---\n\n"
                                    sStat += PrettyText.table(lstOut, True, False) + "\n\n"
                                else:
                                    sStat += "All entries reached gold status\n\n"
            elif self.__model.subsystem_getMapper().all_isTableLookup(self.__sTab):
                # *** Lookups
                if self.__sTab == "taxon":
                    # *** Lookup: Taxon
                    sStat = "*** Report on: Taxon List\n\n"
                    sTitle = "Report on: Taxon List"
                    # Alle Taxa
                    sSqlCmd = "select count(*) from " + self.__model.subsystem_getDb().account_single_getPart("schema") + "." + self.__sTab + ";"
                    cursor.execute(sSqlCmd)
                    lstContent = cursor.fetchall()
                    try:
                        iNumberEntries = lstContent[0][0]
                    except Exception as ex:
                        iNumberEntries = 0
                    sStat += "Number of entries: " + str(iNumberEntries) + "\n\n"
                    # Benutzte Taxa
                    # Anzahl
                    sSqlCmd = 'select count(distinct tax.aid) from ' + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + '.biota bio, ' + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + '.taxon tax where tax.id=bio.taxon_id;'
                    cursor.execute(sSqlCmd)
                    lstContent = cursor.fetchall()
                    iNumberUsedTaxa = lstContent[0][0]
                    if iNumberUsedTaxa > 0:
                        # Taxaliste
                        sSqlCmd = 'select count(tax.id) as "number", tax.aid as "Aphia Id", tax.scientificname as "scientificname", tax.aname as "accepted scientificname" from ' + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + '.biota bio, ' + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + '.taxon tax where tax.id=bio.taxon_id group by tax.scientificname, tax.aname, tax.aid  order by "number" desc;'
                        cursor.execute(sSqlCmd)
                        lstContent = cursor.fetchall()
                        lstHeader = [head[0] for head in cursor.description]
                        lstOut = [lstHeader]
                        lstOut += lstContent
                        sTaxUse = PrettyText.table(lstOut, True, False)
                        sStat += "--- Used taxa (" + str(iNumberUsedTaxa) + "x) ---\n\n"
                        sStat += sTaxUse + "\n\n"
                    # Unbenutzte Taxa
                    # Anzahl
                    sSqlCmd = 'select count(distinct tax.aid) from ' + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + '.taxon tax left join ' + self.__model.subsystem_getDb().account_single_getPart(
                        "schema") + '.biota bio on tax.id=bio.taxon_id where bio.id is null;'
                    cursor.execute(sSqlCmd)
                    lstContent = cursor.fetchall()
                    iNumberUnusedTaxa = lstContent[0][0]
                    if iNumberEntries > 0:
                        # Liste
                        sSqlCmd = 'select tax.aid as "Aphia Id", tax.scientificname as "Scientificname", tax.aname as "Accepted Scientificname" from ' + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + '.taxon tax left join ' + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + '.biota bio on tax.id=bio.taxon_id where bio.id is null'
                        cursor.execute(sSqlCmd)
                        lstContent = cursor.fetchall()
                        lstHeader = [head[0] for head in cursor.description]
                        sStat += "--- Unused taxa (" + str(iNumberUnusedTaxa) + "x) ---\n\n"
                        lstOut = [lstHeader]
                        lstOut += lstContent
                        sTabUse = PrettyText.table(lstOut, True, False)
                        sStat += sTabUse + "\n\n"
                else:
                    # *** Alle Lookups (ausser Taxon)
                    sSheet = str(self.__model.subsystem_getMapper().lookup_table2sheet(self.__sTab))
                    sTitle = "Report on: " + sSheet
                    sStat = "*** Report on: " + sSheet + "\n\n"

                    if self.__sTab == "dataset" and self.__model.memory_load("doi") is None:
                        sStat += "Attention: No further information on DOIs available (first start automatic DOI check manually)\n\n"

                    # Übersicht (Bar Plot)
                    sStat += "--- Overview\n\n"
                    lstBarchartData = [("Status", "Count")]
                    lstBarchartData.append(("all records", iLookup_all))
                    if iLookup_used > 0:
                        lstBarchartData.append(("used", iLookup_used))
                    if iLookup_usedNot > 0:
                        lstBarchartData.append(("not-used", iLookup_usedNot))
                    if iLookup_used_goldNot > 0:
                        lstBarchartData.append(("gold check [?]", iLookup_used_goldNot))
                    if iLookup_usedNot_goldNot > 0:
                        lstBarchartData.append(("purge soft [?]", iLookup_usedNot_goldNot))
                    if self.__sTab == "dataset":
                        lstBarchartData.append(("online", iLookup_dataset_usedAccessPossible))
                        lstBarchartData.append(("offline", iLookup_dataset_usedAccessRestricted))
                        packDoi = self.__model.memory_load("doi")
                        if packDoi is not None:
                            dicDoiOk = packDoi.get("ok")
                            dicDoiNo = packDoi.get("no")
                            dicDoiErr = packDoi.get("err")
                            iNotPangaea = packDoi.get("cluster-not-pangaea")
                            setUsed_free_request_not_ok_ids = set(dicDoiErr.keys()).union(set(dicDoiNo.keys())).intersection(setLookup_dataset_usedAccessPossible_ids)
                            iUsed_free_request_not_ok = len(setUsed_free_request_not_ok_ids)
                            if iUsed_free_request_not_ok > 0:
                                lstBarchartData.append(("bad doi online [!]", iUsed_free_request_not_ok))
                            if iNotPangaea > 0:
                                lstBarchartData.append(("no PANGAEA DOI [?]", iNotPangaea))
                    if self.__sTab == "gear":
                        if iLookup_gear_double > 0:
                            lstBarchartData.append(("double config [?]", iLookup_gear_double))
                    if iLookup_all > 0:
                        sStat += PrettyText.barchart(lstBarchartData, maxbarsize=50, header=True) + "\n"
                    else:
                        sStat += "... no records\n"

                    # Eintraege
                    if iLookup_all > 0:
                        sStat += "--- Usage\n\n"
                        lstTab = [("State", "Count", "All", "%")]
                        lstTab.append(("used", iLookup_used, iLookup_all, iLookup_used_proz))
                        lstTab.append(("not-used", iLookup_usedNot, iLookup_all, 100 - iLookup_used_proz))
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    if iLookup_all > 0:
                        sStat += "--- Gold\n\n"
                        lstTab = [("State","Count", "All", "%")]
                        lstTab.append(("gold", iLookup_gold, iLookup_all, iLookup_gold_proz))
                        lstTab.append(("not-gold", iLookup_all-iLookup_gold, iLookup_all, 100 - iLookup_gold_proz))
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    if iLookup_all > 0 and iLookup_used_goldNot > 0:
                        sStat += "--- Candidates to check for gold status\n\n"
                        lstTab = [("State", "Count", "Used", "%")]
                        lstTab.append(("used not-gold", iLookup_used_goldNot, iLookup_used, iLookup_used_goldNot_proz))
                        sStat += PrettyText.table(lstTab) + "\n\n"
                        lstTab = []
                        lstTab = [[sName] for iId, sName in dicLookup_all.items() if iId in setLookup_used_goldNot_ids]
                        lstTab.sort(key=lambda x: str(x[0]).lower())
                        lstTab.insert(0, ["Name"])
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    if iLookup_all > 0 and iLookup_usedNot_goldNot > 0:
                        sStat += "--- Candidates to purge\n\n"
                        lstTab = [("State", "Count", "Not-used", "%")]
                        lstTab.append(("not-used not-gold", iLookup_usedNot_goldNot, iLookup_usedNot, iLookup_usedNot_goldNot_proz))
                        sStat += PrettyText.table(lstTab) + "\n\n"
                        lstTab = [["Name"]]
                        lstTmp = [[sName] for iId, sName in dicLookup_all.items() if iId in setLookup_usedNot_goldNot_ids]
                        lstTab += lstTmp
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    if iUse_ing_files > 0:
                        sStat += "--- Usage in Ingests\n\n"
                        lstTab = []
                        for iId, lstFiles in dicUse_ing_files.items():
                            if iId in dicLookup_all:
                                sName = dicLookup_all[iId]
                            else:
                                sName = "<Uuups!>"
                            lstTab.append([sName, len(lstFiles), PrettyText.lst2strC(lstFiles)])
                        lstTab.sort(key=lambda x: str(x[0]).lower())
                        lstTab.insert(0, ["Name", "Usage", "Files"])
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    if iUse_lup_names > 0:
                        sStat += "--- Usage in other Lookups\n\n"
                        lstTab = []
                        for iId, dicLups in dicUse_lup_names.items():
                            for sLupTable, lstLupName in dicLups.items():
                                if iId in dicLookup_all:
                                    sName = dicLookup_all[iId]
                                else:
                                    sName = "<Uuups!>"
                            lstTab.append([sName, sLupTable, len(lstLupName), PrettyText.lst2strC(lstLupName)])
                        lstTab.sort(key=lambda x: str(x[0]).lower())
                        lstTab.insert(0, ["Name", "Lookup", "Usage", "Name"])
                        sStat += PrettyText.table(lstTab) + "\n\n"

                    # *** Zusatz: Dataset
                    if self.__sTab == "dataset":
                        if iLookup_all > 0:
                            sStat += "--- Security level: Data\n\n"
                            lstTab = [("State","Count", "All", "%")]
                            lstTab.append(("free", iLookup_dataset_datafree, iLookup_all, iLookup_dataset_datafree_proz))
                            lstTab.append(("request", iLookup_dataset_dataRequest, iLookup_all, iLookup_dataset_dataRequest_proz))
                            lstTab.append(("confidential", iLookup_dataset_dataConfidential, iLookup_all, iLookup_dataset_dataConfidential_proz))
                            sStat += PrettyText.table(lstTab) + "\n\n"

                            sStat += "--- Security level: Data products\n\n"
                            lstTab = [("State","Count", "All", "%")]
                            lstTab.append(("free", iLookup_dataset_productfree, iLookup_all, iLookup_dataset_productfree_proz))
                            lstTab.append(("request", iLookup_dataset_productRequest, iLookup_all, iLookup_dataset_productRequest_proz))
                            lstTab.append(("confidential", iLookup_dataset_productConfidential, iLookup_all, iLookup_dataset_productConfidential_proz))
                            sStat += PrettyText.table(lstTab) + "\n\n"

                            if iLookup_dataset_usedAccessPossible > 0:
                                sStat += "--- Security level: Online accessable (in some way)\n\n"
                                lstTab = [("State", "Count", "Used", "%")]
                                lstTab.append(("used (free/request) (data/product)", iLookup_dataset_usedAccessPossible, iLookup_used, iLookup_dataset_usedAccessPossible_proz))
                                sStat += PrettyText.table(lstTab) + "\n\n"
                                lstTab = []
                                for iId, sName in dicLookup_all.items():
                                    if iId in setLookup_dataset_usedAccessPossible_ids:
                                        sData, sProduct = getAccessLevel(iId)
                                        lstTab.append([sName, sData, sProduct])
                                lstTab.sort(key=lambda x: str(x[0]).lower())
                                lstTab.insert(0, ["Name","Data","Product"])
                                sStat += PrettyText.table(lstTab) + "\n\n"

                            if iLookup_dataset_usedAccessRestricted > 0:
                                sStat += "--- Security level: Online not accessable at all\n\n"
                                lstTab = [("State", "Count", "Used", "%")]
                                lstTab.append(("used confidential (data&product)", iLookup_dataset_usedAccessRestricted, iLookup_used,
                                               iLookup_dataset_usedAccessRestricted_proz))
                                sStat += PrettyText.table(lstTab) + "\n\n"
                                lstTab = []
                                for iId, sName in dicLookup_all.items():
                                    if iId in setLookup_dataset_usedAccessRestricted_ids:
                                        lstTab.append([sName])
                                lstTab.sort(key=lambda x: str(x[0]).lower())
                                lstTab.insert(0, ["Name"])
                                sStat += PrettyText.table(lstTab) + "\n\n"

                        # DOIS
                        packDoi = self.__model.memory_load("doi")
                        if packDoi is not None:
                            dicDoiOk = packDoi.get("ok")
                            dicDoiNo = packDoi.get("no")
                            dicDoiErr = packDoi.get("err")
                            iNotPangaea = packDoi.get("cluster-not-pangaea")
                            setNotPangaea = packDoi.get("cluster-not-pangaea-ids")
                            # Dataset reachable and DOI not OK
                            setUsed_free_request_not_ok_ids = set(dicDoiErr.keys()).union(set(dicDoiNo.keys())).intersection(setLookup_dataset_usedAccessPossible_ids)
                            iUsed_free_request_not_ok = len(setUsed_free_request_not_ok_ids)
                            if iUsed_free_request_not_ok > 0:
                                sStat += "--- Bad DOI online ("+str(iUsed_free_request_not_ok)+"x)\n\n"
                                lstTab = []
                                for iId, sName in dicLookup_all.items():
                                    if iId in setUsed_free_request_not_ok_ids:
                                        sDataAccessLevel, sProductAccessLevel = getAccessLevel(iId)
                                        if iId in dicDoiErr:
                                            sName = dicDoiErr[iId][0]
                                            sDoi = dicDoiErr[iId][1]
                                            sProblem = dicDoiErr[iId][2]
                                            lstTab.append([sName, sDataAccessLevel, sProductAccessLevel, sDoi, sProblem])
                                        elif iId in dicDoiNo:
                                            sName = dicDoiNo[iId]
                                            sDoi = "-"
                                            sProblem = "No DOI given"
                                            lstTab.append([sName, sDataAccessLevel, sProductAccessLevel, sDoi, sProblem])
                                lstTab.sort(key=lambda x: str(x[0]).lower())
                                lstTab.insert(0, ["Name", "Data", "Product", "DOI", "Problem"])
                                sStat += PrettyText.table(lstTab) + "\n\n"
                            if iNotPangaea> 0:
                                sStat += "--- No PANGAEA DOI - really raw data available? (" + str(iNotPangaea) +")\n\n"
                                lstTab = []
                                for iId, sName in dicLookup_all.items():
                                    if iId in setNotPangaea:
                                        lstTab.append([sName])
                                lstTab.sort(key=lambda x: str(x[0]).lower())
                                lstTab.insert(0, ["Name"])
                                sStat += PrettyText.table(lstTab) + "\n\n"

                    # *** Zusatz: Gear
                    if self.__sTab == "gear":
                        # Doppelte Configurationen Definitionen
                        tpGroupHeader = lstLookup_gear_doubleHeader[0]
                        tpConfHeader = lstLookup_gear_doubleHeader[1]
                        sStat += "--- Check for double configurations\n\n"
                        sStat += "* All gear groups (" + str(iLookup_gear_double_groups) + "x)\n\n"
                        lstTab = [tpGroupHeader]
                        lstTab += list(setLookup_gear_double_groups)
                        sStat += PrettyText.table(lstTab) + "\n\n"
                        if iLookup_gear_double > 0:
                            sStat += "* Double configurations found (" + str(iLookup_gear_double) + "x)\n\n"
                            for tpGroup, dicConf in dicLookup_gear_double.items():
                                lstGroup = []
                                for k, v in dict(zip(tpGroupHeader, tpGroup)).items():
                                    lstGroup.append(str(k) + "=" + (str(v if v is not None else "-")) + "")
                                sGroup = PrettyText.lst2strC(lstGroup)
                                for tpConf, lstNames in dicConf.items():
                                    sStat += "Entries ("+str(len(lstNames))+"x): "+PrettyText.lst2strC(lstNames)+"\n"
                                    sStat += "Group: "+sGroup+"\n"
                                    lstTab = [tpConfHeader]
                                    lstTmp = [it if it is not None else "" for it in tpConf]
                                    lstTab.append(lstTmp)
                                    sStat += PrettyText.table(lstTab) + "\n\n"
                        else:
                            sStat += "... no double configurations found\n\n"

                        # Anzahl der Benutzungen
                        sSqlCmd = 'select gea.name as "Name", dat.name as "Dataset", count(gear_id) as "Count" from <schema>.subsample subsmp, <schema>.sample smp, <schema>.gear gea, <schema>.subset sub, <schema>.dataset dat where gea.id=smp.gear_id and sub.subsample_id=subsmp.id and subsmp.sample_id=smp.id and dat.id=sub.dataset_id group by smp.gear_id, gea.name, dat.name order by "Name" asc, "Dataset" asc, "Count" asc'
                        sSqlCmd = sSqlCmd.replace("<schema>", self.__model.subsystem_getDb().account_single_getPart("schema"))
                        cursor.execute(sSqlCmd)
                        sSqlQuery = cursor.query.decode("utf-8")
                        lstContent = cursor.fetchall()
                        lstHeader = [head[0] for head in cursor.description]
                        sStat += "--- Usage in datasets\n\n"
                        lstOut = [lstHeader]
                        lstOut += lstContent
                        sTabUse = PrettyText.table(lstOut, True, False)
                        sStat += sTabUse + "\n\n"

                    # Zusatz: TaxCover
                    if self.__sTab == "taxcover":
                        # --- Anzahl der Benutzungen
                        sSqlCmd = 'select cov.name as "Name", count(taxcover_id) as "Count", dat.name as "Dataset" from <schema>.taxcover cov, <schema>.subset sub, <schema>.dataset dat where cov.id=sub.taxcover_id and dat.id=sub.dataset_id group by sub.taxcover_id, cov.name, dat.name order by "Name" asc, "Dataset" asc, "Count"'
                        sSqlCmd = sSqlCmd.replace("<schema>", self.__model.subsystem_getDb().account_single_getPart("schema"))
                        cursor.execute(sSqlCmd)
                        lstContent = cursor.fetchall()
                        lstHeader = [head[0] for head in cursor.description]
                        # Baue Ergebnis um (Fasse Datasets zusammen)
                        dicContent = {}
                        for row in lstContent:
                            sTaxcover = row[0]
                            iNumber = row[1]
                            sDataset = row[2]
                            if sTaxcover in dicContent:
                                dicContent[sTaxcover][0] += iNumber
                                dicContent[sTaxcover][1].append(sDataset + " (" + str(iNumber) + "x)")
                            else:
                                dicContent[sTaxcover] = [iNumber, [sDataset + " (" + str(iNumber) + "x)"]]
                        odicContentNew = OrderedDict(sorted(dicContent.items(), key=lambda x: str(x[1]).lower(), reverse=True))
                        lstContentNew = []
                        for sTaxcover, (iNumber, lstDataset) in odicContentNew.items():
                            lstContentNew.append([sTaxcover, iNumber, PrettyText.lst2strC(lstDataset)])
                        sStat += "--- Usage in datasets\n\n"
                        lstOut = [lstHeader]
                        lstOut += lstContentNew
                        sTabUse = PrettyText.table(lstOut, True, False)
                        sStat += sTabUse + "\n\n"
                        # --- Unbenutze Gears
                        sSqlCmd = "select cov.name as Name from " + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + ".taxcover cov left join " + self.__model.subsystem_getDb().account_single_getPart(
                            "schema") + ".subset sub on cov.id=sub.taxcover_id where sub.id is null"
                        cursor.execute(sSqlCmd)
                        lstContent = cursor.fetchall()
                        lstHeader = [head[0] for head in cursor.description]

            # Antwort
            result.addData("title", sTitle)
            result.addData("text", sStat)
            result.addData("html", bHtml)
            result.addData("browser", bBrowser)

            # Db disconnect
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Unexpected error", add=str(ex), kat="Statistic on Table", sys="Admin")

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_ShowActionlog(Worker):
    """
    Hole ein ActionLog und sende es an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model,  iId):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__iId = iId

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Loading action log number: "+str(self.__iId))
        progress("on-pulse")
        workflow(top="Database", sub="Waiting for answer")

        try:
            # DB-Verbindung
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # SQL-Command
            sSqlCmd = "select log from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".actionlog where id = %s;"
            cursor.execute(sSqlCmd, [str(self.__iId)])

            # Ergebnis
            lstContent = cursor.fetchall()
            sContent = lstContent[0][0]

            # Antwort
            result.addData("content", sContent)

            # Db disconnect
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Sql error: ", add=str(ex), kat="Show Action Log", sys="Admin")

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_CheckDois(Worker):
    """
    Check DOIs
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Checking DOIs")
        progress("on-pulse")

        workflow(top="Checking DOIs")
        resDoi = self.__model.subsystem_getDb().lookup_dataset_checkDois()

        if resDoi.isOk():
            bHtml = True
            sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)
            sColor_awi_blue_mid = Colors.getColor("awi_blue_mid", html=True)
            sColor_awi_blue_dark = Colors.getColor("awi_blue_dark", html=True)
            iPreview = 5
            sLog = ""
            dicOk = resDoi.getData("ok")
            dicNone = resDoi.getData("no")
            dicErr = resDoi.getData("err")
            dicDOIs_clustered_byPublisher = resDoi.getData("cluster-journal")
            dicDatasets_clustered_byDoi = resDoi.getData("cluster-doi")
            iPangaea_nr = resDoi.getData("cluster-pangaea")
            setPangaea_ids = resDoi.getData("cluster-pangaea-ids")
            iPangaea_proz = resDoi.getData("cluster-pangaea-proz")
            iPangaeaNot_nr = resDoi.getData("cluster-not-pangaea")
            setPangaeaNot_ids = resDoi.getData("cluster-not-pangaea-ids")
            iPangaeaNot_proz = resDoi.getData("cluster-not-pangaea-proz")
            # ---
            if (len(dicOk) + len(dicErr) + len(dicNone)) > 0:
                if bHtml:
                    sLog += '<img height="67" src="https://recruitingapp-5442.de.umantis.com/pubhtml/logo.gif" alt="AWI Logo">'
                    sLog  += '<div layout="block" styleclass="DynForm_DfAttributes" style="z-index:0;position:absolute;left:0px;top:68px;right:0px;height:25px;overflow:hidden;text-align:left;background-color:' + sColor_awi_blue_light + ';border:medium none;-moz-border-colors:none;border-radius:0; margin-top: 5px; margin-bottom: 5px; ;"></div>'
                    sLog  += '<br>'
                    sLog  += "\n\n"
                    sLog  += "&nbsp;"
                    sLog  += '<h1 style="color:' + sColor_awi_blue_dark + ';">DOI Report</h1>\n\n'
                    sLog += "*** Reachability\n\n"
                else:
                    sLog += "*** Reachability of DOIs\n\n"
                if len(dicOk) > 0:
                    lstTable = []
                    for iId, (sName, sDoi, sDoi_redirect) in dicOk.items():
                        lstTable.append((sName, PrettyText.html_link(sDoi, html=bHtml), PrettyText.html_link(sDoi_redirect, html=bHtml)))
                    lstTable.sort(key=lambda x: str(x[0]).lower())
                    lstTable.insert(0, ("Dataset", "DOI", "DOI redirected"))
                    sLog += "* DOI working (" + str(len(dicOk)) + "x)\n\n"
                    sLog += PrettyText.table(lstTable, header=True, html=bHtml, preview=iPreview)
                    sLog += "\n\n"
                if len(dicErr) > 0:
                    lstTable = []
                    for iId, (sName, sDoi, sProblem) in dicErr.items():
                        lstTable.append((sName, PrettyText.html_link(sDoi, html=bHtml), sProblem))
                    sLog += "* DOI not working (" + str(len(dicErr)) + "x)\n\n"
                    lstTable.sort(key=lambda x: str(x[0]).lower())
                    lstTable.insert(0, ("Dataset", "DOI", "Problem"))
                    sLog += PrettyText.table(lstTable, header=True, html=bHtml, preview=iPreview)
                    sLog += "\n\n"
                if len(dicNone) > 0:
                    sLog += "* DOI not given (" + str(len(dicNone)) + "x)\n\n"
                    lstTable = []
                    for iId, sName in dicNone.items():
                        lstTable.append([sName])
                    lstTable.sort(key=lambda x: str(x).lower())
                    lstTable.insert(0, ["Dataset"])
                    sLog += PrettyText.table(lstTable, header=True, html=bHtml, preview=iPreview)
                    sLog += "\n\n"
                if (len(dicDOIs_clustered_byPublisher)) > 0:
                    sLog += "*** DOIs clusterd by publisher\n\n"
                    for sJournal_name, dicDois_journal in dicDOIs_clustered_byPublisher.items():
                        sLog += "--- " + sJournal_name + "\n"
                        lstTab = []
                        for iId, (sName, sDoi, sDoi_redirect) in dicDois_journal.items():
                            sDoi_effective = sDoi_redirect if sDoi_redirect not in [None, ""] else sDoi
                            lstTab.append((sName, PrettyText.html_link(sDoi_effective, html=bHtml)))
                        lstTab.sort(key=lambda x: str(x[0]).lower())
                        lstTab.insert(0, ["Dataset", "DOI"])
                        sLog += PrettyText.table(lstTab, header=True, html=bHtml, preview=iPreview) + "\n\n"
                if (len(dicDatasets_clustered_byDoi)) > 0:
                    sLog += "*** Datasets with same DOI\n\n"
                    for sJournal_name, dicDois in dicDatasets_clustered_byDoi.items():
                        sTmp = ""
                        for sDoi, lstDatasets in dicDois.items():
                            if len(lstDatasets) > 1:
                                sTmp += PrettyText.html_format("DOI: ", italic=True, bold=True,html=bHtml) + PrettyText.html_link(sDoi, html=bHtml) + "\n"
                                lstDatasets.sort(key=lambda x: str(x[0]).lower())
                                sTmp += PrettyText.table(lstDatasets, header=False, html=bHtml, preview=iPreview) + "\n\n"
                        if sTmp != "":
                            sLog += "--- " + sJournal_name + "\n"
                            sLog += sTmp
                if bHtml:
                    sLog = PrettyText.html_txt2html(sLog, title="DOI Report")
            else:
                sLog += "Nothing to check, no DOIs given ... \n\n"

        # Antwort
        result.addData("err", dicErr)
        result.addData("ok", dicOk)
        result.addData("no", dicNone)
        result.addData("log", sLog)
        result.addData("cluster-journal", dicDOIs_clustered_byPublisher)
        result.addData("cluster-doi", dicDatasets_clustered_byDoi)
        result.addData("cluster-pangaea", iPangaea_nr)
        result.addData("cluster-pangaea-ids", setPangaea_ids)
        result.addData("cluster-pangaea-proz", iPangaea_proz)
        result.addData("cluster-not-pangaea", iPangaeaNot_nr)
        result.addData("cluster-not-pangaea-ids", setPangaeaNot_ids)
        result.addData("cluster-not-pangaea-proz", iPangaeaNot_proz)
        progress("off")
        workflow()

        self.sendResult(result)

class Worker_ShowIngestLog(Worker):
    """
    Hole das Log eines Ingests und sende es an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model,  iId):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__iId = iId

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Loading ingest log number: "+str(self.__iId))
        progress("on-pulse")
        workflow(top="Show Ingest log", sub="Waiting for answer from database")

        try:
            # DB-Verbindung
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # SQL-Command
            sSqlCmd = "select log from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".ingest where id = %s;"
            cursor.execute(sSqlCmd, [str(self.__iId)])

            # Ergebnis
            lstContent = cursor.fetchall()
            sContent = lstContent[0][0]

            resFile = self.__model.subsystem_getDb().ingest_getIngestFilenames(self.__iId)
            sFile = resFile.getData("file")

            if sFile is None:
                sTitle = "Ingest log: " + str(self.__iId)
            else:
                sTitle = "Ingest log: " + str(sFile)

            # Antwort
            result.addData("title", sTitle)
            result.addData("text", sContent)

            # Db disconnect
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Sql error", add=str(ex), kat="Show Ingest Log", sys="Admin")

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_GoldEdit(Worker):
    """
    Setze Gold Einträge und sende Ergebnis an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model,  sTab, lstIds, bBool, **kwargs):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__sTab = sTab
        self.__lstIds = lstIds
        self.__bBool = bBool
        self.__closeWin = kwargs['closeWin'] if "closeWin" in kwargs else None

    def run(self):
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        result = Result(name="Delete ingests")

        progress("on-pulse")
        workflow(top="Edit Gold Status")

        # ---

        # DB-Verbindung
        conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
        cursor = conn.cursor()

        iEdit = 0
        iRecs =len(self.__lstIds)

        workflow(top="Edit Status", sub="to: "+("gold" if self.__bBool else "not-gold"))
        for iRec, iId in enumerate(self.__lstIds):
            try:
                bOk = self.__model.subsystem_getDb().table_updateInTable(self.__sTab, iId, {"gold": self.__bBool}, cursor=cursor).isOk()
                bOk = True
                if not bOk:
                    result.addErr(
                        "Error editing " + ("gold" if self.__bBool else "not-gold") + " status for entry with id=" + str(
                            iId) + ": " + str(ex))
                else:
                    iEdit += 1
                progress(iRec, iRecs, 50, proz=True, abs=True, err=result.getErrNo(), warn=result.getWarnNo(), info=result.getInfoNo())
                status("Setting status for entry: " + str(iRec + 1) + "/" + str(iRecs))
                if not self.__model.conf_isDry():
                    conn.commit()
                else:
                    conn.rollback()
            except Exception as ex:
                result.addErr(
                    "Error editing " + ("gold" if self.__bBool else "not-gold") + " status for entry with id=" + str(
                        iId) + ": " + str(ex))
                conn.rollback()

        # Action Log
        if not self.__model.conf_isDry():
            sInfoOut = ""
            # Action
            if self.__sTab == "wormsbase":
                sActionLookup = "gold-tax-"
            else:
                sActionLookup = "gold-lup-"
            if self.__bBool:
                sActionLookup += "yes"
            else:
                sActionLookup += "no"
            # Log
            if self.__sTab == "wormsbase":
                sInfoOut += "TaxonWizard: " + self.__sTab + "\n"
            else:
                sInfoOut += "Lookup: " + self.__sTab + "\n"
            # Details
            sOutDetails = ""
            sOutDetailsTaxa = ""
            if self.__sTab == "wormsbase":
                progress("on-pulse")
                workflow(top="Edit Gold Status", sub="Assembling Taxa data")
                # Base List
                resUpdate = self.__model.subsystem_getDb().table_getWholeTable("wormsbase", order="scientificname", ids=self.__lstIds)
                if resUpdate.isErr():
                    result.addResult(resUpdate, err=True)
                lstHeaders = resUpdate.getData("header")
                lstContent = resUpdate.getData("content")
                lstTable = []
                lstTable.append(lstHeaders)
                lstTable += lstContent
                sOutDetails += PrettyText.table(lstTable) + "\n"
                # Special Taxa Übersicht
                sOutDetailsTaxa += "\n"
                sOutDetailsTaxa += "--- Taxa --\n"
                lstAid = []
                lstTaxa = []
                for sCol in lstContent:
                    iAid = sCol[1]
                    lstAid.append(iAid)
                workflow(top="Edit Gold", sub="Asking MyWoRMS", fmt=False)
                # Ask MyWoRMS
                iRecs = len(lstAid)
                for iRec, iAid in enumerate(lstAid):
                    resMyWorms = self.__model.subsystem_getMyWorms().getTaxon(iAid)
                    progress(iRec, iRecs, 50, proz=True, abs=True, err=result.getErrNo(), warn=result.getWarnNo(),info=result.getInfoNo())
                    if resMyWorms.isOk() and resMyWorms.getData("found"):
                        t = resMyWorms.getData("taxon")
                        lstTaxa.append(t)
                        status("Asking MyWoRMS: " + str(t.signature()), fmt=False)
                # Statistic
                sOutDetailsTaxa += self.__model.subsystem_getMyWorms().getStatisticProblemTaxa(lstTaxa)
            else:
                progress("on-pulse")
                workflow(top="Edit Gold Status", sub="Log")
                # Lookup
                resUpdate = self.__model.subsystem_getDb().table_getWholeTable(self.__sTab, ids=self.__lstIds)
                if resUpdate.isErr():
                    result.addResult(resUpdate, err=True)
                lstHeaders = resUpdate.getData("header")
                lstContent = resUpdate.getData("content")
                lstTable = []
                lstTable.append(lstHeaders)
                lstTable += lstContent
                sOutDetails += PrettyText.table(lstTable)

            # Log
            sInfoOut += "Set status to: " + ("gold" if self.__bBool else "no-gold") + "\n"
            sInfoOut += "Entries: " + str(iEdit) + "\n"
            sInfoOut += "\n"
            sInfoOut += "--- Entries ---\n"
            sInfoOut += sOutDetails
            sInfoOut += sOutDetailsTaxa

            result_actionlog = self.__model.subsystem_getDb().actionlog_new(sActionLookup, self.__sTab, sInfoOut, cursor=cursor)
            if result_actionlog.isErr():
                result.addResult(result_actionlog, err=True)

        # Antwort
        result.addData("bool", self.__bBool)
        result.addData("edit", iEdit)
        result.addData("tab", self.__sTab)
        conn.commit()

        # UI
        progress("off")
        workflow()

        # DB-Verbindung schliessen
        cursor.close()
        conn.close()

        self.sendResult(result)

class Worker_DelEntry(Worker):
    """
    Lösche Einträge in Tabelle und sende Ergebnis an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model, sTab, lstIds, **kwargs):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__lstIds = lstIds
        self.__sTab = sTab


    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Delete entries")

        # DB-Verbindung
        conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
        cursor = conn.cursor()

        # Daten
        iIdStart = self.__lstIds[0] -1 # Auf gut Glück die Id vor der ersten zu löschenden Id herausfischen
        iDel = 0
        iDelFail = 0
        iRecs = 0

        # Progress
        progress("on-pulse")
        workflow(top="Deleting entries")

        if self.__sTab == "wormsbase":
            # *** Wormsbase
            resDel = self.__model.subsystem_getTaxonWizard().delBaseTaxa(ids=self.__lstIds)
            if resDel.isErr():
                result.addResult(resDel)
            else:
                iRecs = len(self.__lstIds)
                iDel = iRecs
                iDelFail = 0
        else:
            # *** Rest
            # Loop: Einträge
            iRecs = len(self.__lstIds)
            for iRec, iId in enumerate(self.__lstIds):
                try:
                    resDel = self.__model.subsystem_getDb().table_delTableEntry(self.__sTab, "id", iId, cursor=cursor)
                    if resDel.isErr():
                        result.addResult(resDel)
                        iDelFail += 1
                    else:
                        iDel += 1
                    # Progress
                    progress(iRec, iRecs, 50, proz=True, abs=True, err=result.getErrNo(), warn=result.getWarnNo(),info=result.getInfoNo())
                    status("Deleting entry: " + str(iRec + 1) + "/" + str(iRecs))

                except Exception as ex:
                    result.addErr("Entry", add=str(ex), sys="Database", kat="Delete entry in table")
                    conn.commit()

            # Action Log
            if not self.__model.conf_isDry():
                sLog = ""
                sInfoOut = "Table: " + self.__sTab + "\n"
                sInfoOut += "Entries: " + str(iDel) + "\n"
                if self.__model.subsystem_getMapper().all_isTableTaxon(self.__sTab):
                    # TaxonWizard
                    sActionLookup = "tax-adm-del"
                else:
                    # Lookup
                    sActionLookup = "lup-del"
                result_actionlog = self.__model.subsystem_getDb().actionlog_new(sActionLookup, self.__sTab, sLog, cursor=cursor)
                result.addResult(result_actionlog)

        conn.commit()

        workflow()
        progress("off")

        # DB-Verbindung schliessen
        cursor.close()
        conn.close()

        # Ergebnis
        result.addData("del-total", iRecs)
        result.addData("del-ok", iDel)
        result.addData("del-err", iDelFail)
        result.addData("tab", self.__sTab)
        result.addData("id-start", iIdStart)
        # Loeschen
        self.sendResult(result)


class Worker_PurgeLookup(Worker):
    """
    Purge Lookups
    """

    def __init__(self, model,  sLookup, bFull):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__sLookup = sLookup
        self.__bFull = bFull

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        result = Result(name="Purge Lookups")

        progress("on-pulse")
        workflow(top="Purge lookup", sub="hard purge")
        # Loeschen
        result = self.__model.subsystem_getDb().lookup_purge(self.__sLookup, self.__bFull)
        progress("off")
        workflow()

        self.sendResult(result)

class Worker_DelIngest(Worker):
    """
    Loesche ingests und sende Ergebnis an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model,  lstIngests):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__lstIngests = lstIngests

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow

        result = Result(name="Delete ingests")

        progress("on-pulse")
        workflow(top="Deleting ingests")
        # Loeschen
        result = self.__model.subsystem_getDb().ingest_delIngest(self.__lstIngests)
        progress("off")
        workflow()

        self.sendResult(result)

class Worker_PopActionLog(Worker):
    """
    Hole Action Log und sende es an View zurück (um es dort anzuzeigen)
    """

    def __init__(self, model,  iId):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__iId = iId

    def run(self):
        pn = self.__smartprinter.normal
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        result = Result(name="Loading Action Log Number: "+str(self.__iId))
        progress("on-pulse")
        workflow(top="Action Log", sub="Waiting for answer from database")

        try:
            # DB-Verbindung
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()

            # SQL-Command
            sSqlCmd = "select log from " + self.__model.subsystem_getDb().account_single_getPart("schema") + ".actionlog where id = %s;"
            cursor.execute(sSqlCmd, [str(self.__iId)])

            # Ergebnis
            lstContent = cursor.fetchall()
            sContent = lstContent[0][0]

            # Antwort
            result.addData("title", "Action Log Number: "+str(self.__iId))
            result.addData("text", sContent)

            # Db disconnect
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Sql error: " + str(ex))

        progress("off")
        workflow()
        self.sendResult(result)

class Worker_ShowWholeTable(Worker):
    """
    Läd eine Tabelle und sende den Inhalt an View zurück (um es dort anzuzeigen)
    Hier wird selbständig erkannt, was für eine Tabelle es ist
    und wie sie angezeigt werden soll
    """

    def __init__(self, model, sTab, **kwargs):
        """
        Args:
            (gui.model.Model): Das Modell
            (str) : Tabellenname
        kwargs:
            type (str): Art der Tabelle (ingest-raw, ingest, lookup, taxon)
            id (int): Datensatz Id auf die fokussiert werden soll
            gold (bool): Nur Gold oder Nicht-Gold Einträge anzeigen
            doi (bool): DOI ok
            doi_pangaea (bool): DOI führt zu PANGAEA
            used (bool): Lookup benutzt
            access (bool): Zugang gewährleistet (irgendwie)
            gear_double (bool): Gear Double Config
        """
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__sTab = sTab  # Tabelle
        self.__title = None  # Titel des Dialogfensters
        self.__sType = None  # Art der Tabelle
        self.__bEdit = False  # Soll die Tabelle editierbar sein (in der View)
        self.__bDel = False  # Soll die Tabelle löschbar sein (in der View)
        self.__bNoIds = False  # Sollen die Ids entfernt werden
        self.__bHuman = False  # : Sollen die Daten humanisiert werden
        self.__bSave = False  # Soll die Tabelle speicherbar sein (in der View)
        self.__iFokusId = None  # Datensatz Id auf die fokussiert werden soll
        self.__bGoldEdit = False # Gold Status Edit
        self.__bUsed = None # Used/Nicht-Used Datensätze anzeigen - None: keine Einschränkung
        self.__bGold = None # Gold/Nicht-Gold Datensätze anzeigen - None: keine Einschränkung
        self.__bAccess = None # Online Zugang irgendwie möglich - None: keine Einschränkung
        self.__bDoi = None # DOI vorhanden und funktioniert - None: keine Einschränkung
        self.__bDoiPangaea = None # DOI führt zu PANGAEA - None: keine Einschränkung
        self.__bGear_double = None # Gears mit doppelter Configuration - None: keine Einschränkung
        self.__lstSelect = None

        # Konfiguration
        # Kwargs
        if "id" in kwargs:
            self.__iFokusId = kwargs["id"]
        if "type" in kwargs:
            self.__sType = kwargs["type"]
        if "used" in kwargs:
            self.__bUsed = kwargs["used"]
        if "doi" in kwargs:
            self.__bDoi = kwargs["doi"]
        if "doi_pangaea" in kwargs:
            self.__bDoiPangaea = kwargs["doi_pangaea"]
        if "access" in kwargs:
            self.__bAccess = kwargs["access"]
        if "gold" in kwargs:
            self.__bGold = kwargs["gold"]
        if "gear_double" in kwargs:
            self.__bGear_double = kwargs["gear_double"]
        if "gold_edit" in kwargs:
            self.__bGoldEdit = kwargs["gold_edit"]
        # Type
        if self.__sType is None:
            if self.__sTab in self.__map.lookup_getDbTables():
                self.__sType = "lookup"
            elif self.__sTab in self.__map.taxon_getDbTables():
                self.__sType = "taxon"
            elif self.__sTab in self.__map.ingest_getDbTables():
                self.__sType = "ingest"

        # Title
        if self.__sType == "ingest":
            self.__sTitle = self.__map.ingest_table2sheet(sTab)
        elif self.__sType == "lookup":
            self.__sTitle = self.__map.lookup_table2sheet(sTab)
        elif self.__sType == "taxon":
            self.__sTitle = self.__map.taxon_table2sheet(sTab)
        else:
            self.__sTitle = self.__sTab.replace("_"," ").capitalize()

        # Edit/Del
        if self.__sType in ["lookup", "taxon"]:
            # Taxon Table ist nicht direkt änderbar
            if self.__sTab == "taxon":
                self.__bEdit = False
                self.__bDel = True
            else:
                self.__bDel = True
                self.__bEdit = True

        # FAO Region
        if self.__sTab == "region":
            self.__bDel = True
            self.__bEdit = False

        # Human
        if self.__sType in ["lookup"]:
            self.__bHuman = True

        # Save
        if self.__sType in ["lookup"]:
            self.__bSave = True

    def run(self):
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        progress("on-pulse")
        result = Result(name="Show table "+self.__sTab)

        # Id Untermengen herausfiltern
        lstIds = None
        if self.__model.subsystem_getMapper().all_isTableLookup(self.__sTab) and self.__sTab != "taxon":
            resLupStatus = self.__model.subsystem_getDb().lookup_getStatus(self.__sTab)
            result.addResult(resLupStatus, msg_only=True)
            if resLupStatus.isOk():
                setLookup_goldNot_ids = resLupStatus.getData("not-gold-ids")
                setLookup_gold_ids = resLupStatus.getData("gold-ids")
                setLookup_used_ids = resLupStatus.getData("used-ids")
                setLookup_usedNot_ids = resLupStatus.getData("not-used-ids")
                setLookup_used_goldNot_ids = resLupStatus.getData("used-not-gold-ids")
                setLookup_usedNot_goldNot_ids = resLupStatus.getData("not-used-not-gold-ids")
                setLookup_dataset_AccessPossible_all = resLupStatus.getData("dataset-access-possible-ids")
                setLookup_dataset_AccessRestricted_all = resLupStatus.getData("dataset-access-restricted-ids")
                setLookup_dataset_UsedAccessPossible_all = resLupStatus.getData("dataset-used-access-possible-ids")
                setLookup_dataset_UsedAccessNotPossible_all = resLupStatus.getData("dataset-used-access-restricted-ids")
                setLookup_dataset_UsedAccessPossibleGoldNot_all = resLupStatus.getData("dataset-used-not-gold-access-possible-ids")
                setLookup_gear_double = resLupStatus.getData("gear-double-ids")
                setDoi_Ok = set()
                setDoi_NotOk = set()
                setDoi_UsedOk = set()
                setDoi_UsedNotOk = set()
                setDoi_UsedNotOkAccessPossible = set()
                setDoi_NotPangaea = set()
                packDoi = self.__model.memory_load("doi")
                if packDoi is not None:
                    try:
                        dicDoi_Ok = packDoi.get("ok")
                        dicDoi_Err = packDoi.get("err")
                        dicDoi_No = packDoi.get("no")
                        setDoi_Ok = set(dicDoi_Ok.keys())
                        setDoi_NotOk = set(dicDoi_Err.keys()).union(set(dicDoi_No.keys()))
                        setDoi_UsedOk = set(dicDoi_Ok.keys()).intersection(setLookup_used_ids)
                        setDoi_UsedNotOk = set(dicDoi_Err.keys()).union(set(dicDoi_No.keys())).intersection(setLookup_used_ids)
                        setDoi_UsedNotOkAccessPossible = set(dicDoi_Err.keys()).union(set(dicDoi_No.keys())).intersection(setLookup_dataset_UsedAccessPossible_all)
                        setDoi_NotPangaea = packDoi.get("cluster-not-pangaea-ids")
                    except:
                        pass
                if (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, True, None, None, None, None):
                    lstIds = list(setLookup_gold_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, False, None, None, None, None):
                    lstIds = list(setLookup_goldNot_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, None, None, None, None):
                    lstIds = list(setLookup_used_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (False, None, None, None, None, None):
                    lstIds = list(setLookup_usedNot_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, False, None, None, None, None):
                    lstIds = list(setLookup_used_goldNot_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (False, False, None, None, None, None):
                    lstIds = list(setLookup_usedNot_goldNot_ids)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, True, None, None, None):
                    lstIds = list(setLookup_dataset_AccessPossible_all)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, False, None, None, None):
                    lstIds = list(setLookup_dataset_AccessRestricted_all)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, True, None, None, None):
                    lstIds = list(setLookup_dataset_UsedAccessPossible_all)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, False, None, None, None):
                    lstIds = list(setLookup_dataset_UsedAccessNotPossible_all)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, False, True, None, None, None):
                    lstIds = list(setLookup_dataset_UsedAccessPossibleGoldNot_all)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, None, None, None, True):
                    lstIds = list(setLookup_gear_double)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, None, True, None, None):
                    lstIds = list(setDoi_Ok)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, None, False, None, None):
                    lstIds = list(setDoi_NotOk)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, None, True, None, None):
                    lstIds = list(setDoi_UsedOk)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, None, False, None, None):
                    lstIds = list(setDoi_UsedNotOk)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (True, None, True, False, None, None):
                    lstIds = list(setDoi_UsedNotOkAccessPossible)
                elif (self.__bUsed, self.__bGold, self.__bAccess, self.__bDoi, self.__bDoiPangaea, self.__bGear_double) == (None, None, None, True, False, None):
                    lstIds = list(setDoi_NotPangaea)

        try:
            # Hole Daten aus der Datenbank
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()
            workflow(top="Show Data", sub="Getting table "+self.__sTab)
            resTable = self.__model.subsystem_getDb().table_getWholeTable(self.__sTab, ids=lstIds, cursor=cursor, select=self.__lstSelect)
            lstHeader = resTable.getData("header")
            lstContent = resTable.getData("content")
            if resTable.isErr():
                result.addResult(resTable)
                raise NameError("Problems getting Table from database")
            # Header anpassen (DB->EXCEL)
            if self.__sType == "lookup":
                # Lookup
                lstTmp = []
                for sHead in lstHeader:
                    sExcel, sSheet = self.__map.lookup_db2excel(self.__sTab, sHead)
                    if sSheet is not None:
                        lstTmp.append(sSheet)
                    else:
                        lstTmp.append(sHead)
                lstHeader = lstTmp.copy()
            elif self.__sType == "taxon":
                # Taxon
                lstTmp = []
                for sHead in lstHeader:
                    sExcel, sSheet = self.__map.taxon_db2excel(self.__sTab, sHead)
                    if sSheet is not None:
                        lstTmp.append(sSheet)
                    else:
                        lstTmp.append(sHead)
                lstHeader = lstTmp.copy()
            elif self.__sType == "ingest":
                # Ingest
                lstTmp = []
                for sHead in lstHeader:
                    if sHead in ["ingest_id", "cruise_id", "station_id", "sample_id", "subset_id","biota_id"]:
                        lstTmp.append("<"+sHead.replace("_","").upper()+">")
                    else:
                        sExcel, sSheet = self.__map.ingest_db2excel(self.__sTab, sHead)
                        if sSheet is not None:
                            lstTmp.append(sSheet)
                        else:
                            lstTmp.append(sHead)
                lstHeader = lstTmp.copy()

            # Entferne Ids
            if self.__bNoIds:
                # Entferne gesamte Spalte "id" (Header und Inhalt)
                # Inhalt
                progress("on")
                workflow(top="Removing ids")
                iRowMax = len(lstContent)
                for iRow, row in enumerate(lstContent):
                    status(str(iRow) + "/" + str(iRowMax))
                    progress(iRow, iRowMax, 50, proz=True, abs=False)  # Fortschritt
                    for iCol, col in enumerate(row):
                        if lstHeader[iCol] == "id":
                            del lstContent[iRow][iCol]
                # Header
                lstHeader.remove("id")

            # Datenbank schliessen
            cursor.close()
            conn.close()

        except Exception as ex:
            result.addErr("Error during getting data from database table", add=str(ex), kat="Show Table", sys="Admin")
            progress("off")
            workflow()
            self.sendResult(result)
            return

        # Verpacke in Multisheet
        ms = MultiSheet(name=self.__sTab)
        ms.setSheet(self.__sTab, lstHeader, lstContent)

        # Antwort
        result.addData("tabName", self.__sTab)
        result.addData("diaTitle", self.__sTitle)
        result.addData("noids", self.__bNoIds)
        result.addData("edit", self.__bEdit)
        result.addData("del_entry", self.__bDel)
        result.addData("gold_edit", self.__bGoldEdit)
        result.addData("save", self.__bSave)
        result.addData("human", self.__bHuman)
        result.addData("id", self.__iFokusId)
        result.addData("gold", self.__bGold)
        result.addData("used", self.__bUsed)
        result.addData("access", self.__bAccess)
        result.addData("gear_double", self.__bGear_double)
        result.addData("doi", self.__bDoi)
        result.addData("doi_pangaea", self.__bDoiPangaea)
        result.addData("multisheet", ms)
        progress("off")
        workflow()
        self.sendResult(result)

class Worker_SaveMap(Worker):
    """
    KLasse speichert Map eine Datei ab
    """

    def __init__(self, model, sFilename):
        super().__init__()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__model = model
        self.__sFile = sFilename

    def run(self):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        progress("on-pulse")
        workflow(top="saving map")
        result = Result(name="Save map to file")

        try:
            # Speichere Karte
            copyfile(self.__model.conf_getMap(), self.__sFile)
        except Exception as ex:
            result.addErr("Error during saving map to file", add=str(ex), kat="Write Map To File", sys="Admin")

        progress("off")
        workflow()
        self.sendResult(result)


class Worker_VisUpdate(Worker):
    """
    Der Visuelle Explorer wird geupdated
    """

    def __init__(self, model, dicOpt):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__dicOpt = dicOpt

    def run(self):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status

        iSubsets = 0  # Anzahl der anzuzeigenden Samples
        iSamplesWarn = 1000  # Grenze ab der gewarnt werden soll
        bSamplesWarn = False  # Warnung bei zu vielen Samples

        # --- Helper
        # Json -> List LonLat, Popup
        def getLonLat(geoJson):
            try:
                lstPoints = []
                lstPopups = []
                lstFeatures = geoJson["features"] if "features" in geoJson else []
                if lstFeatures is None:
                    lstFeatures = []
                    # Progress
                for i, dicFeature in enumerate(lstFeatures):
                    # Progress (bei vielen Samples)
                    if len(lstFeatures) > iSamplesWarn:
                        progress(len(lstFeatures)+i, 2*len(lstFeatures), 100, cli=False)
                        status("Building marker: " + str(i + 1) + "/" + str(len(lstFeatures)))
                    # Hole Daten aus GeoJson
                    iSpatialPrecision = dicFeature["properties"]["precision"]
                    fLat = dicFeature["geometry"]["coordinates"][1]
                    fLon = dicFeature["geometry"]["coordinates"][0]

                    try:
                        sImage = dicFeature["properties"]["bio_image"]
                    except:
                        sImage = None
                    sDoi = dicFeature["properties"]["dataset_doi"]

                    # Wende Precission auf lat/lon an
                    sLon = PrettyText.geo_reformat(fLon, precision=iSpatialPrecision, dummy_brace=False)
                    sLat = PrettyText.geo_reformat(fLat, precision=iSpatialPrecision, dummy_brace=False)

                    # Enferne Anführungszeichen
                    sDataset = str(dicFeature["properties"]["dataset_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sCruise = str(dicFeature["properties"]["cruise_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sStation = str(dicFeature["properties"]["station_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sSample = str(dicFeature["properties"]["sample_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sSubSample = str(dicFeature["properties"]["subsample_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sSubset = str(dicFeature["properties"]["subset_name"]).replace("'","&apos;").replace("\"","&quot;")
                    sSampleStartOn = dicFeature["properties"]["sample_date_start"]

                    # Coordinates
                    lstPoints.append((fLat, fLon))
                    # Popup
                    sPopup = \
"""
<html>
<head>
<style>
td {padding-right: 5px; text-align: left;}
</style>
</head>
<body><table>"""
                    sPopup += f'<tr><td><b>Dataset</b></td><td><span style="white-space: nowrap;">{sDataset}</span></td></tr>'
                    sPopup += f'<tr><td><b>Cruise</b></td><td><span style="white-space: nowrap;">{sCruise}</span></td></tr>'
                    sPopup += f'<tr><td><b>Station</b></td><td><span style="white-space: nowrap;">{sStation}</span></td></tr>'
                    sPopup += f'<tr><td><b>Sample</b></td><td><span style="white-space: nowrap;">{sSample}</span></td></tr>'
                    sPopup += f'<tr><td><b>SubSample</b></td><td><span style="white-space: nowrap;">{sSubSample}</span></td></tr>'
                    sPopup += f'<tr><td><b>Subset</b></td><td><span style="white-space: nowrap;">{sSubset}</span></td></tr>'
                    sPopup += f'<tr><td><b>Date</b></td><td><span style="white-space: nowrap;">{sSampleStartOn}</span></td></tr>'
                    sPopup += f'<tr><td><b>Location</b></td><td><span style="white-space: nowrap;">lat:{sLat} lon:{sLon}</span></td></tr>'
                    sPopup += f'<tr><td><b>Precision</b></td><td><span style="white-space: nowrap;">{iSpatialPrecision} decimal digits</span></td></tr>'
                    if sImage is not None:
                        sPopup += f'<tr><td><b>Image</b></td><td><img width="200" src="{sImage}"></td></tr>'
                    if sDoi not in [None, ""]:
                        sDoi_show = sDoi.replace("https://", "")
                        sDoi_show = sDoi_show.replace("http://", "")
                        sPopup += f'<tr><td><b>DOI</b></td><td><span style="white-space: nowrap;"><a href={sDoi}>{sDoi_show}</a></span></td></tr>'
                    sPopup += "</table></body></html>"
                    lstPopups.append(sPopup)
                    if self.isAbort():
                        result.addErr("Aborted by user ...", kat="Visualisation", sys="Admin")
                        return ([], [])
                return (lstPoints, lstPopups)
            except:
                pn("* Internal error: " + str(traceback.print_exc()))
                return ([], [])

        # Helper: Test format markers
        style1 = lambda x: {
            'radius': 4,
            'fillColor': '#ff0000',
            'color': '#900',
            'weight': 1,
            'opacity': 1,
            'fillOpacity': 0.8
        }
        def style2(feature):
            return {
                'fillColor': 'green'
            }

        # Helper: create empty Map
        def createEmptyMap():
            if not bWorldWrap:
                map = folium.Map(no_wrap=True, control_scale=True, tiles=None)
                map.zoom_start = 1
            else:
                map = folium.Map(control_scale=True, tiles=None)
            # Kacheln
            folium.TileLayer('openstreetmap', name="Open Street Map").add_to(map)
            folium.TileLayer(
                tiles='https://server.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/{z}/{y}/{x}',
                name="National Geography World Map (Esri)",
                maxZoom=16,
                attr='Tiles &copy; Esri &mdash; National Geographic, Esri, DeLorme, NAVTEQ, UNEP-WCMC, USGS, NASA, ESA, METI, NRCAN, GEBCO, NOAA, iPC'
                ).add_to(map)
            folium.TileLayer(
                tiles='https://server.arcgisonline.com/ArcGIS/rest/services/Ocean_Basemap/MapServer/tile/{z}/{y}/{x}',
                name="Ocean Basemap (Esri)",
                attr='Tiles &copy; Esri &mdash; Sources: GEBCO, NOAA, CHS, OSU, UNH, CSUMB, National Geographic, DeLorme, NAVTEQ, and Esri',
                maxZoom=13
                ).add_to(map)
            folium.TileLayer(
                tiles='https://basemap.nationalmap.gov/arcgis/rest/services/USGSTopo/MapServer/tile/{z}/{y}/{x}',
                name="The National Map (USGS)",
                attr='Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>',
                maxZoom=16
                ).add_to(map)
            return map

        # ----

        result = Result(name="Update Visual Explorer")
        progress("on-abort")

        sMarkerType = self.__dicOpt["map_marker_type"]
        sProjectionType = self.__dicOpt["map_projection_type"]
        dicInputSelector = self.__dicOpt["selector"]
        bForceDraw = self.__dicOpt["force_draw"]
        bWorldWrap = True if self.__dicOpt["world_wrap"] != 0 else False

        # --- Selector einstellen
        progress(1, 4, 4, cli=False)
        workflow(top="Visual Explorer", sub="Updating selector")
        self.__model.ui_selector_getSelector().update(dicInputSelector)
        if self.__model.ui_selector_getSelector().isErr():
            pn("* Error in selector configuration")
            pn(self.__model.ui_selector_getSelector().getResult().getErrWarnInfoString(short=True))
            result.addResult(self.__model.ui_selector_getSelector().getResult())
        if self.isAbort():
            result.addErr("Aborted by user ...", kat="Abort", sys="Admin")

        # --- Statistik Selector Subset holen
        if result.isOk():
            workflow(top="Visual Explorer", sub="Updating statistics")
            progress(2, 4, 4, cli=False)
            resultStat = self.__model.subsystem_getDb().selector_getSubsetStatsFromDb()
            if resultStat.isErr():
                pn("* Internal error: " + resultStat.getErrWarnInfoString(short=True))
                result.addResult(resultStat, msg_only=True)
            else:
                dicStatistics = resultStat.getData("stat")
                self.__model.ui_selector_setStatistics(dicStatistics)
                if self.isAbort():
                    result.addErr("Aborted by user ...", kat="Abort", sys="Admin")

        # --- Erzeuge die Map
        map = createEmptyMap()

        # --- Regionen
        # Sammle alle aktiven Regionen ein
        dicRegions = self.__model.ui_selector_getSelector().getEntries("region")
        lstRegions = []
        for sRegion, (bActive, bEffective) in dicRegions.items():
            if bActive and bEffective:
                lstRegions.append(sRegion)
        # Trage sie in Map ein
        iParts = len(lstRegions)
        colGradient = ColorGradient(Colors.getColor("yellow"), Colors.getColor("purple"), iParts)
        style = lambda sCol: lambda feature: dict(weight=1) # color=sCol, fillColor=sCol
        for iNr, sReg in enumerate(lstRegions):
            sCol = colGradient.getHtml(iNr)
            shape, iSize = self.__model.subsystem_getRegions().regions_getShape(sReg)
            if shape is not None:
                folium.GeoJson(shape, name="Region: "+sReg, tooltip=sReg, style_function=style(sCol)).add_to(map)

        # --- Handbremse bei zu vielen Punkten ziehen
        if result.isOk():
            # Hole Anzahl samples
            if "subsets" in dicStatistics:
                iSubsets = dicStatistics["subsets"]
            if iSubsets > iSamplesWarn:
                bSamplesWarn = True
            result.addData("samples_warn", bSamplesWarn)
            if bSamplesWarn and not bForceDraw:
                folium.LayerControl().add_to(map)
                map.save(self.__model.conf_getMap())
                self.sendResult(result)
                return

        # --- Erzeuge Map (als Datei)
        if result.isOk():

            sSamplesWarn = ""
            if iSubsets > iSamplesWarn:
                sSamplesWarn = " (way too much data, please be patient ...)"

            # Info
            workflow(top="Visual Explorer", sub="Getting map data")
            resultGeo = self.__model.subsystem_getDb().selector_getGeoJsonMap()

            # Draw map
            workflow(top="Visual Explorer", sub="Building map")
            if resultGeo.isOk():
                geoJson = resultGeo.getData("geojson")
                if geoJson is not None and "features" in geoJson and geoJson["features"] is not None:
                    # Markers
                    if sMarkerType == "clustered pins":
                        # ClusterMarker
                        lstPoints, lstPopups = getLonLat(geoJson)
                        MarkerCluster(name="Samples", locations=lstPoints, popups=lstPopups).add_to(map)
                    elif sMarkerType == "pins":
                        # Normale Pins
                        # Tooltips
                        ttip = folium.GeoJsonTooltip(
                            fields=["dataset_name", "cruise_name", "station_name", "subset_name", "sample_date_start"],
                            aliases=["Dataset", "Cruise", "Station", "Subset", "Date"],
                            sticky=True,
                            style=('background-color: white; color: black;font-family:courier new; font-size: 9px; padding: 0 px;')
                        )
                        # GeoJson -> Marker
                        points_gjson = folium.features.GeoJson(geoJson, tooltip=ttip, name="Samples")
                        points_gjson.add_to(map)

            else:
                result.addResult(resultGeo, msg_only=True)

        # Layer Control
        folium.LayerControl().add_to(map)

        # Save Map
        if result.isOk():
            progress("on-pulse")
            workflow(top="Visual Explorer", sub="Saving map")
            map.save(self.__model.conf_getMap())

        self.sendResult(result)

class Worker_ServerReconnect(Worker):
    """
    Reconnect Server
    """
    def __init__(self, model, **kwargs):
        super().__init__()
        self.__start = kwargs.get("start")
        self.__status = kwargs.get("status")
        self.__model = model
        self.__sp = model.ui_getSmartprinter()

    def abort(self):
        pass

    def run(self):
        res  = self.__model.subsystem_getConnector().reconnect(start=self.__start, status=self.__status)
        self.sendResult(res)

# -------------------------------------------------------------
# Etc
# -------------------------------------------------------------

class Worker_Generic(Worker):
    """
    Unbenutzt: Spezialisierter Thread Container: WorkflowMultisheetTesting
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()

    def abort(self, bAbort=True):
        pass
        #self.__model.getObjektIhrerWahl().abort(bAbort)

    def run(self):

        workflow = self.__smartprinter.workflow
        progress = self.__smartprinter.progress

        lstAktionen = ["test"]

        def errMsg(sInfo="-"):
            return "Internal error during starting workflow: " + str(sInfo)

        # Werte Aktion aus
        sAktion = self.getInit("aktion")
        if sAktion is None:
            sErr = errMsg("Missing action")
            self.sendResult(Result(err=sErr))
            return
        elif sAktion not in lstAktionen:
            sErr = errMsg("Unknown action")
            self.sendResult(Result(err=sErr))
            return

        # Starte Aktion
        if sAktion == "test":
            # Init
            try:
                print("Test")
                self.sendResult(Result("Test"))
                return
            except Exception as ex:
                sErr = errMsg(ex)
                self.sendResult(Result(err=sErr))
        else:
            sErr = errMsg("Internal Error: Wrong Parameters for worker")
            self.sendResult(Result(err=sErr))
