# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools, ExcelNavigator
from tools.text import PrettyText
from tools.parallel import Stoppable
from logik.converter_antarctic import ConverterAntarctic
from logik.converter_bsh import ConverterBsh
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.geocruncher import GeoCruncher
from pathlib import Path

# -------------------------------------------------------------
# Import: Converter: Konvertiere Excel ins eigene Format
# -------------------------------------------------------------

class Converter(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

        self.setReady(True)
        self.setWorking(True)

        # Converter
        self.__bsh = ConverterBsh(model)
        self.__ant = ConverterAntarctic(model)

        # Init
        self.__ant.init() # Das geht schon offline

    def init(self, sConverter=None):
        res = Result("Starting converter")
        if sConverter is None:
            resBsh = self.__bsh.init()
            resAnt = self.__ant.init()
            res.addResult(resBsh, msg_only=True)
            res.addResult(resAnt, msg_only=True)
        elif sConverter == "bsh":
            resBsh = self.__bsh.init()
            res.addResult(resBsh, msg_only=True)
        elif sConverter == "antmat":
            resAnt = self.__ant.init()
            res.addResult(resAnt, msg_only=True)
        return res

    def isWorking_submodule(self, sConverter):
        if sConverter == "bsh":
            return self.__bsh.isWorking()
        elif sConverter == "antmat":
            self.__ant.isWorking()

    def isReady_submodule(self, sConverter):
        if sConverter == "bsh":
            return self.__bsh.isReady()
        elif sConverter == "antmat":
            self.__ant.isReady()

    def setReady_submodule(self, sConverter, bBool):
        if sConverter == "bsh":
            return self.__bsh.setReady(bBool)
        elif sConverter == "antmat":
            self.__ant.setReady(bBool)

    # Info
    def info(self, sConverter=None, **kwargs):
        bHtml = kwargs["html"] if "html" in kwargs else False
        if sConverter is None:
            sOut = "*** Status of Converter System\n\n"
            sOut += "Sub modules (2x): Antarctic, BSH\n\n"
            sOut += "---------------------------------------------\n"
            sOut += "Antarctic (" + ("working" if self.__ant.isReady() else "not working") + ")\n"
            sOut += "---------------------------------------------\n\n"
            sOut += self.__ant.info(bHtml) +"\n"
            sOut += "---------------------------------------------\n"
            sOut += "BSH (" + ("working" if self.__bsh.isReady() else "not working ... connect to db") +")\n"
            sOut += "---------------------------------------------\n\n"
            sOut += self.__bsh.info(bHtml) +"\n"
            return sOut
        elif sConverter == "bsh":
            return self.__bsh.info(bHtml)
        elif sConverter == "antmat":
            return self.__ant.info(bHtml)

    # Converter: Export
    def start_converter_export(self, sWf, **kwargs):
        """
        Hier werden nur die SQL Befehle zurueckgeliefert
        Args:
            sWf (str): Workflow
        Kwargs:
            -
        Return:
            Result
                data (odic)
                stat (odic)
                log (str)
        """
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        res = Result(name="Excel Converter: Export")

        # Log
        sLog = ""
        if sWf == "mar_trans":
            # BSH
            resBSH = self.__bsh.convert_export_create_sql()
            res.addResult(resBSH)
        else:
            res.addErr("Unknown Conversiontype", add=str(sWf), kat="Configuration", sys="Converter")
        return res


    # Converter Import
    def start_converter_import(self, wb, **kwargs):
        """
        Args:
            wb (workbook): Workbook
        Kwargs:
            result_in: Result in
            ingest_wf (str): Ingest Workflow
            batch (str): Batch Info die weitergegeben wird
        Return:
            Result
                log (str)
        """

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose

        res = Result(name="Excel Converte: Import")

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        sIngestWf = kwargs["ingest_wf"] if "ingest_wf" in kwargs else None

        # Parameter Test
        if wb is None:
            res.addErr("No Excel Workbook given", kat="I/O", sys="Converter")

        # Log
        sLog = ""
        if sIngestWf == "mar_trans":
            # BSH
            resBSH = self.__bsh.convert_import(wb, result_in=resultIn)
            res.addResult(resBSH)

        elif sIngestWf == "ant_mat":
            # Antarktis
            resAntarcticMatrix = self.__ant.convert_import_antmat2mat(wb, result_in=resultIn)
            res.addResult(resAntarcticMatrix)
        else:
            res.addErr("Unknown Conversiontype", add=str(sIngestWf), kat="Configuration", sys="Converter")
        return res

    # Converter Export: Erzeuge Ingestdaten in Excel Sheet
    def create_converter_export_excel(self, sConverter, **kwargs):
        """
        Args:
            sConverter (str)
        Kwargs:
            wb (workbench)
            data (dic): Ingest Data
        Return:
            Result:
                wb (workbench)
        """
        res = Result("Create Converter Excel Workbench")
        wb = kwargs.get("wb")
        dicData = kwargs.get("data")
        if sConverter == "mar_trans":
            # BSH
            res = self.__bsh.convert_export_create_excel_wb(wb=wb, data=dicData)
        return res

    # Converter Import: Schreibe Template
    def write_converter_import_template(self, sWf, sFile):
        """
        Args:
            sWf (str): Workflow
            sFile (str): File
        Return:
            Result
        """

        res = Result("Write Converter Template")
        if sWf == "mar_trans":
            # BSH
            res = self.__bsh.convert_export_write_excel_template(sFile)
        elif sWf == "ant_mat":
            # Antarktis
            res = self.__ant.template_import_antmat2mat(sFile)
        return res


