# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict
import os, sys, csv, getpass, platform, datetime, time, copy, re, traceback
from difflib import SequenceMatcher
import openpyxl.styles.numbers
import warnings
from tools.coding import Result, WorkingObject
from tools.fileio import FileTools
from tools.excel import ExcelNavigator, ExcelTools
from tools.text import PrettyText
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.misc import JsonConverter
from tools.fileio import FileTools
from tools.geocruncher import GeoCruncher
from psycopg2 import sql
from tools.taxon import Taxon

"""
TODO: 
- Uniqueness of Combination of Aphia Id, Lifestage, Sieve and Specification (in DB - Unique funktioiert nicht, da zwei NULL als nicht gleich angesehen werden)
- PK der nicht OBL ist oder nicht PK/FK gleichzeitig -> PK nicht erzeugen (dann ist es so gewollt) -> z.B. Biota
"""

# Interne Datenstruktur für Excel-Tabelle
class Data():

    # Row
    class Row:
        def __init__(self, sSheet, sPk, sFk, odicContent, sExtra):
            self.__sSheet = sSheet
            self.__sPk = sPk
            self.__sFk = sFk
            self.__odicContent = odicContent
            self.__sExtra = sExtra
        def __str__(self):
            sOut = self.__sSheet +": pk[" + str(self.__sPk) + "] fk[" + str(self.__sFk) + "] content[" + PrettyText.dic2str(self.__odicContent) + "] extra[" + str(self.__sExtra) + "]"
            return sOut
        def getHeader(self):
            return list(self.__odicContent.keys())
        def getHeadIdx(self, sHead):
            try:
                return list(self.__odicContent.keys()).index(sHead)
            except:
                return None
        def getSheet(self):
            return self.__sSheet
        def getPk(self):
            return self.__sPk
        def getFk(self):
            return self.__sFk
        def get(self, sKey):
            if sKey in self.__odicContent:
                return self.__odicContent[sKey]
        def getAll(self):
            return self.__odicContent
        def getExtra(self):
            return self.__sExtra

    # Konstruktor
    def __init__(self, odicWb):
        self.__resWorking = Result("Init ingest data structure")
        self.__resLog = Result("Logging ingest data structure")

        self.__odicHeader = OrderedDict()
        self.__odicRows = OrderedDict()
        self.__dicPk = {}
        self.__dicFk = {}

        self.__sIterSheetName = None
        self.__odicIterSheet = None

        try:
            for sSheet, lstHeader in odicWb.items():
                self.__odicHeader[sSheet] = lstHeader
                self.__odicRows[sSheet] = OrderedDict()
                self.__dicPk[sSheet] = set()
                self.__dicFk[sSheet] = set()
        except Exception as ex:
            self.__resWorking.addErr(str(ex), kat="Ingest Data Structure", sys="Ingestor")

    def __str__(self):
        sOut = "*** Status of ingest data structure\n\n"
        if not self.isWorking():
            sOut += "not working\n"
            sOut += self.getWorkingLog().getErrWarnInfoString()
            return sOut
        sOut +="--- Header\n"
        for sSheet, lstHeader in self.__odicHeader.items():
            sOut += sSheet +": "
            lstOutHeader = []
            for sHead in lstHeader:
                iIdx = self.getHeadIdx(sSheet, sHead)
                lstOutHeader.append(sHead +" ("+str(iIdx)+")")
            sOut += PrettyText.lst2strC(lstOutHeader) +"\n"
        if self.getRowNumber() > 0:
            sOut += "--- Content\n"
            for sSheet, odicSheet in self.__odicRows.items():
                if len(odicSheet) > 0:
                    sOut += "* " + sSheet + " ("+str(self.getRowNumber(sSheet))+")"+"\n"
                    lstTable = ["Fk", "Pk"] + self.getHeader(sSheet)
                    lstTable.append("Extra Info")
                    lstTable = [lstTable]
                    for iRow, (sPk, (sFk, lstRow, sInfo)) in enumerate(odicSheet.items()):
                        if iRow < 50:
                            lstTmp = [sFk, sPk] + lstRow
                            lstTmp.append(sInfo)
                            lstTable.append(lstTmp)
                    sOut += PrettyText.table(lstTable) + "\n"
                if len(odicSheet) > 50:
                    sOut += "... cutting of remaining "+str(len(odicSheet)-50)+" rows ...\n"
        if self.__sIterSheetName is not None:
            sOut += "* Coding stuff\n"
            sOut += "- Actually set Iterator sheet: " + str(self.__sIterSheetName) + "\n"
        if self.__resLog.isAnything():
            sOut += "* Log\n"
            sOut += self.__resLog.getErrWarnInfoString()+"\n"
        return sOut

    def __iter__(self):
        return self

    def __call__(self, **kwargs):
        sSheet = kwargs["iter_sheet"] if "iter_sheet" in kwargs else None
        if sSheet not in self.getSheets():
            self.__sIterSheetName = None
            self.__odicIterSheet = None
        else:
            self.__sIterSheetName = sSheet
            self.__odicIterSheet = iter(self.__odicRows[sSheet].items())
        return self

    def __next__(self):
        if self.__odicIterSheet is None:
            raise StopIteration
        sPk, (sFk, lstRow, sInfo) = next(self.__odicIterSheet)
        dicRow = dict(zip(self.getHeader(self.__sIterSheetName), lstRow))
        return Data.Row(self.__sIterSheetName, sPk, sFk, dicRow, sInfo)

    # ---

    def isWorking(self):
        return self.__resWorking.isOk()

    def getWorkingLog(self):
        return self.__resWorking

    def isErr(self):
        return self.__resLog.isErr()

    def getLog(self):
        return self.__resLog.getErrWarnInfoString()

    def getSheets(self):
        if not self.isWorking():
            return None
        return self.__odicHeader.keys()

    def getHeader(self, sSheet):
        if not self.isWorking():
            return None
        try:
            return self.__odicHeader[sSheet]
        except:
            return None

    def getHeadIdx(self, sSheet, sHead):
        try:
            return self.__odicHeader[sSheet].index(sHead)
        except:
            return None

    def getPks(self, sSheet):
        try:
            return self.__dicPk[sSheet]
        except:
            return None

    def getFks(self, sSheet):
        try:
            return self.__dicFk[sSheet]
        except:
            return None

    def addRow(self, sSheet, sPk, sFk, dicRow, sInfo=None):
        res = Result("Add row to ingest data structure")
        if not self.isWorking():
            res.addErr("Not working", kat="Ingest Data Structure", sys="Ingestor")
            return res
        try:
            # TODO: Fehler! Nur Biota hat keinen eindeutigen PK
            if sPk in self.__odicRows[sSheet]:
                res.addErr("PK already used", add=str(sPk), sheer=sSheet, kat="PK and FK", sys="Ingestor")
            lstHeader = self.getHeader(sSheet)
            lstRow = []
            for k, v in dicRow.items():
                if k not in lstHeader:
                    self.__resLog.addErr("Column not known", sheet=sSheet, col=str(k), kat="Header", sys="Ingestor")
            for sHead in lstHeader:
                if sHead in dicRow:
                    lstRow.append(dicRow[sHead])
                else:
                    lstRow.append(None)
                    res.addInfo("Adding row to sheet: column not set", col=str(k), sheet=sSheet, kat="Header", sys="Ingestor")
            self.__odicRows[sSheet][sPk] = (sFk, lstRow, sInfo)
            self.__dicPk[sSheet].add(sPk)
            self.__dicFk[sSheet].add(sFk)
        except Exception as ex:
            res.addErr(str(ex), kat="Internal", sys="Ingestor")
        if res.isAnything():
            self.__resLog.addResult(res)
        return res

    def getRowNumber(self, sSheet=None):
        if not self.isWorking():
            return None
        if sSheet is not None:
            return len(self.__odicRows[sSheet])
        else:
            iRows = 0
            for sSheet, odicSheet in self.__odicRows.items():
                iRows += len(odicSheet)
            return iRows

    def getIterSheet(self):
        return self.__sIterSheetName

# -------------------------------------
# Import: Ingestor: Excel Memory -> Ingest Data
# -------------------------------------

class Ingestor(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)
        self.__model = model
        self.__print = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__lup = model.subsystem_getLookup()
        self.__comment = model.conf_getDbSchema()

        # Fehlercodes in DB notices
        self.__dicDbMessageCodes = {
            "[warn000]": "Warning dummy 000",
            "[info000]": "Information dummy 000"}

        # Formattests
        self.__lstExcelDateFormat = []
        self.__lstExcelTimeFormat = []
        lstTimeTmp = ["hh:mm:ss", "h:mm:ss"]  # Erwartetes Standard Zeitformat der Excel-Datei
        lstDateTmp = ["yyyy/mm/dd", "yyyy-mm-dd"]  # Erwartetes Standard Datumsformat der Excel-Datei
        lstTmpPrae = ["", "[$-809]", "[$-409]"]
        lstTmpPost = ["", ";@"]
        for sFmt in lstDateTmp:
            for sPrae in lstTmpPrae:
                for sPost in lstTmpPost:
                    self.__lstExcelDateFormat.append(sPrae + sFmt + sPost)
        for sFmt in lstTimeTmp:
            for sPrae in lstTmpPrae:
                for sPost in lstTmpPost:
                    self.__lstExcelTimeFormat.append(sPrae + sFmt + sPost)

        # GeoCruncher
        self.__geoCruncher = GeoCruncher()

        # Rerset all
        self.__initData()

    # Info
    def info(self):
        sOut = "*** Status of Ingestor System\n\n"
        sOut += "* Testing\n"
        sOut += "--- Taxa\n"
        sOut += "Biota.ScientificName matching ratio: " + str(self.__model.subsystem_getTaxonWizard().getScientificNameMatchRatio()) + "\n"
        sOut += "Special Taxa: " + PrettyText.dic2str(self.__model.subsystem_getTaxonWizard().getAllSpecialTaxa()) + "\n"
        sOut += "--- Date/Time\n"
        sOut += "Allowed date formats: " + PrettyText.lst2strC(self.__lstExcelDateFormat) + "\n"
        sOut += "Allowed time formats: " + PrettyText.lst2strC(self.__lstExcelTimeFormat) + "\n"
        return sOut

    # Reset
    def __initData(self):
        self.__resultAll = None # Result der gesamten Pipeline
        self.__conn = None
        self.__cursor = None
        self.__wb = None
        self.__sWf = None
        self.__data = None
        self.__lstSheetsSelected = None
        self.__lstSheetsUse = None
        self.__dicLookup = {}
        self.__dicPkIds = {}
        self.__sFile = None
        self.__dicBiotaAll = {}  # Alle Benutzten Biota in Ingest (aid:scientificname)
        self.__setBiotaAll_check_problem_all = set () # Probleme bei Biota
        self.__setBiotaAll_check_problem__no_name = set()
        self.__setBiotaAll_check_problem_no_aname_and_accepted = set()
        self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted = set()
        self.__setBiotaAll_check_problem_no_name_at_all = set()
        self.__setBiotaAll_check_problem_status = set()
        self.__setBiotaAll_check_problem_aphia = set()
        self.__setBiotaAll_check_problem_limbo = set()
        self.__iBiotaAll_check_name_diff = 0
        self.__iBiotaAll_check_name_diff_special = 0
        self.__setBiotaAll_check_accepted = set()
        self.__setBiotaAll_check_unaccepted = set()
        self.__setBiotaAll_check_special = set()
        self.__lstBiotaAll_check_misspellingNames = []
        self.__lstBiotaAll_check_misspellingNames_special = []
        self.__lstBiotaAll_check_notFound = []

        self.__dicBiotaNew = {}  # Unbekannte Biota in Ingest (aid:scientificname)
        self.__cntRecNew = Counter()  # Neue Rec in DB
        self.__cntRecExt = Counter()  # Extended Rec in Db
        self.__cntRecExt_diff = Counter()  # Diff bei Extended Rec in Db
        self.__dicRecExt_diff = {} # Diff bei Extended Rec in Db (Info)
        self.__iIngestId = None  # Ingest Id
        self.__setIngestDepend = set()  # Abhängig von folgenden Ingests
        self.__dicIngestDependFiles = {}
        self.__setDatasetDepend = set()  # Datasets mit gleichen Abhängigkeiten
        self.__setCruiseExtended = set()  # Erweiterungen
        self.__setStationExtended = set()  # Erweiterungen
        self.__setSampleExtended = set()  # Erweiterungen
        self.__setSubSampleExtended = set()  # Erweiterungen
        # Init Lookup
        for sSheet in self.__map.ingest_getExcelSheets(self.__sWf):
            if self.__map.ingest_hasLookup(sSheet):
                self.__dicLookup[sSheet] = {}
                for sCol in self.__map.ingest_getLookupCols(sSheet):
                    self.__dicLookup[sSheet][sCol] = set()

    # Getter
    def __getSheets(self):
        return self.__odicHeader.keys()

    def __getData(self):
        return self.__data

    # Statistik
    def __createStat(self):
        sLogStat = ""
        sLogStat += "--- Ingest ---\n"
        sLogStat += "Date: " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") + "\n"
        sLogStat += "User: " + str(getpass.getuser()) + "\n"
        sLogStat += "OS: " + str(platform.system()) + " " + str(platform.release()) + "\n"
        sLogStat += "Host: " + str(platform.node()) + "\n"
        sLogStat += "File: " + FileTools.getFileWithoutPath(self.__sFile) + "\n"
        sLogStat += "Workflow: " + str(self.__map.workflow_short2long(self.__sWf)) + "\n"

        if len(self.__cntRecNew) > 0:
            sLogStat += "\n"
            sLogStat += "--- Ingest: Created records ---\n"
            sLogStat += "New records: " + str(sum(self.__cntRecNew.values())) + "\n"
            for sTable, iNr in self.__cntRecNew.items():
                sLogStat += sTable + ": " + str(iNr) + "\n"

        if len(self.__cntRecExt) > 0:
            sLogStat += "\n"
            sLogStat += "--- Ingest: Extended records ---\n"
            sLogStat += "Extended records: " + str(sum(self.__cntRecExt.values())) + "\n"
            for sTable, iNr in self.__cntRecExt.items():
                sLogStat += sTable + ": " + str(iNr) + "\n"
            if len(self.__setCruiseExtended) > 0:
                sLogStat += "Extended Cruises: " + str(len(self.__setCruiseExtended)) + "\n" + PrettyText.lst2strR(self.__setCruiseExtended) + "\n"
            if len(self.__setStationExtended) > 0:
                sLogStat += "Extended Stations: " + str(len(self.__setStationExtended)) + "\n" + PrettyText.lst2strR(self.__setStationExtended) + "\n"
            if len(self.__setSampleExtended) > 0:
                sLogStat += "Extended Samples: " + str(len(self.__setSampleExtended)) + "\n" + PrettyText.lst2strR(self.__setSampleExtended) + "\n"
            if len(self.__setSubSampleExtended) > 0:
                sLogStat += "Extended SubSamples: " + str(len(self.__setSubSampleExtended)) + "\n" + PrettyText.lst2strR(self.__setSubSampleExtended) + "\n"
            sLogStat += "\n"
            if sum(self.__cntRecExt_diff.values()) > 0:
                sLogStat += "--- Ingest: Hiccups while extending ---\n"
                sLogStat += "Differing record values: " + str(sum(self.__cntRecExt_diff.values())) + "\n"
                for sTable, iNr in self.__cntRecExt_diff.items():
                    sLogStat += sTable + ": " + str(iNr) + "\n"
                sLogStat += "\n"
            sLogStat += "--- Ingest: Dependencies ---\n"
            if len(self.__setIngestDepend) > 0:
                sLogStat += "Depending on other ingests: " + str(len(self.__setIngestDepend)) + "\n"
                for iIngId, sFile in self.__dicIngestDependFiles.items():
                    sLogStat += str(iIngId) + " (" + sFile + ")\n"
            if len(self.__setDatasetDepend) > 0:
                sLogStat += "Other datasets with same dependancies: " + str(len(self.__setDatasetDepend)) + "\n" + PrettyText.lst2strR(self.__setDatasetDepend) + "\n"
        if len(self.__dicBiotaAll) > 0:
            sLogStat += "\n"
            sLogStat += "--- Ingest: Taxa ---\n"
            # --- Overview
            if len(self.__dicBiotaAll) > 0:
                sLogStat += "All used taxa: " + str(len(self.__dicBiotaAll)) + "\n"
            if len(self.__setBiotaAll_check_accepted) > 0:
                sLogStat += "Accepted taxa: " + str(len(self.__setBiotaAll_check_accepted)) + "\n"
            if len(self.__setBiotaAll_check_unaccepted) > 0:
                sLogStat += "Unaccepted taxa: " + str(len(self.__setBiotaAll_check_unaccepted)) + "\n"
            if len(self.__setBiotaAll_check_special) > 0:
                sLogStat += "Special taxa: " + str(len(self.__setBiotaAll_check_special)) + "\n"
            if len(self.__lstBiotaAll_check_notFound) > 0:
                sLogStat += "Taxon Not Found in MyWoRMS/WoRMS: " + str(len(self.__lstBiotaAll_check_notFound)) + "\n"
            # --- Ne Taxa to CB
            if len(self.__dicBiotaNew) > 0:
                sLogStat += "\n"
                sLogStat += "--- Ingest: New Taxa in TaxonWizard ---\n"
                sLogStat += "New taxa: " + str(len(self.__dicBiotaNew)) + " <- Please verify!\n"
            # ---  Misspellings
            if len(self.__lstBiotaAll_check_misspellingNames) > 0 or len(self.__lstBiotaAll_check_misspellingNames_special):
                sLogStat += "\n"
                sLogStat += "--- Ingest: Misspellings in Biota Sheet ---\n"
                if len(self.__lstBiotaAll_check_misspellingNames) > 0:
                    sLogStat += "ScientificNames: " + str(len(self.__lstBiotaAll_check_misspellingNames)) + "\n"
                if len(self.__lstBiotaAll_check_misspellingNames_special) > 0:
                    sLogStat += "Special Taxa Names: " + str(len(self.__lstBiotaAll_check_misspellingNames_special)) + "\n"
            # --- Misfits
            if len(self.__setBiotaAll_check_problem_all) > 0:
                sLogStat += "\n"
                sLogStat += "--- Ingest: Misfits ---\n"
            if len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted) > 0:
                sLogStat += "Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(
                    len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted)) + "\n"
            if len(self.__setBiotaAll_check_problem_limbo) > 0:
                sLogStat += "Taxa in Limbo: " + str(len(self.__setBiotaAll_check_problem_limbo)) + "\n"
            if len(self.__setBiotaAll_check_problem_aphia) > 0:
                sLogStat += "Taxa with Aphia Id problems: " + str(
                    len(self.__setBiotaAll_check_problem_aphia)) + "\n"
            if len(self.__setBiotaAll_check_problem__no_name) > 0:
                sLogStat += "Taxa with missing ScientificNames: " + str(
                    len(self.__setBiotaAll_check_problem__no_name)) + "\n"
            if len(self.__setBiotaAll_check_problem_no_name_at_all) > 0:
                sLogStat += "Taxa with no names at all: " + str(
                    len(self.__setBiotaAll_check_problem_no_name_at_all)) + "\n"
            if len(self.__setBiotaAll_check_problem_no_aname_and_accepted) > 0:
                sLogStat += "Accepted Taxa with missing AcceptedScientificName: " + str(
                    len(self.__setBiotaAll_check_problem_no_aname_and_accepted)) + "\n"
            if len(self.__setBiotaAll_check_problem_status) > 0:
                sLogStat += "Taxa with status problems: " + str(
                    len(self.__setBiotaAll_check_problem_status)) + "\n"
            sLogStat += "\n"
        return sLogStat

    # Check FK/PK
    def __check_fkpk(self):

        # Helper
        def cleanUp():
            progress("off")
            status()

        res = Result(name="Checking FK-PK Pairs")

        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        status = self.__print.status

        status()
        progress("on-pulse")

        sLog = ""
        sLog += "------------------------------------\n"
        sLog += "Check FK-PK Pairings\n"
        sLog += "------------------------------------\n\n"
        sLog += "\n"

        # --- Checking FK-PK
        pn("* Checking FK->PK References")
        sLog += "--- Check of FK->PK References ---\n"
        lstSheets = self.__map.ingest_getExcelSheets(self.__sWf)
        pv("Sheets to check: "+PrettyText.lst2strC(lstSheets))

        # EWI End-Status des letzten Modules
        iErrStart = self.__resultAll.getErrNo()
        iWarnStart = self.__resultAll.getWarnNo()
        iInfoStart = self.__resultAll.getInfoNo()

        # Rows bestimmen
        iFkTotalMax = 0
        iFkTotalNow = 0
        for sSheetFk, sSheetPk in self.__map.ingest_getFkPkPartnerAll().items():
            setPk = self.__data.getPks(sSheetPk)
            setFk = self.__data.getFks(sSheetFk)
            if setPk is not None and setFk is not None:
                iFkTotalMax += len(setFk)
        # Check
        for sSheetFk, sSheetPk in self.__map.ingest_getFkPkPartnerAll().items():
            setPk = self.__data.getPks(sSheetPk)
            setFk = self.__data.getFks(sSheetFk)
            if setPk is not None and setFk is not None:
                iFkNotFound = 0
                for sFk in setFk:
                    iFkTotalNow += 1
                    # Progress
                    iErr = res.getErrNo() + iErrStart
                    iWarn = res.getWarnNo() + iWarnStart
                    iInfo = res.getInfoNo() + iInfoStart
                    progress(iFkTotalNow, iFkTotalMax, 100, proz=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                    status("Processing #"+str(iFkTotalNow)+"/"+str(iFkTotalMax))
                    if sFk not in setPk:
                        iFkNotFound += 1
                        res.addErr("FK not found in corresponding PK set", add=sSheetFk + "->" + sSheetPk + ": " + sFk, sheet=sSheetFk, sys="Ingestor", kat="PK and FK")
                sInfo = sSheetFk + " (FK) -> " + sSheetPk + " (PK): "
                if iFkNotFound > 0:
                    sTxtTmp = sInfo + str(iFkNotFound) + "x FK not found"
                    pv(sTxtTmp)
                    sLog += sTxtTmp+"\n"
                else:
                    sTxtTmp = sInfo + "ok"
                    pv(sTxtTmp)
                    sLog += sTxtTmp+"\n"
        sLog += "\n"

        # --- PK Benutzung
        pn("* Checking PK Usage")
        pv("Sheets to check: "+PrettyText.lst2strC(lstSheets))
        sLog += "--- Check of PK Usage ---\n"

        # Wer benutzt einen Pk
        lstSheets = self.__map.ingest_getExcelSheets(self.__sWf)
        odicPk = OrderedDict()
        for sSheetFk, sSheetPk in self.__map.ingest_getFkPkPartnerAll().items():
            if (sSheetPk in lstSheets) and (sSheetFk in lstSheets):
                if sSheetPk not in odicPk:
                    odicPk[sSheetPk] = []
                odicPk[sSheetPk].append(sSheetFk)

        # Check
        for sSheetPk, lstSheetsFk in odicPk.items():
            setPk = self.__data.getPks(sSheetPk)
            if setPk is not None:
                for sSheetFk in lstSheetsFk:
                    setFk = self.__data.getFks(sSheetFk)
                    if setFk is not None:
                        setPk = setPk - setFk
                if sSheetPk == "Biota":
                    pass # Bei BIota ist es nicht bemerkenswert wenn PKs nicht benutzt wurden von IndividualMeasure
                else:
                    sInfo = sSheetPk + " (PK) <- " + PrettyText.lst2strC(lstSheetsFk) + " (FK): "
                    if len(setPk) > 0:
                        sTxtTmp = sInfo + str(len(setPk)) + "x PK not used: "+PrettyText.lst2strC(setPk)
                        pv(sTxtTmp+"\n")
                        res.addWarn("PK not used in corresponding FK sets", add=sSheetPk + "<-" + PrettyText.lst2strC(lstSheetsFk) + ": "+PrettyText.lst2strC(setPk), sheet=sSheetPk, sys="Ingestor", kat="PF and FK")
                    else:
                        sTxtTmp = sInfo + "all PK used"
                        pv(sTxtTmp)
                        sLog += sTxtTmp+"\n"
        sLog += "\n"

        cleanUp()
        res.addData("log", sLog)

        return res

    # Check Lookup
    def __check_lookup(self):

        # Helper
        def cleanUp():
            progress("off")
            status()

        res = Result(name="Checking lookup data")
        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        status = self.__print.status

        iFound = 0
        iFoundCache = 0
        iFoundDb = 0
        iNotFound = 0
        odicNotFound = OrderedDict()

        iLookupsTotal = 0
        iLookupNow = 0

        iErrStart = self.__resultAll.getErrNo()
        iWarnStart = self.__resultAll.getWarnNo()
        iInfoStart = self.__resultAll.getInfoNo()

        # Lade Cache des Lookupsystems neu
        pn("* Checking Lookup")
        pv("* Resetting Lookup System")
        self.__lup.reset()
        if self.__bLupSearchSingle:
            pv("* Search mode: ask db for every lookup")
        else:
            pv("* Search mode: load full cache at start")
        if self.__bLupSearchSingle is False:
            pv("* Reloading Lookup Cache")

            self.__lup.update(cursor=self.__cursor)
        # Info
        sInfo = ""
        bInfo = False
        for sSheet, dicCols in self.__dicLookup.items():
            for sCol, setVals in dicCols.items():
                if len(setVals) > 0:
                    iLookupsTotal += len(setVals)
                    bInfo = True
                    sInfo += sSheet +"." + sCol + " ("+ str(len(setVals)) + "): " + PrettyText.lst2strC(setVals) +"\n"

        if bInfo:
            pv("* All Lookups Used in Ingest")
            sInfo +=  "Total number: "+str(iLookupsTotal)
            pv(sInfo)

        # Log
        sLog = ""
        sLog += "------------------------------------\n"
        sLog += "Check Lookup Data\n"
        sLog += "------------------------------------\n\n"
        if bInfo:
            sLog += "--- All Lookups Used By Ingest ---\n"
            sLog += sInfo
            sLog += "\n"

        # Loop
        lstTaxa = []
        for sSheet, dicCols in self.__dicLookup.items():
            for sCol, setVals in dicCols.items():
                if len(setVals) > 0:
                    for sVal in setVals:
                        sVal = str(sVal)
                        iLookupNow += 1
                        # Progress
                        iErr = res.getErrNo() + iErrStart
                        iWarn = res.getWarnNo() + iWarnStart
                        iInfo = res.getInfoNo() + iInfoStart
                        status("Processing #" + str(iLookupNow) + "/" + str(iLookupsTotal))
                        progress(iLookupNow, iLookupsTotal, 100, proz=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                        # Check
                        sTable = self.__map.ingest_getLookup(sSheet, sCol)
                        resCheck = self.__lup.getLookup(sTable, sVal, cursor=self.__cursor, update=self.__bLupSearchSingle)
                        bFound = resCheck.getData("found")
                        bFoundCache = resCheck.getData("cache")
                        bFoundDb = resCheck.getData("db")
                        if not bFound:
                            #res.addWarn("Lookup not found in database",add=sVal, sheet=sSheet, col=sCol, sys="Ingestor", kat="Lookup used in Ingest data")
                            if (sSheet,sCol) in odicNotFound:
                                odicNotFound[(sSheet,sCol)].add(sVal)
                            else:
                                odicNotFound[(sSheet, sCol)] = {sVal}
                        else:
                            iFound += 1
                            if bFoundCache:
                                iFoundCache += 1
                            if bFoundDb:
                                iFoundDb += 1
                        # Merke Aphia Ids
                        if (sSheet, sCol) == ("Biota", "AphiaId"):
                            lstTaxa.append(sVal)

        # Gesamtanzahl der nicht gefundenen Lookups
        for (sSheet,sCol), setNotFound in odicNotFound.items():
            iNotFound += len(setNotFound)

        # Alle Lookups gefunden
        if iNotFound == 0:
            # Ja -> Alle Lookups sind OK
            sLog += "All Lookups used in ingest found in database\n"
            pv("* All lookups found in database")
            if iFoundDb > 0:
                pv("Found in db: " +str(iFoundDb))
            if iFoundCache > 0:
                pv("Found in cache: " +str(iFoundCache))
        else:
            # Nein -> es fehlen Lookups
            pn("* Some lookups not found in database")
            # Gesamtanzahl der nicht gefundenen Lookups
            iNotFound = 0
            for (sSheet, sCol), setNotFound in odicNotFound.items():
                for sVal in setNotFound:
                    res.addErr("Lookup not found in database", add=sVal, sheet=sSheet, col=sCol, sys="Ingestor", kat="Lookup used in Ingest data")
                iNotFound += len(setNotFound)
            if iNotFound > 0:
                pn("* Some lookups are missing anyway")

        # Sauber machen
        cleanUp()

        # Log
        sLog += "\n"
        res.addData("log", sLog)
        return res

    # Check Taxa
    def __check_taxa__mywormsOnthefly(self):
        res = Result(name="Checking taxa")
        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        workflow = self.__print.workflow
        status = self.__print.status
        myworms = self.__model.subsystem_getMyWorms()

        # Daten
        iMaxAgeAllowed = self.__model.subsystem_getMyWorms().puffer_getMaxAgeAllowed()

        lstTaxaUsed = list(self.__dicBiotaAll.keys())
        lstTaxaTaxonlistFound = []
        lstTaxaTaxonlistNotFound = []
        lstTaxaMyWormsFound = []
        lstTaxaMyWormsFoundButTooOld = []

        # EWI End-Status des letzten Modules
        iErrStart = self.__resultAll.getErrNo()
        iWarnStart = self.__resultAll.getWarnNo()
        iInfoStart = self.__resultAll.getInfoNo()

        # Info/Log
        pn("* MyWoRMS: Check Taxa (on-the-fly)", fmt=False)
        sLog = ""
        sLog += "------------------------------------\n"
        sLog += "MyWoRMS: Check Taxa (on-the-fly)\n"
        sLog += "------------------------------------\n\n"


        # *** Übersicht aller benutzten Taxa
        # Taxon Liste: Alle Taxa, die in Taxon Liste sind
        for iAid, dat in self.__dicBiotaAll.items():
            resTaxLup = self.__model.subsystem_getLookup().getLookup("taxon", str(iAid))
            if resTaxLup.getData("found"):
                lstTaxaTaxonlistFound.append(iAid)
            else:
                lstTaxaTaxonlistNotFound.append(iAid)

        # MyWoRMS: Alle Taxa, die online neu abgerufen werden müssen (Neu oder De-Age)
        for iAid, sName in self.__dicBiotaAll.items():
            if not self.__model.subsystem_getTaxonWizard().isSpecialTaxon(iAid):
                resMyWorms = myworms.getTaxon(iAid, puffer=True, worms=False)
                if resMyWorms.isErr():
                    res.addResult(resMyWorms)
                else:
                    bFound = resMyWorms.getData("found")
                    if bFound:
                        t = resMyWorms.getData("taxon")
                        iAge = t.getAge()
                        if iAge > iMaxAgeAllowed:
                            lstTaxaMyWormsFoundButTooOld.append(iAid)
                        else:
                            lstTaxaMyWormsFound.append(iAid)
                    else:
                        self.__lstBiotaAll_check_notFound.append(iAid)
        iWormsContacts = len(self.__lstBiotaAll_check_notFound) + len(lstTaxaMyWormsFoundButTooOld)

        # Info/Log
        sLog += "--- Initialisation ---\n"
        sLog += "Ingest: used taxa: " + str(len(lstTaxaUsed))+ "\n"
        sLog += "Taxon List: taxa found: " + str(len(lstTaxaTaxonlistFound)) + "\n"
        sLog += "Taxon List: taxa not found : " + str(len(lstTaxaTaxonlistNotFound)) +"\n"
        pv("Ingest: used taxa: " + str(len(lstTaxaUsed)))
        pv("Taxon List: taxa found: " + str(len(lstTaxaTaxonlistFound)))
        pv("Taxon List: taxa not found: " + str(len(lstTaxaTaxonlistNotFound)))
        if iWormsContacts == 0:
            pn("* MyWoRMS: all taxa found (all younger "+str(iMaxAgeAllowed)+" days)", fmt=False)
            sLog += "MyWoRMS: all taxa found (all younger "+str(iMaxAgeAllowed)+" days)" + "\n"
        else:
            if len(lstTaxaMyWormsFound) > 0:
                pv("MyWoRMS: taxa found (younger "+str(iMaxAgeAllowed)+" days): " + str(len(lstTaxaMyWormsFound)))
                sLog += "MyWoRMS: taxa found (younger "+str(iMaxAgeAllowed)+" days): " + str(len(lstTaxaMyWormsFound))+"\n"
            if len(lstTaxaMyWormsFoundButTooOld) > 0:
                pn("MyWoRMS: taxa found (older "+str(iMaxAgeAllowed)+" days): " + str(len(lstTaxaMyWormsFoundButTooOld)))
                sLog += "MyWoRMS: taxa found (older "+str(iMaxAgeAllowed)+" days): " + str(len(lstTaxaMyWormsFoundButTooOld)) + "\n"
            if len(self.__lstBiotaAll_check_notFound) > 0:
                pn("MyWoRMS: Taxa not found: "+str(len(self.__lstBiotaAll_check_notFound)))
                sLog += "MyWoRMS: Taxa not found: "+str(len(self.__lstBiotaAll_check_notFound)) + "\n"
        sLog += "\n"

        # *** MyWoRMS: Neue Taxa laden und veraltete Taxa anit-agen
        pv("* Preparing MyWoRMS for on-the-fly Testing", fmt=False)
        sLog += "--- Preparing MyWoRMS for on-the-fly Testing ---\n"
        if iWormsContacts >0:
            pn("* MyWoRMS has to contact WoRMS: " + str(iWormsContacts) + "x", fmt=False)
            sLog += "MyWoRMS has to contact WoRMS: "+str(iWormsContacts)+"x\n"
        # 1) Neue Taxa laden
        if len(self.__lstBiotaAll_check_notFound) >0 :
            pn("* Getting new taxa from WoRMS: "+str(len(self.__lstBiotaAll_check_notFound)), fmt=False)
            progress("on-pulse")
            workflow(top="Ingestion", sub="Asking WoRMS", fmt=False)
            iTaxonNr = 1
            iTaxonMax = len(self.__lstBiotaAll_check_notFound)
            for iAid in self.__lstBiotaAll_check_notFound:
                # Progress
                iErr = res.getErrNo() + iErrStart
                iWarn = res.getWarnNo() + iWarnStart
                iInfo = res.getInfoNo() + iInfoStart
                progress(iTaxonNr, iTaxonMax, 100, proz=True, abs=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                status("Downloading taxon with Aphia Id: "+str(iAid))

                resTax = myworms.getTaxon(iAid)
                iTaxonNr += 1
            myworms.puffer_save() # Speichern
            sLog += "All required new Taxa are gathered from WoRMS now\n"
        # 2) Veraltete Taxa de-agen
        if len(lstTaxaMyWormsFoundButTooOld) > 0:
            pn("* DeAge taxa in MyWoRMS: " + str(len(lstTaxaMyWormsFoundButTooOld)), fmt=False)
            workflow(top="MyWORMS", sub="DeAging Taxa", fmt=False)
            myworms.puffer_update(taxa=lstTaxaMyWormsFoundButTooOld)
            sLog += "All required outdated taxa are de-aged now\n"
        pn("* MyWoRMS: up-to-date and ready to use", fmt=False)
        sLog += "MyWoRMS: up-to-date and ready to use\n"
        sLog += "\n"

        # *** Alle Taxa gegen MyWoRMS testen
        pn("* MyWoRMS: Checking all Taxa (on-the-fly)", fmt=False)
        workflow(top="Ingestion", sub="MyWoRMS (on-the-fly)", fmt=False)
        iMyWormsOTF_checkNow = 0
        iMyWormsOTF_checkTotal = len(self.__dicBiotaAll)
        for iAid, setScientificNames in self.__dicBiotaAll.items():

            # Progress
            iErr = res.getErrNo() + iErrStart
            iWarn = res.getWarnNo() + iWarnStart
            iInfo = res.getInfoNo() + iInfoStart
            progress(iMyWormsOTF_checkNow, iMyWormsOTF_checkTotal, 100, proz=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
            status("MyWoRMS (on-the-fly): "+str(iAid))

            if self.__model.subsystem_getTaxonWizard().isSpecialTaxon(iAid):
                # Sonderfall: Spezial-Taxa
                for sScientificName in setScientificNames:
                    sScientificNameSpecial = self.__model.subsystem_getTaxonWizard().getSpecialTaxonName(iAid)
                    if sScientificName is not None:
                        if SequenceMatcher(None, sScientificName, sScientificNameSpecial).ratio() < self.__model.subsystem_getTaxonWizard().getScientificNameMatchRatio():
                            res.addWarn("Misspelling: Special Taxa Name Differes Too Much From Norm", sheet="Biota",
                                        col="ScientiticName",
                                        add=str(iAid) + ": " + sScientificName + " != " + sScientificNameSpecial,
                                        sys="Ingestor", kat="Checking Taxa")
                            self.__lstBiotaAll_check_misspellingNames_special.append(str(iAid) + ": " + str(sScientificName) + " != " + str(sScientificNameSpecial))
                            self.__iBiotaAll_check_name_diff_special += 1
                        if iAid in lstTaxaTaxonlistNotFound:
                            self.__dicBiotaNew[iAid] = Taxon(blueprint={"aid": iAid, "scientificname": sScientificNameSpecial})
                    else:
                        res.addErr("Uups... Internal Problem: Missing Name of Special Taxon",
                                    sheet="Biota",
                                    col="AphiaId", add=str(iAid),sys="Ingestor", kat="Checking Taxa")
            else:
                # Kein Sonderfall
                resTaxon = myworms.getTaxon(iAid, puffer=True, worms=False)
                if resTaxon.isErr():
                    # Irgend ein andere Fehler
                    res.addResult(resTaxon)
                else:
                    if resTaxon.getData("found") is True:
                        # Taxon Gefunden
                        taxon = resTaxon.getData("taxon")
                        # Teste Taxon auf Probleme
                        myworms.checkProblemTaxon(taxon,
                            self.__setBiotaAll_check_problem_all,
                            self.__setBiotaAll_check_problem__no_name,
                            self.__setBiotaAll_check_problem_no_aname_and_accepted,
                            self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted,
                            self.__setBiotaAll_check_problem_no_name_at_all,
                            self.__setBiotaAll_check_problem_status,
                            self.__setBiotaAll_check_problem_aphia,
                            self.__setBiotaAll_check_problem_limbo)
                        # Test: Special Taxon
                        if taxon.isSpecial():
                           self.__setBiotaAll_check_special.add(taxon)
                        # Test: Accepted
                        if taxon.isAccepted():
                            self.__setBiotaAll_check_accepted.add(taxon)
                        # Test: Unaccepted
                        if taxon.isUnaccepted():
                            self.__setBiotaAll_check_unaccepted.add(taxon)
                        # Test: Namen auf relative Ähnlichkeit mit MyWoRMS-Eintrag
                        for sScientificName in setScientificNames:
                            if sScientificName is not None:
                                if taxon.hasAnyName():
                                    # Taxon hat in MyWoRMS irgendeinen Namen
                                    sName_taxon, sName_taxon_origin = taxon.getAnyName()
                                    if SequenceMatcher(None, sScientificName, sName_taxon).ratio() < self.__model.subsystem_getTaxonWizard().getScientificNameMatchRatio():
                                        res.addWarn("Misspelling: ScientificName differs too much from MyWoRMS", sheet="Biota", col="ScientificName", add=str(iAid)+": "+ str(sScientificName) + " != " + str(taxon.signature()), sys="Ingestor", kat="Checking Taxa")
                                        self.__lstBiotaAll_check_misspellingNames.append(str(iAid) + " " + str(sScientificName) + " != " + str(taxon.signature()))
                                        self.__iBiotaAll_check_name_diff += 1
                                    if iAid in lstTaxaTaxonlistNotFound:
                                        self.__dicBiotaNew[iAid] = taxon
                                else:
                                    # Taxon hat in MYWoRMS keinen Namen
                                    res.addWarn("Misspelling: Taxon in MyWoRMS has no names at all",
                                                sheet="Biota", col="ScientificName",
                                                add=str(iAid) + " " + str(sScientificName) + " ?= " + taxon.signature(), sys="Ingestor", kat="Checking Taxa")
                                    self.__lstBiotaAll_check_misspellingNames.append(str(iAid) + ": " + str(sScientificName) + " ?= " + str(taxon.signature()))
                                    self.__iBiotaAll_check_name_diff += 1
                            else:
                                res.addWarn("Uups... Internal Problem: Missing Scientificname", sheet="Biota",col="AphiaId", add=str(iAid), sys="Ingestor", kat="Checking Taxa")
                    else:
                        # Taxon Nicht gefunden
                        self.__lstBiotaAll_check_notFound.append(str(iAid)+": "+ PrettyText.lst2strC(list(setScientificNames)))
                        res.addErr("AphiaId not found in MyWoRMS", add=str(iAid), sheet="Biota", col="AphiaId", sys="Ingestor", kat="Checking Taxa")


        pn("* MyWoRMS: All Taxa checked (on-the-fly)", fmt=False)

        # *** Log: Taxa Check
        sLogTaxaCheckOverview = ""
        sLogTaxaCheckDetails = ""

        # --- Overview
        if len(self.__dicBiotaNew) > 0:
            sLogTaxaCheckOverview += "New Taxa: "+str(len(self.__dicBiotaNew))+"\n"
        if len(self.__setBiotaAll_check_accepted) > 0:
            sLogTaxaCheckOverview += "Accepted Taxa: "+str(len(self.__setBiotaAll_check_accepted))+"\n"
        if len(self.__setBiotaAll_check_unaccepted) > 0:
            sLogTaxaCheckOverview += "Unaccepted Taxa: " + str(len(self.__setBiotaAll_check_unaccepted)) + "\n"
        if len(self.__setBiotaAll_check_special) > 0:
            sLogTaxaCheckOverview += "Special Taxa: " + str(len(self.__setBiotaAll_check_special)) + "\n"
        if len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted) > 0:
            sLogTaxaCheckOverview += "Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted)) + "\n"
        if len(self.__setBiotaAll_check_problem_limbo) > 0:
            sLogTaxaCheckOverview += "Taxa in Limbo: " + str(len(self.__setBiotaAll_check_problem_limbo)) + "\n"
        if len(self.__setBiotaAll_check_problem_aphia) > 0:
            sLogTaxaCheckOverview += "Taxa with Aphia Id problems: " + str(len(self.__setBiotaAll_check_problem_aphia)) + "\n"
        if len(self.__setBiotaAll_check_problem__no_name) > 0:
            sLogTaxaCheckOverview += "Taxa with missing ScientificNames: " + str(len(self.__setBiotaAll_check_problem__no_name)) + "\n"
        if len(self.__setBiotaAll_check_problem_no_name_at_all) > 0:
            sLogTaxaCheckOverview += "Taxa with no names at all: " + str(len(self.__setBiotaAll_check_problem_no_name_at_all)) + "\n"
        if len(self.__setBiotaAll_check_problem_no_aname_and_accepted) > 0:
            sLogTaxaCheckOverview += "Accepted Taxa with missing AcceptedScientificName: " + str(len(self.__setBiotaAll_check_problem_no_aname_and_accepted)) + "\n"
        if len(self.__setBiotaAll_check_problem_status) > 0:
            sLogTaxaCheckOverview += "Taxa with status problems: " + str(len(self.__setBiotaAll_check_problem_status)) + "\n"
        if len(self.__lstBiotaAll_check_misspellingNames) > 0:
            sLogTaxaCheckOverview += "Misspelling of ScientificNames: "+str(len(self.__lstBiotaAll_check_misspellingNames))+"\n"
        if len(self.__lstBiotaAll_check_misspellingNames_special) > 0:
            sLogTaxaCheckOverview += "Misspelling of Special Taxa Name: "+str(len(self.__lstBiotaAll_check_misspellingNames_special))+"\n"
        if len(self.__lstBiotaAll_check_notFound) > 0:
            sLogTaxaCheckOverview += "Taxon Not Found in MyWoRMS/WoRMS: "+str(len(self.__lstBiotaAll_check_notFound))+"\n"
        # --- Details
        if len(self.__dicBiotaNew) > 0:
            sLogTaxaCheckDetails += "* New Taxa: " + str(len(self.__dicBiotaNew)) + "\n"
            for iAid, t in self.__dicBiotaNew.items():
                sLogTaxaCheckDetails += t.signature()+ "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_accepted) > 0:
            sLogTaxaCheckDetails += "* Accepted Taxa: "+str(len(self.__setBiotaAll_check_accepted))+"\n"
            for t in self.__setBiotaAll_check_accepted:
                sLogTaxaCheckDetails+= t.signature()+ "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_unaccepted) > 0:
            sLogTaxaCheckDetails += "* Unaccepted Taxa: " + str(len(self.__setBiotaAll_check_unaccepted)) + "\n"
            for t in self.__setBiotaAll_check_unaccepted:
                sLogTaxaCheckDetails += t.signature() + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_special) > 0:
            sLogTaxaCheckDetails += "* Special Taxa: " + str(len(self.__setBiotaAll_check_special)) + "\n"
            for t in self.__setBiotaAll_check_special:
                sLogTaxaCheckDetails += t.signature() + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted) > 0:
            sLogTaxaCheckDetails += "* Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(
                len(self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted)) + "\n"
            for t in self.__setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_limbo) > 0:
            sLogTaxaCheckDetails += "* Taxa in Limbo: " + str(len(self.__setBiotaAll_check_problem_limbo)) + "\n"
            for t in self.__setBiotaAll_check_problem_limbo:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_aphia) > 0:
            sLogTaxaCheckDetails += "* Taxa with Aphia Id problems: " + str(len(self.__setBiotaAll_check_problem_aphia)) + "\n"
            for t in self.__setBiotaAll_check_problem_aphia:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem__no_name) > 0:
            sLogTaxaCheckDetails += "* Taxa with missing ScientificNames: " + str(len(self.__setBiotaAll_check_problem__no_name)) + "\n"
            for t in self.__setBiotaAll_check_problem__no_name:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_no_name_at_all) > 0:
            sLogTaxaCheckDetails += "* Taxa with no names at all: " + str(len(self.__setBiotaAll_check_problem_no_name_at_all)) + "\n"
            for t in self.__setBiotaAll_check_problem_no_name_at_all:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_no_aname_and_accepted) > 0:
            sLogTaxaCheckDetails += "* Accepted Taxa with missing AcceptedScientificName: " + str(len(self.__setBiotaAll_check_problem_no_aname_and_accepted)) + "\n"
            for t in self.__setBiotaAll_check_problem_no_aname_and_accepted:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__setBiotaAll_check_problem_status) > 0:
            sLogTaxaCheckDetails += "* Taxa with status problems: " + str(len(self.__setBiotaAll_check_problem_status)) + "\n"
            for t in self.__setBiotaAll_check_problem_status:
                sLogTaxaCheckDetails += str(t.signature()) + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__lstBiotaAll_check_misspellingNames) > 0:
            sLogTaxaCheckDetails += "* Misspelling: ScientificNames: "+str(len(self.__lstBiotaAll_check_misspellingNames))+"\n"
            for sTxt in self.__lstBiotaAll_check_misspellingNames:
                sLogTaxaCheckDetails += sTxt + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__lstBiotaAll_check_misspellingNames_special) > 0:
            sLogTaxaCheckDetails += "* Misspelling: Special Taxa Names: "+str(len(self.__lstBiotaAll_check_misspellingNames_special))+"\n"
            for sTxt in self.__lstBiotaAll_check_misspellingNames_special:
                sLogTaxaCheckDetails += sTxt + "\n"
            sLogTaxaCheckDetails += "\n"
        if len(self.__lstBiotaAll_check_notFound) > 0:
            sLogTaxaCheckDetails += "* Taxon Not Found in MyWoRMS/WoRMS: "+str(len(self.__lstBiotaAll_check_notFound))+"\n"
            for iAid in self.__lstBiotaAll_check_notFound:
                sLogTaxaCheckDetails += str(iAid) + "\n"
            sLogTaxaCheckDetails += "\n"

        sLogTaxaCheckPrae = "------------------------------------\n"
        sLogTaxaCheckPrae += "MyWoRMS: Result Taxa Check\n"
        sLogTaxaCheckPrae += "------------------------------------\n\n"
        sLogTaxaCheckOverviewHead = "--- Overview ---\n"
        sLogTaxaCheckDetailsHead = "--- Details ---\n"

        if sLogTaxaCheckDetails == "":
            sLogTaxaCheck = sLogTaxaCheckPrae + "Everything is ok ...\n"
        else:
            sLogTaxaCheck = sLogTaxaCheckPrae + sLogTaxaCheckOverviewHead + sLogTaxaCheckOverview + "\n" + sLogTaxaCheckDetailsHead + sLogTaxaCheckDetails

        # Log: Gesamt
        sLog += sLogTaxaCheck

        # Antwort
        res.addData("log", sLog)
        return res

    # TaxonWizard: Update TaxonList
    def __check_taxa__quickUpdateTaxonLists(self):
        res = Result(name="TaxonWizard: Update TaxonList")
        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        workflow = self.__print.workflow
        status = self.__print.status
        myworms = self.__model.subsystem_getMyWorms()

        # Daten
        sLog = ""
        sLogAddBase = ""
        sLogAddTaxonList = ""

        # Info/Log
        pn("* TaxonWizard: Quick Update Taxon List", fmt=True)
        sLog += "------------------------------------\n"
        sLog += "TaxonWizard: Quick Update Taxon List\n"
        sLog += "------------------------------------\n\n"
        if len(self.__dicBiotaNew) > 0:
            # *** Adding new Taxa to Base List
            pv("TaxonWizard: Quick Update Base List and Taxon List:" + str(len(self.__dicBiotaNew)))
            lstNewBase = []
            resAddBase = self.__model.subsystem_getTaxonWizard().addBaseTaxa(self.__dicBiotaNew.values(), cursor=self.__cursor, quick_update_taxon_list=True)
            if resAddBase.isAnyMessage():
                res.addResult(resAddBase, msg_only=True)
            sLogAddBase = resAddBase.getData("log") if resAddBase.getData("log") is not None else ""

            # *** Update Lookup TaxonList
            pv("* Lookup System: Reload Taxon List")
            self.__model.subsystem_getLookup().update("taxon", cursor=self.__cursor)

            # Log
            sLog += "--- Quick Changes ---\n"
            sLog += sLogAddBase + "\n"
        else:
            sLog += "Taxon List is up-to-date\n\n"
            pv("Taxon List is up-to-date")

        # Messages merken für folgende Progress-Bars im Ingestor
        if res.isAnyMessage():
            self.__resultAll.addResult(res, msg_only=True)

        # Antwort
        res.addData("log", sLog)
        return res

    # Lade Daten aus den Sheets in interne Speicherstrukturen
    def __load(self):

        # Helper
        def cleanUi():
            progress("off")
            status()

        # Hole Wert aus Zeile
        def getCell(data, row, sSheet, sCol):
            iColIdx = data.getHeadIdx(sSheet, sCol)
            cell = row[iColIdx]
            val = cell.value
            sVal = (str(cell.value) if cell.value is not None else "")
            return (cell, val, sVal)

        # Check Cell
        def checkCell(sSheet, sCol, cell, iRow):

            # Helper
            def cleanString(s):
                if s is None:
                    return s
                s = s.replace("\\", "")  # Backslashes entfernen
                s = s.lower()  # Lower
                s = s.strip()  # Whitespaces weg
                return s

            res = Result("Check row while loading")

            # Zeile
            sRow = str(iRow)

            # Zell-Typ: Vorgabe
            bMapIsDate = self.__map.ingest_isDate(sSheet, sCol)
            bMapIsTime = self.__map.ingest_isTime(sSheet, sCol)
            sMapGeoType = self.__map.ingest_getGeo(sSheet, sCol)
            bMapObligat = self.__map.ingest_isObl(sSheet, sCol)
            bMapIsLookup = self.__map.ingest_isLookup(sSheet, sCol)

            resSchemaDb = self.__model.conf_getDbSchema().getDbSchemaInfo("ingest", sSheet, sCol)
            if resSchemaDb.isErr():
                res.addErr("Problems getting data type from Database Schema", col=sCol, sheet=sSheet, row=sRow,
                           sys="Ingestor", kat="DataType")
            sDbDataType = resSchemaDb.getData("datatype")

            # Zell-Typ: Aktueller Wert
            val = cell.value
            sVal = (str(val) if val is not None else "")
            bValIsDate = cell.is_date and isinstance(val, datetime.datetime)
            bValIsTime = cell.is_date and isinstance(val, datetime.time)
            if bValIsDate:
                sValType = "date"
            elif bValIsTime:
                sValType = "time"
            else:
                sValType = cell.data_type

            # Test auf Daten Typ in Db Schems Schema

            # ------------------------

            # *** Lookup (erstmal alle aussortieren)
            if sVal != "" and bMapIsLookup:
                try:
                    # Kein Taxon -> Das wird separat abgespeichert
                    if (sSheet, sCol) != ("Biota", "AphiaId"):
                        self.__dicLookup[sSheet][sCol].add(sVal)
                except Exception as ex:
                    res.addErr("Lookup Detection", add=str(ex), sheet=sSheet, col=sCol, row=sRow, kat="Lookup",
                               sys="Ingestor")

            # *** Geo (->dD)
            if sMapGeoType is not None and sVal is not None and sVal != "":
                resGeoRefCheck = self.__geoCruncher.crunch(sVal, sMapGeoType)
                if resGeoRefCheck.isErr():
                    res.addResult(resGeoRefCheck, err=True, warn=True)
                else:
                    iValDd = resGeoRefCheck.getData("value")
                    sVal = str(iValDd)

            # *** Datum: Sonderfall 00:00:00
            # !!!! Pain-in-The-Ass-Workaround
            # Feld mit Datumstyp und eigenartigem Inhalt gefunden
            # Verbiege auf 00:00:00
            if bValIsDate:
                if sVal in ["1899-12-30 00:00:00", "1899-12-29 00:00:00"]:
                    res.addWarn("Suspicious date detected", add=sVal + "-forge to-> 00:00:00", sheet=sSheet,
                                col=sCol, row=sRow, sys="Ingestor", kat="Special Date")
                    bValIsDate = False
                    bValIsTime = True
                    sVal = "00:00:00"

            # *** Datum: Is Wert wirklich Datum?
            if val is not None:
                if bMapIsDate and not bValIsDate:
                    res.addWarn("No date", add=str(val) + ", type: " + str(cell.data_type), sheet=sSheet, col=sCol,
                                row=sRow, sys="Ingestor", kat="Date")

            # *** Time: Ist Wert wirklich Zeit
            if val is not None:
                if bMapIsTime and not bValIsTime:
                    res.addWarn("No time", add=str(val) + ", type: " + str(cell.data_type), sheet=sSheet, col=sCol,
                                row=sRow, sys="Ingestor", kat="Time")

            # *** Datum: Format entspricht Vorgabe?
            if bMapIsDate and bValIsDate:
                sFmt = cleanString(cell.number_format)
                if sFmt not in self.__lstExcelDateFormat:
                    res.addWarn("Wrong date format", add=str(sFmt), sheet=sSheet, col=sCol, row=sRow,
                                sys="Ingestor", kat="Date")

            # *** Time: Format entspricht Vorgabe?
            if bMapIsTime and bValIsTime:
                sFmt = cleanString(cell.number_format)
                if sFmt not in self.__lstExcelTimeFormat:
                    res.addWarn("Wrong Time format", add=str(sFmt), sheet=sSheet, col=sCol, row=sRow,
                                sys="Ingestor", kat="Time")

            # *** Datum: Bei Datum Uhrzeit immer abschneiden: Datetime -> Date (ISO-8601 Date)
            if bValIsDate:
                sVal = str(cell.value.date().isoformat())  # ISO-8601 Date YYYY-MM-DD

            # *** Datum: Liegt Datum in der Zukunft
            if val is not None:
                if bMapIsDate and bValIsDate:
                    if val > datetime.datetime.now():
                        res.addErr("Future date found", add=sVal, sheet=sSheet, col=sCol, row=sRow, sys="Ingestor",
                                   kat="Date")

            # *** Obligat
            if sVal in ["", None] and bMapObligat:
                res.addErr("Missing obligate data", col=sCol, sheet=sSheet, row=sRow, sys="Ingestor",
                           kat="Obligate data")

            # *** Float
            if val is not None:
                if "float" in sDbDataType or "numeric" in sDbDataType:
                    try:
                        fTest = float(sVal)
                    except:
                        res.addErr("Float expected", add=sVal, sheet=sSheet, row=iRow, col=sCol, sys="Ingestor",
                                   kat="Wrong data type")

            # *** Integer
            if val is not None:
                if "int" in sDbDataType:
                    try:
                        sValCheckedTest = str(sVal)
                        iTest = int(sVal)
                    except:
                        res.addErr("Integer expected", add=sVal, sheet=sSheet, row=iRow, col=sCol, sys="Ingestor",
                                   kat="Wrong data type")

            # Antwort
            res.addData("dbtype", sDbDataType)
            res.addData("type", sValType)
            res.addData("val", sVal)
            return res

        # -----------------

        res = Result(name="Loading Excel data")

        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        status = self.__print.status

        # Info
        pn("* Loading ingest data, check data types and build new data structure")

        # Rows
        iRowsTotalMax = 0
        iRowsTotalNow = 0
        iRowsSheetMax = 0

        # Rows Total
        resRows = ExcelTools.getRowsTotal(self.__wb)
        if resRows.isErr():
            return resRows
        iRowsTotalMax = resRows.getData("total")

        # Log
        sLog = ""
        sLog += "------------------------------------\n"
        sLog += "Load Excel Workbook and\n"
        sLog += "Check data on-on-the-fly\n"
        sLog += "------------------------------------\n\n"
        sLog += "Accepted date format: yyyy/mm/dd\n"
        sLog += "Accepted time format: hh:mm:ss\n"
        sLog += "\n"

        # Sheets/Headers
        odicWb = OrderedDict()
        for sheet in self.__wb:
            sSheet = sheet.title
            if sSheet in self.__lstSheetsUse:
                resHeader = ExcelTools.getHeadIndex(sheet)
                if resHeader.isErr():
                    cleanUi()
                    return resHeader
                lstHead = resHeader.getData("header")
                odicWb[sSheet] = lstHead

        # Ingest Daten Struktur
        self.__data = Data(odicWb)

        # Indizes für direkten Zugriff auf Taxa in Biota-Tabelle: Aid/Scientificname
        iIdxBiotaAphiaId = self.__data.getHeadIdx("Biota", "AphiaId")
        iIdxBiotaScientificName = self.__data.getHeadIdx("Biota", "ScientificName")
        iIdxBiotaLifestage = self.__data.getHeadIdx("Biota", "Stage")
        iIdxBiotaSpecification = self.__data.getHeadIdx("Biota", "Specification")
        iIdxBiotaSieve = self.__data.getHeadIdx("Biota", "SieveFraction")

        # EWI End-Status des letzten Modules
        iErrStart = self.__resultAll.getErrNo()
        iWarnStart = self.__resultAll.getWarnNo()
        iInfoStart = self.__resultAll.getInfoNo()

        # Separator beim Konkatenieren der PK/FK-Parts
        sPkFkSep = "|"

        # Loop: Sheets
        iRowsTotalNow = 0
        sLog += "--- Loaded Sheets ---\n"
        for sheet in self.__wb:
            sSheet = sheet.title
            if sSheet in self.__lstSheetsUse:
                pn("--- " + sSheet)
                iErr = 0
                iWarn = 0
                iInfo = 0
                iRowsSheetMax = sheet.max_row
                dicUniqProSubset__AidLifestageSieveSpecification = {}

                # Loop: Rows
                for row in sheet.iter_rows():
                    lstDebug = []
                    # Abort
                    if self.isAbort():
                        res.addErr("Aborted by user...", sys="Ingestor", kat="Abort")
                        return res

                    iRowNr = row[0].row
                    sRowNr = str(iRowNr)

                    iRowsTotalNow += 1

                    status("Reading sheet: "+sSheet+" #"+str(iRowNr)+"/"+str(iRowsSheetMax))
                    # Progress
                    iErrGes = res.getErrNo()+iErrStart
                    iWarnGes = res.getWarnNo()+iWarnStart
                    iInfoGes = res.getInfoNo()+iInfoStart
                    progress(iRowsTotalNow, iRowsTotalMax, 100, proz=True, cli=True,
                             err=iErrGes,
                             warn=iWarnGes,
                             info=iInfoGes)

                    # Header -> Überspringen
                    if iRowNr > 1:
                        if ExcelTools.isRowEmpty(row):
                            # Zeile leer
                            res.addInfo("Skipping empty row", sheet=str(sSheet), row=iRowNr, sys="Ingestor", kat="Idle")
                            iInfo += 1
                        else:
                            # Zeile nicht leer
                            # *** Hole PK
                            sPk = None
                            lstPk = self.__map.ingest_getPk(sSheet, self.__sWf)
                            if lstPk is not None and len(lstPk) > 0: # Sind für Sheet Pks definiert?
                                sPk = ""
                                for sCol in lstPk:
                                    cell, cellVal, sCellVal = getCell(self.__data, row, sSheet, sCol)
                                    # Fehlender PK-Teil -> Fehler
                                    if self.__map.ingest_isObl(sSheet, sCol):
                                        if cellVal is None:
                                            res.addErr("Missing (part of) Primary Key", sheet=sSheet, row=sRowNr, col=sCol, sys="Ingestor", kat="Extracting PK and FK")
                                    # Separator-Zeichen in FK gefunden
                                    if sPkFkSep in str(cellVal):
                                        res.addErr("Forbidden Character found in part of Foreign Key: " + sPkFkSep,
                                            sheet=sSheet, row=sRowNr, col=sCol, sys="Ingestor", kat="Extracting PK and FK")
                                    sPk+=sCellVal+sPkFkSep
                                sPk = sPk[:-1]

                            # *** Hole FK
                            sFk = None
                            lstFk = self.__map.ingest_getFk(sSheet, self.__sWf)
                            if lstFk is not None and len(lstFk) > 0: # Sind für Sheet FKs definiert?
                                sFk = ""
                                for sCol in lstFk:
                                    cell, cellVal, sCellVal = getCell(self.__data, row, sSheet, sCol)
                                    # Fehlender FK -> Fehler
                                    if cellVal is None:
                                        res.addErr("Missing (part of) Foreign Key",sheet=sSheet, row=sRowNr, col=sCol, sys="Ingestor", kat="Extracting PK and FK")
                                    # Separator-Zeichen in FK gefunden
                                    if sPkFkSep in str(cellVal):
                                        res.addErr("Forbidden Character found in part of Foreign Key: " +sPkFkSep,
                                                   sheet=sSheet, row=sRowNr, col=sCol, sys="Ingestor", kat="Extracting PK and FK")
                                    sFk+=sCellVal+sPkFkSep
                                sFk = sFk[:-1]

                            # *** Hole Daten
                            odicRow = OrderedDict()
                            for sCol in self.__map.ingest_getExcelCols(sSheet,self.__sWf):
                                cell, cellVal, sCellVal = getCell(self.__data, row, sSheet, sCol)

                                # *** Check: Teste jede Zelle on-the-fly
                                resCheck = checkCell(sSheet, sCol, cell, iRowNr)
                                if resCheck.isAnyMessage():
                                    res.addResult(resCheck, msg_only=True)
                                if resCheck.isWarn():
                                    iWarn += resCheck.getWarnNo()
                                if resCheck.isInfo():
                                    iInfo += resCheck.getInfoNo()
                                if resCheck.isErr():
                                    iErr += resCheck.getErrNo()
                                sDbType = resCheck.getData("dbtype")
                                sCellValType = resCheck.getData("type")
                                sValChecked = resCheck.getData("val")
                                odicRow[sCol]=sValChecked
                                if self.__model.conf_isDebug():
                                    lstDebug.append(sValChecked+" ["+str(sCellValType)+"/"+str(sDbType)+"]")

                            # *** Biota
                            if sSheet == "Biota":
                                # --- Extrahiere alle nötigen Daten
                                try:
                                    iAid = row[iIdxBiotaAphiaId].value
                                except:
                                    iAid = None
                                try:
                                    sScientificName = row[iIdxBiotaScientificName].value
                                except:
                                    sScientificName = None
                                try:
                                    sLifeStage= row[iIdxBiotaLifestage].value
                                except:
                                    sLifeStage = None
                                try:
                                    sSieve = row[iIdxBiotaSieve].value
                                except:
                                    sSieve = None
                                try:
                                    sSpecification = row[iIdxBiotaSpecification].value
                                except:
                                    sSpecification = None

                                # --- Sichere AphiaId mit Scientificname für den folgenden MyWoRMS-On-The-Fly-Test
                                try:
                                    iTaxonAid = int(row[iIdxBiotaAphiaId].value)
                                    # Speichern
                                    if iTaxonAid not in self.__dicBiotaAll:
                                        self.__dicBiotaAll[iTaxonAid] = set()
                                    self.__dicBiotaAll[iTaxonAid].add(sScientificName)
                                except:
                                    res.addErr("Aphia Id is no Integer",
                                               add=str(iAid), sheet=sSheet,
                                               col="AphiaId", row=iRowNr,
                                               kat="Biota Detection",
                                               sys="Ingestor")
                                    iErr += 1

                                # *** Test: Uniq pro Subset (Nur nicht Special Taxa Aid>0): Tupel: Aid-Lifestage-Sieve-Specification
                                if not self.__model.subsystem_getTaxonWizard().isSpecialTaxon(iAid):
                                    if sFk not in dicUniqProSubset__AidLifestageSieveSpecification:
                                        dicUniqProSubset__AidLifestageSieveSpecification[sFk] = set()
                                    sUniqTupel = str(iAid)+"-"+str(sLifeStage)+"-"+str(sSieve)+"-"+str(sSpecification)
                                    if sUniqTupel in dicUniqProSubset__AidLifestageSieveSpecification[sFk]:
                                        res.addErr("Biota with tuple: AphiaId/LifeStage/Sieve/Specification not uniq per subset", add="subset:"+str(sFk)+" -> tupel:"+str(sUniqTupel), sheet=sSheet, row=iRowNr, kat="Biota Check", sys="Ingestor")
                                        iErr += 1
                                    else:
                                        dicUniqProSubset__AidLifestageSieveSpecification[sFk].add(sUniqTupel)

                            # Fehlenden Pk durch Proxy ersetzen (wird nicht weiter gebraucht; dient nur zur Speicherung im Dictionary)
                            # Problem: Biota -> Man kann ja nicht irgendeinen
                            if sPk is None:
                                sId = "row-"+sRowNr
                            else:
                                sId = sPk

                            # Daten Speichern in Sheet-Data-Map
                            resData = self.__data.addRow(sSheet, sId, sFk, odicRow, None)
                            if resData.isErr():
                                res.addResult(resData, msg_only=True)

                            # Debug: Row
                            sRowDebug = sRowNr+" pk["+sId+"] fk["+(sFk if sFk is not None else "-")+"] row["+PrettyText.lst2strC(lstDebug)+"]"
                            pd(sRowDebug)

                # Debug: All PK/FK
                if self.__model.conf_isDebug():
                    if self.__map.ingest_getPk(sSheet, self.__sWf) is not None:
                        pd("pk_all[" + (PrettyText.lst2str(self.__map.ingest_getPk(sSheet, self.__sWf),",") + "] [" + PrettyText.lst2strC(self.__data.getPks(sSheet)))+"]")
                    if self.__map.ingest_getFk(sSheet, self.__sWf) is not None:
                        pd("fk_all[" + (PrettyText.lst2str(self.__map.ingest_getFk(sSheet, self.__sWf),",") + "] [" + PrettyText.lst2strC(self.__data.getFks(sSheet)))+"]")

                # Info
                lstLogSheetErrWarnInfo = []
                if iErr > 0:
                    lstLogSheetErrWarnInfo.append(str(iErr)+"x errors")
                if iWarn > 0:
                    lstLogSheetErrWarnInfo.append(str(iWarn)+"x warnings")
                if iInfo > 0:
                    lstLogSheetErrWarnInfo.append(str(iInfo)+"x infos")
                if len(lstLogSheetErrWarnInfo) > 0:
                    sLogSheetErrWarnInfo = ": "+PrettyText.lst2strC(lstLogSheetErrWarnInfo)
                else:
                    sLogSheetErrWarnInfo = ""
                sLog += sSheet + sLogSheetErrWarnInfo+"\n"

        # Info
        if res.isAnyMessage():
            sLog += "\n"
            sLog += "ATTENTION: some incidents have occurred, see overview section at the beginning of this protocol\n"

        cleanUi()
        sLog += "\n"
        res.addData("log", sLog)

        return res

    # Füttere Daten in Datenbank
    def __feed(self):

        # Helper
        def cleanUi():
            progress("off")
            status()

        # Create SQL Cmd
        def createSql(row):
            res = Result("Create SQL ingest table")
            # Sheet, Table, Schema
            sSheet = row.getSheet()
            sTable = self.__map.ingest_sheet2table(sSheet)
            sSchema = self.__model.subsystem_getDb().account_active_getSchema()
            # PK/FKs/FK(main)
            lstPk = self.__map.ingest_getPk(sSheet, self.__sWf)
            lstFk = self.__map.ingest_getFk(sSheet, self.__sWf)
            sFkMainSheet, sFkMainTable, sFkMainCol = self.__map.ingest_getFkMain(sSheet)
            # SQL: Daten
            lstAtr = []
            dicVal = {}
            # Loop: Cols in Row
            for sCol in self.__map.ingest_getExcelCols(sSheet, self.__sWf):
                if self.__map.ingest_excel2db(sSheet,sCol) != (None, None):
                    # Col existiert
                    sVal = row.get(sCol)
                    if sVal not in ["", None]:
                        # Zelle nicht leer
                        if sCol == sFkMainSheet:
                            # Col: FK-Main (hole iD aus dem Puffer)
                            try:
                                iIdFk = self.__dicPkIds[sFkMainSheet][row.getFk()]
                                sVal = iIdFk
                            except:
                                res.addErr("No FK-id found in buffer", add=str(sSheet)+"."+str(sCol)+": " +str(sVal), sheet=sSheet, row=iRow, sys="Ingestor", kat="SQL")
                        elif self.__map.ingest_isLookup(sSheet, sCol):
                            # Col: Lookup (Hole die Id aus dem Lookup System)
                            sLupTable = self.__map.ingest_getLookup(sSheet, sCol)
                            resLup = self.__lup.getLookup(sLupTable, sVal, cursor=self.__cursor, update=self.__bLupSearchSingle)
                            if resLup.isErr():
                                res.addResult(resLup, msg_only=True)
                            sValId = resLup.getData("id")
                            sVal = sValId
                        # SQL Daten: füge Zellen-Inhalte hinzu
                        sTableTmp, sAtr = self.__map.ingest_excel2db(sSheet, sCol)
                        lstAtr.append(sAtr)
                        dicVal[sAtr] = sVal
            # SQL Daten: füge IngestId hinzu
            lstAtr.append("ingest_id")
            dicVal["ingest_id"] = self.__iIngestId
            # SQL: Cmd
            sqlCmd = sql.SQL(
                "INSERT INTO {schema}.{table} ({attr}) VALUES ({val}) RETURNING id").format(
                schema=sql.Identifier(sSchema),
                table=sql.Identifier(sTable),
                attr=sql.SQL(', ').join(map(sql.Identifier, lstAtr)),
                val=sql.SQL(', ').join(map(sql.Placeholder, lstAtr))
            )
            # Antwort
            res.addData("sql", sqlCmd)
            res.addData("data", dicVal)
            return res

        # Create new Record
        def createRec(row, iRow):
            resCreate = Result("Create new record")
            # Create Cmd
            resSql_create = createSql(row)
            if resSql_create.isErr():
                # Cmd Fehler
                resCreate.addResult(resSql_create, msg_only=True, sheet=row.getSheet(), row=iRow, sys="Ingestor", kat="Db Feeder")
            else:
                # Cmd Ok
                sSql = resSql_create.getData("sql")
                dicVal = resSql_create.getData("data")
                try:
                    self.__cursor.execute(sSql, dicVal)
                    iId = self.__cursor.fetchone()[0]
                    sSqlQuery = self.__cursor.query.decode("utf-8")
                    resCreate.addData("sql", sSqlQuery)
                    # Gab es Nachrichten von der Db?
                    lstNotices = self.__cursor.connection.notices
                    if lstNotices is not None and len(lstNotices) != 0:
                        # Durchlaufe alle Meldungen
                        for sNotice in lstNotices:
                            for sSqlNoticeCode in self.__dicDbMessageCodes:
                                # Enthält der String ein Meldungskürzel -> Auflösen
                                if sSqlNoticeCode in sNotice:
                                    sNotice = sNotice.replace(sSqlNoticeCode, "")
                                    sKatMsg = self.__dicDbMessageCodes[sSqlNoticeCode]
                            if "WARNING: " in sNotice:
                                # Warnung -> in Warnungen verpacken
                                sNoticeWarn = sNotice.replace("WARNING:", "")
                                sNoticeWarn = sNoticeWarn.strip()
                                resCreate.addWarn("Database message", add=sNoticeWarn, sheet=row.getSheet(), row=iRow, kat="Db Feeder", sys="Ingestor")
                            if "INFO: " in sNotice:
                                # Info -> in Info verpacken
                                sNoticeInfo = sNotice.replace("INFO:", "")
                                sNoticeInfo = sNoticeInfo.strip()
                                resCreate.addInfo("Database message", add=sNoticeInfo, sheet=row.getSheet(), row=iRow, kat="Db Feeder", sys="Ingestor")
                            else:
                                # Rest
                                sNoticeRest = sNotice.replace("NOTICE:", "")  # NOTICE aus String entfernen
                                sNoticeRest = sNotice.strip()
                                resCreate.addInfo("Database message",add=sNoticeRest, sheet=row.getSheet(), row=iRow, kat="Db Feeder", sys="Ingestor")
                        self.__cursor.connection.notices = list() # Notices wieder löschen

                    # PK -> Puffer
                    if sSheet not in self.__dicPkIds:
                        self.__dicPkIds[sSheet] = {}
                    self.__dicPkIds[sSheet][row.getPk()] = iId

                    # Stat
                    self.__cntRecNew[sSheet] += 1
                except Exception as ex:
                    resCreate.addErr("Create new record", add=str(ex), sheet=sSheet, sys="Ingestor", kat="Db Feeder")
            return resCreate

        # Teste Gleichheit: rec in == rec db
        def recDiff(row, iId):
            resDiff = Result("Check diff rec in==db")
            # Sheet, Table, Schema
            sSheet = row.getSheet()
            sTable = self.__map.ingest_sheet2table(sSheet)
            sSchema = self.__model.subsystem_getDb().account_active_getSchema()
            # PK/FKs/FK(main)
            lstPk = self.__map.ingest_getPk(sSheet, self.__sWf)
            lstFk = self.__map.ingest_getFk(sSheet, self.__sWf)
            sFkMainSheet, sFkMainTable, sFkMainCol = self.__map.ingest_getFkMain(sSheet)
            # Row (Excel)-> Dic (Db)
            dicTable = {"id": iId}
            # Loop: Cols in Row
            for sCol in self.__map.ingest_getExcelCols(sSheet, self.__sWf):
                if self.__map.ingest_excel2db(sSheet, sCol) != (None, None):
                    # Col existiert
                    sVal = row.get(sCol)
                    if sVal not in ["", None]:
                        # Zelle nicht leer
                        if sCol == sFkMainSheet:
                            # Col: FK-Main (hole iD aus dem Puffer)
                            try:
                                iIdFk = self.__dicPkIds[sFkMainSheet][row.getFk()]
                                sVal = iIdFk
                            except:
                                res.addErr("Record Extension: Diff Check with DB failed", add="No FK-id found in buffer", sys="Ingestor", kat="DB Feeder")
                        elif self.__map.ingest_isLookup(sSheet, sCol):
                            # Col: Lookup (Hole die Id aus dem Lookup System)
                            sLupTable = self.__map.ingest_getLookup(sSheet, sCol)
                            resLup = self.__lup.getLookup(sLupTable, sVal, cursor=self.__cursor, update=self.__bLupSearchSingle)
                            if resLup.isErr():
                                res.addResult(resLup, msg_only=True)
                            sValId = resLup.getData("id")
                            sVal = sValId
                        # Baue Dictionary zuammen
                        sTableTmp, sAtr = self.__map.ingest_excel2db(sSheet, sCol)
                        dicTable[sAtr] = sVal
            # Text: Row == Db
            resDbDiff = self.__model.subsystem_getDb().check_isRecordExistingInTable(sTable,"id",dicTable,True,cursor=self.__cursor)

            # Antwort
            resDiff.addResult(resDbDiff)

            return resDiff

        # Extend Record
        def extendRec(row):
            resExtend = Result("Extend record in database")
            bExtent = False
            if row.getSheet() in ("Cruise", "Station", "Sample", "SubSample"):
                # Test: Metadaten Ast - exisitert Ast bereits?
                resExtensionTest = self.__model.subsystem_getDb().check_isLeafExisting(row=row, cursor=self.__cursor)
                if resExtensionTest.isOk():
                    # Kein Fehler bei Ast überprüpfung
                    sExtendedLeaf_string = resExtensionTest.getData("leaf")
                    sExtendedLeaf_string_verbose = resExtensionTest.getData("leaf_verbose")
                    sExtendedLeaf_level = resExtensionTest.getData("leaf_level")

                    if sExtendedLeaf_level == "subset":
                        # *** Metadaten-Ast: Cruise-Station-Sample-SubSample-Subset -> kann nicht erweitert werden -> nur Überprüfung ob es ihn schon gibt.
                        lstUsedSubsetNames = resExtensionTest.getData("subset_names_used")
                        resExtend.addErr("Subset already existing in Database", add=str(sExtendedLeaf_string) + " -> already used subset names for sample: " + PrettyText.lst2strC(lstUsedSubsetNames))

                    elif sExtendedLeaf_level in ("cruise", "station", "sample", "subsample"):
                        # *** Metadaten-Ast: Cruise[[-Station]-Sample] -> kann erweitert werden -> Test ob eine Erweiterung stattfinden soll
                        # Hole Id für gesuchten Ast
                        lstMap = (("Cruise", "cruise_id"), ("Station", "station_id"), ("Sample", "sample_id"), ("SubSample", "subsample_id"))
                        for sSheetMap, sSheetMapTag in lstMap:
                            if sSheet == sSheetMap:
                                iId = resExtensionTest.getData(sSheetMapTag)
                                if iId is not None:
                                    # Id wurde gefunden -> Extension is möglich
                                    bExtent = True
                                    # *** Test: rec in==db
                                    resDiff = recDiff(row, iId)
                                    if resDiff.isErr():
                                        # Fehler -> raus
                                        resExtend.addResult(resDiff, err=True)
                                        break
                                    if not resDiff.getData("found"): # Eintrag nicht gefunden -> raus (darf eigentlich nicht passieren)
                                        resExtend.addErr("Record Extension: Diff Check with DB failed (this is strange)", add="No Record found in Db", sys="Ingestor", kat="DB Feeder")
                                        break
                                    bEqual= resDiff.getData("equal")
                                    dicSearch = resDiff.getData("search")
                                    dicAnswerFull = resDiff.getData("answer-full")
                                    dicAnswer = resDiff.getData("answer")
                                    dicDiff = resDiff.getData("diff")

                                    if bEqual is not True:
                                        lstDiff = []
                                        for k, v_db in dicDiff.items():
                                            v_in = dicSearch[k]
                                            lstDiff.append(
                                                str(k) + "{in=\'" + str(v_in) + "\' db=\'" + str(v_db) + "\'}")
                                        # *** Nicht gleich: Error/Warn/Info
                                        if self.__sExtendReact== "info":
                                            resFkt = resExtend.addInfo
                                        elif self.__sExtendReact == "warning":
                                            resFkt = resExtend.addWarn
                                        else:
                                            resFkt = resExtend.addErr
                                        resFkt("Record extension: Input Data varies from Db Data", add=sExtendedLeaf_string_verbose+": "+PrettyText.lst2strC(lstDiff), kat="Db Feeder", sys="Ingestor")
                                        self.__cntRecExt_diff[sSheet] += 1 # Log
                                        if sSheet not in self.__dicRecExt_diff:
                                            self.__dicRecExt_diff[sSheet] = []
                                        self.__dicRecExt_diff[sSheet].append(sExtendedLeaf_string_verbose+": "+PrettyText.lst2strC(lstDiff))

                                    # *** Save Puffer: PK (Id des zu benutzenden/erweiternden Eintrages - Max Level des Astes: Cruise, Station, Sample)
                                    if sSheet not in self.__dicPkIds:
                                        self.__dicPkIds[sSheet] = {}
                                    self.__dicPkIds[sSheet][row.getPk()] = iId

                                    # *** Protokoll
                                    # --- Protokolliere den erweiterten Ast
                                    if sExtendedLeaf_level == "cruise":
                                        self.__setCruiseExtended.add(sExtendedLeaf_string)
                                    elif sExtendedLeaf_level == "station":
                                        self.__setStationExtended.add(sExtendedLeaf_string)
                                    elif sExtendedLeaf_level == "sample":
                                        self.__setSampleExtended.add(sExtendedLeaf_string)
                                    elif sExtendedLeaf_level == "subsample":
                                        self.__setSubSampleExtended.add(sExtendedLeaf_string)
                                    # --- Abhängigen Ingests
                                    self.__setIngestDepend = self.__setIngestDepend.union(
                                        resExtensionTest.getData("ingests"))
                                    # --- Datasets mit gleichen Abhängigkeiten
                                    self.__setDatasetDepend = self.__setDatasetDepend.union(
                                        resExtensionTest.getData("datasets"))
                                    # --- Extended Records
                                    self.__cntRecExt[sSheet] += 1
                                    break
                                else:
                                    # Id wurde nicht gefunden -> Keine Extension möglich
                                    pass
                    else:
                        # *** Metadaten-Ast: Länge ist nicht deinfiert -> Kann niemals passieren (entweder ist länge definierbar oder Fehler)
                        resExtend.addErr("That is strange: Leaf level is not determinable", add=str(sExtendedLeaf_string))
                else:
                    # *** Metadaten-Ast ist fehlerhaft
                    # - Entweder ist der Ast unvollständig definiert
                    # - oder die Datenbank Struktur ist kaputt (darf eigentlich niemals passieren!)
                    resExtend.addResult(resExtensionTest, err=True)

            resExtend.addData("extended", bExtent)
            return resExtend

        # --------

        res = Result(name="Feeding Db")

        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug
        progress = self.__print.progress
        status = self.__print.status

        sLog = ""
        sLog += "------------------------------------\n"
        sLog += "Feeding Database\n"
        sLog += "------------------------------------\n\n"

        pn("* Feeding Db")

        # Rows
        iRow = 0
        iRowsTotal = self.__data.getRowNumber()

        # EWI End-Status des letzten Modules
        iErrStart = self.__resultAll.getErrNo()
        iWarnStart = self.__resultAll.getWarnNo()
        iInfoStart = self.__resultAll.getInfoNo()

        # ---

        # Sheets
        for sSheet in self.__map.ingest_getExcelSheets(self.__sWf):

            # Abort
            if self.isAbort():
                res.addErr("Aborted by user...", sys="Ingestor", kat="DB Feeder")
                return res

            # Rows
            iRowSheet = 0
            for row in self.__data(iter_sheet=sSheet):

                # *** Extend DB Record (etwas teuer, da es vor jedem Create gemacht werden muss, Leider geht es nicht anders, da ein Create mit Fehler die Session versaut.)
                bExtended = False
                if res.isOk():
                    if self.__bExtend: # Erlaubt?
                        resExtend = extendRec(row)
                        if resExtend.isAnyMessage():
                            res.addResult(resExtend, msg_only=True, sheet=sSheet, row=iRowSheet, sys="Ingestor", kat="DB Feeder")
                        bExtended = resExtend.getData("extended")
                        bExtended = bExtended if (bExtended is not None) else False

                # *** Create new Record
                if res.isOk():
                    if not bExtended:
                        resCreate = createRec(row, iRowSheet)
                        if resCreate.isAnyMessage():
                            res.addResult(resCreate, msg_only=True, sheet=sSheet, row=iRowSheet, sys="Ingestor", kat="DB Feeder")
                # Progress
                iErr = res.getErrNo() + iErrStart
                iWarn = res.getWarnNo() + iWarnStart
                iInfo = res.getInfoNo() + iInfoStart
                progress(iRow, iRowsTotal, 100, proz=True, cli=True, err=iErr, warn=iWarn, info=iInfo)
                status("Feeding db: " + sSheet + " #" + str(iRowSheet) + "/" + str(iRowsTotal))

                iRow += 1
                iRowSheet += 1
        # ---

        # Hole alle Dateinnamen zu Ingest
        resFilename = self.__model.subsystem_getDb().ingest_getIngestFilenames(self.__setIngestDepend)
        self.__dicIngestDependFiles = resFilename.getData("dic_files")

        # Info
        if len(self.__cntRecNew) > 0:
            sLog += "\n"
            sLog += "--- Created records ---\n"
            sLog += "New records: " + str(sum(self.__cntRecNew.values())) + "\n"
            for sTable, iNr in self.__cntRecNew.items():
                sLog += sTable + ": " + str(iNr) + "\n"

        if len(self.__cntRecExt) > 0:
            sLog += "\n"
            sLog += "--- Extended records ---\n"
            sLog += "Extended records: " + str(sum(self.__cntRecExt.values())) + "\n"
            for sTable, iNr in self.__cntRecExt.items():
                sLog += sTable + ": " + str(iNr) + "\n"
            if len(self.__setCruiseExtended) > 0:
                sLog += "Extended Cruises: " + PrettyText.lst2strC(self.__setCruiseExtended) + "\n"
            if len(self.__setStationExtended) > 0:
                sLog += "Extended Stations: " + PrettyText.lst2strC(self.__setStationExtended) + "\n"
            if len(self.__setSampleExtended) > 0:
                sLog += "Extended Samples: " + PrettyText.lst2strC(self.__setSampleExtended) + "\n"
            sLog += "\n"
            if sum(self.__cntRecExt_diff.values()) > 0:
                sLog += "--- Hiccups while extending ---\n"
                sLog += "Differing record values while extending: "+str(len(self.__cntRecExt_diff.values()))+"\n"
            for sTable, lstDiff in self.__dicRecExt_diff.items():
                sLog += "* "+sTable+": "+str(len(lstDiff))+"\n"
                for sDiff in lstDiff:
                    sLog += sDiff +"\n"
            sLog += "\n"
            sLog += "--- Dependencies ---\n"
            if len(self.__setIngestDepend) > 0:
                sLog += "* Depending on other ingests: " + str(len(self.__setIngestDepend)) + "\n"
                for iIngId, sFile in self.__dicIngestDependFiles.items():
                    sLog += str(iIngId) + " (" + sFile + ")\n"
                sLog += "\n"
            if len(self.__setDatasetDepend) > 0:
                sLog  += "* Other datasets with same dependancies: " + str(
                    len(self.__setDatasetDepend)) + "\n" + PrettyText.lst2strR(self.__setDatasetDepend) + "\n"

        # Aufräumen
        cleanUi()
        # Antwort
        res.addData("log", sLog)
        return res

    #  ------------------------------------------------------------------------------------------------------------------

    # Start Ingestion
    def startIngest(self, wb, sWf, sFile, **kwargs):
        """
        Args:
            wb: Workbench
            sWf (str): Workflow
        Kwargs:
            result_in: Result In
            cursor: Db Cursor
            extend (bool): Extension allowed
            extend_react (str): Extension reaction if not matching 100%
            lup_search_single (bool): Teste Lookup -> Befrage einzeln (oder am Anfang alles herunterladen)
        Returns:
            Result:
        """

        # Helper: Aufraeumen
        def cleanUi():
            # Gui
            progress("off")
            status()
        def closeDb(res):
            # Dry
            if self.__model.conf_isDry():
                self.__conn.rollback()
            elif res.isErr():
                self.__conn.rollback()
            else:
                self.__conn.commit()
            self.__cursor.close()
            self.__conn.close()

        # --------

        res = Result(name="Ingestor")

        pn = self.__print.normal
        pv = self.__print.verbose
        pd = self.__print.debug

        workflow = self.__print.workflow
        progress = self.__print.progress
        status = self.__print.status

        # Parameter
        self.__initData()
        self.__resultAll = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        self.__wb = wb
        self.__sFile = sFile
        self.__sWf = sWf
        self.__lstSheetsSelected = kwargs["sheets"] if "sheets" in kwargs else None
        self.__lstSheetsUse = []
        self.__bLupSearchSingle = kwargs["lup_search_single"] if "lup_search_single" in kwargs else False
        self.__bExtend = kwargs["extend"] if "extend" in kwargs else self.__model.conf_isRecordExtensionAllowed()
        self.__sExtendReact = kwargs["extend_react"] if "extend_react" in kwargs else self.__model.conf_getRecordExtensionProblemReactionlevel()
        self.__cursor = kwargs["cursor"] if "cursor" in kwargs else None
        self.__conn = None

        progress("on-pulse")
        pn("* Importing Ingestion Data")
        workflow(top="Ingestion", sub="Importing Data", fmt=False)
        status()

        # Test
        if self.__bExtend and self.__sExtendReact not in ["error","warning","info"]:
            res.addErr("Extension reaction level unknown", add=str(self.__sExtendReact), sys="Ingestor", kat="Configuration")
        if wb is None:
            res.addErr("Excel Workbench not set", sys="Ingestor", kat="Configuration")
        if sWf not in self.__map.workflow_getAll():
            res.addErr("Workflow unknown", add=str(sWf), sys="Ingestor", kat="Configuration")
        if sFile is None:
            res.addErr("Filename unknown", add=str(sFile), sys="Ingestor", kat="Configuration")

        # Eigene DB Connection
        if self.__cursor is None:
            bOwnConnection = True
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                return res
            self.__conn = resConn.getData("connection")
            self.__cursor = resConn.getData("cursor")
        else:
            bOwnConnection = False

        # Bestimme die Sheets die effektiv benutzt werden sollen
        self.__lstSheetsUse = [sSheet for sSheet in self.__map.ingest_getExcelSheets(self.__sWf) if ((self.__lstSheetsSelected is None) or (sSheet in self.__lstSheetsSelected))]

        # ---

        # Log
        sLog = "************************************\n"
        sLog += "Ingestion\n"
        sLog += "************************************\n"
        sLog += "\n"
        sLog += "--- Configuration ---\n"
        sLog += "Workflow: " + str(self.__map.workflow_short2long(sWf)) + "\n"
        sLog += "Allowed sheets: " + PrettyText.lst2strC(self.__map.ingest_getExcelSheets(self.__sWf)) + "\n"
        if self.__lstSheetsSelected is not None:
            sLog += "Selected sheets: " + PrettyText.lst2strC(self.__lstSheetsSelected) + "\n"
        sLog += "Use sheets: " + PrettyText.lst2strC(self.__lstSheetsUse) + "\n"
        sLog += "Lookup search mode: " + ("ask db for every single lookup" if self.__bLupSearchSingle else "load full cache at start") + "\n"
        sLog += "Update of records allowed: " + ("yes" if self.__bExtend else "no") + "\n"
        sLog += "Update of records - reaction level if differences occurr: " + str(self.__sExtendReact) + "\n"
        sLog += "\n"
        iSteps = 7

        # Info
        pn("* Ingestion")
        if bOwnConnection:
            pd("Using own Db connection")
        else:
            pd("Reusing Db connection")
        pv("Workflow: " + str(self.__map.workflow_short2long(sWf)))
        pv("Allowed sheets: " + PrettyText.lst2strC(self.__map.ingest_getExcelSheets(self.__sWf)))
        if self.__lstSheetsSelected is not None:
            pv("Selected sheets: " + PrettyText.lst2strC(self.__lstSheetsSelected))
        pv("Use sheets: " + PrettyText.lst2strC(self.__lstSheetsUse))

        # Load
        if res.isOk():
            progress("on-pulse-abort")
            workflow(top="Ingestion", sub="Chew data", fmt=False)
            resLoad = self.__load()
            if resLoad.isOk():
                sLog += resLoad.getData("log")
            if resLoad.isAnyMessage():
                res.addResult(resLoad, msg_only=True)
                self.__resultAll = res
            progress("on-pulse-abort")

        # Check Lookup
        if res.isOk():
            workflow(top="Ingestion", sub="Checking Lookup", fmt=False)
            resCheckLookup = self.__check_lookup()
            if resCheckLookup.isOk():
                sLog += resCheckLookup.getData("log")
            if resCheckLookup.isAnyMessage():
                res.addResult(resCheckLookup, msg_only=True)
                self.__resultAll = res

        # Check Taxa
        if res.isOk():
            workflow(top="Ingestion",  sub="MyWoRMS (on-the-fly)", fmt=False)
            resCheckTaxa = self.__check_taxa__mywormsOnthefly()
            if resCheckTaxa.isOk():
                sLog += resCheckTaxa.getData("log")
            if resCheckTaxa.isAnyMessage():
                res.addResult(resCheckTaxa, msg_only=True)
                self.__resultAll = res

        # TaxonWizard: Update Taxon List / Base List
        if res.isOk():
            workflow(top="Ingestion", sub="TaxonWizard update", fmt=False)
            resAddBaseUpdateTaxonlist = self.__check_taxa__quickUpdateTaxonLists()
            if resAddBaseUpdateTaxonlist.isOk():
                sLog += resAddBaseUpdateTaxonlist.getData("log")
            if resAddBaseUpdateTaxonlist.isAnyMessage():
                res.addResult(resAddBaseUpdateTaxonlist, msg_only=True)
                self.__resultAll = res

        # Check PK-FK
        if res.isOk():
            progress("on-pulse-abort")
            workflow(top="Ingestion", sub="Checking PK/FK",fmt=False)
            resCheckFkPk = self.__check_fkpk()
            if resCheckFkPk.isOk():
                sLog += resCheckFkPk.getData("log")
            if resCheckFkPk.isAnyMessage():
                res.addResult(resCheckFkPk, msg_only=True)
                self.__resultAll = res

        # Log Ingest Entry (create new)
        if res.isOk():
            progress("on-pulse-abort")
            workflow(top="Ingestion", sub="Create Ingest Log", fmt=False)
            resCreateIngestLog = self.__model.subsystem_getDb().inglog_new(cursor=self.__cursor, wf=self.__sWf)
            self.__iIngestId = resCreateIngestLog.getData("id")
            if resCreateIngestLog.isAnyMessage():
                res.addResult(resCreateIngestLog, msg_only=True)
                self.__resultAll = res

        # Feed Db
        if res.isOk():
            progress("on-pulse-abort")
            workflow(top="Ingestion", sub="Feeding Db", fmt=False)
            resFeed = self.__feed()
            if resFeed.isOk():
                sLog += resFeed.getData("log")
            if resFeed.isAnyMessage():
                res.addResult(resFeed, msg_only=True)
                self.__resultAll = res

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb(res)

        # Antwort
        res.addData("ingest-biota-new", len(self.__dicBiotaNew))
        res.addData("ingest-counter-new", self.__cntRecNew)
        res.addData("ingest-counter-update", self.__cntRecExt)
        res.addData("ingest-ingest-id", self.__iIngestId)
        res.addData("ingest-ingest-depend", self.__setIngestDepend)
        res.addData("ingest-dataset-depend", self.__setDatasetDepend)
        res.addData("ingest-cruise-extend", self.__setCruiseExtended)
        res.addData("ingest-station-extend", self.__setStationExtended)
        res.addData("ingest-sample-extend", self.__setSampleExtended)

        # Log
        res.addData("log", sLog)
        res.addData("log_stat", self.__createStat())
        return res


