# coding=utf-8

import openpyxl
from collections import Counter, OrderedDict, namedtuple
import csv, time, psycopg2, warnings, traceback
from tools.coding import Result, WorkingObject, DataTools
from tools.excel import ExcelTools
from tools.misc import JsonConverter
from tools.text import PrettyText
from tools.excel import ExcelTools
from tools.parallel import Stoppable

# -------------------------------------
# Lookuo (Hält die Lookup Daten)
# -------------------------------------

class Lookup(Stoppable, WorkingObject):

    # Konstruktor
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        self.__model = model
        self.__map = model.subsystem_getMapper()
        self.__smartprinter = model.ui_getSmartprinter()
        self.__lookupDbCache = None
        self.__lookupDbMissName = None
        self.__lookupDbMissId = None

        # Abhängigkeiten Testen
        if not self.__model.subsystem_getMapper().isReady():
            self.setWorking(False)

        # Init
        self.reset()
        self.setWorking(True)

    # To String
    def __str__(self):
        return self.info()

    # ----------------------------------------------------------------------------------------------------

    # Info
    def info(self):

        sOut = "*** Status of Lookup System\n\n"
        bErr = False

        sConn = self.__model.subsystem_getDb().conn_getConnectionInfo()
        bConnected = self.__model.subsystem_getConnector().db_isConnected()
        sOut += "Server: " + str(sConn) + "\n"
        sOut += "Connected: " + ("yes" if bConnected else "no") + "\n\n"

        if bConnected:
            resCache = self.getStatusCache()
            resDb = self.getStatusDb()
            resDbMiss = self.getStatusDbMiss__search_name2id()
            resDbMissRev = self.getStatusDbMiss__search_id2name()

            dicCache = resCache.getData("status")
            dicDb = resDb.getData("status")
            setDbMiss = resDbMiss.getData("status")
            setDbMissRev = resDbMissRev.getData("status")

            if resCache.isErr():
                bErr = True
                sOut += resCache.getErr()+"\n"
            if resDb .isErr():
                bErr = True
                sOut += resDb.getErr() + "\n"
            if dicCache is None:
                bErr = True
                sOut += "No Data received: status cache"
            if dicDb is None:
                bErr = True
                sOut += "No Data received: status DB"
            if setDbMiss is None:
                bErr = True
                sOut += "No Data received: DB miss name2id"
            if setDbMissRev is None:
                bErr = True
                sOut += "No Data received: DB miss id2name"

            if not bErr:
                try:
                    lstStatus = [["Lookup", "DB-cache", "DB", "DB-miss name2id", "DB-miss id2name"]]
                    for sLookup in self.__model.subsystem_getMapper().lookup_getDbTables():
                        lstStatus.append([sLookup, dicCache[sLookup], dicDb[sLookup], setDbMiss[sLookup], setDbMissRev[sLookup]])
                    sOut += PrettyText.table(lstStatus, True)
                except Exception as ex:
                    sOut += str(ex)
        return sOut

    # -----------------------------------------------------------------------------

    # Hole alle Lookup aus DB
    def __getLookupDbAll__name2id(self, sLookup, **kwargs):
        """
        Lookup: Hole ID zu name
        args:
            (str) sLookup: Lookup Table
        kwargs:
            (Cursor) cursor: DB Cursor
        Return:
           Result
               (dic) lookup: dictionary mit lookup name->id
        """
        res = Result(naem="Get all lookups from table: "+sLookup)
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnConneciton = False if cursor is not None else True
        try:
            if bOwnConneciton:
                # Eigene Connection: An
                conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
                cursor = conn.cursor()

            sKeyAttrib = self.__model.subsystem_getMapper().lookup_getKey(sLookup)
            if sKeyAttrib is not None:
                sSqlCmd = "SELECT " + sKeyAttrib + ", id from " + self.__model.subsystem_getDb().account_single_getPart("schema") + "." + sLookup + ";"
                cursor.execute(sSqlCmd)
                lstAntwort = cursor.fetchall()
                dicLookup = dict((str(name), str(id)) for name, id in lstAntwort)
                res.addData("lookup", dicLookup)
        except Exception as e:
            res.addErr("Connection to database", add=str(e), sys="Lookup Import", kat="Database Connection")
        # Eigene Connection: Aus
        if bOwnConneciton:
            try:
                cursor.close()
                conn.close()
            except:
                pass
        return res

    # Lookup DB: Hole Id zu einem Namen
    def __getLookupDbSingle__name2id(self, sLookup, sName, **kwargs):
        """
        Lookup: Hole ID zu name
       args:
            (str) sLookup: Lookup Table
            (str) sNaame: Lookup Name
       kwargs:
            (Cursor) cursor: DB Cursor
       Return:
           Result
               (int) id: id zu lookup name
               (bool) found: gefunden
       """
        res = Result(name="Lookup single: Get (name->)id: " + sLookup)
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnConnection = True if cursor is None else False

        # Lookup muss in DB gesucht werden
        try:
            if bOwnConnection:
                # Eigene Connection: An
                conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
                cursor = conn.cursor()
            sKeyAttrib = self.__model.subsystem_getMapper().lookup_getKey(sLookup)
            if sKeyAttrib is not None:
                sSqlCmd = "SELECT id from " + self.__model.subsystem_getDb().account_single_getPart("schema") + "." + sLookup + " where " + sKeyAttrib + "='" + str(sName) + "'"
                cursor.execute(sSqlCmd)
                dbAntwort = cursor.fetchone()
                iId = dbAntwort[0] if dbAntwort is not None else None
                if iId is not None:
                    res.addData("id", iId)
                    res.addData("found", True)
                else:
                    res.addData("found", False)
        except Exception as e:
            res.addErr("Connection to database", add=str(e), sys="Lookup Import", kat="Database Connection")
            res.addData("found", False)

        # Eigene Connection: Aus
        if bOwnConnection:
            try:
                cursor.close()
                conn.close()
            except:
                pass

        return res

    # Lookup DB: Hole Namen zu einer Id
    def __getLookupDbSingle__id2name(self, sLookup, iId, **kwargs):
        """
        Lookup: Hole ID zu name
        args:
            (str) sLookup: Lookup Table
            (str) iId: Lookup Id
        kwargs:
            (Cursor) cursor: DB Cursor
        Return:
            Result
                (str) name: name zu lookup id
                (bool) found: gefunden
        """
        res = Result(naem="Lookup single: get (id->)name: " + sLookup)
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        bOwnCOnnection = True if cursor is None else False

        # Lookup muss in DB gesucht werden
        try:
            if bOwnCOnnection:
                # Eigene  Connection: an
                conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
                cursor = conn.cursor()
            sKeyAttrib = self.__model.subsystem_getMapper().lookup_getKey(sLookup)
            if sKeyAttrib is not None:
                sSqlCmd = "SELECT "+sKeyAttrib+" from " + self.__model.subsystem_getDb().account_single_getPart(
                    "schema") + "." + sLookup + " where id ='" + str(iId) + "'"
                cursor.execute(sSqlCmd)
                dbAntwort = cursor.fetchone()
                sName = dbAntwort[0] if dbAntwort is not None else None
                if sName is not None:
                    res.addData("name", sName)
                    res.addData("found", True)
                else:
                    res.addData("found", False)
        except Exception as e:
            res.addErr("Connection to database", add=str(e), sys="Lookup Import", kat="Database Connection")
            res.addData("found", False)

        # Eigene Connection: aus
        if bOwnCOnnection:
            try:
                cursor.close()
                conn.close()
            except:
                pass
        return res

    # Checke Ingest-Daten
    def __checkImportDataBulk(self, excelFile, lstTabs=None):
        pv = self.__smartprinter.verbose
        res = Result(name="Check import bulk data")

        # --- Check: Sheets
        pv("- Checking sheets")
        lstLookupSheets = self.__model.subsystem_getMapper().lookup_getExcelSheets()
        lstExcelNames = excelFile.sheetnames
        lstSheetFound = []
        lstSheetMissing = []
        lstSheetsImport = []
        for sSheet in lstLookupSheets:
            if sSheet in lstExcelNames:
                lstSheetFound.append(sSheet)
                if (len(lstTabs) == 0) or (sSheet in lstTabs):
                    lstSheetsImport.append(sSheet)
            else:
                lstSheetMissing.append(sSheet)
        pv("Allowed sheets: " + PrettyText.lst2strC(lstLookupSheets))
        pv("Found allowed sheets in file: " + PrettyText.lst2strC(lstSheetFound))
        if len(lstTabs) > 0:
            pv("Only use these sheets: " + PrettyText.lst2strC(lstTabs))
        pv("These Sheets will be imported: " + PrettyText.lst2strC(lstSheetsImport))
        res.addData("sheetsImport", lstSheetsImport)


        # Vergleiche alle Attribute mit Headern aus entsprechendem Sheet
        dicHeader = {}
        dicHeaderIndex = {}
        for sSheet in lstSheetsImport:
            resExcelHeader = ExcelTools.getHeadIndex(excelFile[sSheet])
            lstExcelHeader = resExcelHeader.getData("header")
            dicExcelHeaderIndex = resExcelHeader.getData("index")
            if resExcelHeader.isErr() or lstExcelHeader is None:
                res.addResult(resExcelHeader)
                return res
            dicHeader[sSheet] = lstExcelHeader
            dicHeaderIndex[sSheet] = dicExcelHeaderIndex
            resCheck = DataTools.cmpLists(lstExcelHeader, self.__model.subsystem_getMapper().lookup_getExcelCols(sSheet))
            if resCheck.getData("istFehlt") is not None:
                res.addErr("Header not found in sheet", sheet=str(sSheet), add=PrettyText.lst2strC(
                    resCheck.getData("istFehlt")), sys="Lookup Import", kat="Header")
            if resCheck.getData("istDoppelt") is not None:
                res.addErr("Found double headers in sheet", sheet=sSheet, add=PrettyText.lst2strC(
                    resCheck.getData("istDoppelt")), sys="Lookup Import", kat="Header")
            if resCheck.getData("istPlus") is not None:
                pv("Found surplus headers in sheet (" + str(sSheet) + "): " + PrettyText.lst2strC(
                    resCheck.getData("istPlus")))
        res.addData("header", dicHeader)
        res.addData("headerIndex", dicHeaderIndex)
        # Bestimme Anzahl der gesamt einzulesenden Zeilenanzahl
        pv("- Getting sum of new entries")
        iRowsTotal = 0
        for sSheet in lstSheetsImport:
            pv(sSheet + ": " + str(excelFile[sSheet].max_row - 1))
            iRowsTotal += excelFile[sSheet].max_row - 1
        pv("total: " + str(iRowsTotal))
        res.addData("entriesTotal", iRowsTotal)
        return res

    # -----------------------------------------------------------------------------

    # Lösche Lookupsystem
    def reset(self, lookup=None):
        pd = self.__smartprinter.debug
        res = Result("Lookup Reset")
        lstLookupDbTables = self.__model.subsystem_getMapper().lookup_getDbTables()

        # Init
        if self.__lookupDbCache in ({}, None):
            self.__lookupDbCache = {}
            self.__lookupDbMissName = {}
            self.__lookupDbMissId = {}
            pd("* Init lookup: " + PrettyText.lst2strC(lstLookupDbTables))
            for sLookup in lstLookupDbTables:
                if sLookup not in self.__lookupDbCache:
                    self.__lookupDbCache[sLookup] = {}
                    self.__lookupDbMissName[sLookup] = set()
                    self.__lookupDbMissId[sLookup] = set()
            return res

        # Reset
        lstRstInfo = []
        if type(lookup) is str:
            # Reset 1 Lookup
            if lookup in lstLookupDbTables:
                self.__lookupDbCache[lookup] = {}
                self.__lookupDbMissName[lookup] = set()
                self.__lookupDbMissId[lookup] = set()
                lstRstInfo.append(lookup)
        elif type(lookup) is list:
            # Reset mehrere Lookups
            for sTab in lookup:
                if sTab in lstLookupDbTables:
                    self.__lookupDbCache[sTab] = {}
                    self.__lookupDbMissName[sTab] = set()
                    self.__lookupDbMissId[sTab] = set()
                    lstRstInfo.append(sTab)
        elif lookup is None:
            # Reset alle Lookups
            self.__lookupDbCache = {}
            for sTab in lstLookupDbTables:
                self.__lookupDbCache[sTab] = {}
                self.__lookupDbMissName[sTab] = set()
                self.__lookupDbMissId[sTab] = set()
                lstRstInfo.append(sTab)
        pd("* Reset lookups: "+PrettyText.lst2strC(lstRstInfo))
        return res

    # Erstelle Lookup System
    def update(self, lookup=None, **kwargs):
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow

        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        res = Result(name="Update lookup system")
        lstLookupDbTables = self.__model.subsystem_getMapper().lookup_getDbTables()
        lstUpdate = []
        if type(lookup) is str:
            # Update 1 Lookup
            if lookup in lstLookupDbTables:
                lstUpdate = [lookup]
        elif type(lookup) is list:
            # Update mehrere Lookups
            for sTab in lookup:
                if sTab in lstLookupDbTables:
                    lstUpdate.append(sTab)
        elif lookup is None:
            # Update alle Lookups
            lstUpdate = lstLookupDbTables.copy()

        if len(lstUpdate) == 0:
            res.addInfo("Nothing to update", sys="Lookup System", kat="Idle")
            return res

        # Hole alle Lookups
        pd("* Update lookups: "+PrettyText.lst2strC(lstUpdate))
        workflow(top="Lookup", sub="update")
        progress("pulse-on")
        for iNr, sTab in enumerate(lstUpdate):
            progress(iNr+1, len(lstUpdate), 100, text=sTab, abs=True)  # Fortschritt
            resLookup = self.__getLookupDbAll__name2id(sTab, cursor=cursor)
            if resLookup.isErr():
                res.addResult(resLookup, err=True)
            elif resLookup.getData("lookup") is not None:
                self.__lookupDbCache[sTab] = resLookup.getData("lookup")
                self.__lookupDbMissName[sTab] = set()
                self.__lookupDbMissId[sTab] = set()
        progress("off")
        return res

    # -----------------------------------------------------

    # Is Cache empty
    def isCacheEmpty(self):
        sRec = 0
        for sLookup, dic in self.__lookupDbCache.items():
            sRec += len(dic)
        if sRec == 0:
            return True
        else:
            return False

    # Status Cache
    def getStatusCache(self):
        res = Result("Lookup: Get Status Cache")
        try:
            dicStatus = {sLookup: len(dic) for sLookup, dic in self.__lookupDbCache.items()}
            res.addData("status", dicStatus)
        except Exception as ex:
            res.addErr("Status", add=str(ex), sys="Lookup System", kat="Cache")
        return res

    # Status DB
    def getStatusDb(self):
        res = Result(naem="Lookup: Get Status DB")
        dicStatus = {}
        lstLookupDbTables = self.__model.subsystem_getMapper().lookup_getDbTables()
        try:
            # Connection
            conn = psycopg2.connect(self.__model.subsystem_getDb().conn_getConnStrPsql())
            cursor = conn.cursor()
            for sLookup in lstLookupDbTables:
                sSqlCmd = "SELECT count(id) from " + self.__model.subsystem_getDb().account_single_getPart("schema") + "." + sLookup
                cursor.execute(sSqlCmd)
                dbAntwort = cursor.fetchone()
                erg = dbAntwort[0] if dbAntwort is not None else None
                dicStatus[sLookup] = erg
        except Exception as e:
            res.addErr("Connetion to database", add=str(e), sys="Lookup Import", kat="Database Connection")
        # Connection schliessen
        try:
            cursor.close()
            conn.close()
        except:
            pass
        res.addData("status", dicStatus)
        return res

    # Status DB Miss Name search
    def getStatusDbMiss__search_name2id(self):
        res = Result("Lookup: Status DB-Miss (search by name)")
        try:
            dicStatus = {sLookup: len(setMiss) for sLookup, setMiss in self.__lookupDbMissName.items()}
            res.addData("status", dicStatus)
        except Exception as ex:
            res.addErr("Status Miss", add=str(ex), sys="Lookup System", kat="Cache")
        return res

    # Status DB Miss Id Search
    def getStatusDbMiss__search_id2name(self):
        res = Result("Get Lookup Status DB-Miss (search by id)")
        try:
            dicStatus = {sLookup: len(setMiss) for sLookup, setMiss in self.__lookupDbMissId.items()}
            res.addData("status", dicStatus)
        except Exception as ex:
            res.addErr("Status Miss", add=str(ex), sys="Lookup System", kat="Cache")
        return res

    # -----------------------------------------------------

    # Hole alle Name->Ids zu einem Lookup
    def getLookupCacheAll(self, sLookup, **kwargs):
        bReverse = True if "reverse" in kwargs and kwargs["reverse"] in ["true", "True", True] else False
        res = Result(name="Get all Lookups from Cache for Table: "+str(sLookup))
        if not self.isReady():
            res.addErr("Lookup system is not working", sys="Lookup System", kat="Internal")
        if sLookup is None:
            res.addErr("Lookup table not given", sys="Lookup System", kat="Configuration")
        if sLookup not in self.__model.subsystem_getMapper().lookup_getDbTables():
            res.addErr("Lookup table not found", add=str(sLookup), sys="Lookup System", kat="Configuration")
        if res.isErr():
            return res
        try:
            if not bReverse:
                dic = self.__lookupDbCache[sLookup].copy()
            else:
                dic = {v: k for k, v in self.__lookupDbCache[sLookup].items()}
            res.addData("lookup", dic)
        except Exception as ex:
            res.addErr("Internal error", add=str(ex), sys="Lookup System", kat="Internal")
            return res
        return res

    # Hole Id zu einem Lookup Namen
    def getLookup(self, sLookup, name, **kwargs):
        """
        Hole id zu lookup name
        Args:
            (str) lookup: Lookup Table
            (str) name: Lookup Name
        Kwargs:
            (bool) update: Suche auch in Datenbank
            (Cursor) cursor: DB Cursor
        Return:
            Result
                (int) id: id of lookup name
                (bool) found: gefunden
                (bool) cache: gefunden in cache
                (bool) db: gefunden in datenbank
                (bool) db-miss-hit: nicht gefunden in cache und db -> aber in db-miss cache gefunden
                (bool) db-miss-new: nicht gefunden in cache und db -> eintrag in db-miss cache
        """
        bUpdate = True if "update" in kwargs and kwargs["update"] in ["true", "True", True] else False
        cursor = kwargs["cursor"] if "cursor" in kwargs else None

        res = Result(name="Lookup (name->id) Table["+str(sLookup)+"], name["+str(name) +"]:  search in "+("cache only" if not bUpdate else "cache and db"))
        if not self.isReady():
            res.addErr("Lookup system is not working", sys="Lookup System", kat="Internal")
            return res
        if sLookup is None:
            res.addErr("Lookup table not given", sys="Lookup System", kat="Configuration")
        if name is None:
            res.addErr("Name not given", kat="Lookup System")
        if sLookup not in self.__model.subsystem_getMapper().lookup_getDbTables():
            res.addErr("Lookup table not found", add=str(sLookup), sys="Lookup System", kat="Configuration")

        if res.isOk():
            if name in self.__lookupDbCache[sLookup]:
                # Key in DB Cache gefunden
                res.addData("id", self.__lookupDbCache[sLookup][name])
                res.addData("found", True)
                res.addData("cache", True)
            elif not bUpdate:
                res.addData("found", False)
            else:
                # Nicht gefunden im Cache -> Update aktiv -> Suche in DB
                if name in self.__lookupDbMissName[sLookup]:
                    # Key wurde bereits erfolglos in DB gesucht
                    res.addData("found", False)
                    res.addData("db-miss-hit", True)
                else:
                    # Key wurde noch nicht erfolglos in DB gesucht -> Suche in DB
                    resSingle = self.__getLookupDbSingle__name2id(sLookup, name, cursor=cursor)
                    if resSingle.isErr():
                        res.addResult(resSingle, err=True)
                    else:
                        if resSingle.getData("found"):
                            # Key in DB gefunden
                            res.addData("id", resSingle.getData("id"))
                            res.addData("found", True)
                            res.addData("db", True)
                            self.__lookupDbCache[sLookup][name] = resSingle.getData("id")
                        else:
                            # Key nicht in DB gefunden
                            res.addData("found", False)
                            res.addData("db-miss-new", True)
                            self.__lookupDbMissName[sLookup].add(name)
        return res

    # Hole den Namen zu einer Lookup Id
    def getLookupRev(self, sLookup, iId, **kwargs):
        """
        Hole name zu lookup id
        Args:
            (str) lookup: Lookup Table
            (int) id: Lookup id
        Kwargs:
            (bool) update: Suche auch in Datenbank
            (Cursor) cursor: DB Cursor
        Return:
            Result
                (name) name: name of lookup with id
                (bool) found: gefunden
                (bool) cache: gefunden in cache
                (bool) db: gefunden in datenbank
                (bool) db-miss-hit: nicht gefunden in cache und db -> aber in db-miss cache gefunden
                (bool) db-miss-new: nicht gefunden in cache und db -> eintrag in db-miss cache
        """
        bUpdate = True if "update" in kwargs and kwargs["update"] in ["true", "True", True] else False
        cursor = kwargs["cursor"] if "cursor" in kwargs else None

        res = Result(name="Lookup Reverse (id->name)")

        erg = None
        bFound = False
        bCache = False
        bDb = False

        if sLookup is None:
            res.addErr("Lookup table not given", sys="Lookup System", kat="Configuration")
        if iId is None:
            res.addErr("Id is not given", sys="Lookup System", kat="Configuation")
        if type(iId) != int:
            res.addErr("Id is not an integer", sys="Lookup System", kat="Configuration")
        if sLookup not in self.__map.lookup_getDbTables():
            res.addErr("Lookup table not found" , add=str(sLookup), sys="Lookup System", kat="Configuration")

        if res.isOk():
            # Suche in Cache
            for k, v in self.__lookupDbCache[sLookup].items():
                if v == iId:
                    res.addData("name", k)
                    res.addData("found", True)
                    res.addData("cache", True)
                    break
            if not bFound and bUpdate:
                if iId in self.__lookupDbMissId[sLookup]:
                    # Suche in DB Rev Miss Cache
                    res.addData("found", False)
                    res.addData("db-miss-hit", True)
                else:
                    # Suche in DB
                    try:
                        resSingle = self.__getLookupDbSingle__id2name(sLookup, iId, cursor=cursor)
                        if resSingle.isErr():
                            # Fehler
                            res.addData("found", False)
                            res.addResult(resSingle, err=True)
                        elif not resSingle.getData("found"):
                            # Nicht gefunden
                            res.addData("found", False)
                            res.addData("db-miss-new", True)
                            self.__lookupDbMissId[sLookup].add(iId)
                        else:
                            # Gefunden
                            res.addData("name", resSingle.getData("name"))
                            res.addData("found", True)
                            res.addData("db", True)
                            self.__lookupDbCache[sLookup][resSingle.getData("name")] = iId
                    except Exception as ex:
                        res.addErr("Get Lookup Reverse", add=str(ex), sys="Lookup System", kat="Internal")
        return res

    # -----------------------------------------------------

    # Lade neue Lookup Daten ein
    def addLookupBatch(self, wb, **kwargs):
        """
        Ingest von Lookup-Tables

        Args:
            wb (workbook): Excel Workbbook
        Kwargs:
            result_in: Result in
            cursor: Db Cursor
            sheets (lst): Liste aller hinzuzufügenden Sheets
            import_excel (lst): Zu importierende Einträge aus Excel: all/gold/not-gold
            import_excel_goldrush (bool): Erlaube nur gold importe
            import_mode (lst): Import modus: new/update/new-update
            import_db (lst): Nur folgende Einträge in DB updaten: all/gold/not-gold
            reuse (str): Reaktion wenn existierender Eintrag in DB nicht 100% matched: error/warning/info
            batch (str): Batch Info die weitergegeben wird
        Returns:
             Result
        """

        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")
        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status
        res = Result(name="Add Lookup Batch")

        # Parameter
        resultIn = kwargs["result_in"] if "result_in" in kwargs else Result("Empty Result Input Dummy")
        lstImportTables = kwargs["sheets"] if "sheets" in kwargs else None
        sCfgExcelInput = kwargs["import_excel"] if "import_excel" in kwargs else None
        sCfgDbMode = kwargs["import_mode"] if "import_mode" in kwargs else None
        sCfgDbUpdate = kwargs["import_db"] if "import_db" in kwargs else None
        bGoldRush = kwargs["import_excel_goldrush"] if ("import_excel_goldrush" in kwargs and kwargs["import_excel_goldrush"] in [True,False])  else False

        progress("on-pulse")
        workflow(top="Lookup", sub="Importing Data")
        pn("* Importing Lookup Data")
        status()

        self.update()
        self.abort(False)

        sReuse= kwargs["reuse"] if "reuse" in kwargs else None
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # None -> Default
        if sCfgExcelInput is None:
            sCfgExcelInput = self.__model.conf_getLookupImport_excel()
        if sCfgDbMode is None:
            sCfgDbMode = self.__model.conf_getLookupImport_db()
        if sCfgDbUpdate is None:
            sCfgDbUpdate = self.__model.conf_getLookupImport_db_update()

        # Teste Konfiguration
        if sCfgExcelInput not in ["all", "gold", "not-gold"]:
            res.addErr("Type of import data unkwnown", add=str(sCfgExcelInput), sys="Lookup Import", kat="Configuration")
        if sReuse not in ["error", "warning", "info"]:
            res.addErr("Type of reusing level if rec in database not 100% matches", add=str(sReuse), sys="Lookup Import", kat="Configuration")
        if sCfgDbMode not in ["new", "update", "both"]:
            res.addErr("Type of import mode unkwnown" , add=str(sCfgDbMode), sys="Lookup Import", kat="Configuration")
        if sCfgDbUpdate not in [None, "all", "gold", "not-gold"]:
            res.addErr("Type of database update data unkwnown", add=str(sCfgDbUpdate), sys="Lookup Import", kat="Configuration")
        if sCfgDbMode in ["update","both"] and sCfgDbUpdate is None:
            res.addErr("Missing type of update data", sys="Lookup Import", kat="Configuration")
        if sCfgExcelInput in ["gold", "not-gold"] and bGoldRush:
            bGoldRush = False

        if res.isErr():
            cleanUi()
            return res
        if sCfgExcelInput == "not-gold" and sCfgDbMode == "update" and sCfgDbUpdate == "gold":
            cleanUi()
            res.addInfo("Nothing could happen: Excel:not-gold cannot update DB:gold", sys="Lookup Import", kat="Configuration")
            return res

        # Info String für Konfiguration
        sCfgInfo = "Excel: Import"
        if sCfgExcelInput == "all":
            sCfgInfo += " all records"
        elif sCfgExcelInput == "gold":
            sCfgInfo += " only gold records"
        elif sCfgExcelInput == "not-gold":
            sCfgInfo += " only not-gold records"
        else:
            sCfgInfo +="? "
        if bGoldRush:
            sCfgInfo += "GoldRush!"
        if sCfgDbMode == "new":
            sCfgInfo += "\nDatabase: create only new records"
        else:
            sCfgInfo += "\nDatabase: "
            if sCfgDbMode == "both":
                sCfgInfo += "create new records and update "
            elif sCfgDbMode == "update":
                sCfgInfo += "update "
            else:
                sCfgInfo += "? "
            if sCfgDbUpdate == "all":
                sCfgInfo += "all found matching records"
            elif sCfgDbUpdate == "gold":
                sCfgInfo += "only found matching gold records"
            elif sCfgDbUpdate == "not-gold":
                sCfgInfo += "only found matching not-gold records"
            else:
                sCfgInfo += "? "
        if sCfgDbMode == "new":
            sCfgInfo += "\nReaction if reused record in databse is not 100% matching: " +str(sReuse)

        sCfgInfo += "\n"

        # ---------

        if bOwnConnection:
            pv("Using own connection")
        else:
            pv("Reusing connection")
        pv("Import records: " + str(sCfgExcelInput))
        pv("Import mode: " + str(sCfgDbMode))
        if sCfgDbMode in ["new","both"]:
            pv("Reuse reaction level if rec not 100% matching: " + str(sReuse))
        if sCfgDbMode != "new":
            pv("Update matching records in DB: " + str(sCfgDbUpdate))
        workflow(top="Import", sub="Checking data")

        # --- Check Parameter
        if wb is None:
            res.addErr("No import data given (Excel workbook)", sys="Lookup Import", kat="Configuration")
        if lstImportTables is None:
            lstImportTables = []
        for sSheet in lstImportTables:
            if sSheet not in self.__model.subsystem_getMapper().lookup_getExcelSheets():
                res.addErr("Sheet not valid", sheet=str(sSheet), sys="Lookup Import", kat="Configuration")
        if res.isErr():
            cleanUi()
            return res

        # --- Teste Eingangsdaten
        resCheck = self.__checkImportDataBulk(wb, lstImportTables)
        res.addResult(resCheck)
        if res.isErr():
            cleanUi()
            return res

        # --- Starte Ingest
        lstSheets = resCheck.getData("sheetsImport")
        dicHeader = resCheck.getData("header")
        dicHeaderIndex = resCheck.getData("headerIndex")
        iRowsTotal = resCheck.getData("entriesTotal")

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Log Total
        sLogTotal = "------------------------------------\n"\
                    "Detailed logs per lookup table\n"\
                    "------------------------------------\n\n"
        if self.__model.conf_isDry():
            sLogTotal += "Dry mode activated.\n\n"

        lstSheetsUsed = []
        lstTablesChanged = []
        iExcelRowsTotal = 0
        iRowAktuell = 0
        iRecImportNo = 0
        iRecCreateNo = 0
        iRecUpdateNotNecessaryEqual = 0
        iRecUpdateDeniedByRights = 0
        iRecUpdateDeniedByDbGold = 0
        iRecUpdateDeniedByDbNotGold = 0
        iRecUpdateDeniedByNewOnly = 0
        iRecUpdateDeniedByTaxon = 0
        iRecUpdateDiff = 0

        iRecImportNoTotal = 0
        iRecUpdateTotal = 0
        iRecUpdateNoTotal = 0
        iRecNewTotal = 0
        iRecNewNoTotal = 0
        iRecErrTotal = 0
        odicRecErrTotal = OrderedDict()
        lstiImportNotGold_total = []

        # Kein direktes Schreiben in die Taxon Liste per Lookup mehr möglich (entferne sicherheits halber Taxon Sheet - Dürfte aber nie reingekommen sein)
        if "Taxon" in lstSheets:
            lstSheets.remove("Taxon")  # Erntferne Taxon
            sLogTotal += "Importing Taxa data from sheet Taxon is not allowed -> discarding Sheet\n"

        # Loop Sheets
        for sSheet in lstSheets:
            status("Add Lookup: "+sSheet)
            sTable = self.__map.lookup_sheet2table(sSheet)

            # Erlaubte Header für Lookup
            lstHeadersAllowed = self.__model.subsystem_getMapper().lookup_getExcelCols(sSheet)

            # Log: Single
            lstImportNotGold = []
            lstRecImportNo = []
            lstRecCreateNo = []
            lstRecUpdateDeniedByNewOnly = []
            lstRecUpdateNotNecessaryEqual = []
            lstRecUpdateDeniedByRights = []
            lstRecUpdateDeniedByDbGold = []
            lstRecUpdateDeniedByDbNotGold = []
            lstRecUpdateDeniedByTaxon = []
            lstRecUpdate = []
            lstRecNew = []
            lstRecErr = []

            ws = wb[sSheet]
            bSheetEmpty = True
            # Rows
            for iRow, row in enumerate(ws.iter_rows()):
                if iRow > 0:
                    odicRowExcel = ExcelTools.row2dic(row, dicHeaderIndex[sSheet], conv_val_2_str=True)
                    if odicRowExcel is not None:
                        # Zeile nicht leer
                        # Entferne: Unerlaubte Header aus Lookup
                        for sCol in list(odicRowExcel.keys()):
                            if sCol not in lstHeadersAllowed:
                                del odicRowExcel[sCol]

                        # Teste: Leere Zeile?
                        bEmpty = True
                        for col, val in odicRowExcel.items():
                            if val not in [None, ""]:
                                bEmpty = False

                        # Translate Namen: Excel -> DB
                        odicRowDb = OrderedDict()
                        for sCol, sVal in odicRowExcel.items():
                            sTable, sAttribute = self.__model.subsystem_getMapper().lookup_excel2db(sSheet, sCol)
                            odicRowDb[sAttribute] = sVal

                        # Wenn ein Gold-Attribut existiert -> None -> False
                        if "gold" in odicRowDb:
                            if str(odicRowDb["gold"]).strip().lower() in ["true", "t", "1"]:
                                odicRowDb["gold"] = True
                            else:
                                odicRowDb["gold"] = False

                        if not bEmpty:
                            # Zeile is voll
                            bSheetEmpty = False
                            iExcelRowsTotal += 1
                            sTestKey = self.__model.subsystem_getMapper().lookup_getKey(sTable)
                            sTestVal = odicRowDb[sTestKey]
                            bRowGold = None # Rec In Gold
                            if sTable != "taxon":
                                bRowGold = True if str(odicRowDb["gold"]).strip().lower() in ["true","1","t"] else False
                                if bGoldRush and bRowGold is False:
                                    res.addErr("GoldRush mode is actice and lookup is not gold", sheet=sSheet, col=sTestKey, row=iRow, sys="Lookup Import", kat="Lookup gold check")

                            # Teste ob Test-Value leer ist
                            if sTestVal in ["", None]:
                                res.addErr("Lookup Key is empty", sheet=sSheet, col=sTestKey, row=iRow, sys="Lookup Import", kat="Lookup Structure")

                            # Teste ob Import Record benutzt werden soll?
                            if sTable == "taxon" or \
                                    sCfgExcelInput == "all" or \
                                    (sCfgExcelInput == "gold" and bRowGold is True) or \
                                    (sCfgExcelInput == "not-gold" and bRowGold is False):

                                # Eintrag soll benutzt werden
                                if not bRowGold:  # Log: Benutzung von not gold Einträgen
                                    lstImportNotGold.append(str(sTestVal))
                                    res.addInfo("Import of not-gold entries", sheet=sSheet, col=sTestKey,
                                                add=str(sTestVal), sys="Lookup Import", kat="Gold System")

                                # ---------------------------------------------------------------
                                # TWEAK/SPEZIALISIERUNG
                                # ---------------------------------------------------------------
                                bErrTweak = False
                                # Dataset
                                if sTable == "dataset":
                                    # contact_person_id
                                    key = odicRowDb["contact_person_id"]
                                    resLookupSearch = self.getLookup("person", key, update=True, cursor=cursor)
                                    if resLookupSearch.getData("found"):
                                        odicRowDb["contact_person_id"] = resLookupSearch.getData("id")
                                    else:
                                        bErrTweak = True
                                        lstRecErr.append("#"+str(iRow)+": person " + str(key) + " not found in lookup table country")
                                        res.addErr("Entry not found", sheet="Dataset", col="ContactPerson", add=str(key), row=iRow, sys="Lookup Import", kat="Search")
                                        iRecErrTotal += 1
                                    # country_id
                                    key = odicRowDb["owner_country_id"]
                                    resLookupSearch = self.getLookup("country", key, update=True, cursor=cursor)
                                    if resLookupSearch.getData("found"):
                                        odicRowDb["owner_country_id"] = resLookupSearch.getData("id")
                                    else:
                                        bErrTweak = True
                                        lstRecErr.append("#"+str(iRow)+": country " + str(key) + " not found in lookup table country")
                                        res.addErr("Entry not found", sheet="Dataset", col="OwnerCountry", add=str(key), row=iRow, sys="Lookup Import", kat="Search")
                                        iRecErrTotal += 1

                                if not bErrTweak:
                                    # Test: Eintrag schon in DB
                                    resExist = self.__model.subsystem_getDb().check_isRecordExistingInTable(sTable, sTestKey, dict(odicRowDb), cursor=cursor)
                                    if resExist.getData("found"):
                                        iId = resExist.getData("id")
                                        # --- Record bereits in DB gefunden
                                        if resExist.getData("equal"):
                                            # --- Record gleich
                                            lstRecUpdateNotNecessaryEqual.append("#" + str(iRow) +": "+ sTestKey + "=" + str(sTestVal))
                                            res.addInfo("No Update - exactly same entry in Db", sheet=sSheet, col=sTestKey, add=str(sTestVal),row=iRow, sys="Lookup Import", kat="Update Db")
                                            iRecUpdateNotNecessaryEqual += 1
                                            iRecUpdateNoTotal += 1
                                        else:
                                            # --- Record differiert
                                            dicDiff = resExist.getData("diff")
                                            sDiff = ""
                                            for kDb, vDb in dicDiff.items():
                                                if kDb in odicRowDb:
                                                    vIn = odicRowDb[kDb]
                                                else:
                                                    vIn = "-"
                                                sDiff += str(kDb) + "{in=\'" + str(vIn) + "\' db=\'" + str(vDb) + "\'} "
                                            if sTable == "taxon":
                                                # Update bei Taxon -> Nein, das besitzt kein Gold -und-> Updates über Wizard durchführen)
                                                lstRecUpdateDeniedByTaxon.append("#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal))
                                                res.addInfo("No Update possible in Taxon table (Hint: Use TaxonWizard)", sheet="Taxon", row=iRow, sys="Lookup Import", kat="Update Db")
                                                iRecUpdateDeniedByTaxon += 1
                                                iRecUpdateNoTotal += 1
                                            else:
                                                bDbGold = resExist.getData("answer-full")["gold"] if "gold" in resExist.getData("answer-full") else False
                                                # Nur Update wenn ...
                                                # 0) mode != new
                                                # 1) db-update: all (excel: gold | excel: not-gold und db: not-gold)
                                                # 2) db-update: not-gold (excel: gold/not-gold)
                                                # 3) db-update: gold (excel: gold)

                                                if sReuse == "info":
                                                    res_addReact = res.addInfo
                                                elif sReuse == "warning":
                                                    res_addReact = res.addWarn
                                                else:
                                                    res_addReact = res.addErr

                                                # -- Mode = New und nur gold differiert (in not-gold und db gold) -> Auge zudrücken (das hat schonmal jemand gecheckt -> OK)
                                                if sCfgDbMode == "new" and len(dicDiff) == 1 and "gold" in dicDiff and not bRowGold is bDbGold:
                                                    res.addInfo("Using equal gold db entry (another person has already checked the entry)",
                                                                 sheet=sSheet,
                                                                 col=sTestKey,
                                                                 add=str(sTestVal)+", diff:"+sDiff, row=iRow,
                                                                 sys="Lookup Import", kat="Update Db")
                                                # --- Mode = Update
                                                elif sCfgDbMode != "new" and \
                                                        ((sCfgDbUpdate == "all" and (bRowGold or (not bRowGold and not bDbGold))) or \
                                                        (sCfgDbUpdate == "not-gold" and not bDbGold) or \
                                                        (sCfgDbUpdate == "gold" and (bDbGold and bRowGold))):
                                                        resUpdate = self.__model.subsystem_getDb().table_updateInTable(sTable, iId, odicRowDb, cursor=cursor)
                                                        if resUpdate.isErr():
                                                            lstRecErr.append(("#" + str(iRow) + ": " + resUpdate.getResult()))
                                                            iRecErrTotal += 1
                                                            res.addResult(resUpdate, err=True)
                                                        else:
                                                            lstRecUpdate.append("#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal) + ", diff=" + sDiff)
                                                            iRecUpdateTotal += 1
                                                            res_addReact("Update entry in Db", sheet=sSheet, col=sTestKey, add=str(sTestVal)+", diff:"+sDiff, row=iRow, sys="Lookup Import", kat="Update Db")
                                                            if not bRowGold: # Log: Benutzung von not gold Einträgen
                                                                lstImportNotGold.append(sTestVal)
                                                else:
                                                    # Kein Update
                                                    sNoUpdate = "#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal) + ", gold(in="+str(bRowGold)+" db="+str(bDbGold)+"), diff=" + sDiff
                                                    sNoUpdateShort = sTestKey + "=" + str(
                                                        sTestVal) + ", gold(in=" + str(bRowGold) + " db=" + str(
                                                        bDbGold) + "), diff=" + sDiff
                                                    if sCfgDbMode == "new":
                                                        lstRecUpdateDeniedByNewOnly.append(sNoUpdate)
                                                        iRecUpdateDeniedByNewOnly += 1
                                                        iRecUpdateNoTotal += 1
                                                        res_addReact("Update Denied By Rule: New Only", sheet=sSheet, col=sTestKey,
                                                                    add=str(sNoUpdateShort) + ", diff:" + sDiff, row=iRow,
                                                                    sys="Lookup Import", kat="Update Db")
                                                    elif (sCfgDbUpdate == "not-gold" and bDbGold):
                                                        lstRecUpdateDeniedByDbGold.append(sNoUpdate)
                                                        iRecUpdateDeniedByDbGold += 1
                                                        iRecUpdateNoTotal += 1
                                                        res_addReact("Update Denied By Rule: Gold Only", sheet=sSheet, col=sTestKey,
                                                                    add=str(sNoUpdateShort) + ", diff:" + sDiff, row=iRow,
                                                                    sys="Lookup Import", kat="Update Db")
                                                    elif (sCfgDbUpdate == "gold" and not bDbGold):
                                                        lstRecUpdateDeniedByDbNotGold.append(sNoUpdate)
                                                        iRecUpdateDeniedByDbNotGold += 1
                                                        iRecUpdateNoTotal += 1
                                                        res_addReact("Update Denied By Rule: Not Gold Only", sheet=sSheet, col=sTestKey,
                                                                    add=str(sNoUpdateShort) + ", diff:" + sDiff, row=iRow,
                                                                    sys="Lookup Import", kat="Update Db")
                                                    elif (not bRowGold and bDbGold):
                                                        lstRecUpdateDeniedByRights.append(sNoUpdate)
                                                        iRecUpdateDeniedByRights += 1
                                                        iRecUpdateNoTotal += 1
                                                        res_addReact("Update Denied By Rule: Not Gold is Weaker Than Gold", sheet=sSheet,
                                                                    col=sTestKey,
                                                                    add=str(sNoUpdateShort) + ", diff:" + sDiff, row=iRow,
                                                                    sys="Lookup Import", kat="Update Db")
                                                    else:
                                                        res_addReact("Update Denied By Unknwown Cause",
                                                                    sheet=sSheet,
                                                                    col=sTestKey,
                                                                    add=str(sNoUpdateShort) + ", diff:" + sDiff, row=iRow,
                                                                    sys="Lookup Import", kat="Update Db")
                                                        lstRecErr.append("unknown block of update")
                                                        iRecErrTotal += 1
                                    else:
                                        # --- Record noch nicht in DB
                                        if sCfgDbMode in ["new", "both"]:
                                            # --- Create new
                                            resCreate = self.__model.subsystem_getDb().table_insertIntoTable(sTable, odicRowDb, cursor=cursor)
                                            if resCreate.isErr():
                                                res.addResult(resCreate, err=True, sys="Lookup Import", sheet=sSheet, col=sTestKey, kat="Create new Lookup", row=iRow)
                                                lstRecErr.append(("#" + str(iRow) + ": " + resCreate.getErr()))
                                                iRecErrTotal += 1
                                            else:
                                                sSqlQuery = resCreate.getData("sql")
                                                iId = resCreate.getData("id")
                                                lstRecNew.append("#" + str(iRow) + ": "+str(sTestKey) + "=" + str(sTestVal))
                                                iRecNewTotal +=1
                                        else:
                                            # Kein Create
                                            lstRecCreateNo.append("#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal))
                                            iRecCreateNo += 1
                                            iRecNewNoTotal += 1
                            else:
                                # --- Kein EXCEL import
                                # Info
                                sNoImportInfo = ""
                                if (sCfgExcelInput == "gold" and bRowGold is False):
                                    sNoImportInfo = "(not gold)"
                                elif (sCfgExcelInput == "not-gold" and bRowGold is True):
                                    sNoImportInfo = "(gold)"
                                lstRecImportNo.append("#" + str(iRow) + ": " + sTestKey + "=" + str(sTestVal)+" "+sNoImportInfo)
                                iRecImportNo += 1
                                iRecImportNoTotal += 1

                    # Info
                    iRowAktuell += 1
                    lstStatusEWI = list()
                    sStatus = ""
                    if resultIn.isErr():
                        lstStatusEWI.append("+Err:" + str(resultIn.getErrNo()))
                    if resultIn.isWarn():
                        lstStatusEWI.append("+Wrn:" + str(resultIn.getWarnNo()))
                    if resultIn.isInfo():
                        lstStatusEWI.append("+Inf:" + str(resultIn.getInfoNo()))
                    if res.isErr() or iRecErrTotal > 0:
                        lstStatusEWI.append("Err:" + str(iRecErrTotal + res.getErrNo()))
                    if iRecImportNo > 0:
                        lstStatusEWI.append("ImpBlk:" + str(iRecImportNo))
                    if iRecNewTotal > 0:
                        lstStatusEWI.append("New:" + str(iRecNewTotal))
                    if iRecCreateNo > 0:
                        lstStatusEWI.append("NewBlk:" + str(iRecCreateNo))
                    if iRecUpdateTotal > 0:
                        lstStatusEWI.append("Upd:" + str(iRecUpdateTotal))
                    if (iRecUpdateDeniedByRights + iRecUpdateDeniedByDbNotGold + iRecUpdateDeniedByDbGold + iRecUpdateDeniedByTaxon) > 0:
                        lstStatusEWI.append("UpdBlk:" + str(iRecUpdateDeniedByRights + iRecUpdateDeniedByDbNotGold + iRecUpdateDeniedByDbGold + iRecUpdateDeniedByNewOnly))
                    if iRecUpdateNotNecessaryEqual > 0:
                        lstStatusEWI.append("UpdEq:" + str(iRecUpdateNotNecessaryEqual))
                    if len(lstStatusEWI) != 0:
                            sStatus = "[" + (" ".join(lstStatusEWI)) + "]"
                    progress(iRowAktuell, iRowsTotal, 100, proz=True, abs=True, cli=True, status=sStatus)
                    status("Reading Sheet: "+sSheet+" #"+str(iRowAktuell)+"/"+str(iRowsTotal)+" "+sStatus)
                    # Abbruch durch Benutzer
                    if self.isAbort():
                        res.addErr("Aborted by user ...", sys="Lookup Import", kat="Abort")
                        cleanUi()
                        closeDb()
                        return res

            # Statistik für eine Lookup Table
            sLog = "--- "+str(sTable).capitalize() + " ---\n"

            sLog += "Selected mode:\n" + sCfgInfo + "\n"
            if len(lstRecNew) >0 or \
                    len(lstRecUpdateNotNecessaryEqual) > 0 or \
                    len(lstRecUpdateDeniedByRights) > 0 or \
                    len(lstRecUpdateDeniedByDbNotGold) > 0 or \
                    len(lstRecUpdateDeniedByDbGold) > 0 or \
                    len(lstRecUpdateDeniedByNewOnly) > 0 or \
                    len(lstRecUpdateDeniedByTaxon) > 0 or \
                    len(lstRecErr) > 0 or \
                    len(lstRecImportNo) > 0 or \
                    len(lstRecCreateNo) > 0 or \
                    len(lstImportNotGold) > 0 or \
                    len(lstRecUpdate) > 0:
                # Bearbeitete Sheets
                lstSheetsUsed.append(sSheet)
                # Veränderte Tabellen
                if len(lstRecNew) > 0 or len(lstRecUpdate) > 0:
                    lstTablesChanged.append(sSheet)
                if len(lstImportNotGold) != 0:
                    sLog += "- Imported not gold entries: " + str(len(lstImportNotGold)) + "\n"
                    for sTxt in lstImportNotGold:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                    lstiImportNotGold_total.append(sSheet+" ("+str(len(lstImportNotGold))+")")
                if len(lstRecImportNo) != 0:
                    sLog += "- Import blocked by rule: " + str(len(lstRecImportNo)) + "\n"
                    for sTxt in lstRecImportNo:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecErr) != 0:
                    sLog += "- Failed creations: " + str(len(lstRecErr)) + "\n"
                    for sTxt in lstRecErr:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecCreateNo) > 0:
                    sLog += "- Creation blocked (due to selected mode: not allowed to create new records in db): " + str(len(lstRecCreateNo)) + "\n"
                    for sTxt in lstRecCreateNo:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecNew) > 0:
                    sLog += "- Created records: " + str(len(lstRecNew)) + "\n"
                    for sTxt in lstRecNew:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateNotNecessaryEqual) != 0:
                    sLog += "- Update not necessary (record existing exactly in db: " + str(len(lstRecUpdateNotNecessaryEqual)) + "\n"
                    for sTxt in lstRecUpdateNotNecessaryEqual:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedByNewOnly) != 0:
                    sLog += "- Update denied (due to rule: updates not allowed at all): " + str(len(lstRecUpdateDeniedByNewOnly)) + "\n"
                    for sTxt in lstRecUpdateDeniedByNewOnly:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedByRights) != 0:
                    sLog += "- Update denied (due to rule: not-gold cannot override gold): " + str(len(lstRecUpdateDeniedByRights)) + "\n"
                    for sTxt in lstRecUpdateDeniedByRights:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedByDbNotGold) != 0:
                    sLog += "- Update denied (due to selected mode: not allowed to update not-gold entries in db: " + str(len(lstRecUpdateDeniedByDbNotGold)) + "\n"
                    for sTxt in lstRecUpdateDeniedByDbNotGold:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedByDbGold) != 0:
                    sLog += "- Update denied (due to selected mode: not allowed to update gold entries in db): " + str(len(lstRecUpdateDeniedByDbGold)) + "\n"
                    for sTxt in lstRecUpdateDeniedByDbGold:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdateDeniedByTaxon) != 0:
                    sLog += "- Update for taxon denied (use TaxonWizard instead): " + str(len(lstRecUpdateDeniedByTaxon)) + "\n"
                    for sTxt in lstRecUpdateDeniedByTaxon:
                        sLog += sTxt + "\n"
                    sLog += "\n"
                if len(lstRecUpdate) != 0:
                    sLog += "- Updated records in db: " + str(len(lstRecUpdate)) + "\n"
                    for sTxt in lstRecUpdate:
                        sLog += sTxt + "\n"
                    sLog += "\n"

            else:
                sLog += "... nothing to do\n\n"

            # Aktuelles Lookup Log an gesamtes Log anhängen
            sLogTotal += sLog

            # Fehlerliste an Gesamtfehlerliste herandhängen
            if len(lstRecErr) > 0:
                odicRecErrTotal[sSheet] = lstRecErr

            # Aktuelles Lookup Log im Action Log speichern
            sActionLookup = "lup-add"
            result_actionlog = self.__model.subsystem_getDb().actionlog_new(sActionLookup, sTable, sLog, cursor=cursor)
            if result_actionlog.isErr():
                res.addResult(result_actionlog, err=True, sys="Lookup Import", sheet=sSheet, col=sTestKey,
                              kat="Writing to Action Log", row=iRow)

        # Sind Fehler aufgetreten -> Setze diese an Gesamt Log an den Anfang
        if iRecErrTotal > 0:
            sLogTmp =  "------------------------------------\n"\
                      "Overview of occurred errors ("+str(iRecErrTotal)+")\n"\
                      "------------------------------------\n\n"
            for sSheet, lstErr in odicRecErrTotal.items():
                sLogTmp += "- "+str(sSheet)+"\n"
                for sErr in lstErr:
                    sLogTmp += str(sErr) + "\n"
            sLogTotal = sLogTmp + "\n" + sLogTotal

        # Log Präefix
        sLogPrae = "************************************\n"
        sLogPrae += "Import Lookup Data\n"
        sLogPrae += "************************************\n"
        sLogPrae += "\n"
        sLogPrae += "------------------------------------\n" \
          "Overview\n" \
        "------------------------------------\n\n"
        sLogPrae += sCfgInfo
        if bGoldRush:
            sLogPrae += "GoldRush!\n"
        sLogPrae += "Processed Excel sheets: " + PrettyText.lst2strC(lstSheetsUsed) + "\n"
        sLogPrae += "Proccessed EXCEL records: " + str(iExcelRowsTotal) +"\n"
        sLogPrae += "Tables changed: "+ PrettyText.lst2strC(lstTablesChanged) +"\n"
        sLogPrae += "Denied imports: " + str(iRecImportNoTotal) +"\n"
        sLogPrae += "Created records in db: " + str(iRecNewTotal) +"\n"
        sLogPrae += "Denied creations: " + str(iRecNewNoTotal) +"\n"
        sLogPrae += "Updating records in db: " + str(iRecUpdateTotal) +"\n"
        sLogPrae += "Denied Updates: " + str(iRecUpdateNoTotal) +"\n"
        sLogPrae += "Imported not-gold entries: " + PrettyText.lst2strC(lstiImportNotGold_total)+"\n\n"
        sLogTotal = sLogPrae + sLogTotal + "\n"

        # Log: Summary Stat
        sLogSummary = ""
        if bGoldRush:
            sLogSummary += "GoldRush!\n"
        if iExcelRowsTotal > 0:
            sLogSummary += "Proccessed EXCEL records: " + str(iExcelRowsTotal) + "\n"
        if iRecImportNoTotal > 0:
            sLogSummary += "Denied imports: " + str(iRecImportNoTotal) + "\n"
        if len(lstTablesChanged) > 0:
            sLogSummary += "Tables changed: " + PrettyText.lst2strC(lstTablesChanged) + "\n"
        if iRecNewTotal > 0:
            sLogSummary += "Created records in db: " + str(iRecNewTotal) + "\n"
        if iRecNewNoTotal > 0:
            sLogSummary += "Denied creations: " + str(iRecNewNoTotal) + "\n"
        if iRecUpdateTotal > 0:
            sLogSummary += "Updated records in db: " + str(iRecUpdateTotal) + "\n"
        if iRecUpdateNoTotal > 0:
            sLogSummary += "Denied Updates: " + str(iRecUpdateNoTotal) + "\n"
        if len(lstiImportNotGold_total) > 0:
            sLogSummary += "Imported not-gold entries: " + PrettyText.lst2strC(lstiImportNotGold_total)+"\n\n"
        if sLogSummary != "":
            sLogSummary = "--- Lookup --\n" + sLogSummary + "\n"

        # Antwort
        res.addData("lookup-excel-sheets-used", lstSheetsUsed)
        res.addData("lookup-tables-changed", lstTablesChanged)
        res.addData("log", sLogTotal)
        res.addData("log_stat", sLogSummary)
        res.addData("lookup-excel-recs", iExcelRowsTotal)
        res.addData("lookup-import-no", iRecImportNoTotal)
        res.addData("lookup-new", iRecNewTotal)
        res.addData("lookup-new-no", iRecNewNoTotal)
        res.addData("lookup-update", iRecUpdateTotal)
        res.addData("lookup-update-no", iRecUpdateNoTotal)
        res.addData("lookup-err", iRecErrTotal)

        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()
        
        # Ende
        return res






