# coding=utf-8


# -------------------------------------------------
# My WoRMS
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from tools.taxon import Taxon
from shutil import copyfile

class MyWorms(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()

        # Db
        self.__conn = None
        self.__cursor = None

        # Puffer
        self.__bPufferHasUnsavedData = False # Besitzt der Puffer auf HDD-ungespeicherte Daten?
        self.__iAgeMaxAllowed = 30
        self.__sFileNamePufferWorms = "worms.dat"
        self.__dicPufferWorms = {}
        self.__odicPufferWorms_ageTaxa = OrderedDict()
        self.__fPufferWorms_meanAge = 0
        self.__iPufferWorms_overMaxAge = 0
        self.__dicPufferWorms_query = {}

        # Problem Taxa
        self.__setProblem_problems = set()
        self.__setProblem_no_name = set()
        self.__setProblem_no_aname_and_accepted = set()
        self.__setProblem_not_accepted_and_not_leading_to_accepted = set()
        self.__setProblem_no_name_at_all = set()
        self.__setProblem_status = set()
        self.__setProblem_aphia = set()
        self.__setProblem_limbo = set()

    # Info
    def info(self):
        sOut = "*** Status of MyWoRMS\n\n"
        if self.puffer_getSizeTaxa() > 0:
            sOut += self.puffer_getStatusString() + "\n"
        return sOut

    def puffer_hasUnsavedData(self):
        return self.__bPufferHasUnsavedData

    def puffer_getMaxAgeAllowed(self):
        return(self.__iAgeMaxAllowed)

    def puffer_getMeanAge(self):
        return self.__fPufferWorms_meanAge

    def puffer_getOverMaxAge(self):
        return(self.__iPufferWorms_overMaxAge)

    def puffer_getOverMaxAgeProzent(self):
        try:
            fProz = self.puffer_getOverMaxAge() / self.puffer_getSizeTaxa()
        except:
            fProz = 0
        if fProz < 0:
            fProz = 0
        return fProz

    def puffer_getProblemTaxa_limbo(self):
        return self.__setProblem_limbo

    def puffer_getProblemTaxa_noScientificName(self):
        return self.__setProblem_no_name

    def puffer_getProblemTaxa_noNameAtAll(self):
        return self.__setProblem_no_name_at_all

    def puffer_getProblemTaxa_noAcceptedScientificNameAndAccepted(self):
        return self.__setProblem_no_aname_and_accepted

    def puffer_getProblemTaxa_notAcceptedAndNotLeadingToAccepted(self):
        return self.__setProblem_not_accepted_and_not_leading_to_accepted

    def puffer_getProblemTaxa_status(self):
        return self.__setProblem_status

    def puffer_getProblemTaxa_aphia(self):
        return self.__setProblem_aphia

    def puffer_getProblemTaxa_all(self):
        return self.__setProblem_problems

    def puffer_getSizeTaxa(self):
        return(len(self.__dicPufferWorms))

    def puffer_getSizeQuery(self):
        return len(self.__dicPufferWorms_query)

    def puffer_getSizeProblems(self):
        return len(self.puffer_getProblemTaxa_all())

    def puffer_getSizeQueryMultiAnswer(self):
        iMulti = 0
        for sSn, lstTaxon in self.__dicPufferWorms_query.items():
            if len(lstTaxon) > 1:
                iMulti += 1
        return iMulti

    def puffer_cleanDisk(self, **kwargs):
        self.puffer_cleanRam(**kwargs)
        self.puffer_save()

    def puffer_cleanRam(self, **kwargs):
        # Parameter
        if "taxa" in kwargs or "query" in kwargs:
            bTaxa = kwargs["taxa"] if "taxa" in kwargs else False
            bSearch = kwargs["query"] if "query" in kwargs else False
        else:
            bTaxa = True
            bSearch = True
        # Taxa
        if bTaxa:
            self.__dicPufferWorms = {}
            self.__odicPufferWorms_ageTaxa = OrderedDict()
            self.__fPufferWorms_meanAge = 0
            self.__iPufferWorms_overMaxAge = 0
            self.__bPufferHasUnsavedData = True
            self.__setProblem_problems = set()
            self.__setProblem_no_name = set()
            self.__setProblem_no_aname_and_accepted = set()
            self.__setProblem_no_name_at_all = set()
            self.__setProblem_status = set()
            self.__setProblem_aphia = set()
            self.__setProblem_limbo = set()
        # Name Search
        if bSearch:
            self.__dicPufferWorms_query = {}
            self.__bPufferHasUnsavedData = True

    def puffer_save(self):
        res = Result("Save MyWoRMS buffer")
        if self.puffer_hasUnsavedData():
            sFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferWorms
            try:
                date = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))
                data = (date, self.__dicPufferWorms, self.__dicPufferWorms_query)
                with open(sFile, 'wb') as handle:
                    pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
                    res.addData("file", sFile)
                    res.addData("date", date)
                self.__bPufferHasUnsavedData = False
            except Exception as ex:
                res.addErr("Problems saving buffer", add=str(ex), kat="I/O", sys="My WoRMS")
        else:
            res.addInfo("Saving Buffer", add="Buffer has no unsaved data -> Saving buffer not needed ", kat="I/O", sys="My WoRMS")
            res.addData("save_not_needed", True)
        return res

    def puffer_import(self, sFile):
        pn = self.__model.ui_getSmartprinter().normal
        res = Result("Import WoRMS buffer")
        try:
            # Restore Backup
            if FileTools.getFileType(sFile) == "zip":
                bWormsRestoreFileZip = True
                sWormsRestoreFileZip = sFile
                sWormsRestoreFile = FileTools.getFileWithoutType(sWormsRestoreFileZip)
                sWormsRestoreFileName = FileTools.getFileWithoutPath(sWormsRestoreFile)
                sWormsRestoreFilePath = FileTools.getFilePath(sWormsRestoreFile)
                resCheckFileExists = FileTools.prepareSaveFile(sWormsRestoreFile)
                if resCheckFileExists.getData("bak") is not None:
                    pn("* Backup existing file")
                    pn(str(sFile) + " -> " + str(resCheckFileExists.getData("bak")))
                if sWormsRestoreFile in ["", None] or sWormsRestoreFileZip in ["", None]:
                    res.addErr("Problems unzipping file", add=str(sWormsRestoreFileZip), kat="Zip", sys="Admin")
                # Unzip
                pn("* Unzipping file")
                pn(str(sWormsRestoreFileZip))
                pn("->")
                pn(str(sWormsRestoreFile))
                lstZipContent = []
                try:
                    lstZipContent = zipfile.ZipFile(sWormsRestoreFileZip).namelist()
                    zipfile.ZipFile(sWormsRestoreFileZip).extract(sWormsRestoreFileName, path=sWormsRestoreFilePath)
                    pn("done")
                except Exception as ex:
                    res.addErr("File not found in zip package",
                                  add=str(sWormsRestoreFileName) + " -> found only this files: " + PrettyText.lst2strC(
                                      lstZipContent), kat="Zip", sys="Admin")
                    pn("error")
            else:
                bWormsRestoreFileZip = False
                sWormsRestoreFileZip = None
            pass

            # Alter Puffer löschen
            self.puffer_cleanRam()

            # Puffer kopieren
            sWorms = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferWorms
            if bWormsRestoreFileZip:
                copyfile(sWormsRestoreFile, sWorms)
            else:
                copyfile(sFile, sWorms)

            # Die enpackte Datei loeschen
            if bWormsRestoreFileZip:
                pn("* Removing unzipped file")
                pn(str(sWormsRestoreFile))
                os.remove(sWormsRestoreFile)
                pn("done")

        except Exception as ex:
            res.addErr("Problems saving copy of My WoRMS", add=str(ex), kat="I/O", sys="My WoRMS")
        return res

    def puffer_export(self, sFile):
        res = Result("Snapshot WoRMS buffer")
        try:
            # Puffer kopieren
            sWorms = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferWorms
            sFileName = FileTools.getFileWithoutPath(sFile)
            sFilePath = FileTools.getFilePath(sFile)

            copyfile(sWorms, sFile)
            # Kopie zippen und Original löschen
            zipfile.ZipFile(sFile+ ".zip", mode='w', compression=zipfile.ZIP_DEFLATED).write(sFile, sFileName)
            os.remove(sFile)
            res.addData("file", str(sFileName)+".zip")
            res.addData("ordner", str(sFilePath))
        except Exception as ex:
            res.addErr("Problems saving copy of My WoRMS", add=str(ex), kat="I/O", sys="My WoRMS")
        return res

    def puffer_snapshot(self):
        res = Result("Snapshot WoRMS buffer")
        sOrigFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferWorms
        sSnapshotFolder = self.__model.conf_getAppSnapshotDirMyWorms()
        sSnapshotFileName = "worms_"+PrettyText.createTimestamp() + ".dat"
        sSnapshotFile = sSnapshotFolder +"/" + sSnapshotFileName
        try:
            res.addData("snapshot-filename", sSnapshotFileName+".zip")
            res.addData("snapshot-folder", sSnapshotFolder)
            res.addData("snapshot-file", sSnapshotFile)
            # Puffer kopieren
            copyfile(sOrigFile, sSnapshotFile)
            # Kopie zippen und Original löschen
            zipfile.ZipFile(sSnapshotFile + ".zip", mode='w', compression=zipfile.ZIP_DEFLATED).write(sSnapshotFile, sSnapshotFileName)
            os.remove(sSnapshotFile)
        except Exception as ex:
            res.addErr("Problems Taking Buffer Snapshot", add=str(ex), kat="I/O", sys="My WoRMS")
        return res

    def puffer_load(self):
        res = Result("Reload WoRMS buffer")
        sFile = self.__model.conf_getUserDataDir() + "/" + self.__sFileNamePufferWorms
        try:
            with open(sFile, 'rb') as handle:
                self.puffer_cleanRam()
                date, self.__dicPufferWorms, self.__dicPufferWorms_query = pickle.load(handle)
                self.puffer_createStats()
                res.addData("file", sFile)
                res.addData("date", date)
                # # Scatter
                # for iAid, t in self.__dicPufferWorms.items():
                #     datum = t.getDateCreate()
                #     t.debug__setDateCreate(datum - datetime.timedelta(days=random.randint(1,90)))
                #     self.__dicPufferWorms[iAid] = t
        except Exception as ex:
            self.puffer_cleanDisk()
            res.addErr("Problems loading buffer -> Made a new empty buffer", add=str(ex), kat="I/O", sys="My WoRMS")
        return res

    def __puffer_createStats__checkProblemTaxon(self, taxon):
        self.checkProblemTaxon(taxon,
                               self.__setProblem_problems,
                               self.__setProblem_no_name,
                               self.__setProblem_no_aname_and_accepted,
                               self.__setProblem_not_accepted_and_not_leading_to_accepted,
                               self.__setProblem_no_name_at_all,
                               self.__setProblem_status,
                               self.__setProblem_aphia,
                               self.__setProblem_limbo
                               )

    def puffer_createStats(self):
        pn = self.__model.ui_getSmartprinter().normal
        self.__odicPufferWorms_ageTaxa = OrderedDict()
        self.__fPufferWorms_meanAge = 0
        self.__iPufferWorms_overMaxAge = 0
        if self.__dicPufferWorms is None or len(self.__dicPufferWorms) == 0:
            return

        # Alle Taxa nach Alter in Tagen gruppieren
        iMaxAge = 0
        dateNow = datetime.datetime.now()
        lstDays = []
        dicAgeTaxa = {}
        for iAid, taxon in self.__dicPufferWorms.items():
            # Error -> Remove Taxon (dürfte aber nie passieren)
            if taxon.isErr() or not taxon.isFound():
                pn("Error in Taxon in Buffer -> Removing item:", str(iAid))
                self.__dicPufferWorms.popitem(iAid)
                self.__bPufferHasUnsavedData = True
            # Age
            iDaysTaxon = taxon.getAge()
            lstDays.append(iDaysTaxon)
            if iDaysTaxon not in dicAgeTaxa:
                dicAgeTaxa[iDaysTaxon] = []
            if iDaysTaxon > self.__iAgeMaxAllowed:
                self.__iPufferWorms_overMaxAge += 1
            dicAgeTaxa[iDaysTaxon].append(iAid)
            # Problem Taxa
            self.__puffer_createStats__checkProblemTaxon(taxon)

        # Sortiere nach Alter
        self.__odicPufferWorms_ageTaxa = OrderedDict(sorted(dicAgeTaxa.items(), reverse=True))
        # Mittleres Alter aller Taxa
        self.__fPufferWorms_meanAge = sum(lstDays) / len(lstDays)


    def puffer_getStatusString(self, **kwargs):
        # Parameter
        if "taxa" in kwargs or "search" or "misfits" in kwargs:
            bTaxa = kwargs["taxa"] if "taxa" in kwargs else False
            bSearch = kwargs["search"] if "search" in kwargs else False
            bMisfits = kwargs["misfits"] if "misfits" in kwargs else False
        else:
            bTaxa = True
            bSearch = True
            bMisfits = True
        sOut = ""


        # Taxa
        if bTaxa:
            lstBufferStatus = [["Age", "Taxa"]]
            # Bar Plot
            lstBarchartData = [("Age Group", "Number of Taxa")]
            for iDay, lstTaxa in self.__odicPufferWorms_ageTaxa.items():
                lstBarchartData.append((str(iDay)+" days", len(lstTaxa)))

            sOut += "--- Buffered WoRMS Taxa ---\n\n"

            sOut += "* Age\n"
            sOut += "Allowed max. age (days): " + str(self.__iAgeMaxAllowed) + "\n"
            sOut += "Taxa: " + str(self.puffer_getSizeTaxa()) + "\n"
            sOut += "Median age: " + f'{self.puffer_getMeanAge():.1f}' + " days\n\n"
            sOut += PrettyText.barchart(lstBarchartData, maxbarsize=25, header=True) +"\n"

        # Misfits
        if bMisfits:
            setTaxaProblems_noName = self.puffer_getProblemTaxa_noScientificName()
            setTaxaProblems_noNameAtAll = self.puffer_getProblemTaxa_noNameAtAll()
            setTaxaProblems_noAnameAndAccpeted = self.puffer_getProblemTaxa_noAcceptedScientificNameAndAccepted()
            setTaxaProblems_notAcceptedAndNotLeadingToAccepted = self.puffer_getProblemTaxa_notAcceptedAndNotLeadingToAccepted()
            setTaxaProblems_status = self.puffer_getProblemTaxa_status()
            setTaxaProblems_aphia = self.puffer_getProblemTaxa_aphia()
            setTaxaProblems_limbo = self.puffer_getProblemTaxa_limbo()
            setTaxaProblemTaxa = self.puffer_getProblemTaxa_all()
            iTaxaProblems = len(setTaxaProblemTaxa)
            if iTaxaProblems > 0:
                sOut += "--- Misfits ("+str(iTaxaProblems)+") ---\n"
                if len(setTaxaProblems_notAcceptedAndNotLeadingToAccepted) > 0:
                    sOut += "* Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(len(setTaxaProblems_notAcceptedAndNotLeadingToAccepted)) + "\n"
                    for t in setTaxaProblems_notAcceptedAndNotLeadingToAccepted:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_limbo) >0:
                    sOut += "* Taxa in Limbo: " + str(len(setTaxaProblems_limbo)) + "\n"
                    for t in setTaxaProblems_limbo:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_aphia) >0:
                    sOut += "* Taxa with Aphia Id problems: " + str(len(setTaxaProblems_aphia)) + "\n"
                    for t in setTaxaProblems_aphia:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_noName) >0:
                    sOut += "* Taxa with missing ScientificNames: " + str(len(setTaxaProblems_noName)) + "\n"
                    for t in setTaxaProblems_noName:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_noNameAtAll) > 0:
                    sOut += "* Taxa with no names at all: " + str(len(setTaxaProblems_noNameAtAll)) + "\n"
                    for t in setTaxaProblems_noNameAtAll:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_noAnameAndAccpeted) > 0:
                    sOut += "* Accepted Taxa with missing AcceptedScientificName: " + str(len(setTaxaProblems_noAnameAndAccpeted)) + "\n"
                    for t in setTaxaProblems_noAnameAndAccpeted:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                if len(setTaxaProblems_status) > 0:
                    sOut += "* Taxa with status problems: " + str(len(setTaxaProblems_status)) + "\n"
                    for t in setTaxaProblems_status:
                        sOut += str(t.signature()) + "\n"
                    sOut += "\n"
                sOut += "\n"

        # Suchnamen
        if bSearch:
            sOut += "--- Buffered WoRMS Queries by ScientificName ---\n\n"
            sOut += "Queries: " + str(self.puffer_getSizeQuery()) + "\n"
            iMulti = self.puffer_getSizeQueryMultiAnswer()
            if iMulti > 0:
                sOut += "Queries with multiple answers: " + str(iMulti) +"\n"
                lstTable = [("ScientificName", "Answers")]
                for sSn, lstTaxa in self.__dicPufferWorms_query.items():
                    if len(lstTaxa) > 1:
                        lstTable.append((sSn, len(lstTaxa)))
                sOut += PrettyText.table(lstTable) + "\n"
        return sOut

    def puffer_update(self, **kwargs):
        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug
        progress = self.__smartprinter.progress
        workflow = self.__smartprinter.workflow
        status = self.__smartprinter.status
        mywormsUpdate = self.__smartprinter.updateMyworms

        res = Result(name="Update MyWoRMS buffer")

        # Parameter
        lstTaxaSubset = kwargs["taxa"] if "taxa" in kwargs else None
        bTaxaSubset = True if lstTaxaSubset is not None else False

        self.puffer_cleanRam()
        self.puffer_load()
        self.puffer_createStats()
        self.abort(False)

        # Log
        progress("on-pulse-abort")
        sLog = ""
        sLog += "************************************\n"
        sLog += "Anti-aging MyWoRMS\n"
        sLog += "************************************\n\n"

        # Startzeit
        iZeitStart = time.time()
        sZeitStart = str(datetime.datetime.now().strftime("%Y-%m-%d %H:%M"))

        iTaxaTotal = self.puffer_getSizeTaxa()
        iTaxaMaxAgeAllowed = self.puffer_getMaxAgeAllowed()

        # Status Start
        fTaxaMeanAge = self.puffer_getMeanAge()
        iTaxaOverAge = self.puffer_getOverMaxAge()
        fTaxaOverAgeProz = self.puffer_getOverMaxAgeProzent()
        # Status Ende
        fTaxaMeanAge_now = self.puffer_getMeanAge()
        iTaxaOverAge_now = self.puffer_getOverMaxAge()
        fTaxaOverAgeProz_now = self.puffer_getOverMaxAgeProzent()

        # Kopie des Alters-Klassen Puffers
        odicPufferWorms_ageTaxa = self.__odicPufferWorms_ageTaxa.copy()

        # Max Taxa älter als 0 Tage bestimmen / On-The-Fly-Stat Dictionary erzeugen
        iNrMax = 0
        dicStatOnTheFly = {}
        for iDay, lstTaxa in odicPufferWorms_ageTaxa.items():
            dicStatOnTheFly[iDay] = len(lstTaxa)
            if iDay > 0:
                iNrMax += len(lstTaxa)

        if bTaxaSubset:
            sLog += "--- Only DeAging Subset ---\n\n"
            sLog += "Number: " + str(len(lstTaxaSubset)) + "\n"
            sLog += "Aphia Ids: "+PrettyText.lst2strC(lstTaxaSubset)+"\n"
            iNrMax = len(lstTaxaSubset)

        # Loop
        pv("* DeAging Taxa", fmt=False)
        iNr = 0
        iNrDay = 0
        iAgeGroups =len(odicPufferWorms_ageTaxa.keys())
        iAgeGroup = 0
        sStatus = ""
        sLog += "--- DeAging Age Groups ---\n\n"
        for iDay, lstTaxa in odicPufferWorms_ageTaxa.items():
            if self.isAbort():
                break
            iAgeGroup += 1
            if iDay > 0:
                sLogErr = ""
                iNrDay = 0
                iMaxDay = len(lstTaxa)
                workflow(top="MyWoRMS", sub="DeAging", fmt=False)
                if self.isAbort():
                    break
                for iAid in lstTaxa:
                    if lstTaxaSubset is None or iAid in lstTaxaSubset: # Nur Subset verjüngen
                        if self.isAbort():
                            break
                        # Update Taxon
                        iNr += 1
                        iNrDay += 1
                        if not bTaxaSubset:
                            text = "Age group: " + str(iDay) + " days " + "(" + str(iAgeGroup) + "/" + str(
                                iAgeGroups) + "):"
                        progress(iNr, iNrMax, 100, proz=True, abs=True, cli=True, status=sStatus, text = "Age group: "+str(iDay)+" days "+"("+str(iAgeGroup)+"/"+str(iAgeGroups)+"):")
                        status("DeAging Taxon with AphiaId: " + str(iAid) + " ("+str(iNr)+"/"+str(iNrMax)+")")
                        taxon = Taxon(iAid)
                        pv(str(taxon.signature()))
                        if taxon.isErr():
                            # Fehler
                            sLogErr += "Error: "+str(taxon)+"\n"
                        else:
                            # Ok
                            if self.isAidInPuffer(iAid):
                                self.__addTaxonToPuffer(taxon, update=True)
                            else:
                                sLogErr += "Internal Problem: Taxon de-aged but it was not in buffer: "+str(iAid) +"\n"
                        # Stats on the fly anpassen
                        if self.__iPufferWorms_overMaxAge > 0:
                            self.__iPufferWorms_overMaxAge -= 1
                        self.__fPufferWorms_meanAge -= (iDay / iTaxaTotal)
                        # Update MyWorms
                        mywormsUpdate()

                sLog += ">" + str(iDay) + " days: "+str(iNrDay)+" taxa\n"
                sLog +=sLogErr

        # Log Zeit
        iSekTotal = time.time() - iZeitStart
        iStd = int(iSekTotal // 3600)
        iMin = int((iSekTotal % 3600) // 60)
        iSek = int(iSekTotal % 60)

        # Result
        lstStats = [("", "Before", "After")]
        lstStats.append(("Age mean.", f'{fTaxaMeanAge:.1f}'+" days", f'{self.puffer_getMeanAge():.1f}'+" days"))
        lstStats.append(("Over max. age", str(iTaxaOverAge)+" taxa", str(self.puffer_getOverMaxAge())+" taxa"))
        lstStats.append(("Ratio over max. age", f'{fTaxaOverAgeProz*100:.1f}'+" %", f'{self.puffer_getOverMaxAgeProzent()*100:.1f}'+" %"))

        sLog += "\n"
        sLog += "--- Result ---\n\n"

        sLog += "Number of total de-aged taxa: " + str(iNr)+"\n"
        sLog += "Allowed maximum age: " + str(self.puffer_getMaxAgeAllowed()) + " days\n"
        sLog += "Duration: " + str(iStd) + "h" + str(iMin) + "m" + str(iSek) + "s\n"
        sLog += "\n"
        sLog += PrettyText.table(lstStats)+"\n"
        if self.isAbort():
            sLog += "\n<Stopped by user>\n\n"

        # Puffer save an Rebuild Stats
        self.puffer_save()
        self.puffer_createStats()

        res.addData("title", "Update of MyWoRMS buffer")
        res.addData("text", sLog)
        progress("off")
        return res

    # ---

    def __askWormsByScientificname(self, sName, **kwargs):
        """
        Hole AphiaId plus TaxonInfos zu Scientificname vonm WoRMS
        Args
            sSn (str): ScientificName
        Kwargs:
            fuzzy (bool): Fuzzy
            trymax (int): Maximale Anfragen bis abgebrochen wird
            like (bool): Suche nach Ähnlichen Namen (bei der normalen Suche)
            marine_only (bool): Suche nur nach Marinen Tierchen
        Return
            lst((int,str)): Liste mit Tupel von AphiaIds und TaxonInfo
        """

        res = Result("Get Aphia Id from WoRMS")
        status = self.__model.ui_getSmartprinter().status

        # Parameter
        bFuzzy = kwargs["fuzzy"] if ("fuzzy" in kwargs and kwargs["fuzzy"] in (True,False)) else False
        bLike = kwargs["like"] if ("like" in kwargs and kwargs["like"] in (True, False)) else True
        bMarineOnly = kwargs["marine_only"] if ("marine_only" in kwargs and kwargs["marine_only"] in (True, False)) else False
        iTryMax = kwargs["trymax"] if ("trymax" in kwargs) else 5

        # Daten
        iTry = 1
        iTryMax = 5
        bDone = False
        bFound = False
        lstJson = None

        # Scientificnamen fuer die URL umformatierten
        # Enferne einige Sonderzeichen (/.)
        sNameUrl = sName
        sNameUrl = sNameUrl.replace("/", " ")
        sNameUrl = sNameUrl.replace(".", " ")
        sNameUrl = sNameUrl.replace("&", " ")

        # Transformiere restliche Sonderzeichen
        sNameUrl = urllib.parse.quote(sNameUrl)

        if sNameUrl != sName:
            res.addInfo("Scientificname has to be modified for use in URL", add="Orig:" + str(sName) + " -> URL-Version:" + str(sNameUrl), kat="Asking WoRMS by ScientificName", sys="MyWoRMS")

        while not bDone:
            try:
                # Cooldown?
                if iTry > 1:
                    status("Asking WoRMS by ScientificName: give WoRMS a little time to cool down ...", fmt=False)
                    time.sleep(1)
                    status("Asking WoRMS by ScientificName: " + str(iTry) + "/" + str(iTryMax) + " ... ", fmt=False)

                # Request Cmd
                sMarineOnly = "true" if bMarineOnly else "false"
                sLike = "true" if bLike else "false"
                if bFuzzy:
                    sRequest = "http://www.marinespecies.org/rest/AphiaRecordsByMatchNames?scientificnames%5B%5D=" + str(sNameUrl) + "&marine_only="+sMarineOnly
                else:
                    sRequest = "http://www.marinespecies.org/rest/AphiaRecordsByName/" + str(sNameUrl) + "?like="+sLike+"&marine_only="+sMarineOnly

                # Anfrage
                resp = requests.get(sRequest)


                # Response Code
                iStatus = resp.status_code
                if iStatus == 200:
                    # Anfrage Ok -> AphiaId gefunden
                    bFound = True
                elif iStatus == 204:
                    # Anfrage Ok -> AphiaId nicht gefunden
                    bFound = False
                elif iStatus == 400:
                    # Anfrage nicht OK (Anfragedaten fehlerhaft)
                    res.addErr("Missing Properties -> Status: "+str(iStatus),
                               add="Orig:" + str(sName) + " -> URL-Version:" + str(sNameUrl),
                               kat="Asking WoRMS by ScientificName", sys="MyWoRMS")
                else:
                    # Anfrage nicht OK (Andere Probleme fehlerhaft)
                    res.addErr("Other Problems -> Status: " + str(iStatus),
                               add="Orig:" + str(sName) + " -> URL-Version:" + str(sNameUrl),
                               kat="Asking WoRMS by ScientificName", sys="MyWoRMS")
                bDone = True

                # Antwort
                if bFound:
                    try:
                        if bFuzzy:
                            lstJson = resp.json()[0]
                        else:
                            lstJson = resp.json()
                    except Exception as ex:
                        res.addErr("Internal Problem: ", add=str(ex), kat="Asking WoRMS by ScientificName", sys="MyWoRMS")

            except Exception as ex:
                # Irgendwas ging schief -> Nochmal probieren
                iTry += 1
                res.addInfo("Problems occured: with ScientificName: "+str(sName),
                           add="Try "+str(iTry)+": "+str(ex),
                           kat="Asking WoRMS by ScientificName", sys="MyWoRMS")
                if iTry > iTryMax:
                    res.addInfo("Connection to WoRMS failed: " + str(sName),
                               add="Tries: " + str(iTry),
                               kat="Asking WoRMS by ScientificName", sys="MyWoRMS")

        res.addData("sql", sRequest)
        res.addData("found", bFound)
        res.addData("tries", iTry)
        res.addData("json", lstJson)
        return res

    def __getTaxonFromPuffer(self, iAid):
        if iAid in self.__dicPufferWorms:
            taxon = self.__dicPufferWorms[iAid]
            return taxon
        else:
            return None

    def __addTaxonToPuffer(self, t, **kwargs):
        bUpdate = kwargs["update"] if ("update" in kwargs and kwargs["update"] in [True, False]) else True
        if t is not None:
            iAid = t.getAphiaId()
            if bUpdate:
                self.__puffer_createStats__checkProblemTaxon(t)
                self.__dicPufferWorms[iAid] = t
                self.__bPufferHasUnsavedData = True
                return True
            elif iAid not in self.__dicPufferWorms:
                self.__puffer_createStats__checkProblemTaxon(t)
                self.__dicPufferWorms[iAid] = t
                self.__bPufferHasUnsavedData = True
                return True
        return False

    def isAidInPuffer(self, iAid):
        if iAid in self.__dicPufferWorms:
            return True
        else:
            return False

    def isTaxonInPuffer(self, taxon):
        if taxon.getData("aid") is not None:
            return True
        else:
            return False

    def isQueryInPuffer(self, sSn):
        return sSn in self.__dicPufferWorms_query

    def queryByName(self, sSn, **kwargs):
        """
        Hole AphiaId plus TaxonInfos zu ScientificName von WoRMS
        Args
            sSn (str): ScientificName
        Kwargs:
            like (bool): Fuzzy Generic
            fuzzy (bool): Fuzzy Special
            puffer (bool): Benutze Puffer
            marine_only (bool): Suche nur nach marinen Lebewesen
            fallback (bool): Wenn alle ermittelten Taxa unakzeptiert sind und keine akzeptierten ermittelt werden können -> Nimm die unakzetierten Ergebnisse (besser als nichts)
        Return
            Result:
                multi (bool): mehrfache Taxa in Anwort
                taxa_nr (int): Anzahl der Taxa in Antwort
                taxa (lst): Taxa
                taxa_worms_original (lst): Original Antwort von WoRMS
                taxa_discarded_unaccepted (lst): Aus Originalantwort entfernte nicht akzeptierte Taxa
                taxa_added_accepted_traced (lst): Zu Originalantwort aus taxa_discarded_unaccepted hinzugefügte ermittlete akzeptierte Taxa
                tries (int): Verbindungsversuche zu WoRMS
                fuzzy (bool): Namenssuche wird fuzzy durchgeführt
                puffer_used: Soll der Puffer befragt werden, wenn nicht werden alle Ergebnisse trotzdem gepuffert
                found (bool): Ergebnis wurde gefunden (Original-Ergebnis)
                found_in_puffer (bool): Es wurde in Puffer gefunden
                fallback_used (bool): Wurde die Ergebniss mende auf Null reduziert und das Originalergebnis benutzt

        """

        def uniqTaxonList(lstTaxa):
            lstUniq = []
            setAid = set()
            for t in lstTaxa:
                if t.getAphiaId() not in setAid:
                    lstUniq.append(t)
                    setAid.add(t.getAphiaId())
            return lstUniq

        self.abort(False)

        res = Result("Get Taxa by Name")
        status = self.__model.ui_getSmartprinter().status
        mywormsUpdate = self.__smartprinter.updateMyworms

        # Parameter
        bFuzzy = kwargs["fuzzy"] if ("fuzzy" in kwargs and kwargs["fuzzy"] in (True,False)) else False
        bLike = kwargs["like"] if ("like" in kwargs and kwargs["like"] in (True,False)) else False
        bUsePuffer = kwargs["puffer"] if ("puffer" in kwargs and kwargs["puffer"] in (True,False)) else False
        bNoFallback = kwargs["no_fallback"] if ("no_fallback" in kwargs and kwargs["no_fallback"] in (True, False)) else False
        bMarineOnly = kwargs["marine_only"] if ("marine_only" in kwargs and kwargs["marine_only"] in (True, False)) else False

        # Fuzzy (Wenn beides gesetzt ist nimm das schnellere)
        if bFuzzy and bLike:
            bLike = False

        # Daten
        bFallback_used = False
        lstTaxa = None
        lstTaxa_discardedUnacccepted = []
        lstTaxa_addedAcceptedTraced = []
        lstTaxa_worms = []
        lstJson = None
        bFoundInPuffer = False
        bFound = False
        iTries = 0
        bMulti = False

        # Suche im Puffer
        if bUsePuffer:
            lstTaxa = self.__dicPufferWorms_query[sSn] if sSn in self.__dicPufferWorms_query else None
            if lstTaxa is not None:
                bFound = True
                bFoundInPuffer = True
            else:
                bFound = False
                bFoundInPuffer = False

        # Frage WoRMS
        if not bFound:
            resAsk = self.__askWormsByScientificname(sSn, fuzzy=bFuzzy, marine_only=bMarineOnly, like=bLike)
            if resAsk.isErr():
                # Fehler
                res.addResult(resAsk, msg_only=True)
            else:
                # OK
                lstJson = resAsk.getData("json")
                bFound = resAsk.getData("found")
                iTries = resAsk.getData("tries")
                lstTaxa = []
                if bFound:
                    # Taxa zusammenstellen
                    iNrMax = len(lstJson)
                    for iNr, dicJson in enumerate(lstJson):
                        # Index-String
                        if iNrMax > 1:
                            sIndex = " (" + str(iNr+1) + "/" + str(iNrMax)+")"
                        else:
                            sIndex = ""
                        iAid = dicJson["AphiaID"]
                        resAskTaxon = self.getTaxon(iAid, puffer=True, worms=True, deage=True)
                        mywormsUpdate()
                        if resAskTaxon.isErr():
                            # Fehler -> Verwerfen
                            res.addResult(resAskTaxon, msg_only=True)
                        else:
                            t = resAskTaxon.getData("taxon")
                            if t.isAccepted():
                                # Nur Akzeptiertes Taxon benutzen (auch keine Limbos) -> Save
                                lstTaxa_worms.append(t)
                                lstTaxa.append(t)
                                status(sSn +sIndex+": "+ str(t.signature()))
                            else:
                                # Nicht Akzeptiertes Taxon oder Taxon mit Problemen
                                lstTaxa_worms.append(t)
                                lstTaxa_discardedUnacccepted.append(t) # Speichern in Discarded Liste
                                status(sSn +sIndex+": [discarded] <unaccepted>")
                                if self.isAbort():  # Abort
                                    res.addErr("Aborted by user...", sys="Name Query", kat="MyWoRMS")
                                    break
                                if t.getAphiaId_accepted() is not None:
                                    # Hole Akzeptiertes Taxon
                                    tacc = Taxon(t.getAphiaId_accepted())
                                    if tacc.getResult().isErr():
                                        # Fehler -> Verwerfen
                                        res.addResult(tacc.getResult(), msg_only=True)
                                    elif tacc.isAccepted():
                                        # Akzeptiertes Taxon gefunden -> Save
                                        lstTaxa.append(tacc)
                                        lstTaxa_addedAcceptedTraced.append(tacc)
                                        status(sSn + sIndex + ": [traced] " + str(tacc.signature()))
                                    elif tacc.isUnaccepted():
                                        res.addWarn("WoRMS Problem - Taxon should be accepted but is not", add=str(tacc.signature()), kat="Name Query", sys="MyWoRMS")
                # Listen Uniq machen
                lstTaxa = uniqTaxonList(lstTaxa)
                lstTaxa_worms = uniqTaxonList(lstTaxa_worms)
                lstTaxa_discardedUnacccepted = uniqTaxonList(lstTaxa_discardedUnacccepted)
                lstTaxa_addedAcceptedTraced = uniqTaxonList(lstTaxa_addedAcceptedTraced)

                # Test: Gab es eine Antwort und sind alle Antworten unakzeptiert und keine akzeptieren Ergbnisse ermittlebar -> nimm die Original Amtwort (das ist besser als nichts)
                if not bNoFallback:
                    if len(lstTaxa) == 0 and len(lstTaxa_worms) > 0:
                        lstTaxa = lstTaxa_worms
                        bFallback_used = True

                # Multi Taxon
                if len(lstTaxa) > 1:
                    bMulti = True

                # Speichern in Puffer
                if bFound:
                    self.__bPufferHasUnsavedData = True
                    self.__dicPufferWorms_query[sSn] = lstTaxa

        # Antwort
        res.addData("multi", bMulti)
        res.addData("taxa_nr", len(lstTaxa) if lstTaxa is not None else 0)
        res.addData("taxa", lstTaxa)
        res.addData("taxa_worms_original", lstTaxa_worms)
        res.addData("taxa_discarded_unaccepted", lstTaxa_discardedUnacccepted)
        res.addData("taxa_added_accepted_traced", lstTaxa_addedAcceptedTraced)
        res.addData("tries", iTries)
        res.addData("fuzzy", bFuzzy)
        res.addData("marine_only", bMarineOnly)
        res.addData("puffer_used", bUsePuffer)
        res.addData("found", bFound)
        res.addData("found_in_puffer", bFoundInPuffer)
        res.addData("fallback_used", bFallback_used)

        return res

    def getTaxon(self, iAid, **kwargs):
        """
        Hole Taxon
        Args
            iAid (int): Aphia ID
        Kwargs:
            puffer (bool): Befrage Puffer
            worms (bool): Befrage Worms
            deage (bool): De-Age Taxon in Puffer wenn zu alt
            deage_force (bool): De-Age Taxon in Puffer immer (auch wenn nicht zu alt)
        Return
            found (bool)
            asked_puffer (bool)
            puffer_age_too_old (bool)
            puffer_de_aged (bool)
            asked_worms (bool)
            taxon: Taxon
        """
        res = Result("Ask MyWorms by Aphia Id")
        mywormsUpdate = self.__smartprinter.updateMyworms

        # Parameter
        if "puffer" not in kwargs and "worms" not in kwargs:
            bPuffer = True
            bWorms = True
        else:
            bPuffer = kwargs["puffer"] if "puffer" in kwargs else False
            bWorms = kwargs["worms"] if "worms" in kwargs else False
        bDeAge = kwargs["deage"] if (("deage" in kwargs) and (kwargs["deage"]) in [True,False]) else False
        bDeAge_force = kwargs["deage_force"] if (("deage_force" in kwargs) and (kwargs["deage_force"]) in [True,False]) else False
        taxon = None

        # Init
        bFound = False
        bAskedPuffer = False
        iTaxonPuferAge = None
        bTaxonPufferOld = False
        bTaxonPufferUpdated = False
        bAskedWorms = False

        # Taxon weder im Puffer noch bei WoRMS abfragen
        if bPuffer is False and bWorms is False:
            res.addErr("Searching in Puffer and and asking WoRMS deacitvated", add=str(iAid), kat="Buffer",sys="MyWoRMS")

        # Erstmal AphiaId von String zu int konvertieren
        try:
            iAid = int(iAid)
        except:
            res.addErr("AphiaId is not interger", add=str(iAid), kat="Parameter", sys="MyWoRMS")

        if res.isOk():
            # Puffer befragen erlaubt?
            if bPuffer:
                # Puffer wird befragt
                taxon = self.__getTaxonFromPuffer(iAid)
                bAskedPuffer = True
                if taxon is not None:
                    # Taxon in Puffer
                    iTaxonPuferAge = taxon.getAge()
                    bFound = True
                    if iTaxonPuferAge > self.puffer_getMaxAgeAllowed():
                        # Taxon ist zu alt
                        bTaxonPufferOld = True
                elif not bWorms:
                    # Taxon nicht in Puffer und WoRMS soll nicht befragt werden -> Fehler
                    res.addWarn("Taxon not found in MyWorms (and asking WoRMS deacitvated)", add=str(iAid), kat="Buffer", sys="MyWoRMS")

        if res.isOk():
            # WoRMS befragen
            if bDeAge_force: # Force: Deage!
                bDeAge = True
                bTaxonPufferOld = True

            if (not bFound and bWorms) or (bFound and bTaxonPufferOld and bDeAge):
                # WoRMS befragen
                taxon = Taxon(iAid)
                bAskedWorms = True
                if taxon.isOk() and taxon.isFound():
                    # Taxon wurde gefunden
                    bFound = True
                    if bTaxonPufferOld:
                        bTaxonPufferUpdated = True # Taxon wurde geupdated
                    # Statistic on-the-fly anpassen
                    if bTaxonPufferOld:
                        # Taxon wurde geupdated
                        fMeanBefore = self.puffer_getMeanAge()
                        iTaxaBefore = self.puffer_getSizeTaxa()
                        fSumDaysBefore = fMeanBefore * iTaxaBefore
                        self.__fPufferWorms_meanAge = (fSumDaysBefore - iTaxonPuferAge) / iTaxaBefore
                    else:
                        # Taxon war eine neues Taxon
                        fMeanBefore = self.puffer_getMeanAge()
                        iTaxaBefore = self.puffer_getSizeTaxa()
                        fSumDaysBefore = fMeanBefore * iTaxaBefore
                        self.__fPufferWorms_meanAge = fSumDaysBefore / (iTaxaBefore +1)
                    # Taxon in Puffer schreiben
                    self.__addTaxonToPuffer(taxon)
                    mywormsUpdate() # GUI updaten
                else:
                    # Taxon wurde micht gefunden oder sonstiger Fehler ist passiert
                    pass

        res.addData("taxon", taxon)
        res.addData("asked_puffer", bAskedPuffer)
        res.addData("puffer_age_too_old", bTaxonPufferOld)
        res.addData("puffer_de_age_forced", bDeAge_force)
        res.addData("puffer_de_aged", bTaxonPufferUpdated)
        res.addData("asked_worms", bAskedWorms)
        res.addData("found", bFound)
        return res

    def checkProblemTaxon(self, tax,
                          setProblem_problems,
                          setProblem_no_name,
                          setProblem_no_aname_and_accepted,
                          setProblem_not_accepted_and_not_leading_to_accepted,
                          setProblem_no_name_at_all,
                          setProblem_status,
                          setProblem_aphia,
                          setProblem_limbo):
        """
        Teste ein Taxon oder Liste von Taxa auf Probleme
        Jedes Taxon wird einem Set entsprechend des Problems hinzugefügt
        Sets kommen über Parameter per Call-By-Reference
        Args:
            tax (lst/taxon): Taxon oder Liste von Taxay
            setProblem_problems (set): Irgend ein Problem mit Taxon
            setProblem_no_name (set): Kein Scientificname
            setProblem_no_aname_and_accepted (set): Kein Accepted ScientificName und Accepted
            setProblem_not_accepted_and_not_leading_to_accepted (set): Nicht akzeptiert, besitzt Aaid und führt zu einem nicht akzeptierten Taxon
            setProblem_no_name_at_all (set): Gar kein Name
            setProblem_status (set): Probleme mit Status
            setProblem_aphia (set): Probleme mit Aphia Id
            setProblem_limbo (set): Taxa in Limbo
        """
        if type(tax) is not list:
            lstTaxon = [tax]
        else:
            lstTaxon = tax
        for taxon in lstTaxon:
            if taxon.hasProblems():
                setProblem_problems.add(taxon)
            if taxon.getScientificName() is None:
                setProblem_no_name.add(taxon)
            if taxon.isAccepted() and taxon.getScientificName_accepted() is None:
                setProblem_no_aname_and_accepted.add(taxon)
            if taxon.isUnaccepted() and taxon.getAphiaId_accepted() is not None and not taxon.isRedirected_2_accepted():
                setProblem_not_accepted_and_not_leading_to_accepted.add(taxon)
            if taxon.hasNoNameAtAll():
                setProblem_no_name_at_all.add(taxon)
            if taxon.hasProblems_status():
                setProblem_status.add(taxon)
            if taxon.hasProblems_aphia():
                setProblem_aphia.add(taxon)
            if taxon.isLimbo():
                setProblem_limbo.add(taxon)

    def getStatisticProblemTaxa(self, lstTaxa):
        """
        Gib eine Statiistik zu den problematischen pberprüften Taxa aus
        Args:
            lstTaxa (lst): Taxa
        Return
            (str): Statistik
        """

        # Einzel-Taxon -> TaonListe
        if type(lstTaxa) is Taxon:
            lstTaxa = [lstTaxa]

        # Data
        setBiotaAll_check_problem_all = set()
        setBiotaAll_check_problem_all = set()
        setBiotaAll_check_problem__no_name = set()
        setBiotaAll_check_problem_no_aname_and_accepted = set()
        setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted = set()
        setBiotaAll_check_problem_no_name_at_all = set()
        setBiotaAll_check_problem_status = set()
        setBiotaAll_check_problem_aphia  = set()
        setBiotaAll_check_problem_limbo = set()
        setBiotaAll_check_special = set ()
        setBiotaAll_check_accepted  = set ()
        setBiotaAll_check_unaccepted  = set ()
        # Loop
        for t in lstTaxa:
            # Problem
            self.checkProblemTaxon(t,
                setBiotaAll_check_problem_all,
                setBiotaAll_check_problem__no_name,
                setBiotaAll_check_problem_no_aname_and_accepted,
                setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted,
                setBiotaAll_check_problem_no_name_at_all,
                setBiotaAll_check_problem_status,
                setBiotaAll_check_problem_aphia,
                setBiotaAll_check_problem_limbo)
            # Special Taxon
            if t.isSpecial():
                setBiotaAll_check_special.add(t)
            # Accepted
            if t.isAccepted():
                setBiotaAll_check_accepted.add(t)
            # Test: Unaccepted
            if t.isUnaccepted():
                setBiotaAll_check_unaccepted.add(t)
        # Stat
        sStat = ""
        # Overview
        # --- Overview
        sStat += "Taxa (total): " + str(len(lstTaxa)) + "\n"
        if len(setBiotaAll_check_accepted) > 0:
            sStat += "Accepted Taxa: " + str(len(setBiotaAll_check_accepted)) + "\n"
        if len(setBiotaAll_check_unaccepted) > 0:
            sStat += "Unaccepted Taxa: " + str(len(setBiotaAll_check_unaccepted)) + "\n"
        if len(setBiotaAll_check_special) > 0:
            sStat += "Special Taxa: " + str(len(setBiotaAll_check_special)) + "\n"
        if len(setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted) > 0:
            sStat += "Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(
                len(setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted)) + "\n"
        if len(setBiotaAll_check_problem_limbo) > 0:
            sStat += "Taxa in Limbo: " + str(len(setBiotaAll_check_problem_limbo)) + "\n"
        if len(setBiotaAll_check_problem_aphia) > 0:
            sStat += "Taxa with Aphia Id problems: " + str(
                len(setBiotaAll_check_problem_aphia)) + "\n"
        if len(setBiotaAll_check_problem__no_name) > 0:
            sStat += "Taxa with missing ScientificNames: " + str(
                len(setBiotaAll_check_problem__no_name)) + "\n"
        if len(setBiotaAll_check_problem_no_name_at_all) > 0:
            sStat += "Taxa with no names at all: " + str(
                len(setBiotaAll_check_problem_no_name_at_all)) + "\n"
        if len(setBiotaAll_check_problem_no_aname_and_accepted) > 0:
            sStat += "Accepted Taxa with missing AcceptedScientificName: " + str(
                len(setBiotaAll_check_problem_no_aname_and_accepted)) + "\n"
        if len(setBiotaAll_check_problem_status) > 0:
            sStat += "Taxa with status problems: " + str(
                len(setBiotaAll_check_problem_status)) + "\n"
        sStat +="\n"
        # Details
        if len(setBiotaAll_check_accepted) > 0:
            sStat += "* Accepted Taxa: " + str(len(setBiotaAll_check_accepted)) + "\n"
            for t in setBiotaAll_check_accepted:
                sStat += t.signature() + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_unaccepted) > 0:
            sStat += "* Unaccepted Taxa: " + str(len(setBiotaAll_check_unaccepted)) + "\n"
            for t in setBiotaAll_check_unaccepted:
                sStat += t.signature() + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_special) > 0:
            sStat += "* Special Taxa: " + str(len(setBiotaAll_check_special)) + "\n"
            for t in setBiotaAll_check_special:
                sStat += t.signature() + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted) > 0:
            sStat += "* Unaccepted Taxa with Accepted AphiaId leading to not accepted Taxa: " + str(
                len(setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted)) + "\n"
            for t in setBiotaAll_check_problem_not_accepted_and_not_leading_to_accepted:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_limbo) > 0:
            sStat += "* Taxa in Limbo: " + str(
                len(setBiotaAll_check_problem_limbo)) + "\n"
            for t in setBiotaAll_check_problem_limbo:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_aphia) > 0:
            sStat += "* Taxa with Aphia Id problems: " + str(
                len(setBiotaAll_check_problem_aphia)) + "\n"
            for t in setBiotaAll_check_problem_aphia:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem__no_name) > 0:
            sStat += "* Taxa with missing ScientificNames: " + str(
                len(setBiotaAll_check_problem__no_name)) + "\n"
            for t in setBiotaAll_check_problem__no_name:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_no_name_at_all) > 0:
            sStat += "* Taxa with no names at all: " + str(
                len(setBiotaAll_check_problem_no_name_at_all)) + "\n"
            for t in setBiotaAll_check_problem_no_name_at_all:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_no_aname_and_accepted) > 0:
            sStat += "* Accepted Taxa with missing AcceptedScientificName: " + str(
                len(setBiotaAll_check_problem_no_aname_and_accepted)) + "\n"
            for t in setBiotaAll_check_problem_no_aname_and_accepted:
                sStat += str(t.signature()) + "\n"
            sStat += "\n"
        if len(setBiotaAll_check_problem_status) > 0:
            sStat += "* Taxa with status problems: " + str(
                len(setBiotaAll_check_problem_status)) + "\n"
            for t in setBiotaAll_check_problem_status:
                sStat += str(t.signature()) + "\n"
        # Antwort
        return sStat