# coding=utf-8
import argparse, sys, os, textwrap, csv
from tools.coding import Result
from tools.parallel import Stoppable
from tools.text import TextEncoding

# -------------------------------------
# Versions Kontrolle
# -------------------------------------

class VersionControl(Stoppable):

    # Konstruktor
    def __init__(self, sp):
        super().__init__()

        # Instanzvariablen
        self.__smartprinter = sp

    # Init
    def diff(self, sIn1, sIn2, sOut):

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow

        sFileOrig = sIn1
        sFileNew = sIn2
        sFileOut = sOut

        encOrig = TextEncoding.checkout(sFileOrig, self.__smartprinter)
        encNew = TextEncoding.checkout(sFileNew, self.__smartprinter)

        # Info
        pn("* Version control")
        pn("Original version:   " + sFileOrig + " ("+encOrig[0]+", "+str(round(encOrig[1]*100))+"%)")
        pn("Editied version:    " + sFileNew + " ("+encNew[0]+", "+str(round(encNew[1]*100))+"%)")
        pn("Original with logs: " + sFileOut)

        # Einlesen der CSV-Datei
        lstOrig = list()
        lstOrigHeader = None
        lstNew = list()
        lstNewHeader = None
        lstOut = list()
        lstOutHeader = None

        progress("on-pulse")
        workflow(top="Loading data", sub="Working", fmt=False)
        try:
            # Orig
            with open(sFileOrig, encoding='utf-8', newline='') as csvfile:
                csvreader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstOrigHeader = csvreader.__next__()
                for lstZeile in csvreader:
                    lstOrig.append(lstZeile)
            # New
            with open(sFileNew, encoding='utf-8', newline='') as csvfile:
                csvreader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstNewHeader = csvreader.__next__()
                for lstZeile in csvreader:
                    lstNew.append(lstZeile)
        except Exception as ex:
            sErr = "Error while reading: " + str(ex)
            pn(sErr)
            return (False, sErr)

        # Konsistenztests
        if len(lstOrig) != len(lstNew):
            sErr = "Error: Original and edited version differ in row number."
            pn(sErr)
            return (False, sErr)

        if lstOrigHeader != lstNewHeader:
            sErr = "Error: Original and edited version have different headers."
            pn(sErr)
            return (False, sErr)

        # Out
        pn("Logged modifications:")
        lstOrigHeader = lstOrigHeader
        lstOrigHeader.append("Veraenderungen")
        lstOut.append(lstOrigHeader)
        iAnzahlAenderungen = 0

        # Auswertung
        progress("on-pulse-abort")
        workflow(top="Processing data", sub="Working", fmt=False)
        iZeile = 0
        for zeile in lstNew:
            progress(iZeile, len(lstNew),10, proz=True, abs=False, cli=False)
            status("Processing Row: " + str(iZeile) + "/"+str(len(lstNew)))
            iSpalte = 0
            sOut = ""
            for spalte in zeile:
                wertOrig = lstOrig[iZeile][iSpalte]
                wertNeu = lstNew[iZeile][iSpalte]
                if wertNeu != wertOrig:
                    sOut += str(lstOrigHeader[iSpalte]) + ": \"" + str(wertOrig) + "\" -> \"" + str(wertNeu) + "\" | "
                    iAnzahlAenderungen += 1
                iSpalte += 1
            if sOut != "":
                sOut = sOut[: -3]
                pv("Row " + str(iZeile) +": " + sOut)
            zeileOut = zeile
            zeileOut.append(sOut)
            lstOut.append(zeileOut)
            iZeile +=1
            if self.isAbort():
                progress("off")
                return Result(info="Process aborted by user ...")

        pn("Number of modifications:" + str(iAnzahlAenderungen))

        # Speicher Ergebnis
        try:
            with open(sFileOut, 'w', encoding='utf-8', newline='') as csvWriter:
                writer = csv.writer(csvWriter, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
                writer.writerows(lstOut)
        except Exception as ex:
            sErr = "Error while saving: " + str(ex)
            pn(sErr)
            progress("off")
            return Result(err=sErr)

        # Fertig
        progress("off")

        return (Result(name="Version control", info=str(iAnzahlAenderungen)+" modifications have been logged."))

