# coding=utf-8


# -------------------------------------------------
# Virtual Assistant
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
from difflib import SequenceMatcher
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from logik.virtual_assistant_taxareport import TaxonSpotter
from logik.virtual_assistant_taxapolish import TaxaPolish

from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile
from tools.taxon import Taxon

class VirtualAssistant(Stoppable, WorkingObject):

    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__taxaPolishing = TaxaPolish(model)
        self.__taxaReport = TaxonSpotter(model)

        # Subprozesse die bei einem Abort mit abgebrochen werden sollen
        self.addAbortChild(self.__taxaReport)
        self.addAbortChild(self.__taxaPolishing)

    # ----------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of VirtualAssistant\n\n"
        sOut += "TaxaPolishing: OK\n"
        sOut += "TaxonSpotter: OK\n"
        return sOut

    def getTaxonSpotter(self):
        return self.__taxaReport

    def getTaxaPolishing(self):
        return self.__taxaPolishing

    def start_taxaPolishing(self, sFileOrig, sFilePolished=None, **kwargs):
        res = self.__taxaPolishing.start_polishing(sFileOrig, sFilePolished=None, **kwargs)
        return res

    def start_taxaReport(self, **kwargs):
        res = self.__taxaReport.start_report(**kwargs)
        return res


