# coding=utf-8


# -------------------------------------------------
# TaxonSpotter
# -------------------------------------------------

import openpyxl, traceback, pickle, zipfile
from collections import Counter, OrderedDict
from pathlib import Path
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
from difflib import SequenceMatcher
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.ui import Colors
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from tools.taxacruncher import TaxaCruncher
from shutil import copyfile
from tools.taxon import Taxon

class TaxonSpotter(Stoppable, WorkingObject):
    """
    TODO:
    - Legende einfügen
    - Statistik anpassen mit Taxa
    - Taxon: selber befragen nach (in-/direkt un-/akzeptiert)
    """
    def __init__(self, model):
        Stoppable.__init__(self)
        WorkingObject.__init__(self)

        # -- Member
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__map = model.subsystem_getMapper()
        self.__myworms = model.subsystem_getMyWorms()
        self.__wizard = model.subsystem_getTaxonWizard()

        # Taxa
        self.__lstTaxa = None
        self.__lstTaxa_clean = None

        # Db
        self.__conn = None
        self.__cursor = None

        # Subprozesse die bei einem Abort mit abgebrochen werden sollen
        self.addAbortChild(self.__myworms)

    # ----------------------------------------

    # Info
    def info(self):
        sOut = "*** Status of TaxonSpotter\n\n"
        return sOut

    def __reset(self):
        self.abort(False)
        self.__conn = None
        self.__cursor = None
        self.__lstTaxa = None
        self.__lstTax_clean = None

    def start_report(self, **kwargs):
        """
        Report

        Kwargs:
            file (str)
            puffer (bool): Use Buffer for Name Queries
            fuzzy (bool): Use Fuzzy Name Queries
            marine_only (bool): Suche nur nach marinen Lebewesen
            sheet (str)
            aid_given (str)
            sn_given (str)
            aid (str)
            sn (str)
            aid_acc (str)
            sn_acc (str)
            html (bool): HTML Log
        Returns:
             Result
        """
        # Helper: Aufräumen
        def cleanUi():
            workflow()
            progress("off")

        def closeDb():
            if self.__model.conf_isDry() and conn is not None:
                conn.rollback()
            elif res.isErr():
                conn.rollback()
            else:
                conn.commit()
            cursor.close()
            conn.close()

        # ---

        # TODO: DEV
        #kwargs = {'file': '/Users/paul/_temp/taxa_listen/taxa_2.xlsx', 'sheet': 'Taxa', 'aid_given': 'AphiaID given', 'sn_given': 'ScientificName given', 'aid': 'AphiaID', 'sn': 'ScientificName', 'aid_acc': 'AphiaId accepted', 'sn_acc': 'ScientificName accepted', 'fuzzy': False, 'like': True, 'puffer': True, 'marine_only': True, 'html': True, 'reportdir': '/Users/paul/_temp/report'}
        #kwargs = {'file': '/Users/paul/_temp/taxa_listen/taxa_marlin.xlsx', 'sheet': 'Taxa', 'aid_given': 'AphiaId given', 'sn_given': 'ScientificName given', 'aid_acc': 'AphiaId (accepted)', 'sn_acc': 'ScientificName (accepted)', 'fuzzy': False, 'like': True, 'puffer': True, 'marine_only': True, 'html': True, 'reportdir': '/Users/paul/_temp/report'}
        #kwargs = {'file': '/Users/paul/_temp/taxa_marlin.xlsx', 'sheet': 'Taxa', 'aid_given': 'AphiaId given', 'sn_given': 'ScientificName given', 'aid': 'AphiaId', 'sn': 'ScientificName', 'aid_acc': 'AphiaId accepted', 'sn_acc': 'ScientificName accepted', 'fuzzy': False, 'like': False, 'puffer': True, 'marine_only': False, 'html': True}
        #kwargs = {'file': '/Users/paul/_temp/beast/awi_alpha_ventus_beofino_beofino2.xlsx', 'sheet': 'Biota', 'aid_given': 'AphiaId', 'sn_given': 'Species', 'fuzzy': True, 'like': False, 'puffer': True, 'marine_only': True, 'html': True}
        #kwargs = {'file': '/Users/paul/_temp/beast/beast_petrogas_borkumse_stenen_biodiv_offshore_hardsub.xlsx', 'sheet': 'Biota', 'aid_given': 'AphiaId', 'sn_given': 'ScientificName', 'fuzzy': True, 'like': False, 'puffer': True, 'marine_only': True, 'html': True, 'reportdir': '/Users/paul/_temp/report'}

        pn = self.__smartprinter.normal
        pv = self.__smartprinter.verbose
        pd = self.__smartprinter.debug

        progress = self.__smartprinter.progress
        status = self.__smartprinter.status
        workflow = self.__smartprinter.workflow
        userinput = self.__smartprinter.inputAsk__comboMulti
        getInputAnswer = self.__smartprinter.getInputAnswer

        # Abbruch Reset
        self.abort(False)
        res = Result(name="VirtualAssistant: TaxonSpotter")
        pn("* TaxonSpotter", fmt=False)

        # Parameter
        cursor = kwargs["cursor"] if "cursor" in kwargs else None
        sFile = kwargs["file"] if "file" in kwargs else None
        sDirReport = kwargs["reportdir"] if "reportdir" in kwargs else None
        sSheet = kwargs["sheet"] if "sheet" in kwargs else None
        sAid_given = kwargs["aid_given"] if "aid_given" in kwargs else None
        sSn_given = kwargs["sn_given"] if "sn_given" in kwargs else None
        sAid = kwargs["aid"] if "aid" in kwargs else None
        sSn = kwargs["sn"] if "sn" in kwargs else None
        sAid_acc = kwargs["aid_acc"] if "aid_acc" in kwargs else None
        sSn_acc = kwargs["sn_acc"] if "sn_acc" in kwargs else None
        bFuzzy = kwargs["fuzzy"] if ("fuzzy" in kwargs and kwargs["fuzzy"] in [True, False]) else False
        bLike = kwargs["like"] if ("like" in kwargs and kwargs["like"] in [True, False]) else False
        bPuffer = kwargs["puffer"] if ("puffer" in kwargs and kwargs["puffer"] in [True, False]) else False
        bMarineOnly = kwargs["marine_only"] if ("marine_only" in kwargs and kwargs["marine_only"] in [True, False]) else False
        bHTML = kwargs["html"] if ("html" in kwargs and kwargs["html"] in [True, False]) else False

        if cursor is None:
            bOwnConnection = True
        else:
            bOwnConnection = False

        # ---

        sColor_awi_blue_dark = Colors.getColor("awi_blue_dark", html=True)
        sColor_awi_blue_mid = Colors.getColor("awi_blue_mit", html=True)
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)
        sColor_blue = Colors.getColor("blue", html=True)
        sColor_orange = sColor_awi_blue_mid
        sColor_red = Colors.getColor("red", html=True)
        sColor_green = Colors.getColor("green", html=True)
        sColor_yellow = Colors.getColor("yellow", html=True)
        sColor_green_light = Colors.getColor("green_light", html=True)
        sColor_yellow = Colors.getColor("yellow", html=True)
        sColor_purple = Colors.getColor("purple", html=True)
        sColor_rose = Colors.getColor("rose", html=True)
        sColor_white = Colors.getColor("white", html=True)

        sColor_accepted = sColor_green
        sColor_unaccepted_2_accepted = sColor_green_light
        sColor_unaccepted = sColor_rose
        sColor_unaccepted_2_unaccepted = sColor_yellow
        sColor_problems = sColor_red
        sColor_limbo = sColor_purple

        # ---

        progress("on-pulse")
        workflow(top="TaxonSpotter", sub="Starting up ...", fmt=False)

        if bOwnConnection:
            pd("Using own connection")
        else:
            pd("Reusing connection")

        # --- Check Parameter

        if sFile is None:
            res.addErr("No Excel file given", sys="TaxonSpotter", kat="I/O")
        if res.isErr():
            cleanUi()
            return res

        # Result Directory
        if not FileTools.isPathExiting(sDirReport):
            try:
                pn("* Directory for results not found: " + str(sDirReport))
                os.makedirs(sDirReport)
                pn("... now its made.")
            except Exception as ex:
                res.addErr("Create folder for results", add=str(ex), kat="Init", sys="TaxonSpotter")
                return res

        # Eigene DB Connection
        if bOwnConnection:
            resConn = self.__model.subsystem_getDb().conn_getConnection()
            if resConn.isErr():
                res.addResult(resConn)
                cleanUi()
                return res
            # Connection: An
            conn = resConn.getData("connection")
            cursor = resConn.getData("cursor")

        # Check Input
        if sAid == "": sAid = None
        if sSn == "": sSn = None
        if sAid_acc == "": sAid_acc = None
        if sSn_acc == "": sSn_acc = None

        # Log Total
        sLog = ""
        if bHTML:
            sLog += '<img height="67" src="https://recruitingapp-5442.de.umantis.com/pubhtml/logo.gif" alt="AWI Logo">'
            sLog += '<div layout="block" styleclass="DynForm_DfAttributes" style="z-index:0;position:absolute;left:0px;top:68px;right:0px;height:25px;overflow:hidden;text-align:left;background-color:'+sColor_awi_blue_light+';border:medium none;-moz-border-colors:none;border-radius:0; margin-top: 5px; margin-bottom: 5px; ;"></div>'
            sLog += '<br>'
            sLog += "\n\n"
            sLog += "&nbsp;"
            sLog += '<h1 style="color:' + sColor_awi_blue_dark + ';">TaxonSpotter</h1>\n\n'
        else:
            sLog = "*** TaxonSpotter\n\n"

        sLog += "*** Configuration and Information\n\n"
        sLog += "* Taxa File\n"
        lstTab = (
            ("File", str(FileTools.getFileWithoutPath(sFile))),
            ("Sheet", str(sSheet)),
        )
        sLog += PrettyText.table(lstTab, False, bHTML) + "\n"
        sLog += "\n"
        sLog += "* Taxa Data in Sheet\n"
        lstTab = (
            ("Taxa", "Column Name"),
            ("AphiaId (given by user)", str(sAid_given)),
            ("ScientificName (given by user)", str(sSn_given))
        )
        sLog += PrettyText.table(lstTab, True, bHTML) + "\n"
        sLog += "\n"

        sLog += "* Reference Data in Sheet\n"
        lstTab = (
            ("Reference", "Column Name"),
            ("AphiaId", (str(sAid) if sAid is not None else "-")),
            ("ScientificName", (str(sSn) if sSn is not None else "-")),
            ("AphiaId (accepted)", (str(sAid_acc) if sAid_acc is not None else "-")),
            ("ScientificName (accepted)", (str(sSn_acc) if sSn_acc is not None else "-"))
        )
        sLog += PrettyText.table(lstTab, True, bHTML) + "\n"
        sLog += "\n"
        sLog += "* Name Query Mode\n"
        lstTab = (
            ("Mode", "Option"),
            ("Use Buffer", ("yes" if bPuffer else "no")),
            ("Fuzzy Special ", ("yes" if bFuzzy else "no")),
            ("Fuzzy Generic ", ("yes" if bLike else "no"))
        )
        sLog += PrettyText.table(lstTab, True, bHTML) + "\n"
        sLog += "\n"

        if bHTML:
            sLog += "* Taxa Color Codes\n"
            lstTab = (
                ("Color", "Type", "Comment"),
                ('<span style="color:' + sColor_accepted + '">&#11044;</span>', "accepted (direct)", "Taxon itself is accepted"),
                ('<span style="color:' + sColor_unaccepted_2_accepted + '">&#11044;</span>', "accepted (redirected)", "Unaccepted taxon is redirected to accepted taxon"),
                ('<span style="color:' + sColor_unaccepted + '">&#11044;</span>', "unaccepted (direct)", "Taxon itself is unaccepted"),
                ('<span style="color:' + sColor_unaccepted_2_unaccepted + '">&#11044;</span>', "unaccepted (redirected)", "Unaccepted taxon is redirected to unaccepted taxon"),
                ('<span style="color:' + sColor_limbo + '">&#11044;</span>', "Limbo", "Taxon is either accepted nor unaccepted (it's stuck in between states in Limbo)"),
                ('<span style="color:' + sColor_problems + '">&#11044;</span>', "Problem", "Taxon has severe internal problems")
            )
            sLog += PrettyText.table(lstTab, True, bHTML) + "\n"
            sLog += "\n"

        # *** Excel
        pn("* Opening Excel", fmt=False)
        # Excel: Hole Wb
        resWb = ExcelTools.getWb(sFile, rw=True)
        if resWb.isErr():
            res.addResult(resWb, err=True)
            return res
        wb = resWb.getData("wb")

        # Excel: Hole Indizes der Spalte: AphiaId und Scientificname
        odicWb = OrderedDict()
        iColAid_given = None
        iColSn_given = None
        iColAid = None
        iColSn = None
        iColAid_acc = None
        iColSn_acc = None

        # Excel: Hole Indizes
        lstSheets_found = None
        lstHeader_found = None
        try:
            lstSheets_found = wb.sheetnames
            lstHeader_found = [cell.value for cell in wb[sSheet][1]]
            for cell in wb[sSheet][1]:
                if cell.value is not None:
                    sCol = str(cell.value)
                    if sCol == sAid_given:
                        iColAid_given = cell.column -1 # AphiaId
                    elif sCol == sSn_given:
                        iColSn_given = cell.column -1 # Scientificname
                    elif sCol == sAid:
                        iColAid = cell.column - 1  # AphiaId
                    elif sCol == sSn:
                        iColSn = cell.column - 1  # Scientificname
                    elif sCol == sAid_acc:
                        iColAid_acc = cell.column - 1  # AphiaId
                    elif sCol == sSn_acc:
                        iColSn_acc = cell.column - 1  # Scientificname
        except Exception as ex:
            # Sheet ist leer
            pass

        # Excel: Tests
        if sSheet not in wb.sheetnames:
            res.addErr("Sheet with Biota data not not found",
                       add=str(sSheet) +" -> sheets found: "+PrettyText.lst2strC(lstSheets_found), sys="TaxonSpotter", kat="Config")
        if iColAid_given is None:
            res.addErr("Column with AphiaIds (given) not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sAid_given) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")
        if iColSn_given is None:
            res.addErr("Column with ScientificNames (given) not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sSn_given) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")
        if sAid is not None and iColAid is None:
            res.addErr("Reference: Column with AphiaIds not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sAid) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")
        if sSn is not None and iColSn is None:
            res.addErr("Reference: Column with ScientificNames not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sSn) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")
        if sAid_acc is not None and iColAid_acc is None:
            res.addErr("Reference: Column with AphiaIds (accepted) not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sAid_acc) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")
        if sSn_acc is not None and iColSn_acc is None:
            res.addErr("Reference: Column with ScientificNames (accepted) not found in Biota Sheet",
                       add=str(sSheet) + "." + str(sSn_acc) + " -> columns found: " + PrettyText.lst2strC(lstHeader_found), sys="TaxonSpotter", kat="Config")

        # *** Excel: Hole alle Daten aus wb
        workflow(top="TaxonSpotter", sub="Importing EXCEL", fmt=False)
        tc = TaxaCruncher(self.__model)
        lstTaxonList_input = []
        if res.isOk():
            progress("on-abort")
            iRowAll = len(list(wb[sSheet].rows))
            for iRow, row in enumerate(wb[sSheet]):
                iRow += 1
                if self.isAbort(): # Abort
                    res.addErr("Aborted by user...", sys="TaxonSpotter", kat="Abort")
                    break
                if iRow > 1: # Header
                    progress(iRow, iRowAll, 100, proz=True, abs=True, cli=True)
                    if ExcelTools.isRowEmpty(row):
                        # Zeile leer
                        res.addInfo("Empty row", sheet=sSheet, row=iRow, sys="TaxonSpotter", kat="Reading Data")
                    else:
                        # Zeile nicht leer
                        sAid_given = row[iColAid_given].value if row[iColAid_given].value not in [None, ""] else None
                        sSn_given = row[iColSn_given].value if row[iColSn_given].value not in [None, ""] else None
                        if iColAid is not None:
                            sAid = row[iColAid].value if row[iColAid].value not in [None, ""] else None
                        if iColSn is not None:
                            sSn = row[iColSn].value if row[iColSn].value not in [None, ""] else None
                        if iColAid_acc is not None:
                            sAid_acc = row[iColAid_acc].value if row[iColAid_acc].value not in [None, ""] else None
                        if iColSn_acc is not None:
                            sSn_acc = row[iColSn_acc].value if row[iColSn_acc].value not in [None, ""] else None
                        lstTaxonList_input.append((sAid_given, sSn_given, sAid, sSn, sAid_acc, sSn_acc, iRow))
                        tc.addTaxon(sAid_given, sSn_given, iRow)

        # *** Crunch
        taxa = None
        dicName2taxa = {}
        if res.isOk():
            resTc = tc.crunch(fuzzy=bFuzzy, like=bLike, puffer=bPuffer, marine_only=bMarineOnly, html=bHTML)
            res.addResult(resTc, msg_only=True)
            sLog_taxa_crunch = resTc.getData("log")
            taxa = resTc.getData("taxa")
            dicName2taxa = resTc.getData("name2taxa")
            if sLog_taxa_crunch is not None:
                sLog += sLog_taxa_crunch

        # *** Taxa: Statistik
        if taxa is not None:
            sLog_taxa_stat = taxa.getStatistic(html=bHTML, accepted=True)
            if sLog_taxa_stat is not None:
                sLog += sLog_taxa_stat

        # *** Referenzen: Checke gegen Aid/Sn (acc)
        lstRef_diff_aid = []
        lstRef_diff_sn = []
        lstRef_diff_aid_acc = []
        lstRef_diff_sn_acc = []
        for sAid_given, sSn_given, sAid, sSn, sAid_acc, sSn_acc, iRow in lstTaxonList_input:
            taxon = None
            try:
                iAid_given = int(sAid_given)
            except:
                iAid_given = None
            if iAid_given is not None:
                taxon = taxa.getTaxon(iAid_given)
            elif (sSn_given is not None) and (sSn_given in dicName2taxa):
                taxon = dicName2taxa[sSn_given]
            if taxon is not None:
                sAid_given_clean = str(taxon.getAphiaId())
                sSn_given_clean = taxon.getScientificName()
                sAid_given_acc_clean = str(taxon.getAphiaId_accepted())
                sSn_given_acc_clean = taxon.getScientificName_accepted()
                if (sAid is not None) and (sAid_given_clean != str(sAid)):
                    lstRef_diff_aid.append((str(iRow),str(sAid_given_clean),str(sAid)))
                if (sSn is not None) and (SequenceMatcher(None, sSn_given_clean, sSn).ratio() < self.__wizard.getScientificNameMatchRatio()):
                    lstRef_diff_sn.append((str(iRow) , str(sSn_given_clean) , str(sSn)))
                if (sAid_acc is not None) and sAid_given_acc_clean is not None and (sAid_given_acc_clean != str(sAid_acc)):
                    lstRef_diff_aid_acc.append((str(iRow),  str(sAid_given_acc_clean) , str(sAid_acc)))
                if (sSn_acc is not None) and sSn_given_acc_clean is not None and (SequenceMatcher(None, sSn_given_acc_clean, sSn_acc).ratio() < self.__wizard.getScientificNameMatchRatio()):
                    lstRef_diff_sn_acc.append((str(iRow) , str(sSn_given_acc_clean) , str(sSn_acc)))
            else:
                pass
        lstRef = [
            ("AphiaId (given) != AphiaId", "AphiaId (given)", "AphiaId", lstRef_diff_aid),
            ("ScientificName (given) != ScientificName","ScientificName (given)", "ScientificName", lstRef_diff_sn),
            ("AphiaId (given) != AphiaId (accepted)", "AphiaId (given)", "AphiaId (accepted)", lstRef_diff_aid_acc),
            ("ScientificName (given) != ScientificName", "ScientificName (given)", "ScientificName", lstRef_diff_sn_acc)
        ]
        sLogRef = ""
        for sType, sHead1, sHead2, lstRows in lstRef:
            if len(lstRows) > 0:
                sLogRef += "* " + sType + " (" + str(len(lstRows)) + "x)\n"
                lstTab = [("Row", sHead1, sHead2)]
                for sRow, sCol1, sCol2 in lstRows:
                    lstTab.append((sRow, sCol1,sCol2))
                sLogRef += PrettyText.table(lstTab, True, bHTML) + "\n"
        if sLogRef != "":
            sLog += "*** Check against Reference\n\n"
            sLog += sLogRef

        # *** Save Report
        sTimeStamp = PrettyText.createTimestamp()
        sDir_rep_stamp = str(Path(sDirReport, sTimeStamp))
        try:
            os.makedirs(sDir_rep_stamp)
        except Exception as ex:
            res.addErr("Create timesamped folder for reports", add=str(ex), kat="I/O", sys="TaxonSpotter")
        sTmp = FileTools.getFileWithoutPath(sFile)
        sFile_rep_name = FileTools.getFileWithoutType(sTmp)

        if bHTML:
            sLog = PrettyText.html_txt2html(sLog, title="TaxReport: " + FileTools.getFileWithoutPath(sFile))

        progress("on-pulse")
        workflow(top="TaxonSpotter", sub="Writing Reference")
        pn("* Writing Report")
        if bHTML:
            sFile_report = sFile_rep_name + "_report" + ".html"
        else:
            sFile_report = sFile_rep_name + "_report" + ".txt"
        sFile_report = str(Path(sDir_rep_stamp, sFile_report))
        resBackup = FileTools.prepareSaveFile(sFile_report)
        if resBackup.isErr():
            res.addResult(resBackup, sys="TaxonSpotter", kat="I/O", msg_only=True)
        else:
            try:
                with open(sFile_report, "w") as f:
                    f.write(sLog)
            except Exception as ex:
                    res.addErr("Error during saving Report File", add=str(ex), sys="TaxonSpotter", kat="I/O")

        # *** Write Reference Taxon List
        if taxa is not None:
            pn("* Writing Reference Taxon List")
            sFile_reference = sFile_rep_name + "_reference" + ".xlsx"
            sFile_reference = str(Path(sDir_rep_stamp, sFile_reference))
            resBackup = FileTools.prepareSaveFile(sFile_reference)
            if resBackup.isErr():
                res.addResult(resBackup, sys="TaxonSpotter", kat="I/O", msg_only=True)
            else:
                try:
                    wb = openpyxl.Workbook()
                    ws = wb.active
                    ws.title = "Reference Taxon List"
                    ws.append(["AphiaId",
                               "ScientificName",
                               "AphiaId (accepted)",
                               "ScientificName (accepted)",
                               "Not marine",
                               "Errors",
                               "Name Problems",
                               "Status Problems",
                               "Accepted",
                               "Limbo",
                               "Status",
                               "AphiaID Chain",
                               "Tail accepted",
                               "Uaccept reason",
                               "Rank",
                               "Kingdom",
                               "Phylum",
                               "Class",
                               "Order",
                               "Family",
                               "Genus",
                               "Authorty",
                               "Authorty (accepted)"
                               ])
                    for cell in ws["1:1"]:
                        cell.font = openpyxl.styles.Font(bold=True)
                    lstTaxa = taxa.getTaxa(sort="sn")
                    for iRow, t in enumerate(lstTaxa):
                        progress(iRow, iRowAll, 100, proz=True, abs=True, cli=True)
                        ws.append([t.getAphiaId(),
                                   t.getScientificName(),
                                   t.getAphiaId_accepted(),
                                   t.getScientificName_accepted(),
                                   "yes" if not t.isMarine() else "",
                                   "yes" if t.isErr() else "",
                                   "yes" if t.hasProblems_names() else "",
                                   "yes" if t.hasProblems_status() else "",
                                   "yes" if t.isAccepted() else "",
                                   "yes" if t.isLimbo() else "",
                                   t.getStatus(),
                                   PrettyText.lst2str(t.getAcceptedWay(), "->") if t.getAcceptedWay() is not None else "",
                                   "yes" if t.isRedirected_2_accepted() else "",
                                   t.getUnacceptReason(),
                                   t.getRank(),
                                   t.getKingdom(),
                                   t.getPhylum(),
                                   t.getClass(),
                                   t.getOrder(),
                                   t.getFamily(),
                                   t.getGenus(),
                                   t.getAuthority(),
                                   t.getAuthority_accepted()
                                   ])
                    column_letters = tuple(openpyxl.utils.get_column_letter(col_number + 1) for col_number in range(ws.max_column))
                    for column_letter in column_letters:
                        ws.column_dimensions[column_letter].bestFit = True
                    wb.save(filename=sFile_reference)
                except Exception as ex:
                    res.addErr("Error during saving Excel file", add=str(ex), sys="TaxonSpotter", kat="I/O")

        # Alle Fehler/Warnungen vor das Log stellen
        # if res.isAnyMessage():
        #     sLog = res.getErrWarnInfoStringHeader() + res.getErrWarnInfoString() + res.getErrWarnInfoStringFooter() + sLog
        #     sLog = PrettyText.clean_wizard(sLog, max_one_empty_row=True)

        # Antwort
        res.addData("title", "TaxonSpotter")
        res.addData("text", sLog)
        res.addData("html", bHTML)
        # Aufräumen
        cleanUi()
        if bOwnConnection:
            closeDb()

        return res

