--
-- PostgreSQL database dump
--

-- Dumped from database version 13.4
-- Dumped by pg_dump version 13.4

SET statement_timeout = 0;
SET lock_timeout = 0;
SET idle_in_transaction_session_timeout = 0;
SET client_encoding = 'UTF8';
SET standard_conforming_strings = on;
SELECT pg_catalog.set_config('search_path', '', false);
SET check_function_bodies = false;
SET xmloption = content;
SET client_min_messages = warning;
SET row_security = off;

--
-- Name: cb_empty; Type: SCHEMA; Schema: -; Owner: -
--

CREATE SCHEMA cb_empty;


--
-- Name: antwort; Type: TYPE; Schema: cb_empty; Owner: -
--

CREATE TYPE cb_empty.antwort AS (
	err text,
	warn text,
	info text
);


--
-- Name: is_sample_ok(text, integer, boolean); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.is_sample_ok(s_schema text DEFAULT NULL::text, i_sample_id integer DEFAULT NULL::integer, b_debug boolean DEFAULT false) RETURNS text[]
    LANGUAGE plpgsql
    AS $$DECLARE
	f_replicates float; -- Ausgerechnete Replikatanzahl
	f_gsa float; -- GSA: Gear Sampled Area wurde ausgerechnet
	b_gsa boolean; -- GSA: Gear Sampled Area ist berechenbar, wurde aber nicht ausgerechnet
	b_schema_existing boolean; 
	rec_sample record; -- Sample
	rec_gear record; -- Gear
	arr_antwort text ARRAY[3]; -- Antwort String-Array:0:Error/1:Warnung/2:Info
BEGIN	
	
	b_gsa = false;
	
	-- ----------------------------------------------------------------------
	-- HIER WIRD NUR AUF DIREKTE, DEN PROZESS UNTERBRECHENDE FEHLER GETESTET!
	-- WELCHES DATENSZENARIO VORLIEGT WIRD NICHT GETESTET
	-------------------------------------------------------------------------
	
	-- Teste ob Schema existiert
	EXECUTE format('SELECT TRUE FROM information_schema.schemata WHERE schema_name=%L', s_schema) INTO b_schema_existing;
	IF b_schema_existing is null THEN
		arr_antwort[0] = 'Schema not existing: ' ||  s_schema;
		return arr_antwort ;
	END IF;

	-- Hole das gesuchte Sample
	EXECUTE format('SELECT * from %s.sample smp WHERE smp.id=%s',  s_schema, i_sample_id) INTO rec_sample;
	IF rec_sample is null THEN
		arr_antwort[0] = 'Sample not existing: i_sample_id=' || i_sample_id;
		return arr_antwort;
	END IF;

	-- Hole gear data
	EXECUTE format('SELECT name, category, width, area FROM %s.gear gea WHERE gea.id=%s', s_schema, rec_sample.gear_id) INTO rec_gear;
	
	-- ------------------------------------------------------------------------------------------
	-- CHECK GRAB_PENETRATION_DEPTH: Teste ob bei gegebener das Gear auch wirklich ein Grab ist 
	-- ------------------------------------------------------------------------------------------
	
	IF rec_sample.grab_penetration_depth is not NULL and rec_gear.category <> 'grab' THEN
		arr_antwort[0] = 'Grab penetration depth is set (' || rec_sample.grab_penetration_depth || ') and gear ('|| rec_gear.name ||') type is not grab ';
		return arr_antwort;
	END IF;
	
	-- ------------------------------------------------------------------------------------------
	-- CHECK AREA: Teste ob die Gear Area Eingabe konsistent ist (keine Eingabe ist auch korrekt)
	-- ------------------------------------------------------------------------------------------
	
	-- ------------------------------------
	-- Eingangstests (fuer alle Gear-Typen)
	-- ------------------------------------
	
	-- sample.sampled_area und sample.reference_area duerfen nicht gleichzeitig gesetzt sein
	IF rec_sample.sampled_area is not NULL and rec_sample.reference_area is not NULL THEN
		arr_antwort[0] = 'Sampling area SA (' || rec_sample.sampled_area || ') and reference area RA ('|| rec_sample.reference_area ||') are set simultaniously';
		return arr_antwort;
	END IF;
	
	-- -------------------------------------
	-- Haupttests (getrennt nach Gear-Typen)
	-- -------------------------------------
	
	IF rec_gear.category='trawl' THEN
		-- *** Trawl
		-- Sampling Distance darf nicht mehrfach berechenbar sein: sample.sampling_distance <=> sample.sampling_speed + sample.time_start/end) 
		IF rec_sample.sampling_distance is not NULL and (rec_sample.sampling_speed is not NULL and rec_sample.time_start is not NULL and rec_sample.time_end is not NULL) THEN
			arr_antwort[0] = 'Trawling distance could be calculated twice (sampling_distance and sampling_speed with time_start/end are set together)';
			return arr_antwort;
		END IF;
		-- Teste ob GSA berechenbar ist 
		if (rec_sample.sampling_distance is not NULL) or
			(rec_sample.sampling_speed is not NULL and rec_sample.time_start is not NULL and rec_sample.time_end is not NULL) or 
			(rec_sample.lat_start is not NULL and rec_sample.lat_end is not NULL and rec_sample.lon_start is not NULL and rec_sample.lon_end is not NULL) and
			rec_gear.width is not NULL THEN
				-- Ja... GSA ist berechenbar
				b_gsa = true;
		END IF;
	ELSE
		-- ### Grab, Sensor, Video, Scrape, Rocks (Rest)
		-- Area ist in Gear-Tabelle definiert?
		IF (rec_gear.area is not NULL and rec_gear.area <> 0) THEN
			-- Ja... GSA ist berechenbar und wird berechnet
			f_gsa = rec_gear.area * rec_sample.replicates; 
			b_gsa = true;
			-- Teste ob GSA von expliziter Angabe SA (sample.sampled_area) abweicht
			IF (rec_sample.sampled_area is not NULL and rec_sample.sampled_area <> f_gsa) THEN
				arr_antwort[0] = 'Gear (' || rec_gear.name || ') sampling area GSA (sample.replicates*gear.area -> '|| rec_sample.replicates || '*' || rec_gear.area || '=' || f_gsa ||') differing from given sampled area SA in sample (' || rec_sample.sampled_area || ')';
				return arr_antwort; 
			END IF;
		END IF;
	END IF;
	
	-- -------------------------------------
	-- Abschlussbericht
	-- -------------------------------------
	-- Debug Info 
	IF b_debug is true THEN
		RAISE NOTICE '%','* Information: area sample';
		RAISE NOTICE 'schema: %', s_schema;
		RAISE NOTICE 'sample.id: %', i_sample_id;
		RAISE NOTICE 'sample.sampled_area SA: %', rec_sample.sampled_area;
		RAISE NOTICE 'sample.reference_area RA: %', rec_sample.reference_area;
		RAISE NOTICE 'sample.lat_start: %', rec_sample.lat_start;
		RAISE NOTICE 'sample.lon_start: %', rec_sample.lon_start;
		RAISE NOTICE 'sample.lat_end: %', rec_sample.lat_end;
		RAISE NOTICE 'sample.lon_end: %', rec_sample.lon_end;
		RAISE NOTICE 'sample.time_start: %', rec_sample.time_start;
		RAISE NOTICE 'sample.time_end: %', rec_sample.time_end;
		RAISE NOTICE 'sample.sampling_distance: %', rec_sample.sampling_distance;
		RAISE NOTICE 'sample.sampling_speed: %', rec_sample.sampling_speed;
		RAISE NOTICE 'sample.replicates: %', rec_sample.replicates;
		RAISE NOTICE 'gear.name: %', rec_gear.name;
		RAISE NOTICE 'gear.category: %', rec_gear.category;
		RAISE NOTICE 'gear.width: %', rec_gear.width;
		RAISE NOTICE 'gear.area: %', rec_gear.area;
		IF b_GSA = true then
			RAISE NOTICE '%', 'GSA (calculated via gear): area is calcuable';
		ELSE
			RAISE NOTICE 'GSA (calculated via gear): %', f_gsa;
		END IF;
	END IF;
	
	return arr_antwort;
END;$$;


--
-- Name: is_subset_biota(text, integer, boolean); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.is_subset_biota(s_schema_name text DEFAULT NULL::text, i_subset_id integer DEFAULT NULL::integer, b_debug boolean DEFAULT false) RETURNS boolean
    LANGUAGE plpgsql
    AS $_$declare
	b_schema_existing boolean; 
	b_subset_is_biota boolean;
	rec_subset record;
begin

	-- Test: alle Daten angegeben
	if s_schema_name is null or i_subset_id is null then
		raise exception 'missing data: schema=% subset=%', s_schema_name, i_subset_id;
	end if;

	-- Test: Schema existiert
	execute format('select true from information_schema.schemata where schema_name=%L', s_schema_name) into b_schema_existing;
	if b_schema_existing is null then
		raise exception 'schema not existing: %', s_schema_name;
	end if;
	
	-- Test: Subset biota existiert
	execute format('select taxcover_id,taxres,spatiallevel from %1$s.subset where id=%2$L', s_schema_name, i_subset_id) into rec_subset;
	IF (rec_subset.taxcover_id is Null or rec_subset.taxres is Null or rec_subset.spatiallevel is Null) THEN
		b_subset_is_biota = false;
	ELSE
		b_subset_is_biota = true;
	END IF;
	
	-- Debug Info 
	if b_debug is true then
		raise notice '* Information: subset is biota';
		raise notice 'schema: %', s_schema_name;
		raise notice 'taxcover_id: %', rec_subset.taxcover_id;
		raise notice 'taxres: %', rec_subset.taxres;
		raise notice 'spatiallevel: %', rec_subset.spatiallevel;
		if b_subset_is_biota then 
			raise notice 'subset is biota: yes';
		else
			raise notice 'subset is biota: no';
		end if;
	end if;

	-- Antwort
	return b_subset_is_biota;
end;
$_$;


--
-- Name: is_subset_triplet_uniq(text, integer, integer, text, integer, boolean); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.is_subset_triplet_uniq(s_schema_name text DEFAULT NULL::text, i_subsample_id integer DEFAULT NULL::integer, i_taxcover_id integer DEFAULT NULL::integer, s_taxres text DEFAULT NULL::text, i_spatiallevel integer DEFAULT NULL::integer, b_debug boolean DEFAULT false) RETURNS boolean
    LANGUAGE plpgsql
    AS $_$declare
	b_schema_existing boolean; 
	i_count_triplets integer;
begin

	-- Teste ob alle Daten angegeben wurden
	if s_schema_name is null or i_subsample_id is null or i_taxcover_id is null or s_taxres is null or i_spatiallevel is null then
		raise exception 'missing data: schema=% sample=% taxcover=% taxres=% spatiallevel=%', s_schema_name, i_subsample_id, i_taxcover_id, s_taxres, i_spatiallevel;
	end if;

	-- Teste ob Schema existiert
	execute format('SELECT TRUE FROM information_schema.schemata WHERE schema_name=%L', s_schema_name) into b_schema_existing;
	if b_schema_existing is null then
		raise exception 'schema not existing: %', s_schema_name;
	end if;
	
	-- Test Einmaligkeit
	execute format('select count(id) from %1$s.subset sub where sub.subsample_id=%2$L and sub.taxcover_id=%3$L and sub.taxres=%4$L and sub.spatiallevel=%5$L', s_schema_name, i_subsample_id, i_taxcover_id, s_taxres, i_spatiallevel) into i_count_triplets;
	
	-- Debug Info 
	if b_debug is true then
		raise notice '* Information: subset triplet uniq';
		raise notice 'schema: %', s_schema_name;
		raise notice 'subsample_id: %', i_subsample_id;
		raise notice 'taxcover_id: %', i_taxcover_id;
		raise notice 'taxres: %', s_taxres;
		raise notice 'spatiallevel: %', i_spatiallevel;
		raise notice 'found subsets: %', i_count_triplets;
	end if;

	-- Antwort
	if i_count_triplets > 0 then return false; else return true; end if;
end;$_$;


--
-- Name: new_biota_check_consistency(); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.new_biota_check_consistency() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$DECLARE
	
	-- Debug
	b_debug boolean; -- Debug
	
	-- Basis
	i_subset_id integer; -- Subset Id
	i_sample_id integer; -- Sample Id
	i_subsample_id integer; -- SubSample Id
	rec_subsample record; -- SubSample
	rec_sample record; -- Sample
	rec_gear record; -- Gear
	rec_branch record; -- Vollständiger Cruise-Station-Sample-Subset-Ast des Tests 
	
	-- BioDjin
	f_gsa float; -- GSA: Sampled Area berechnet aus Gear und evtl. anderen Quellen
	f_gsa_tmp float; -- GSA Temp
	b_gsa boolean; -- GSA: Sampled Area ist berechenbar aus Gear und evtl. anderen Quellen
	b_subset_is_biota boolean; -- Antwort des Subset-Biota-Test 
	s_szenario_grab text; -- Code Grab-Datenszenario
	
	-- Answer
	s_error text; -- Fehlertext
	s_warn text; -- Warnungstext
	s_branch text;
	s_branch_long text;
	
BEGIN

	b_debug = false;
	b_gsa = false;
	
	--------------------------------------------------------------------------------------------
	-- Vorarbeiten 
	--------------------------------------------------------------------------------------------
	
	-- Subset Id
	i_subset_id = NEW.subset_id;
	
	-- SubSample Id
	EXECUTE format('select subsample_id from %1$s.subset where id=%2$s;',TG_TABLE_SCHEMA, i_subset_id) into i_subsample_id;
	
	-- Hole SubSample
	EXECUTE format('select * from %1$s.subsample where id=%2$s;',TG_TABLE_SCHEMA, i_subsample_id) into rec_subsample;
	
	-- Sample Id
	EXECUTE format('select sample_id from %1$s.subsample where id=%2$s;',TG_TABLE_SCHEMA, i_subsample_id) into i_sample_id;
	
	-- Hole Sample
	EXECUTE format('select * from %1$s.sample where id=%2$s;',TG_TABLE_SCHEMA, i_sample_id) into rec_sample;
	
	-- Hole gear data
	EXECUTE format('SELECT name, category, width, area, smpunit FROM %s.gear gea WHERE gea.id=%s',  TG_TABLE_SCHEMA, rec_sample.gear_id) INTO rec_gear;

	-- Teste ob Subset Biotisches Subset ist (Tax-Cover/Resolution/Spatial-Level sind gesetzt)
	EXECUTE format('select %1$s."is_subset_biota"(%1$L,%2$s,False)', TG_TABLE_SCHEMA, i_subset_id) INTO b_subset_is_biota;
	IF b_subset_is_biota IS FALSE THEN
		s_error = 'You want to attach an biota to a subset that is not a biota subset';
	END IF;
	
	-- Teste ob in number eine Fließkommazahl erlaubt ist (es muss eine Angabe in sample.subsample_share oder biota.subtotal_factor exisiteren)
	if NEW.number is not null and ((trunc(NEW.number) != NEW.number) and rec_sample.subsample_share is NULL and NEW.subtotal_factor is NULL) THEN
		s_error = 'Float in biota.number (' || NEW.number || ') is not allowed (missing data in sample.subsample_share or biota.subtotal_factor)';
	END IF;
		
	--------------------------------------------------------------------------------------------
	-- BioDjinn
	--------------------------------------------------------------------------------------------
			
	-- Bestimme Area (sampled_area, reference_area)
	IF rec_gear.category='trawl' then
		-- Trawl
		-- Teste ob Sampled Area berechenbar ist via Gear und Sample?
		if (rec_sample.sampling_distance is not NULL) or
			(rec_sample.sampling_speed is not NULL and rec_sample.time_start is not NULL and rec_sample.time_end is not NULL) or 
			(rec_sample.lat_start is not NULL and rec_sample.lat_end is not NULL and rec_sample.lon_start is not NULL and rec_sample.lon_end is not NULL) and
			rec_gear.width is not NULL THEN
			-- Sampled Area ist berechenbar - wird aber nicht berechnet
			b_gsa = true;
		END IF;
	ELSE
		-- Sensor/Trap/Grab/Video/Scrape/Rocks
		-- Hole Area aus Gear-Tabelle (gear.area * sample.replicates)
		IF (rec_gear.area is not NULL and rec_gear.area <> 0) THEN
				-- Sampled Area ist berechenbar und wird berechnet
			b_gsa = true;
			f_gsa = rec_gear.area * rec_sample.replicates;
			-- Teste ob Area aus Gear-Tabelle von sample.sampled_area abweicht
			IF (rec_sample.sampled_area is not NULL and rec_sample.sampled_area <> f_gsa) THEN
				s_error = 'BioDjinn says: Gear (' || rec_gear.name || ') sampling area (sample.replicates*gear.area -> '|| rec_sample.replicates || '*' || rec_gear.area || '=' || f_gsa ||') differing from given sampled area in sample (' || rec_sample.sampled_area || ')';
			END IF;
		END IF;
	END IF;
	
	-- Teste: ob Sampled/Reference-Area zu Standard Unit des Gears passt (both,area)
	if (rec_sample.sampled_area IS NOT NULL OR rec_sample.reference_area IS NOT NULL) AND rec_gear.smpunit NOT IN ('both','area') THEN
		s_error = 'BioDjinn says: Sampled/Reference-Area is given and gear standard unit is not set to area (it is set to: ' || rec_gear.smpunit || ')';
	END IF;
	
	-- Teste: ob Sampled/Reference-Vol zu Standard Unit des Gears passt (both,vol)
	if (rec_sample.sampled_vol is not NULL OR rec_sample.reference_vol is not null) AND rec_gear.smpunit not in ('both','vol') THEN
		s_error = 'BioDjinn says: Sampled/Reference-Volume is given and gear standard unit is not set to volume (it is set to: ' || rec_gear.smpunit || ')';
	END IF;
	
	-- Teste: SampledArea/Vol oder ReferenceArea/Vol und Gear Standard Unit = both -> Immer Warnung
	if (rec_sample.sampled_area is NOT NULL or rec_sample.sampled_vol is NOT NULL) and rec_gear.smpunit = 'both' THEN
		s_warn = 'BioDjinn says: SampledArea/Volume is given and gear standard unit is set to both - is that right?';
	END IF;
	if (rec_sample.reference_area is NOT NULL or rec_sample.reference_vol is NOT NULL) AND rec_gear.smpunit = 'both' THEN
		s_warn = 'BioDjinn says: ReferenceArea/Volume is given and gear standard unit is set to both - is that right?';
	END IF;
	
	-- Test: Mutual Exclusion: Number/Abundanz - PresenceAbsence
	IF NEW.number IS NOT NULL and NEW.presence_absence IS NOT NULL THEN
		s_error = 'BioDjinn says: Number (' || NEW.number || ') and Presence/Absence ('|| NEW.presence_absence ||') must not be set at the same time';
	END IF;
	IF NEW.abundance IS NOT NULL and NEW.presence_absence IS NOT NULL THEN
		s_error = 'BioDjinn says: Abundance (' || NEW.abundance || ') and Presence/Absence ('|| NEW.presence_absence ||') must not be set at the same time';
	END IF;

	-- Teste: Daten-Szenario
	-- 0a (1 Sample, 1 Replicate, Sampled Area berechenbar durch replicate * gear.area oder direkt durch sample.sampled_area oder Sampled Volume gegeben in sample.sampled_vol)
	IF ((rec_sample.sampled_area is NOT NULL or f_gsa is not null) OR rec_sample.sampled_vol is not NULL) AND
		(rec_sample.reference_area is NULL AND rec_sample.reference_vol is NULL) AND
		NEW.number is NOT NULL AND
		NEW.abundance is NULL AND
		NEW.presence_absence is NULL AND
		rec_sample.replicates = 1 THEN
			NEW.abundance = NULL; -- Verwerfen -> Kann aus SA und N berechnet werden
			s_szenario_grab = '0a';
	-- 0b (1 Sample, n Replicate, Sampled Area (in Summe) berechenbar durch replicate * gear.area oder direkt durch sample.sampled_area oder Sampled Volume gegeben in sample.sampled_vol)
	ELSIF ((rec_sample.sampled_area is NOT NULL or f_gsa is not null) or rec_sample.sampled_vol is not NULL) AND
		(rec_sample.reference_area is NULL AND rec_sample.reference_vol is NULL) AND
		NEW.number is NOT NULL AND
		NEW.abundance is NULL AND
		NEW.presence_absence is NULL AND
		rec_sample.replicates > 1 THEN
			NEW.abundance = NULL; -- Verwerfen -> Kann aus SA und N berechnet werden
			s_szenario_grab = '0b';
	-- 0c (1 Sample, ? Replikate, Reference Area/Vol, Abundance, Number ->  Sampled Area ist berechenbar (SA = (N*RA)/A) )
	ELSIF (rec_sample.sampled_area is NULL AND rec_sample.sampled_vol is NULL) AND 
		((rec_sample.reference_area is NOT NULL) OR (rec_sample.reference_vol is NOT NULL)) AND 
		NEW.number is NOT NULL AND 
		NEW.abundance is NOT NULL AND
		NEW.presence_absence is NULL THEN
			s_szenario_grab = '0c';			
			-- Teste: Fall 0c bei gegebener Reference Area, Gear!=Trawl, GSA berechenbar -> Gilt die Formel: GSA=(N*RA)/A
			-- Gear=Video/Sensor/Grab und GSA 
			IF (rec_sample.reference_area is NOT NULL) and (rec_gear.category <> 'trawl') and b_gsa is True THEN
				f_gsa = rec_gear.area * rec_sample.replicates;
				f_gsa_tmp = (NEW.number*rec_sample.reference_area)/NEW.abundance; 
				if f_gsa <> f_gsa_tmp THEN
					-- Problem: Vergleich zweier Floats -> das kann häufig falsch positiv zuschlagen.
					s_error = format('BioDjinn says: Problem testing gear calculated sampled_area GSA - equation GSA=(N*RA)/A does not apply: gear (%s),  gear category (%s), GSA (%s), (N*RA)/A (%s), number N (%s), abundance A (%s), reference_area RA (%s)', rec_gear.name, rec_gear.category, f_gsa, f_gsa_tmp, NEW.number, NEW.abundance, rec_sample.reference_area);
				END IF;
			END IF;	
	-- 1a (1 Sample, Abundance, Reference Area/Vol, keine Sampled Area/Vol, GSA (in unser Datenmodell reingequetscht)
	ELSIF (rec_sample.sampled_area is NULL  AND rec_sample.sampled_vol is NULL) AND 
		((rec_sample.reference_area is NOT NULL) OR (rec_sample.reference_vol is NOT NULL)) AND
		b_gsa AND
		NEW.number is NULL AND 
		NEW.abundance is NOT NULL AND
		NEW.presence_absence is NULL THEN
			s_szenario_grab = '1a';
	-- 1b (1 Sample, Abundance, Reference Area/Vol, keine Sampled Area, keine GSA (in unser Datenmodell reingequetscht)
	ELSIF (rec_sample.sampled_area is NULL  AND rec_sample.sampled_vol is NULL) AND 
		((rec_sample.reference_area is NOT NULL) OR (rec_sample.reference_vol is NOT NULL)) and 
		not b_gsa AND
		NEW.number is NULL AND 
		NEW.abundance is NOT NULL AND
		NEW.presence_absence is NULL THEN
		s_szenario_grab = '1b';
	-- 2 (1 Sample, Number)
	ELSIF rec_sample.sampled_area is NULL  AND rec_sample.sampled_vol is NULL AND 
		rec_sample.reference_area is NULL AND rec_sample.reference_vol is NULL and 
		NEW.number is NOT NULL and 
		NEW.abundance is NULL AND
		NEW.presence_absence is NULL THEN
			s_szenario_grab = '2';
	-- 3 (1 Sample, PresenceAbsence, Sampled Area/Vol)
	ELSIF ((rec_sample.sampled_area is NOT NULL or f_gsa is NOT NULL) OR (rec_sample.sampled_vol is not NULL)) AND 
		(rec_sample.reference_area is NULL AND rec_sample.reference_vol is NULL) and 
		NEW.number is NULL AND 
		NEW.abundance is NULL AND
		NEW.presence_absence is NOT NULL THEN
			s_szenario_grab = '3';	
	-- 4 (1 Sample, PresenceAbsence)
	ELSIF (rec_sample.sampled_area is NULL AND rec_sample.sampled_vol is NULL) AND 
		NEW.number is NULL and 
		NEW.presence_absence is NOT NULL THEN
			s_szenario_grab = '4';
	-- Kein Szenario erkannt -> Fehler
	ELSE
		if b_gsa = true then
			-- Sampled Area ist berechenbar (Trawl) wurde aber nicht ausgerechent
			s_error = format('BioDjinn says: Could not determine data szenario: replicates R (%s), number N (%s), abundance A (%s), gear calculated sampled_area GSA (is calcuable), sampled_area SA (%s), reference_area RA (%s) and presence/absence PA (%s)', 
							 rec_sample.replicates, NEW.number, NEW.abundance, rec_sample.sampled_area, rec_sample.reference_area, NEW.presence_absence);
		ELSE
			-- Sampled Area ist berechenbar und wurde bereits ausgerechnet
			s_error = format('BioDjinn says: Could not determine biota data szenario: replicates R (%s), number N (%s), abundance A (%s), gear calculated sampled_area GSA (%s), sampled_area SA (%s), reference_area RA (%s) and presence/absence PA (%s)', 
							 rec_sample.replicates, NEW.number, NEW.abundance, f_gsa, rec_sample.sampled_area, rec_sample.reference_area, NEW.presence_absence);
		END IF;
	END IF;
	
	--------------------------------------------------------------------------------------------
	-- Fehler 
	--------------------------------------------------------------------------------------------
	IF s_error is not NULL or s_warn is not NULL THEN
		-- Bestimme Ast für Meldungen
		EXECUTE format('select cru.name cruise, cru.ingest_id cruise_ing, sta.name station, sta.ingest_id station_ing, smp.name sample, smp.ingest_id sample_ing, subsmp.name subsample, subsmp.ingest_id subsample_ing, sub.name subset, sub.ingest_id subset_ing from %1$s.cruise cru, %1$s.station sta, %1$s.sample smp, %1$s.subsample subsmp, %1$s.subset sub where sta.cruise_id=cru.id and smp.station_id=sta.id and subsmp.sample_id=smp.id and sub.subsample_id=subsmp.id and sub.id=%2$s;',TG_TABLE_SCHEMA, i_subset_id) into rec_branch;
		s_branch_long = format('Cruise=%L (ingest=%s), Station=%L (ingest=%s), Sample=%L (ingest=%s), Subset=%L (ingest=%s)', rec_branch.cruise, rec_branch.cruise_ing, rec_branch.station, rec_branch.station_ing, rec_branch.sample, rec_branch.sample_ing, rec_branch.subset, rec_branch.subset_ing);
		s_branch = format('%s|%s|%s|%s|%s {%s}{%s}{%s}{%s}{%s}', rec_branch.cruise, rec_branch.station, rec_branch.sample, rec_branch.subsample, rec_branch.subset, rec_branch.cruise_ing, rec_branch.station_ing, rec_branch.sample_ing, rec_branch.subsample_ing, rec_branch.subset_ing);
	END IF;
	IF s_error is not NULL THEN
		-- Error
		RAISE EXCEPTION '%', s_error || ' -> ' || s_branch;
	END IF;
	IF s_warn is not NULL THEN
		-- Warnung
		RAISE WARNING '%', s_warn || ' -> ' || s_branch;
	END IF;
	
	--------------------------------------------------------------------------------------------
	-- Alles OK 
	--------------------------------------------------------------------------------------------
	-- Info
	IF b_debug THEN
		RAISE NOTICE '%', '* Information: new biota';
		RAISE NOTICE 'Gear category: %', rec_gear.category;
		RAISE NOTICE 'Gear name: %', rec_gear.name;
		RAISE NOTICE 'Number: %', NEW.number;
		RAISE NOTICE 'Abundance: %', NEW.abundance;
		RAISE NOTICE 'Presence/Absence: %', NEW.presence_absence;
		RAISE NOTICE 'Sample.replicates: %', rec_sample.replicates;
		-- Gear Sampled Area
		IF b_gsa=true THEN
			RAISE NOTICE '%', 'Gear Sampled Area GSA: is calcuable';
			IF f_gsa is not null then
				RAISE NOTICE 'Gear Sampled Area GSA: %', f_gsa;
			END IF;
		END IF;
		-- Sampled Area
		RAISE NOTICE 'Sampled Area SA: %', rec_sample.sampled_area;
		-- Reference Area
		RAISE NOTICE 'Reference area RA: %', rec_sample.reference_area;
		-- Data Szenario
		IF s_szenario_grab IS NOT NULL THEN
			RAISE NOTICE 'Data scenario type: %', s_szenario_grab;
		END IF;
	END IF; 	
	
	-- Antwort
	RETURN NEW;
			
END;$_$;


--
-- Name: new_gear_check_consistency(); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.new_gear_check_consistency() RETURNS trigger
    LANGUAGE plpgsql
    AS $$
declare
	--
	 
begin
	-- Test: Sampling Unit
	if NEW.smpunit not in ('area', 'vol', 'both') then
		raise exception 'Unknown sampling unit: %', NEW.smpunit;
	end if;

	-- Test: Category
	if NEW.category not in ('grab', 'trawl', 'video', 'sensor', 'trap', 'scrape', 'rocks', 'dive', 'census', 'remote sensing') then
		raise exception 'Unknown gear category: %', NEW.category;
	end if;
	
	-- Test: Grab
	if NEW.category = 'grab' then
		IF NEW.meshsize is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	end if;
	
	-- Test: Trawl
	if NEW.category = 'trawl' then
		IF NEW.area is not null then
			raise exception '%', 'Area can not be set at gear with category: trawl';
		END IF;
	end if;
	
	-- Test: Census, Remote Sensing
	 if NEW.category in ('census', 'remote sensing') then
	 	IF NEW.meshsize is not null or NEW.length is not null or NEW.width is not null or NEW.height is not null or NEW.area is not null or NEW.mass is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	 end if;
	
	-- Test: Video
	if NEW.category = 'video' then
		IF NEW.meshsize is not null or NEW.length is not null or NEW.width is not null or NEW.height is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	end if;
	
	-- Test: Sensor
	if NEW.category = 'sensor' then
		IF NEW.meshsize is not null or NEW.length is not null or NEW.width is not null or NEW.height is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	end if;
	
	-- Test: Scrape
	if NEW.category = 'scrape' then
	end if;
	
	-- Test: Rocks
	if NEW.category = 'rocks' then
	end if;
	
	-- Test: Dive
	if NEW.category = 'dive' then
		IF NEW.meshsize is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	end if;
	
	-- Test: Trap
	if NEW.category = 'trap' then  
		IF NEW.meshsize is not null or NEW.length is not null or NEW.width is not null or NEW.height is not null then
			raise exception 'Meshsize can not be set at gear with category: %', NEW.category;
		END IF;
	end if;
	
	return NEW;
end;
$$;


--
-- Name: new_sample_check_consistency(); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.new_sample_check_consistency() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$DECLARE
	arr_area_antwort text ARRAY[3]; -- Antwort des Area-Berechnbarkeitstest 
	rec_branch record; -- Vollständiger Cruise-Station-Sample-Ast des Tests 
	s_branch text; 
BEGIN
	-- Teste Area
	EXECUTE format('select %1$s."is_sample_ok"(%1$L, %2$L,%3$L)', TG_TABLE_SCHEMA, NEW.id, false) into arr_area_antwort;
	
	-- *** Alles lief gut 
	-- Area ist konsistent? (konsistent bestimmbar/nicht bestimmtt
	IF arr_area_antwort[0] is NULL THEN
		RETURN NEW;
	END IF;
	
	-- ** Es gab einen Fehler
	-- Bestimme Ast für Meldungen
	EXECUTE format('select cru.name cruise, cru.ingest_id cruise_ing, sta.name station, sta.ingest_id station_ing, smp.name sample, smp.ingest_id sample_ing from %1$s.cruise cru, %1$s.station sta, %1$s.sample smp where sta.cruise_id=cru.id and smp.station_id=sta.id and smp.id=%2$s;', TG_TABLE_SCHEMA, NEW.id) into rec_branch;
	s_branch = format('Cruise=%L (ingest=%s), Station=%L (ingest=%s), Sample=%L (ingest=%s)', rec_branch.cruise, rec_branch.cruise_ing, rec_branch.station, rec_branch.station_ing, rec_branch.sample, rec_branch.sample_ing);
	RAISE EXCEPTION 'Error while checking area: %', s_branch || ' - ' || arr_area_antwort[0];
	
END;$_$;


--
-- Name: new_subsample_check_consistency(); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.new_subsample_check_consistency() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$DECLARE
	
	-- Basis
	rec_branch record; -- Vollständiger Cruise-Station-Sample-Subset-Ast des Tests 
	f_ratio_sum float;
	i_ratio_null integer;
	
	-- Answer
	s_error text; -- Fehlertext
	s_warn text; -- Warnungstext
	s_branch text;
	
BEGIN
	
	--------------------------------------------------------------------------------------------
	-- Test: Summe aller Ratios aller SubSamples pro Sample <= 1
	--------------------------------------------------------------------------------------------
	
	f_ratio_sum = 0;
	i_ratio_null = 0;
	
	-- Hole Summe aller Ratios aller SubSamples pro Sample
	EXECUTE format('select sum(ratio) from %1$s.subsample subsmp where subsmp.ratio is not NULL and subsmp.sample_id=%2$s',TG_TABLE_SCHEMA, NEW.sample_id) into f_ratio_sum;
	
	-- Sum Ratio > 1 -> Warnung	   
	IF f_ratio_sum > 1 THEN
		s_warn = format('Sum of all SubSample ratios per Sample >1: %1$s', f_ratio_sum);		
	END IF;
	
	-- Sum Ratio > 1 und es sind noch Ratios ohne Wert vorhanden -> Warnung
	IF f_ratio_sum = 1 THEN
		EXECUTE format('select count(id) from %1$s.subsample subsmp where subsmp.ratio is NULL and subsmp.sample_id=%2$s',TG_TABLE_SCHEMA, NEW.sample_id) into i_ratio_null;
		IF i_ratio_null > 0 THEN
			s_warn = format('Sum of all SubSample ratios per Sample = 1 and SubSamples with no ratio value are given: %1$sx', i_ratio_null);		
		END IF;
	END IF;
	
	--------------------------------------------------------------------------------------------
	-- Fehler/Warnung
	--------------------------------------------------------------------------------------------
	IF s_error is not NULL or s_warn is not NULL THEN
		-- Bestimme Ast für Meldungen
		EXECUTE format('select cru.name cruise, cru.ingest_id cruise_ing, sta.name station, sta.ingest_id station_ing, smp.name sample, smp.ingest_id sample_ing, subsmp.name subsample, subsmp.ingest_id subsample_ing from %1$s.cruise cru, %1$s.station sta, %1$s.sample smp, %1$s.subsample subsmp where sta.cruise_id=cru.id and smp.station_id=sta.id and subsmp.sample_id=smp.id and subsmp.id=%2$s;',TG_TABLE_SCHEMA, NEW.id) into rec_branch;
		s_branch = format('%s|%s|%s|%s {%s}{%s}{%s}{%s}', rec_branch.cruise, rec_branch.station, rec_branch.sample, rec_branch.subsample, rec_branch.cruise_ing, rec_branch.station_ing, rec_branch.sample_ing, rec_branch.subsample_ing);
	END IF;
	IF s_error is not NULL THEN
		-- Error
		RAISE EXCEPTION '%', s_error || ' -> ' || s_branch;
	END IF;
	IF s_warn is not NULL THEN
		-- Warnung
		RAISE WARNING '%', s_warn || ' -> ' || s_branch;
	END IF;
	
	-- Antwort
	RETURN NEW;
			
END;$_$;


--
-- Name: new_subset_check_consistency(); Type: FUNCTION; Schema: cb_empty; Owner: -
--

CREATE FUNCTION cb_empty.new_subset_check_consistency() RETURNS trigger
    LANGUAGE plpgsql
    AS $_$DECLARE
	-- Vollständiger Cruise-Station-Sample-SubSample-Subset-Ast des Tests
	rec_branch record; -- als record
	s_branch text; -- als string
	-- Triplett
	s_taxcover text;
	b_is_triplet_uniq boolean;
	-- Error Typen
	b_error boolean; 
	b_error_triplet boolean; -- Triplet tax_cover/tax_res/spatial_level nicht korrekt gesetzt
	b_error_triplet_uniq boolean;
	
BEGIN
	-- *** Tests
	b_error=False;
	b_error_triplet=False;
	b_error_triplet_uniq=True;

	-- Test Triplet
	-- Entweder Taxonomic Coverage, Taxonomic Resolution und Spatial Level angegeben oder nichts
	IF not (NEW.taxcover_id is not Null and NEW.taxres is not Null and NEW.spatiallevel is not Null) and
	   not (NEW.taxcover_id is Null and NEW.taxres is Null and NEW.spatiallevel is Null) THEN
		b_error=True;
		b_error_triplet=True;
	END IF;
	
	--- Test: Einmaligkeit Triplet pro Sample
	IF (b_error is not true and NEW.subsample_id is not null and NEW.taxcover_id is not Null and NEW.taxres is not Null and NEW.spatiallevel is not Null) THEN
		EXECUTE format('select %1$s.is_subset_triplet_uniq(%1$L, %2$s, %3$s, %4$L, %5$s)', TG_TABLE_SCHEMA, NEW.subsample_id, NEW.taxcover_id, NEW.taxres, NEW.spatiallevel) into b_is_triplet_uniq;
		IF b_is_triplet_uniq is false THEN
			b_error=True;
			b_error_triplet_uniq = True;
		END IF;
	END IF;

	-- *** Alles OK -> Zurück
	IF b_error is False THEN RETURN NEW; END IF;

	-- *** Es lief was schief
	-- Meldungen sind notwendig
	
	-- Bestimme Ast für Meldungen
	EXECUTE format('select cru.name cruise, sta.name station, smp.name sample, subsmp.name subsample, sub.name subset from %1$s.cruise cru, %1$s.station sta, %1$s.sample smp, %1$s.subsample subsmp, %1$s.subset sub where subsmp.id=%2$s and subsmp.id=sub.subsample_id and smp.id=subsmp.sample_id and sta.id=smp.station_id and cru.id=sta.cruise_id;',TG_TABLE_SCHEMA, NEW.subsample_id) into rec_branch;
	s_branch = format('%s|%s|%s|%s|%s', rec_branch.cruise, rec_branch.station, rec_branch.sample, rec_branch.subsample, rec_branch.subset, NEW.name);

	-- Fehlermeldungen
	IF b_error_triplet is True THEN
		-- Triplet-Fehler
		RAISE EXCEPTION 'Subset: taxonomic coverage, taxonomic resolution, spatial_level: Either all fields must be set or none of them: %', s_branch;
	ELSIF b_error_triplet_uniq is True THEN
		-- Triplet ist nicht uniq pro subsample
		RAISE EXCEPTION 'Subset: taxonomic coverage, taxonomic resolution, spatial_level: Not uniq: %', s_branch;
	END IF;
END;$_$;


SET default_tablespace = '';

SET default_table_access_method = heap;

--
-- Name: actionlog; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.actionlog (
    id integer NOT NULL,
    username text NOT NULL,
    action text NOT NULL,
    object text,
    date timestamp without time zone NOT NULL,
    host text NOT NULL,
    os text NOT NULL,
    log text,
    ip text NOT NULL
);


--
-- Name: actionlog_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.actionlog_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: actionlog_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.actionlog_id_seq OWNED BY cb_empty.actionlog.id;


--
-- Name: admin; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.admin (
    key text NOT NULL,
    value text NOT NULL,
    id integer NOT NULL
);


--
-- Name: COLUMN admin.key; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.admin.key IS '{"name": "Key"}';


--
-- Name: COLUMN admin.value; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.admin.value IS '{"name": "Value"}';


--
-- Name: admin_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.admin_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: admin_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.admin_id_seq OWNED BY cb_empty.admin.id;


--
-- Name: biota; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.biota (
    id integer NOT NULL,
    subset_id integer NOT NULL,
    taxon_id integer NOT NULL,
    lifestage_id integer,
    ingest_id integer NOT NULL,
    name text NOT NULL,
    number numeric,
    given_taxon_name text NOT NULL,
    biomass_wet double precision,
    biomass_dry double precision,
    biomass_afdm double precision,
    subtotal_factor text,
    remark text,
    specification text,
    abundance numeric,
    presence_absence integer,
    sieve integer,
    image text,
    CONSTRAINT "abundance has to be: >=0" CHECK ((abundance >= (0)::numeric)),
    CONSTRAINT "number has to be: >=0" CHECK ((number >= (0)::numeric)),
    CONSTRAINT "presence_absence has to be: 0/1" CHECK ((presence_absence = ANY (ARRAY[0, 1])))
);


--
-- Name: COLUMN biota.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.id IS '{"name":"Biota"}';


--
-- Name: COLUMN biota.subset_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.subset_id IS '{"name":"Sample"}';


--
-- Name: COLUMN biota.taxon_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.taxon_id IS '{"name":"Taxon", "comment": "Value must be defined in lookup table: Taxon.AphiaId"}';


--
-- Name: COLUMN biota.lifestage_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.lifestage_id IS '{"name":"Lifestage", "comment":"Value must be defined in lookup table: Lifestage.Name"}';


--
-- Name: COLUMN biota.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN biota.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.name IS '{"name":"Biota name"}';


--
-- Name: COLUMN biota.number; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.number IS '{"name":"Number", "unit":"#","comment":"Total number of individuals per taxon in the sample, number refers to sampled area of sampling gear"}';


--
-- Name: COLUMN biota.given_taxon_name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.given_taxon_name IS '{"name":"Scientific name", "comment":"Original scientific name from taxonomic identification in the lab"}';


--
-- Name: COLUMN biota.biomass_wet; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.biomass_wet IS '{"name":"Biomass wet", "unit":"g", "comment":"Total wetmass of individuals per taxon in the sample per sampled area"}';


--
-- Name: COLUMN biota.biomass_dry; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.biomass_dry IS '{"name":"Biomass dry", "unit":"g", "comment":"Total drymass of inviduals per taxon in the sample per sampled area"}';


--
-- Name: COLUMN biota.biomass_afdm; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.biomass_afdm IS '{"name":"Biomass afdm", "unit":"g", "comment":"Total ash free dry mass of individuals per taxon in the sample per sampled area"}';


--
-- Name: COLUMN biota.subtotal_factor; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.subtotal_factor IS '{"name":"Subtotal factor", "comment":"Factor for subsample of taxon number or taxon weight in sample, i.e. Subtotal factor = Samplesize kg/subsample size kg, i.e number and weight must be multiplied by subtotal factor"}';


--
-- Name: COLUMN biota.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.remark IS '{"name":"Remark"}';


--
-- Name: COLUMN biota.specification; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.specification IS '{"name":"Specification", "comment":"additional specification of scientific name, e.g. sp., cf., aff., indet. etc."}';


--
-- Name: COLUMN biota.abundance; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.abundance IS '{"name":"Abundance", "comment":"Abundance: Total number of individuals per taxon per reference area", "unit":"#/m**2"}';


--
-- Name: COLUMN biota.presence_absence; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.presence_absence IS '{"name" : "Presence/Absence", "comment":"Presence/Absence", "unit":"1/0"}';


--
-- Name: COLUMN biota.sieve; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.sieve IS '{"name":"Mesh size", "unit":"µm", "comment":"Sieves used to strip organismes from sediment for population analysis"}';


--
-- Name: COLUMN biota.image; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.biota.image IS '{"name":"Image", "comment":"URI to image file of taxon"}';


--
-- Name: biota_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.biota_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: biota_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.biota_id_seq OWNED BY cb_empty.biota.id;


--
-- Name: country; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.country (
    id integer NOT NULL,
    name text NOT NULL,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN country.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.country.id IS '{"name": "Country"}';


--
-- Name: COLUMN country.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.country.name IS '{"name": "Country name", "comment":"Re-enter exact text/name in corresponding tables: Dataset.OwnerCountry"}';


--
-- Name: COLUMN country.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.country.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: country_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.country_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: country_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.country_id_seq OWNED BY cb_empty.country.id;


--
-- Name: crs; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.crs (
    id integer NOT NULL,
    name text DEFAULT 'WGS84'::character varying NOT NULL,
    srid integer,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN crs.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.crs.id IS '{"name": "CRS"}';


--
-- Name: COLUMN crs.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.crs.name IS '{"name": "Coordinate reference system name", "comment": "Re-enter exact text/name in corresponding table: Cruise.RefSystem"}';


--
-- Name: COLUMN crs.srid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.crs.srid IS '{"name": "SRID", "comment": "Integer value for unique identification of Spatial Reference IDentifier"}';


--
-- Name: COLUMN crs.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.crs.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: crs_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.crs_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: crs_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.crs_id_seq OWNED BY cb_empty.crs.id;


--
-- Name: cruise; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.cruise (
    id integer NOT NULL,
    name text NOT NULL,
    ingest_id integer NOT NULL,
    crs_id integer NOT NULL,
    ship_id integer,
    positioningsystem_id integer,
    lead_person_id integer,
    date_start date,
    date_end date,
    region text,
    remark text
);


--
-- Name: COLUMN cruise.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.id IS '{"name": "Cruise"}';


--
-- Name: COLUMN cruise.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.name IS '{"name": "Cruise name", "pangaea":"134113"}';


--
-- Name: COLUMN cruise.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.ingest_id IS '{"name": "Ingest"}';


--
-- Name: COLUMN cruise.crs_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.crs_id IS '{"name": "CRS", "comment":"Value must be defined in lookup table: CRS.Name"}';


--
-- Name: COLUMN cruise.ship_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.ship_id IS '{"name": "Ship", "comment":"Value must be defined in lookup table: Ship.Name"}';


--
-- Name: COLUMN cruise.positioningsystem_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.positioningsystem_id IS '{"name": "Positioning system", "comment":"Value must be defined in lookup table: PosSys.Name"}';


--
-- Name: COLUMN cruise.lead_person_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.lead_person_id IS '{"name": "Lead person", "comment":"Value must be defined in lookup table: Person.Name"}';


--
-- Name: COLUMN cruise.date_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.date_start IS '{"name": "Date start", "unit": "YYYY-MM-DD","comment": "Cruise start date","pangaea":"152481"}';


--
-- Name: COLUMN cruise.date_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.date_end IS '{"name": "Date end", "unit": "YYYY-MM-DD","comment": "Cruise end date","pangaea":"152481"}';


--
-- Name: COLUMN cruise.region; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.region IS '{"name": "Region", "comment": "Sea area description, not defined e.g. North Sea, German Bight or Doggerbank"}';


--
-- Name: COLUMN cruise.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.cruise.remark IS '{"name":"Remark"}';


--
-- Name: cruise_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.cruise_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: cruise_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.cruise_id_seq OWNED BY cb_empty.cruise.id;


--
-- Name: dataset; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.dataset (
    id integer NOT NULL,
    name text NOT NULL,
    institution text NOT NULL,
    client text,
    reference_person text NOT NULL,
    owner_country_id integer NOT NULL,
    contact_person_id integer NOT NULL,
    realm text,
    doi text,
    description text,
    collection character(1) NOT NULL,
    access_level_data text DEFAULT 'confidential'::text NOT NULL,
    access_level_data_product text DEFAULT 'confidential'::text NOT NULL,
    gold boolean DEFAULT false NOT NULL,
    CONSTRAINT "access_level_data has to be: confidential/free/request" CHECK (((access_level_data = 'confidential'::text) OR (access_level_data = 'free'::text) OR (access_level_data = 'request'::text))),
    CONSTRAINT "access_level_data_product has to be: confidential/free/request" CHECK ((access_level_data_product = ANY (ARRAY['confidential'::text, 'free'::text, 'request'::text]))),
    CONSTRAINT "collection has to be: M/R" CHECK (((collection = 'M'::character(1)) OR (collection = 'R'::character(1))))
);


--
-- Name: COLUMN dataset.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.id IS '{"name":"Dataset"}';


--
-- Name: COLUMN dataset.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.name IS '{"name":"Dataset name", "comment": "Unique name, e.g. combination of project name and specific cruise. Re-enter exact text/name in corresponding table: Subset.Project", "pangaea":"132665"}';


--
-- Name: COLUMN dataset.institution; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.institution IS '{"name":"Institution", "comment": "Dataset origin; Name of project-responsible institution"}';


--
-- Name: COLUMN dataset.client; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.client IS '{"name":"Client", "comment":"Provider of funds"}';


--
-- Name: COLUMN dataset.reference_person; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.reference_person IS '{"name":"Reference person", "comment":"Contact person for dataset"}';


--
-- Name: COLUMN dataset.owner_country_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.owner_country_id IS '{"name":"Owner country", "comment":"Value must be defined in lookup table: Country.Name"}';


--
-- Name: COLUMN dataset.contact_person_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.contact_person_id IS '{"name":"Contact person", "comment":"Value must be defined in lookup table: Person.Name"}';


--
-- Name: COLUMN dataset.realm; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.realm IS '{"name":"Realm", "comment":"Geographic realm, not defined e.g. North Sea, Arctic, Antarctic"}';


--
-- Name: COLUMN dataset.doi; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.doi IS '{"name":"DOI", "comment":"Digital Object Identifier"}';


--
-- Name: COLUMN dataset.description; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.description IS '{"name":"Description"}';


--
-- Name: COLUMN dataset.collection; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.collection IS '{"name":"Collection", "comment":"Long-term monitoring = M; Research-based project = R"}';


--
-- Name: COLUMN dataset.access_level_data; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.access_level_data IS '{"name":"Data accessibility", "comment": "free (free access), request (on request by contact) or confidential (confidential)"}';


--
-- Name: COLUMN dataset.access_level_data_product; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.access_level_data_product IS '{"name":"Data product accessibility",  "comment": "free (free access), request (on request by contact) or confidential (confidential)"}';


--
-- Name: COLUMN dataset.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.dataset.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: dataset_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.dataset_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: dataset_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.dataset_id_seq OWNED BY cb_empty.dataset.id;


--
-- Name: gear; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.gear (
    id integer NOT NULL,
    name text NOT NULL,
    type text,
    category text NOT NULL,
    width double precision,
    height double precision,
    length double precision,
    smpunit text NOT NULL,
    area double precision,
    mass double precision,
    meshsize double precision,
    description text,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN gear.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.id IS '{"name":"Id"}';


--
-- Name: COLUMN gear.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.name IS '{"name": "Gear name", "comment": "Gears name, e.g. Dredge2.0m. Re-enter exact text/name in corresponding table: Sample.Gear.","pangaea":"55539"}';


--
-- Name: COLUMN gear.type; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.type IS '{"name": "Gear type", "comment": "Gears type, e.g. VV"}';


--
-- Name: COLUMN gear.category; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.category IS '{"name": "Gear category", "unit":"grab,trawl scrape,trap,sensor,video", "comment": "Allowed gear categories"}';


--
-- Name: COLUMN gear.width; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.width IS '{"name": "WIdth of gear", "comment": "WIdth of gear", "unit": "m","pangaea":"50144"}';


--
-- Name: COLUMN gear.height; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.height IS '{"name": "Gear heigth", "comment": "Height of gear", "unit": "m","pangaea":"92744"}';


--
-- Name: COLUMN gear.length; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.length IS '{"name": "Gear length", "comment": "Length of gear", "unit": "m","pangaea":"45226"}';


--
-- Name: COLUMN gear.smpunit; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.smpunit IS '{"name": "Sampling unit", "comment": "Standard information of sampling unit (area and/or volume)", "unit": "area,vol,both"}';


--
-- Name: COLUMN gear.area; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.area IS '{"name": "Area of gear", "comment": "Area of gear", "unit": "m**2"}';


--
-- Name: COLUMN gear.mass; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.mass IS '{"name": "Gear mass", "comment": "Mass of gear", "unit": "kg"}';


--
-- Name: COLUMN gear.meshsize; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.meshsize IS '{"name": "Mesh size", "comment":  "Mesh size", "unit": "mm","pangaea":"74895"}';


--
-- Name: COLUMN gear.description; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.description IS '{"name": "Description", "comment": "Description of gear"}';


--
-- Name: COLUMN gear.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.gear.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: gear_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.gear_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: gear_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.gear_id_seq OWNED BY cb_empty.gear.id;


--
-- Name: indmeasure; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.indmeasure (
    id integer NOT NULL,
    biota_id integer NOT NULL,
    ingest_id integer NOT NULL,
    sex text,
    length_total double precision,
    length_st double precision,
    mass_total double precision,
    mass_sh double precision,
    remark text
);


--
-- Name: COLUMN indmeasure.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.id IS '{"name": "Individual measure"}';


--
-- Name: COLUMN indmeasure.biota_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.biota_id IS '{"name": "Biota"}';


--
-- Name: COLUMN indmeasure.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.ingest_id IS '{"name": "Ingest Id"}';


--
-- Name: COLUMN indmeasure.sex; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.sex IS '{"name": "Sex", "unit": "m,f","comment": "Sex of individual"}';


--
-- Name: COLUMN indmeasure.length_total; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.length_total IS '{"name": "Total Length", "unit": "cm","comment": "Total length of animal. Longest extension of animal: fish = body length, crab = carapax width, shrimp = body length"}';


--
-- Name: COLUMN indmeasure.length_st; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.length_st IS '{"name": "Standard length", "unit": "cm","comment": "Standard length of fish: from head to beginning of tail"}';


--
-- Name: COLUMN indmeasure.mass_total; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.mass_total IS '{"name": "Total mass", "unit": "g","comment": "Total mass of animal"}';


--
-- Name: COLUMN indmeasure.mass_sh; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.mass_sh IS '{"name": "Slaughtery mass", "unit": "g","comment": "Slaughtery mass of fish, i.e. without intestines"}';


--
-- Name: COLUMN indmeasure.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.indmeasure.remark IS '{"name": "Remark"}';


--
-- Name: indmeasure_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.indmeasure_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: indmeasure_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.indmeasure_id_seq OWNED BY cb_empty.indmeasure.id;


--
-- Name: ingest; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.ingest (
    id integer NOT NULL,
    name text DEFAULT 'anonymous'::character varying NOT NULL,
    created_on timestamp without time zone DEFAULT now() NOT NULL,
    workflow character varying NOT NULL,
    log character varying,
    rows integer,
    warn integer,
    info integer,
    file text
);


--
-- Name: COLUMN ingest.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.id IS '{"name":"Ingest"}';


--
-- Name: COLUMN ingest.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.name IS '{"name":"Ingest name"}';


--
-- Name: COLUMN ingest.created_on; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.created_on IS '{"name":"Ingest timestamp"}';


--
-- Name: COLUMN ingest.workflow; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.workflow IS '{"name":"Workflow", "comment":"Ingest workflow e.g. North Sea workflow, Arctic workflow"}';


--
-- Name: COLUMN ingest.log; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.log IS '{"name":"Log", "comment":"Log messages of ingest"}';


--
-- Name: COLUMN ingest.rows; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.rows IS '{"name":"Rows", "comment":"Ingested rows"}';


--
-- Name: COLUMN ingest.warn; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.warn IS '{"name":"Warnings", "comment":"Occurred warnings during ingestion process"}';


--
-- Name: COLUMN ingest.info; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.info IS '{"name":"Information", "comment":"Occurred information during ingestion process"}';


--
-- Name: COLUMN ingest.file; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ingest.file IS '{"name":"Filename", "comment":"Filename of ingestion data"}';


--
-- Name: ingest_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.ingest_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: ingest_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.ingest_id_seq OWNED BY cb_empty.ingest.id;


--
-- Name: lifestage; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.lifestage (
    id integer NOT NULL,
    name text NOT NULL,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN lifestage.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.lifestage.id IS '{"name":"Life stage"}';


--
-- Name: COLUMN lifestage.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.lifestage.name IS '{"name":"Life stage name", "comment": "Re-enter exact text/name in corresponding table: Biota.Stage"}';


--
-- Name: COLUMN lifestage.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.lifestage.gold IS '{"name":"Gold standard lookup entry", "comment":"This is a quality checked gold standard lookup entry."}';


--
-- Name: lifestage_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.lifestage_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: lifestage_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.lifestage_id_seq OWNED BY cb_empty.lifestage.id;


--
-- Name: noworms; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.noworms (
    id integer NOT NULL,
    scientificname text NOT NULL,
    target text
);


--
-- Name: COLUMN noworms.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.noworms.id IS '{"name":"Scientific names of taxa not found in WoRMS", comment="Mapping scientific names other services than WoRMS (expert knowledge)"}';


--
-- Name: COLUMN noworms.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.noworms.scientificname IS '{"name":"Scientific name"}';


--
-- Name: COLUMN noworms.target; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.noworms.target IS '{"name":"Target"}';


--
-- Name: noworms_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.noworms_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: noworms_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.noworms_id_seq OWNED BY cb_empty.noworms.id;


--
-- Name: person; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.person (
    id integer NOT NULL,
    name text NOT NULL,
    full_name text,
    affiliation text,
    email text,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN person.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.id IS '{"name":"Person"}';


--
-- Name: COLUMN person.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.name IS '{"name":"Short name", "comment":"Short name in login style. Re-enter exact text/name in corresponding tables: Dataset.ContactPerson, Cruise.SciCruiseLeader, Station.ProjectLeader, Sample.RespScientist"}';


--
-- Name: COLUMN person.full_name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.full_name IS '{"name":"Given name and surname"}';


--
-- Name: COLUMN person.affiliation; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.affiliation IS '{"name":"Affiliation", "comment":"Person affiliation"}';


--
-- Name: COLUMN person.email; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.email IS '{"name":"Email address"}';


--
-- Name: COLUMN person.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.person.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: person_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.person_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: person_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.person_id_seq OWNED BY cb_empty.person.id;


--
-- Name: possys; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.possys (
    id integer NOT NULL,
    name text NOT NULL,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN possys.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.possys.id IS '{"name":"Positioning system"}';


--
-- Name: COLUMN possys.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.possys.name IS '{"name":"Positioning system", "comment":"Re-enter exact text/name in corresponding table: Cruise.PositionSystem"}';


--
-- Name: COLUMN possys.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.possys.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: positioningsystem_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.positioningsystem_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: positioningsystem_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.positioningsystem_id_seq OWNED BY cb_empty.possys.id;


--
-- Name: region; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.region (
    id integer NOT NULL,
    code character varying,
    name character varying,
    level character varying,
    geometry public.geometry
);


--
-- Name: COLUMN region.code; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.region.code IS '{"name":"FAO code"}';


--
-- Name: COLUMN region.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.region.name IS '{"name":"FAO region name"}';


--
-- Name: COLUMN region.level; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.region.level IS '{"name":"FAO region level"}';


--
-- Name: COLUMN region.geometry; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.region.geometry IS '{"name":"FAO region shape"}';


--
-- Name: region2sample; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.region2sample (
    sample_id integer NOT NULL,
    region_id integer NOT NULL
);


--
-- Name: region_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.region_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: region_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.region_id_seq OWNED BY cb_empty.region.id;


--
-- Name: sample; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.sample (
    id integer NOT NULL,
    name text NOT NULL,
    station_id integer NOT NULL,
    gear_id integer NOT NULL,
    ingest_id integer NOT NULL,
    responsible_person_id integer,
    sampled_area double precision,
    lon_start double precision NOT NULL,
    lat_start double precision NOT NULL,
    date_start date NOT NULL,
    time_start time without time zone,
    date_start_error integer DEFAULT 0,
    depth_start double precision,
    lon_end double precision,
    lat_end double precision,
    date_end date,
    time_end time without time zone,
    depth_end double precision,
    sample_mass real DEFAULT 0,
    subsample_mass real,
    subsample_share real,
    sampling_distance double precision,
    method_biomass text,
    method_biomass_dry text,
    method_biomass_afdm text,
    method_conservation text,
    temperature_surfacewater_start double precision,
    temperature_surfacewater_end double precision,
    salinity_surfacewater_start double precision,
    salinity_surfacewater_end double precision,
    remark text,
    reference_area double precision,
    replicates integer DEFAULT 1,
    sampling_speed double precision,
    sediment_properties text,
    sediment_smell text,
    grab_penetration_depth double precision,
    sampled_vol double precision,
    reference_vol double precision,
    "precision" integer DEFAULT 5 NOT NULL,
    CONSTRAINT "sample reference_area has to be: >0" CHECK (((reference_area <> NULL::double precision) AND (reference_area > (0)::double precision))),
    CONSTRAINT "sample replicates has to be: >0" CHECK ((replicates > 0)),
    CONSTRAINT "sample sampled_area hs to be: >0" CHECK (((sampled_area <> NULL::double precision) AND (sampled_area > (0)::double precision)))
);


--
-- Name: COLUMN sample.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.id IS '{"name":"Sample"}';


--
-- Name: COLUMN sample.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.name IS '{"name":"Sample name","pangaea":"16065"}';


--
-- Name: COLUMN sample.station_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.station_id IS '{"name":"Station"}';


--
-- Name: COLUMN sample.gear_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.gear_id IS '{"name":"Gear", "comment":"Value must be defined in lookup table: Gear.Name"}';


--
-- Name: COLUMN sample.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN sample.responsible_person_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.responsible_person_id IS '{"name":"Responsible person", "comment":"Value must be defined in lookup table: Person,Name"}';


--
-- Name: COLUMN sample.sampled_area; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sampled_area IS '{"name":"Sampled area", "unit":"m**2", "comment":"Actually sampled area","pangaea":"55219"}';


--
-- Name: COLUMN sample.lon_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.lon_start IS '{"name": "Longitude start", "comment": "Logged longitude on starting the sampling  - trawl: net on ground", "unit": "dD", "pangaea":"1601"}';


--
-- Name: COLUMN sample.lat_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.lat_start IS '{"name": "Latitude start", "comment": "Logged latitude on starting the sampling - trawl: net on ground", "unit": "dD","pangaea":"1600"}';


--
-- Name: COLUMN sample.date_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.date_start IS '{"name":"Start date", "unit":"YYYY-MM-DD", "comment":"Date of sampling start"}';


--
-- Name: COLUMN sample.time_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.time_start IS '{"name":"Start time", "unit":"HH:MM:SS", "comment":"Time of sampling start - trawl: net on ground","pangaea":"152460"}';


--
-- Name: COLUMN sample.date_start_error; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.date_start_error IS '{"name":"Start date error", "comment":"Refers to the number of days from date deviation, e.g. if only year is given for sample: start date=2017-06-31 and start date error=132 (days), if only year and month is given: start date=2017-04-15 and start date_error=15 (days)"}';


--
-- Name: COLUMN sample.depth_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.depth_start IS '{"name":"Start depth", "unit":"m", "comment":"Depth of sampling start","pangaea":"1619"}';


--
-- Name: COLUMN sample.lon_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.lon_end IS '{"name":"End lon", "unit":"dD", "comment":"Logged longitude of sampling end  - trawl: net off ground","pangaea":"1601"}';


--
-- Name: COLUMN sample.lat_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.lat_end IS '{"name":"End lat", "unit":"dD", "comment":"Logged latitude of sampling end  - trawl: net off ground","pangaea":"1600"}';


--
-- Name: COLUMN sample.date_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.date_end IS '{"name":"End date", "unit":"YYYY-MM-DD", "comment":"Date of sampling end"}';


--
-- Name: COLUMN sample.time_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.time_end IS '{"name":"End time", "unit":"HH:MM:SS", "comment":"Time of sampling end - trawl: net off ground" ,"pangaea":"150986"}';


--
-- Name: COLUMN sample.depth_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.depth_end IS '{"name":"End depth", "unit":"m", "comment":"Depth of sampling end","pangaea":"1619"}';


--
-- Name: COLUMN sample.sample_mass; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sample_mass IS '{"name":"Sample mass", "unit":"kg", "comment":"Total mass of haul/sample, mainly relevant for trawled/net-towed gears"}';


--
-- Name: COLUMN sample.subsample_mass; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.subsample_mass IS '{"name":"Subsample mass", "unit":"kg", "comment":"Total mass of sub-haul/sub-sample of total catch, mainly relevant for trawled/net-towed gears"}';


--
-- Name: COLUMN sample.subsample_share; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.subsample_share IS '{"name":"Subtotal factor", "unit":"kg/kg", "comment":"Factor for extrapolating subsample to total catch size, factor = sample weight/subsample weight"}';


--
-- Name: COLUMN sample.sampling_distance; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sampling_distance IS '{"name":"Sampling distance", "unit":"m", "comment":"Total trawled distance of sample, only relevant for trawled/net-towed gears","pangaea":"46450"}';


--
-- Name: COLUMN sample.method_biomass; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.method_biomass IS '{"name":"Method biomass", "comment":"Wetmass determined with or without shell for molluscs, categories = with shell, without shell"}';


--
-- Name: COLUMN sample.method_biomass_dry; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.method_biomass_dry IS '{"name":"Method DM", "comment":"Method to obtain drymass either by measuring or by converting with species factors, categories = measured, converted"}';


--
-- Name: COLUMN sample.method_biomass_afdm; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.method_biomass_afdm IS '{"name":"Method AFDM", "comment":"Method to obtain ash free drymass either by measuring or by converting with species factors, categories = measured, converted"}';


--
-- Name: COLUMN sample.method_conservation; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.method_conservation IS '{"name":"Method conservation", "comment":"Conservation method for sample, Alk = 70% Alkohol, For = Formaldehyde, Frz = Frozen"}';


--
-- Name: COLUMN sample.temperature_surfacewater_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.temperature_surfacewater_start IS '{"name":"Temperature surfacewater start", "unit":"°C", "comment":"Surface-water temperature of sampling start"}';


--
-- Name: COLUMN sample.temperature_surfacewater_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.temperature_surfacewater_end IS '{"name":"Temperature surfacewater end", "unit":"°C", "comment":"Surface-water temperature of sampling end"}';


--
-- Name: COLUMN sample.salinity_surfacewater_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.salinity_surfacewater_start IS '{"name":"Salinity surfacewater start", "unit":"PSU", "comment":"Surface-water salinity of sampling start"}';


--
-- Name: COLUMN sample.salinity_surfacewater_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.salinity_surfacewater_end IS '{"name":"Salinity surfacewater end", "unit":"PSU", "comment":"Surface-water salinity of sampling end"}';


--
-- Name: COLUMN sample.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.remark IS '{"name":"Remark"}';


--
-- Name: COLUMN sample.reference_area; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.reference_area IS '{"name":"Reference area", "unit":"m**2", "comment":"Fictional/defined area"}';


--
-- Name: COLUMN sample.replicates; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.replicates IS '{"name":"Replicates", "comment":"Number of replicates per sample, not per station."}';


--
-- Name: COLUMN sample.sampling_speed; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sampling_speed IS '{"name":"Sampling speed", "unit":"knots", "comment":"Mean speed of trawling - only usefull for trawling gear", "pangaea":"32561"}';


--
-- Name: COLUMN sample.sediment_properties; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sediment_properties IS '{"name":"Sediment properties", "comment":"General sediment description of texture, appearance and other characteristics, e.g. silty fine sand, strong bioturbation marks etc"}';


--
-- Name: COLUMN sample.sediment_smell; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sediment_smell IS '{"name":"Sediment smell", "comment":"Description of sediment smell, e.g. smell of sulfur"}';


--
-- Name: COLUMN sample.grab_penetration_depth; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.grab_penetration_depth IS '{"name":"Grab penetration depth", "unit":"cm", "comment":"Grab penetration depth into the sediment"}';


--
-- Name: COLUMN sample.sampled_vol; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.sampled_vol IS '{"name":"Sampled volume", "unit":"m**3", "comment":"Sampled volume, refers to sampling volume of the gear - i.e. trawl or grab"}';


--
-- Name: COLUMN sample.reference_vol; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample.reference_vol IS '{"name":"Reference volume", "unit":"m**3", "comment":"Reference volume, could be calculated."}';


--
-- Name: COLUMN sample."precision"; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sample."precision" IS '{"name":"Spatial precision", "comment":"Spatial precision as number of decimal digits for geographic position: lat, lon - range: 1-5 - default when field is left empty is 5"}';


--
-- Name: sample_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.sample_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: sample_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.sample_id_seq OWNED BY cb_empty.sample.id;


--
-- Name: sediment; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.sediment (
    id integer NOT NULL,
    subset_id integer NOT NULL,
    ingest_id integer NOT NULL,
    residue1 double precision,
    meshsize1 double precision,
    residue2 double precision,
    meshsize2 double precision,
    residue3 double precision,
    meshsize3 double precision,
    residue4 double precision,
    meshsize4 double precision,
    residue5 double precision,
    meshsize5 double precision,
    residue6 double precision,
    meshsize6 double precision,
    residue7 double precision,
    meshsize7 double precision,
    residue8 double precision,
    meshsize8 double precision,
    residue9 double precision,
    meshsize9 double precision,
    median double precision,
    gsd double precision,
    weight double precision,
    method_grainsize text,
    temperature double precision,
    c_org double precision,
    c_total double precision,
    loss_on_ignition double precision,
    nitrogen double precision,
    sulfur double precision,
    oxygen_pd double precision,
    remark text,
    fuz_silt integer,
    fuz_vfsand integer,
    fuz_fsand integer,
    fuz_msand integer,
    fuz_csand integer,
    fuz_vcsand integer,
    fuz_gravel integer,
    fuz_stones integer,
    fuz_shell integer,
    fuz_clay integer,
    type text
);


--
-- Name: COLUMN sediment.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.id IS '{"name":"Sediment"}';


--
-- Name: COLUMN sediment.subset_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.subset_id IS '{"name":"Subset"}';


--
-- Name: COLUMN sediment.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN sediment.residue1; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue1 IS '{"name":"Residue 1", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize1; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize1 IS '{"name":"Mesh size 1", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue2; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue2 IS '{"name":"Residue 2", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize2; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize2 IS '{"name":"Mesh size 2", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue3; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue3 IS '{"name":"Residue 3", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize3; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize3 IS '{"name":"Mesh size 3", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue4; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue4 IS '{"name":"Residue 4", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize4; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize4 IS '{"name":"Mesh size 4", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue5; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue5 IS '{"name":"Residue 5", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize5; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize5 IS '{"name":"Mesh size 5", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue6; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue6 IS '{"name":"Residue 6", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize6; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize6 IS '{"name":"Mesh size 6", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue7; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue7 IS '{"name":"Residue 7", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize7; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize7 IS '{"name":"Mesh size 7", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue8; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue8 IS '{"name":"Residue 8", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize8; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize8 IS '{"name":"Mesh size 8", "unit":"µm"}';


--
-- Name: COLUMN sediment.residue9; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.residue9 IS '{"name":"Residue 9", "unit":"%"}';


--
-- Name: COLUMN sediment.meshsize9; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.meshsize9 IS '{"name":"Mesh size 9", "unit":"µm"}';


--
-- Name: COLUMN sediment.median; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.median IS '{"name":"Median", "unit":"µm", "comment":"Median grain size"}';


--
-- Name: COLUMN sediment.gsd; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.gsd IS '{"name":"Gsd", "comment":"Sorting coefficient sigma1 of sediment"}';


--
-- Name: COLUMN sediment.weight; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.weight IS '{"name":"Weight", "unit":"g", "comment":"Total weight of sediment sample for grainsize determination"}';


--
-- Name: COLUMN sediment.method_grainsize; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.method_grainsize IS '{"name":"Method grainsize", "comment":"Method description to determine grainsize"}';


--
-- Name: COLUMN sediment.temperature; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.temperature IS '{"name":"Temperature", "unit":"°C", "comment":"Measured sediment temperature"}';


--
-- Name: COLUMN sediment.c_org; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.c_org IS '{"name":"C org", "unit":"g/kg", "comment":"Organic carbon content of sediment sample"}';


--
-- Name: COLUMN sediment.c_total; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.c_total IS '{"name":"C total", "unit":"%", "comment":"Total carbon content of sediment sample"}';


--
-- Name: COLUMN sediment.loss_on_ignition; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.loss_on_ignition IS '{"name":"Loss on ignition", "unit":"%", "comment":"Loss on ignition as a measure of sediment organic matter"}';


--
-- Name: COLUMN sediment.nitrogen; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.nitrogen IS '{"name":"Nitrogen", "unit":"%", "comment":"Total nitrogen content of sediment sample"}';


--
-- Name: COLUMN sediment.sulfur; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.sulfur IS '{"name":"Sulfur", "unit":"%", "comment":"Total sulfur content of sediment sample"}';


--
-- Name: COLUMN sediment.oxygen_pd; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.oxygen_pd IS '{"name":"Oxygen penetration depth", "unit":"mm", "comment":"Oxygen penetration depth, i.e. oxygen horizon by visual inspection"}';


--
-- Name: COLUMN sediment.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.remark IS '{"name":"Remark"}';


--
-- Name: COLUMN sediment.fuz_silt; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_silt IS '{"name":"Fuz silt", "comment":"Fuzzy coding of sample silt fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_vfsand; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_vfsand IS '{"name":"Fuz vfsand", "comment":"Fuzzy coding of sample very fine sand fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_fsand; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_fsand IS '{"name":"Fuz fsand", "comment":"Fuzzy coding of sample fine sand fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_msand; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_msand IS '{"name":"Fuz msand", "comment":"Fuzzy coding of sample medium sand fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_csand; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_csand IS '{"name":"Fuz csand", "comment":"Fuzzy coding of sample coarse sand fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_vcsand; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_vcsand IS '{"name":"Fuz vcsand", "comment":"Fuzzy coding of sample very coarse sand fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_gravel; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_gravel IS '{"name":"Fuz gravel", "comment":"Fuzzy coding of sample gravel fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_stones; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_stones IS '{"name":"Fuz stones", "comment":"Fuzzy coding of sample stones fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_shell; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_shell IS '{"name":"Fuz shell", "comment":"Fuzzy coding of sample shell fraction, e.g. from Molluscs; numbers: 1-3"}';


--
-- Name: COLUMN sediment.fuz_clay; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.fuz_clay IS '{"name":"Fuz clay", "comment":"Fuzzy coding of sample clay fraction; numbers: 1-3"}';


--
-- Name: COLUMN sediment.type; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.sediment.type IS '{"name":"Type"}';


--
-- Name: sediment_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.sediment_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: sediment_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.sediment_id_seq OWNED BY cb_empty.sediment.id;


--
-- Name: ship; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.ship (
    id integer NOT NULL,
    name text NOT NULL,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN ship.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ship.id IS '{"name":"Vessel"}';


--
-- Name: COLUMN ship.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ship.name IS '{"name":"Research vessel name", "comment":"Re-enter exact text/name in corresponding table: Cruise.ShipName"}';


--
-- Name: COLUMN ship.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.ship.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: ship_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.ship_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: ship_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.ship_id_seq OWNED BY cb_empty.ship.id;


--
-- Name: station; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.station (
    id integer NOT NULL,
    name text NOT NULL,
    cruise_id integer NOT NULL,
    ingest_id integer NOT NULL,
    responsible_person_id integer,
    lon_start double precision,
    lat_start double precision,
    date_start date,
    time_start time without time zone,
    depth_start double precision,
    lon_end double precision,
    lat_end double precision,
    date_end date,
    time_end time without time zone,
    depth_end double precision,
    location text,
    lon_target double precision,
    lat_target double precision,
    replicates integer,
    temperature_air double precision,
    temperature_surfacewater double precision,
    temperature_bottomwater double precision,
    salinity_surfacewater double precision,
    salinity_bottomwater double precision,
    oxygen_bottomwater double precision,
    windspeed_beaufort double precision,
    wind_direction text,
    remark text,
    "precision" integer DEFAULT 5 NOT NULL
);


--
-- Name: COLUMN station.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.id IS '{"name":"Station"}';


--
-- Name: COLUMN station.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.name IS '{"name":"Station name","pangaea":"102005"}';


--
-- Name: COLUMN station.cruise_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.cruise_id IS '{"name":"Cruise"}';


--
-- Name: COLUMN station.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN station.responsible_person_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.responsible_person_id IS '{"name":"Responsible person", "comment":"Value must be defined in lookup table: Person.Name"}';


--
-- Name: COLUMN station.lon_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lon_start IS '{"name":"Start lon", "unit":"dD", "comment":"Logged longitude at station start","pangaea":"1601"}';


--
-- Name: COLUMN station.lat_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lat_start IS '{"name":"Start lat", "unit":"dD", "comment":"Logged latitude at station start","pangaea":"1600"}';


--
-- Name: COLUMN station.date_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.date_start IS '{"name":"Start date", "unit":"YYYY-MM-DD", "comment":"Date at station start","pangaea":"152481"}';


--
-- Name: COLUMN station.time_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.time_start IS '{"name":"Start time", "unit":"HH:MM:SS", "comment":"Time at station start" ,"pangaea":"152460"}';


--
-- Name: COLUMN station.depth_start; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.depth_start IS '{"name":"Start depth", "unit":"m", "comment":"Depth at station start","pangaea":"1619"}';


--
-- Name: COLUMN station.lon_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lon_end IS '{"name":"End lon", "unit":"dD", "comment":"Logged longitude at station end","pangaea":"1601"}';


--
-- Name: COLUMN station.lat_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lat_end IS '{"name":"End lat", "unit":"dD", "comment":"Logged latitude at station end","pangaea":"1600"}';


--
-- Name: COLUMN station.date_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.date_end IS '{"name":"End date", "unit":"YYYY-MM-DD", "comment":"Date at station end","pangaea":"152481"}';


--
-- Name: COLUMN station.time_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.time_end IS '{"name":"End time", "unit":"HH:MM:SS", "comment":"Time at station end" ,"pangaea":"150986"}';


--
-- Name: COLUMN station.depth_end; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.depth_end IS '{"name":"End depth", "unit":"m", "comment":"Depth at station end","pangaea":"1619"}';


--
-- Name: COLUMN station.location; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.location IS '{"name":"Location", "comment":"Dedicated station name, e.g. long-term research station name"}';


--
-- Name: COLUMN station.lon_target; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lon_target IS '{"name":"Target lon", "unit":"dD", "comment":"Target longitude of station, i.e. fixed coordinates of a station specified during cruise planning","pangaea":"1601"}';


--
-- Name: COLUMN station.lat_target; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.lat_target IS '{"name":"Target lat", "unit":"dD", "comment":"Target latitude of station, i.e. fixed coordinates of a station specified during cruise planning","pangaea":"1600"}';


--
-- Name: COLUMN station.replicates; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.replicates IS '{"name":"Replicates", "comment":"Number of replicates per gear of a station"}';


--
-- Name: COLUMN station.temperature_air; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.temperature_air IS '{"name":"Temperature air", "unit":"°C", "comment":"Air temperature in Celsius"}';


--
-- Name: COLUMN station.temperature_surfacewater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.temperature_surfacewater IS '{"name":"Temperature surfacewater", "unit":"°C", "comment":"Sea surface temperature SST"}';


--
-- Name: COLUMN station.temperature_bottomwater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.temperature_bottomwater IS '{"name":"Temperature bottomwater", "unit":"°C", "comment":"Sea bottom temperature SBT"}';


--
-- Name: COLUMN station.salinity_surfacewater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.salinity_surfacewater IS '{"name":"Salinity surfacewater", "unit":"PSU", "comment":"Salinity at sea surface"}';


--
-- Name: COLUMN station.salinity_bottomwater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.salinity_bottomwater IS '{"name":"Salinity bottomwater", "unit":"PSU", "comment":"Salinity at sea bottom"}';


--
-- Name: COLUMN station.oxygen_bottomwater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.oxygen_bottomwater IS '{"name":"Oxygen bottomwater", "unit":"mg", "comment":"Oxgen content of bottom water"}';


--
-- Name: COLUMN station.windspeed_beaufort; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.windspeed_beaufort IS '{"name":"Wind speed", "unit":"Bft", "comment":"Beaufort number for wind condition at sea, assigned after the Beaufort wind force scale. Decimals are allowed"}';


--
-- Name: COLUMN station.wind_direction; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.wind_direction IS '{"name":"Wind direction", "comment":"Wind direction in text format, e.g. NW for north-west"}';


--
-- Name: COLUMN station.remark; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station.remark IS '{"name":"Remark"}';


--
-- Name: COLUMN station."precision"; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.station."precision" IS '{"name":"Spatial precision", "comment":"Spatial precision as number of decimal digits for geographic position: lat, lon - range: 1-5 - default when field is left empty is 5"}';


--
-- Name: station_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.station_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: station_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.station_id_seq OWNED BY cb_empty.station.id;


--
-- Name: subsample; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.subsample (
    id integer NOT NULL,
    ingest_id integer NOT NULL,
    sample_id integer NOT NULL,
    ratio double precision,
    name text NOT NULL
);


--
-- Name: COLUMN subsample.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subsample.id IS '{"name":"SubSample"}';


--
-- Name: COLUMN subsample.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subsample.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN subsample.sample_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subsample.sample_id IS '{"name":"Sample"}';


--
-- Name: COLUMN subsample.ratio; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subsample.ratio IS '{"name":"Ratio", "comment": "Proportion of the sub-sample of the corresponding sample", "unit": "positive floating point number"}';


--
-- Name: COLUMN subsample.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subsample.name IS '{"name":"SubSample name"}';


--
-- Name: subsample_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.subsample_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: subsample_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.subsample_id_seq OWNED BY cb_empty.subsample.id;


--
-- Name: subset; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.subset (
    id integer NOT NULL,
    name text NOT NULL,
    subsample_id integer NOT NULL,
    dataset_id integer NOT NULL,
    ingest_id integer NOT NULL,
    taxcover_id integer,
    taxres text,
    spatiallevel integer,
    CONSTRAINT check_subset_taxonomic_resolution CHECK ((taxres = ANY (ARRAY['domain'::text, 'kingdom'::text, 'phylum'::text, 'class'::text, 'order'::text, 'family'::text, 'genus'::text, 'species'::text])))
);


--
-- Name: COLUMN subset.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.name IS '{"name":"Subset name"}';


--
-- Name: COLUMN subset.subsample_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.subsample_id IS '{"name":"SubSample"}';


--
-- Name: COLUMN subset.dataset_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.dataset_id IS '{"name":"Dataset", "comment":"Value must be defined in lookup table: Dataset.Name"}';


--
-- Name: COLUMN subset.ingest_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.ingest_id IS '{"name":"Ingest"}';


--
-- Name: COLUMN subset.taxcover_id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.taxcover_id IS '{"name": "Taxonomic coverage ", "comment":"Defines the taxonomic coverage or scope, e.g. benthos_complete. Value must be defined in lookup table: TaxCover.Name"}';


--
-- Name: COLUMN subset.taxres; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.taxres IS '{"name": "Taxonomic resolution", "comment": "Taxonomic resolution of biota: domain, kingdom, phylum, class, order, family, genus, species"}';


--
-- Name: COLUMN subset.spatiallevel; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.subset.spatiallevel IS '{"name": "Spatial level ", "comment": "Spatial level of biota: 0-10"}';


--
-- Name: subset_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.subset_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: subset_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.subset_id_seq OWNED BY cb_empty.subset.id;


--
-- Name: taxcover; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.taxcover (
    id integer NOT NULL,
    name text NOT NULL,
    description text,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN taxcover.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxcover.id IS '{"name":"Taxonomic coverage"}';


--
-- Name: COLUMN taxcover.name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxcover.name IS '{"name":"Taxonomic coverage name", "comment":"Re-enter exact text/name in corresponding table: Subset.TaxonomicCoverage"}';


--
-- Name: COLUMN taxcover.description; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxcover.description IS '{"name":"Description"}';


--
-- Name: COLUMN taxcover.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxcover.gold IS '{"name":"Gold standard lookup entry", "unit":"boolean", "comment":"This is a quality check ranking that is filled out while ingest. (Leave blank.)"}';


--
-- Name: taxcover_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.taxcover_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: taxcover_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.taxcover_id_seq OWNED BY cb_empty.taxcover.id;


--
-- Name: taxon; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.taxon (
    id integer NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    valid_name text,
    aaid integer,
    aname text,
    status text,
    unacceptreason text,
    rank text,
    kingdom text,
    phylum text,
    class text,
    "order" text,
    family text,
    genus text,
    vaid integer,
    valid_authority text,
    citation text,
    url text,
    lsid text,
    ismarine integer,
    isbrackish integer,
    isfreshwater integer,
    isterrestrial integer,
    isextinct integer,
    is_colony boolean,
    authority text,
    match_type text,
    modified text
);


--
-- Name: COLUMN taxon.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.id IS '{"name":"Taxon"}';


--
-- Name: COLUMN taxon.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.aid IS '{"name":"AphiaId"}';


--
-- Name: COLUMN taxon.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.scientificname IS '{"name":"Scientific name"}';


--
-- Name: COLUMN taxon.valid_name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.valid_name IS '{"name":"Valid name"}';


--
-- Name: COLUMN taxon.aaid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.aaid IS '{"name":"Accepted AphiaId"}';


--
-- Name: COLUMN taxon.aname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.aname IS '{"name": "Accepted scientific name"}';


--
-- Name: COLUMN taxon.status; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.status IS '{"name":"Status"}';


--
-- Name: COLUMN taxon.unacceptreason; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.unacceptreason IS '{"name":"Unacceptreason"}';


--
-- Name: COLUMN taxon.rank; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.rank IS '{"name":"Rank"}';


--
-- Name: COLUMN taxon.kingdom; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.kingdom IS '{"name":"Kingdom"}';


--
-- Name: COLUMN taxon.phylum; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.phylum IS '{"name":"Phylum"}';


--
-- Name: COLUMN taxon.class; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.class IS '{"name":"Class"}';


--
-- Name: COLUMN taxon."order"; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon."order" IS '{"name":"Order"}';


--
-- Name: COLUMN taxon.family; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.family IS '{"name":"Family"}';


--
-- Name: COLUMN taxon.genus; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.genus IS '{"name":"Genus"}';


--
-- Name: COLUMN taxon.vaid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.vaid IS '{"name":"Valid AphiaId"}';


--
-- Name: COLUMN taxon.valid_authority; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.valid_authority IS '{"name":"Valid authority"}';


--
-- Name: COLUMN taxon.citation; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.citation IS '{"name":"Citation"}';


--
-- Name: COLUMN taxon.url; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.url IS '{"name":"Url"}';


--
-- Name: COLUMN taxon.lsid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.lsid IS '{"name":"Lsid"}';


--
-- Name: COLUMN taxon.ismarine; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.ismarine IS '{"name":"Is marine"}';


--
-- Name: COLUMN taxon.isbrackish; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.isbrackish IS '{"name":"Is brackish"}';


--
-- Name: COLUMN taxon.isfreshwater; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.isfreshwater IS '{"name":"Is freshwater"}';


--
-- Name: COLUMN taxon.isterrestrial; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.isterrestrial IS '{"name":"Is terrestrial"}';


--
-- Name: COLUMN taxon.isextinct; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.isextinct IS '{"name":"Is extinct"}';


--
-- Name: COLUMN taxon.is_colony; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.is_colony IS '{"name":"Is colony"}';


--
-- Name: COLUMN taxon.authority; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.authority IS '{"name":"Authority"}';


--
-- Name: COLUMN taxon.match_type; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.match_type IS '{"name":"Match type"}';


--
-- Name: COLUMN taxon.modified; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.taxon.modified IS '{"name":"Modified"}';


--
-- Name: taxon_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.taxon_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: taxon_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.taxon_id_seq OWNED BY cb_empty.taxon.id;


--
-- Name: wormsaaid; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.wormsaaid (
    id integer NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    accepted_scientificname text,
    aaid integer NOT NULL,
    aaid_scientificname text
);


--
-- Name: COLUMN wormsaaid.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.id IS '{"name":"Accepted Aphia Ids", comment="Overriding the actual accepted Aphia Ids of a taxon (expert knowledge)"}';


--
-- Name: COLUMN wormsaaid.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.aid IS '{"name":"Aphia Id"}';


--
-- Name: COLUMN wormsaaid.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.scientificname IS '{"name":"Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormsaaid.accepted_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.accepted_scientificname IS '{"name":"Accepted scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormsaaid.aaid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.aaid IS '{"name":"Accepted Aphia Id"}';


--
-- Name: COLUMN wormsaaid.aaid_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsaaid.aaid_scientificname IS '{"name":"Scientific name of the Aaid-taxon (field is filled in automatically)"}
';


--
-- Name: wormsaaid_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.wormsaaid_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: wormsaaid_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.wormsaaid_id_seq OWNED BY cb_empty.wormsaaid.id;


--
-- Name: wormsbase; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.wormsbase (
    id integer NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    accepted_scientificname text,
    gold boolean DEFAULT false NOT NULL
);


--
-- Name: COLUMN wormsbase.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsbase.id IS '{"name":"Taxon lists base Aphia Ids", comment="Aphia Ids to build the taxon list using WoRMS"}';


--
-- Name: COLUMN wormsbase.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsbase.aid IS '{"name":"Aphia Id", "comment":"Aphia Ids to build the taxon list using WoRMS"}';


--
-- Name: COLUMN wormsbase.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsbase.scientificname IS '{"name":"Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormsbase.accepted_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsbase.accepted_scientificname IS '{"name":"Accepted scientific name (field is filled in automatically)"}
';


--
-- Name: COLUMN wormsbase.gold; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsbase.gold IS '{"name":"Gold standard entry", "unit":"boolean", "comment":"This is a quality check ranking. True means: An expert certainly agrees that the taxon rightly belongs in the taxon list."}
';


--
-- Name: wormsbase_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.wormsbase_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: wormsbase_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.wormsbase_id_seq OWNED BY cb_empty.wormsbase.id;


--
-- Name: wormschange_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.wormschange_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: wormschange_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.wormschange_id_seq OWNED BY cb_empty.wormsaaid.aaid;


--
-- Name: wormschange; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.wormschange (
    id integer DEFAULT nextval('cb_empty.wormschange_id_seq'::regclass) NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    accepted_scientificname text,
    treated text NOT NULL,
    workflow text,
    date_revision date,
    validity text,
    source text,
    comment text
);


--
-- Name: COLUMN wormschange.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.id IS '{"name":"Taxon grouping", comment="Taxon grouping"}';


--
-- Name: COLUMN wormschange.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.aid IS '{"name": "Aphia Id of the original taxon"}';


--
-- Name: COLUMN wormschange.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.scientificname IS '{"name": "Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormschange.accepted_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.accepted_scientificname IS '{"name": "Accepted Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormschange.treated; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.treated IS '{"name":"Name of taxon to be treaded as"}';


--
-- Name: COLUMN wormschange.workflow; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.workflow IS '{"name": "Workflow", "comment":"Workflow, e.g. Northsea"}';


--
-- Name: COLUMN wormschange.date_revision; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.date_revision IS '{"name": "Date of revision"}';


--
-- Name: COLUMN wormschange.validity; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.validity IS '{"name": "Validity", "comment":"Validity of entry (e.g. datasets)"}';


--
-- Name: COLUMN wormschange.source; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.source IS '{"name": "Source"}';


--
-- Name: COLUMN wormschange.comment; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormschange.comment IS '{"name": "Comment"}
';


--
-- Name: wormscolony; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.wormscolony (
    id integer NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    accepted_scientificname text
);


--
-- Name: COLUMN wormscolony.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormscolony.id IS '{"name":"Colony taxa", comment="Aphia Ids of colony taxa (expert knowledge)"}';


--
-- Name: COLUMN wormscolony.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormscolony.aid IS '{"name":"Aphia Id"}';


--
-- Name: COLUMN wormscolony.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormscolony.scientificname IS '{"name":"Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormscolony.accepted_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormscolony.accepted_scientificname IS '{"name":"Accepted scientific name (field is filled in automatically)"}
';


--
-- Name: wormscolony_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.wormscolony_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: wormscolony_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.wormscolony_id_seq OWNED BY cb_empty.wormscolony.id;


--
-- Name: wormsprivate; Type: TABLE; Schema: cb_empty; Owner: -
--

CREATE TABLE cb_empty.wormsprivate (
    id integer NOT NULL,
    private_name text NOT NULL,
    aid integer NOT NULL,
    scientificname text,
    accepted_scientificname text
);


--
-- Name: COLUMN wormsprivate.id; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsprivate.id IS '{"name":"Alternative scientific names", comment="Mapping scientific names to Aphia Ids (expert knowledge)"}';


--
-- Name: COLUMN wormsprivate.private_name; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsprivate.private_name IS '{"name":"Private name"}';


--
-- Name: COLUMN wormsprivate.aid; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsprivate.aid IS '{"name":"Aphia Id"}';


--
-- Name: COLUMN wormsprivate.scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsprivate.scientificname IS '{"name":"Scientific name (field is filled in automatically)"}';


--
-- Name: COLUMN wormsprivate.accepted_scientificname; Type: COMMENT; Schema: cb_empty; Owner: -
--

COMMENT ON COLUMN cb_empty.wormsprivate.accepted_scientificname IS '{"name":"Accepted scientific name (field is filled in automatically)"}
';


--
-- Name: wormsprivate_id_seq; Type: SEQUENCE; Schema: cb_empty; Owner: -
--

CREATE SEQUENCE cb_empty.wormsprivate_id_seq
    START WITH 1
    INCREMENT BY 1
    NO MINVALUE
    NO MAXVALUE
    CACHE 1;


--
-- Name: wormsprivate_id_seq; Type: SEQUENCE OWNED BY; Schema: cb_empty; Owner: -
--

ALTER SEQUENCE cb_empty.wormsprivate_id_seq OWNED BY cb_empty.wormsprivate.id;


--
-- Name: actionlog id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.actionlog ALTER COLUMN id SET DEFAULT nextval('cb_empty.actionlog_id_seq'::regclass);


--
-- Name: admin id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.admin ALTER COLUMN id SET DEFAULT nextval('cb_empty.admin_id_seq'::regclass);


--
-- Name: biota id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota ALTER COLUMN id SET DEFAULT nextval('cb_empty.biota_id_seq'::regclass);


--
-- Name: country id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.country ALTER COLUMN id SET DEFAULT nextval('cb_empty.country_id_seq'::regclass);


--
-- Name: crs id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.crs ALTER COLUMN id SET DEFAULT nextval('cb_empty.crs_id_seq'::regclass);


--
-- Name: cruise id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise ALTER COLUMN id SET DEFAULT nextval('cb_empty.cruise_id_seq'::regclass);


--
-- Name: dataset id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.dataset ALTER COLUMN id SET DEFAULT nextval('cb_empty.dataset_id_seq'::regclass);


--
-- Name: gear id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.gear ALTER COLUMN id SET DEFAULT nextval('cb_empty.gear_id_seq'::regclass);


--
-- Name: indmeasure id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.indmeasure ALTER COLUMN id SET DEFAULT nextval('cb_empty.indmeasure_id_seq'::regclass);


--
-- Name: ingest id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.ingest ALTER COLUMN id SET DEFAULT nextval('cb_empty.ingest_id_seq'::regclass);


--
-- Name: lifestage id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.lifestage ALTER COLUMN id SET DEFAULT nextval('cb_empty.lifestage_id_seq'::regclass);


--
-- Name: noworms id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.noworms ALTER COLUMN id SET DEFAULT nextval('cb_empty.noworms_id_seq'::regclass);


--
-- Name: person id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.person ALTER COLUMN id SET DEFAULT nextval('cb_empty.person_id_seq'::regclass);


--
-- Name: possys id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.possys ALTER COLUMN id SET DEFAULT nextval('cb_empty.positioningsystem_id_seq'::regclass);


--
-- Name: region id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region ALTER COLUMN id SET DEFAULT nextval('cb_empty.region_id_seq'::regclass);


--
-- Name: sample id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample ALTER COLUMN id SET DEFAULT nextval('cb_empty.sample_id_seq'::regclass);


--
-- Name: sediment id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sediment ALTER COLUMN id SET DEFAULT nextval('cb_empty.sediment_id_seq'::regclass);


--
-- Name: ship id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.ship ALTER COLUMN id SET DEFAULT nextval('cb_empty.ship_id_seq'::regclass);


--
-- Name: station id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station ALTER COLUMN id SET DEFAULT nextval('cb_empty.station_id_seq'::regclass);


--
-- Name: subsample id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subsample ALTER COLUMN id SET DEFAULT nextval('cb_empty.subsample_id_seq'::regclass);


--
-- Name: subset id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset ALTER COLUMN id SET DEFAULT nextval('cb_empty.subset_id_seq'::regclass);


--
-- Name: taxcover id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.taxcover ALTER COLUMN id SET DEFAULT nextval('cb_empty.taxcover_id_seq'::regclass);


--
-- Name: taxon id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.taxon ALTER COLUMN id SET DEFAULT nextval('cb_empty.taxon_id_seq'::regclass);


--
-- Name: wormsaaid id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsaaid ALTER COLUMN id SET DEFAULT nextval('cb_empty.wormsaaid_id_seq'::regclass);


--
-- Name: wormsaaid aaid; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsaaid ALTER COLUMN aaid SET DEFAULT nextval('cb_empty.wormschange_id_seq'::regclass);


--
-- Name: wormsbase id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsbase ALTER COLUMN id SET DEFAULT nextval('cb_empty.wormsbase_id_seq'::regclass);


--
-- Name: wormscolony id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormscolony ALTER COLUMN id SET DEFAULT nextval('cb_empty.wormscolony_id_seq'::regclass);


--
-- Name: wormsprivate id; Type: DEFAULT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsprivate ALTER COLUMN id SET DEFAULT nextval('cb_empty.wormsprivate_id_seq'::regclass);


--
-- Name: actionlog actionlog_pkey; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.actionlog
    ADD CONSTRAINT actionlog_pkey PRIMARY KEY (id);


--
-- Name: admin admin_key_uniq; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.admin
    ADD CONSTRAINT admin_key_uniq UNIQUE (key);


--
-- Name: indmeasure autopsy_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.indmeasure
    ADD CONSTRAINT autopsy_id_pk PRIMARY KEY (id);


--
-- Name: biota biota_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT biota_id_pk PRIMARY KEY (id);


--
-- Name: country country name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.country
    ADD CONSTRAINT "country name has to be unique" UNIQUE (name);


--
-- Name: country country_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.country
    ADD CONSTRAINT country_id_pk PRIMARY KEY (id);


--
-- Name: crs crs name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.crs
    ADD CONSTRAINT "crs name has to be unique" UNIQUE (name);


--
-- Name: crs crs_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.crs
    ADD CONSTRAINT crs_id_pk PRIMARY KEY (id);


--
-- Name: cruise cruise name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT "cruise name has to be unique" UNIQUE (name);


--
-- Name: cruise cruise_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT cruise_id_pk PRIMARY KEY (id);


--
-- Name: sample cruise_station_sample_name_uniq; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT cruise_station_sample_name_uniq UNIQUE (name, station_id);


--
-- Name: biota cruise_station_sample_subset_biota_name_uniq; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT cruise_station_sample_subset_biota_name_uniq UNIQUE (subset_id, id);


--
-- Name: subset cruise_station_sample_subset_name_uniq; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset
    ADD CONSTRAINT cruise_station_sample_subset_name_uniq UNIQUE (subsample_id, name);


--
-- Name: dataset dataset name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.dataset
    ADD CONSTRAINT "dataset name has to be unique" UNIQUE (name);


--
-- Name: dataset dataset_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.dataset
    ADD CONSTRAINT dataset_id_pk PRIMARY KEY (id);


--
-- Name: wormsaaid expert knowledge aaid: Aphia ID in has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsaaid
    ADD CONSTRAINT "expert knowledge aaid: Aphia ID in has to be unique" UNIQUE (aid);


--
-- Name: wormsbase expert knowledge base list: Aphia ID has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsbase
    ADD CONSTRAINT "expert knowledge base list: Aphia ID has to be unique" UNIQUE (aid);


--
-- Name: wormscolony expert knowledge colony: Aphia ID has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormscolony
    ADD CONSTRAINT "expert knowledge colony: Aphia ID has to be unique" UNIQUE (aid);


--
-- Name: wormsprivate expert knowledge private names: scientificname has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.wormsprivate
    ADD CONSTRAINT "expert knowledge private names: scientificname has to be unique" UNIQUE (private_name);


--
-- Name: sediment fuz_clay has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_clay has to be: 1/2/3" CHECK ((fuz_clay = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_csand  has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_csand  has to be: 1/2/3" CHECK ((fuz_csand = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_fsand  has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_fsand  has to be: 1/2/3" CHECK ((fuz_fsand = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_gravel  has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_gravel  has to be: 1/2/3" CHECK ((fuz_gravel = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_msand has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_msand has to be: 1/2/3" CHECK ((fuz_msand = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_shell has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_shell has to be: 1/2/3" CHECK ((fuz_shell = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_silt has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_silt has to be: 1/2/3" CHECK ((fuz_silt = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_stones has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_stones has to be: 1/2/3" CHECK ((fuz_stones = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: sediment fuz_vfsand has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "fuz_vfsand has to be: 1/2/3" CHECK ((fuz_vfsand = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: gear gear name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.gear
    ADD CONSTRAINT "gear name has to be unique" UNIQUE (name);


--
-- Name: gear gear_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.gear
    ADD CONSTRAINT gear_id_pk PRIMARY KEY (id);


--
-- Name: ingest ingest_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.ingest
    ADD CONSTRAINT ingest_id_pk PRIMARY KEY (id);


--
-- Name: lifestage lifestage name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.lifestage
    ADD CONSTRAINT "lifestage name has to be unique" UNIQUE (name);


--
-- Name: lifestage lifestage_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.lifestage
    ADD CONSTRAINT lifestage_id_pk PRIMARY KEY (id);


--
-- Name: person name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.person
    ADD CONSTRAINT "name has to be unique" UNIQUE (name);


--
-- Name: noworms noworms scientificname has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.noworms
    ADD CONSTRAINT "noworms scientificname has to be unique" UNIQUE (scientificname);


--
-- Name: person person_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.person
    ADD CONSTRAINT person_id_pk PRIMARY KEY (id);


--
-- Name: possys positioning system name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.possys
    ADD CONSTRAINT "positioning system name has to be unique" UNIQUE (name);


--
-- Name: possys positioningsystem_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.possys
    ADD CONSTRAINT positioningsystem_id_pk PRIMARY KEY (id);


--
-- Name: station precision has to be in: [1,5] (integer); Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.station
    ADD CONSTRAINT "precision has to be in: [1,5] (integer)" CHECK ((("precision" > 0) AND ("precision" < 6))) NOT VALID;


--
-- Name: subsample ratio has to be in: [0-1]; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.subsample
    ADD CONSTRAINT "ratio has to be in: [0-1]" CHECK (((ratio > (0)::double precision) AND (ratio <= (1)::double precision))) NOT VALID;


--
-- Name: region region_name_key; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region
    ADD CONSTRAINT region_name_key UNIQUE (name);


--
-- Name: region region_pkey; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region
    ADD CONSTRAINT region_pkey PRIMARY KEY (id);


--
-- Name: sample sample_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT sample_id_pk PRIMARY KEY (id);


--
-- Name: sediment sediment_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sediment
    ADD CONSTRAINT sediment_id_pk PRIMARY KEY (id);


--
-- Name: ship ship name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.ship
    ADD CONSTRAINT "ship name has to be unique" UNIQUE (name);


--
-- Name: ship ship_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.ship
    ADD CONSTRAINT ship_id_pk PRIMARY KEY (id);


--
-- Name: subset spatial level has to be in: [0-10] (integer); Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.subset
    ADD CONSTRAINT "spatial level has to be in: [0-10] (integer)" CHECK (((spatiallevel >= 0) AND (spatiallevel <= 10))) NOT VALID;


--
-- Name: sample precision has to be in: [1,5] (integer); Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sample
    ADD CONSTRAINT "precision has to be in: [1,5] (integer)" CHECK ((("precision" > 0) AND ("precision" < 6))) NOT VALID;


--
-- Name: station station name has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station
    ADD CONSTRAINT "station name has to be unique" UNIQUE (name, cruise_id);


--
-- Name: station station_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station
    ADD CONSTRAINT station_id_pk PRIMARY KEY (id);


--
-- Name: subsample subsample_id; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subsample
    ADD CONSTRAINT subsample_id PRIMARY KEY (id);


--
-- Name: subset subset_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset
    ADD CONSTRAINT subset_id_pk PRIMARY KEY (id);


--
-- Name: sediment subset_id_uniq  has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sediment
    ADD CONSTRAINT "subset_id_uniq  has to be unique" UNIQUE (subset_id);


--
-- Name: taxcover taxcover_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.taxcover
    ADD CONSTRAINT taxcover_id_pk PRIMARY KEY (id);


--
-- Name: taxon taxon Aphia ID has to be unique; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.taxon
    ADD CONSTRAINT "taxon Aphia ID has to be unique" UNIQUE (aid);


--
-- Name: taxon taxon_id_pk; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.taxon
    ADD CONSTRAINT taxon_id_pk PRIMARY KEY (id);


--
-- Name: region uniq_code; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region
    ADD CONSTRAINT uniq_code UNIQUE (code);


--
-- Name: region2sample uniq_sample_id_region_id; Type: CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region2sample
    ADD CONSTRAINT uniq_sample_id_region_id UNIQUE (sample_id, region_id);


--
-- Name: sediment vcsand has to be: 1/2/3; Type: CHECK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE cb_empty.sediment
    ADD CONSTRAINT "vcsand has to be: 1/2/3" CHECK ((fuz_vcsand = ANY (ARRAY[1, 2, 3, NULL::integer]))) NOT VALID;


--
-- Name: fki_sample2region_region; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX fki_sample2region_region ON cb_empty.region2sample USING btree (sample_id);


--
-- Name: fki_subsample_ingest_fk; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX fki_subsample_ingest_fk ON cb_empty.subsample USING btree (ingest_id);


--
-- Name: fki_subset_subsample_fk; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX fki_subset_subsample_fk ON cb_empty.subset USING btree (subsample_id);


--
-- Name: idx_autopsy; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_autopsy ON cb_empty.indmeasure USING btree (biota_id);


--
-- Name: idx_autopsy_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_autopsy_0 ON cb_empty.indmeasure USING btree (ingest_id);


--
-- Name: idx_cruise_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_cruise_0 ON cb_empty.cruise USING btree (positioningsystem_id);


--
-- Name: idx_cruise_1; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_cruise_1 ON cb_empty.cruise USING btree (crs_id);


--
-- Name: idx_cruise_2; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_cruise_2 ON cb_empty.cruise USING btree (ship_id);


--
-- Name: idx_cruise_3; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_cruise_3 ON cb_empty.cruise USING btree (lead_person_id);


--
-- Name: idx_cruise_4; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_cruise_4 ON cb_empty.cruise USING btree (ingest_id);


--
-- Name: idx_dataset_1; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_dataset_1 ON cb_empty.dataset USING btree (contact_person_id);


--
-- Name: idx_population; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_population ON cb_empty.biota USING btree (taxon_id);


--
-- Name: idx_population_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_population_0 ON cb_empty.biota USING btree (subset_id);


--
-- Name: idx_population_2; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_population_2 ON cb_empty.biota USING btree (lifestage_id);


--
-- Name: idx_population_3; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_population_3 ON cb_empty.biota USING btree (ingest_id);


--
-- Name: idx_sample; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_sample ON cb_empty.sample USING btree (station_id);


--
-- Name: idx_sample_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_sample_0 ON cb_empty.sample USING btree (ingest_id);


--
-- Name: idx_sample_1; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_sample_1 ON cb_empty.sample USING btree (responsible_person_id);


--
-- Name: idx_sample_2; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_sample_2 ON cb_empty.sample USING btree (gear_id);


--
-- Name: idx_sediment_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_sediment_0 ON cb_empty.sediment USING btree (ingest_id);


--
-- Name: idx_station; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_station ON cb_empty.station USING btree (cruise_id);


--
-- Name: idx_station_0; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_station_0 ON cb_empty.station USING btree (ingest_id);


--
-- Name: idx_station_2; Type: INDEX; Schema: cb_empty; Owner: -
--

CREATE INDEX idx_station_2 ON cb_empty.station USING btree (responsible_person_id);


--
-- Name: gear gear_check_consistency; Type: TRIGGER; Schema: cb_empty; Owner: -
--

CREATE TRIGGER gear_check_consistency BEFORE INSERT OR UPDATE ON cb_empty.gear FOR EACH ROW EXECUTE FUNCTION cb_empty.new_gear_check_consistency();


--
-- Name: biota trigger_check_consistency; Type: TRIGGER; Schema: cb_empty; Owner: -
--

CREATE TRIGGER trigger_check_consistency AFTER INSERT OR UPDATE ON cb_empty.biota FOR EACH ROW EXECUTE FUNCTION cb_empty.new_biota_check_consistency();


--
-- Name: sample trigger_check_consistency; Type: TRIGGER; Schema: cb_empty; Owner: -
--

CREATE TRIGGER trigger_check_consistency AFTER INSERT OR UPDATE ON cb_empty.sample FOR EACH ROW EXECUTE FUNCTION cb_empty.new_sample_check_consistency();


--
-- Name: subsample trigger_check_consistency; Type: TRIGGER; Schema: cb_empty; Owner: -
--

CREATE TRIGGER trigger_check_consistency AFTER INSERT OR UPDATE ON cb_empty.subsample FOR EACH ROW EXECUTE FUNCTION cb_empty.new_subsample_check_consistency();


--
-- Name: subset trigger_check_consistency; Type: TRIGGER; Schema: cb_empty; Owner: -
--

CREATE TRIGGER trigger_check_consistency BEFORE INSERT OR UPDATE ON cb_empty.subset FOR EACH ROW EXECUTE FUNCTION cb_empty.new_subset_check_consistency();


--
-- Name: indmeasure autopsy_biota_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.indmeasure
    ADD CONSTRAINT autopsy_biota_fk FOREIGN KEY (biota_id) REFERENCES cb_empty.biota(id);


--
-- Name: indmeasure autopsy_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.indmeasure
    ADD CONSTRAINT autopsy_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: biota biota_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT biota_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: biota biota_lifestage_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT biota_lifestage_fk FOREIGN KEY (lifestage_id) REFERENCES cb_empty.lifestage(id);


--
-- Name: biota biota_subset_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT biota_subset_fk FOREIGN KEY (subset_id) REFERENCES cb_empty.subset(id);


--
-- Name: biota biota_taxon_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.biota
    ADD CONSTRAINT biota_taxon_fk FOREIGN KEY (taxon_id) REFERENCES cb_empty.taxon(id);


--
-- Name: cruise cruise_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT cruise_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: cruise cruise_person_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT cruise_person_fk FOREIGN KEY (lead_person_id) REFERENCES cb_empty.person(id);


--
-- Name: cruise cruise_possys_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT cruise_possys_fk FOREIGN KEY (positioningsystem_id) REFERENCES cb_empty.possys(id);


--
-- Name: cruise cruise_ship_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT cruise_ship_fk FOREIGN KEY (ship_id) REFERENCES cb_empty.ship(id);


--
-- Name: subsample cruise_station_sample_subsampe_name_uniq; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subsample
    ADD CONSTRAINT cruise_station_sample_subsampe_name_uniq FOREIGN KEY (sample_id) REFERENCES cb_empty.sample(id) NOT VALID;


--
-- Name: dataset dataset_country_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.dataset
    ADD CONSTRAINT dataset_country_fk FOREIGN KEY (owner_country_id) REFERENCES cb_empty.country(id);


--
-- Name: dataset dataset_person_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.dataset
    ADD CONSTRAINT dataset_person_fk FOREIGN KEY (contact_person_id) REFERENCES cb_empty.person(id);


--
-- Name: cruise fk_cruise_0; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.cruise
    ADD CONSTRAINT fk_cruise_0 FOREIGN KEY (crs_id) REFERENCES cb_empty.crs(id);


--
-- Name: region2sample sample2region_region; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region2sample
    ADD CONSTRAINT sample2region_region FOREIGN KEY (region_id) REFERENCES cb_empty.region(id) NOT VALID;


--
-- Name: region2sample sample2region_sample; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.region2sample
    ADD CONSTRAINT sample2region_sample FOREIGN KEY (sample_id) REFERENCES cb_empty.sample(id) NOT VALID;


--
-- Name: sample sample_gear_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT sample_gear_fk FOREIGN KEY (gear_id) REFERENCES cb_empty.gear(id);


--
-- Name: sample sample_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT sample_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: sample sample_person_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT sample_person_fk FOREIGN KEY (responsible_person_id) REFERENCES cb_empty.person(id);


--
-- Name: sediment sample_sediment_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sediment
    ADD CONSTRAINT sample_sediment_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: sample sample_station_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sample
    ADD CONSTRAINT sample_station_fk FOREIGN KEY (station_id) REFERENCES cb_empty.station(id);


--
-- Name: sediment sediment_subset_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.sediment
    ADD CONSTRAINT sediment_subset_fk FOREIGN KEY (subset_id) REFERENCES cb_empty.subset(id);


--
-- Name: station station_cruise_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station
    ADD CONSTRAINT station_cruise_fk FOREIGN KEY (cruise_id) REFERENCES cb_empty.cruise(id);


--
-- Name: station station_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station
    ADD CONSTRAINT station_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: station station_person_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.station
    ADD CONSTRAINT station_person_fk FOREIGN KEY (responsible_person_id) REFERENCES cb_empty.person(id);


--
-- Name: subsample subsample_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subsample
    ADD CONSTRAINT subsample_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id) NOT VALID;


--
-- Name: subsample subsample_sample_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subsample
    ADD CONSTRAINT subsample_sample_fk FOREIGN KEY (sample_id) REFERENCES cb_empty.sample(id);


--
-- Name: subset subset_dataset_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset
    ADD CONSTRAINT subset_dataset_fk FOREIGN KEY (dataset_id) REFERENCES cb_empty.dataset(id);


--
-- Name: subset subset_ingest_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset
    ADD CONSTRAINT subset_ingest_fk FOREIGN KEY (ingest_id) REFERENCES cb_empty.ingest(id);


--
-- Name: subset subset_taxcover_fk; Type: FK CONSTRAINT; Schema: cb_empty; Owner: -
--

ALTER TABLE ONLY cb_empty.subset
    ADD CONSTRAINT subset_taxcover_fk FOREIGN KEY (taxcover_id) REFERENCES cb_empty.taxcover(id);


--
-- PostgreSQL database dump complete
--

