"""
Modul mit verschiedensten allgemeinen Hilfsklassen
"""

# coding=utf-8
from collections import Counter, OrderedDict
from tools.text import PrettyText

class WorkingObject:
    """
    Generische Klasse: Arbeitsfähigkeit und Bereitschft eines Objektes
    """

    def __init__(self):
        self.__bWorking = True
        self.__bReady = True

    def setWorking(self, bBool):
        self.__bWorking = bBool

    def isWorking(self):
        return self.__bWorking

    def setReady(self, bBool):
        self.__bReady = bBool

    def isReady(self):
        return self.__bReady and self.__bWorking

class DataTools:
    """
    Hilfsklasse für allgemeine Funktionen mit Daten
    """
    def tab2dic (lstHeader, lstContent, sKeyCol, lstExcludeCols=None):
        """
        Konversion einer Tabelle in ein Dictionary
            Args:
                lstHeader: Header
                lstContent: Content
                sKeyCol: Key Column
                lstExcludeCols: Columns die entfernt werden sollen
        """

        res = Result(name="Converting Table to Dictionary")

        # Header to Index
        dictHead2idx = {}
        lstHeaderEffective = []
        if lstExcludeCols is None:
            lstIdxExcludeHeaders = []
        lstIdxExcludeHeaders = []
        for iIdx, sHead in enumerate(lstHeader):
            if sHead in dictHead2idx:
                res.addErr("Error while processing header: double column " + str(sHead))
                return res
            dictHead2idx[sHead] = iIdx
            if sHead in lstExcludeCols:
                lstIdxExcludeHeaders.append(iIdx)
            else:
                lstHeaderEffective.append(sHead)

        # Baue Dictionary zusammen
        dic = {}
        for lstRow in lstContent:
            dicRow = {}
            sKey = lstRow[dictHead2idx[sKeyCol]]
            for iIdx, sVal in enumerate(lstRow):
                if iIdx != dictHead2idx[sKeyCol] and iIdx not in lstIdxExcludeHeaders:
                    dicRow[lstHeader[iIdx]] = sVal
            if sKey in dic:
                res.addErr("Error while converting table to dictionary: Key "+sKeyCol+" found mulitple times: " + str(sKey))
                return res
            dic[sKey] = dicRow

        res.addData("dic", dic)
        res.addData("cols", lstHeaderEffective)
        res.addData("key", sKeyCol)
        return res

    def cmpLists(lstIst, lstSoll):
        """
        Vergleiche zwei Listen miteinander
            Args:
                lstIst: Ist
                lstSoll: Soll
            Return:
                result (tools.coding.Result): istPlus (set), istFehlt (set). istDoppelt (cnt), sollDoppelt (cnt)

        """
        if lstIst is None:
            lstIst = []
        if lstSoll is None:
            lstSoll = []
        res = Result(name="Compare to lists")
        lstIstPlus = list(set(lstIst).difference(set(lstSoll)))
        lstIstFehlt = list(set(lstSoll).difference(set(lstIst)))
        lstIstDoppelt = [sName for sName, iCnt in Counter(lstIst).items() if iCnt>1]
        lstSollDoppelt = [sName for sName, iCnt in Counter(lstSoll).items() if iCnt>1]
        if len(lstIstPlus) > 0:
            res.addData("istPlus", lstIstPlus)
        if len(lstIstFehlt) > 0:
            res.addData("istFehlt", lstIstFehlt)
        if len(lstIstDoppelt) > 0:
            res.addData("istDoppelt", lstIstDoppelt)
        if len(lstSollDoppelt) > 0:
            res.addData("sollDoppelt", lstSollDoppelt)
        return res

class ProgressString:
    """
    Hilfsklasse zum Erzeugen von Progress Strings
    """

    def __init__(self, **kwargs):
        """

        Args:
            **kwargs:
                akt: aktueller Wert
                max: Maximaler Wert
                prae: Praefix
                post: Postfix
        """

        self.__iMax = None
        self.__sPrae = None
        self.__sPost = None
        self.__iNow = None
        if "max" in kwargs:
            self.__iMax = kwargs["max"]
        if "now" in kwargs:
            self.__iNow = kwargs["now"]
        if "prae" in kwargs:
            self.__sPrae = kwargs["prae"]
        if "post" in kwargs:
            self.__sPost = kwargs["post"]

    def __getProgress(self):
        if self.__iMax is not None and self.__iNow is not None :
            return " [" + str(self.__iNow) + "/" + str(self.__iMax) + "]"
        else:
            return ""

    def __getTitle(self, sTrenner):
        sTitle = ""
        if self.__sPrae is  not None:
            sTitle += str(self.__sPrae)
        if self.__sPrae is not None and self.__sPost is not None :
            sTitle += " "+str(sTrenner)+" "
        if self.__sPost is not None:
            sTitle += str(self.__sPost)
        return sTitle

    def setPrae(self, sPrae):
        self.__sPrae = sPrae

    def setPost(self, sPost):
        self.__sPost = sPost

    def getPrae(self):
        return self.__sPrae

    def getPost(self):
        return self.__sPost

    def up(self, sPrae=None, **kwargs):
        if sPrae is not None:
            self.__sPrae = sPrae
        else:
            if "prae" in kwargs:
                self.__sPrae = kwargs["prae"]
        if "post" in kwargs:
            self.__sPost = kwargs["post"]
        self.__iNow += 1

    def gui(self):
        return  self.__getTitle("|")+ self.__getProgress()

    def cli(self):
        return "* "+self.__getTitle("-") + self.__getProgress()

    def log(self):
        return "* "+self.__getTitle("-")

class MultiSheet:
    """
    MultiSheet Datenmodell
    """

    def __init__(self, **kwargs):
        self.__sWbName = kwargs["name"] if "name" in kwargs else OrderedDict()
        self.__odicWb = OrderedDict()
        self.__sOpenSheet = None
        self.__bError = False
        self.__bEmpty = True
        self.__result = Result("Multi Sheet")

    def __str__(self):
        sOut = ""
        if self.__sWbName is not None:
            sOut += self.__sWbName.upper()+ " (#"+str(self.getRows())+"), open sheet="+str(self.getOpenSheetName())+"\n"
        for sSheet, dicData in self.__odicWb.items():
            sOut += sSheet + " (#"+str(self.getRows(sSheet))+")\n"
            lstHeader = dicData["header"]
            lstContent = dicData["content"]
            lstTable = [lstHeader]
            lstTable += lstContent
            sOut += PrettyText.table(lstTable, True) + "\n\n"
        return sOut

    def getResult(self):
        return self.__result

    def isOk(self):
        return not self.__bEmpty and not self.isErr()

    def isErr(self):
        return self.__result.isErr()

    def isEmpty(self):
        return self.__bEmpty

    def setOpenSheetName(self, sSheet):
        if sSheet in self.__odicWb:
            self.__sOpenSheet = sSheet

    def setOpenSheetIdx(self, iIdx):
        try:
            self.__sOpenSheet = list(self.__odicWb.keys())[iIdx]
        except:
            pass

    def getOpenSheetName(self):
        if self.__sOpenSheet is not None:
            return self.__sOpenSheet
        try:
            return self.getSheetNames()[0]
        except:
            return None

    def getOpenSheetIdx(self):
        if self.__sOpenSheet is not None:
            return list(self.__odicWb.keys()).index(self.__sOpenSheet)
        else:
            return 0

    def setSheet(self, sSheet, lstHeaders, lstContent):
        res = Result("Add Sheet")
        if lstContent is None:
            lstContent = []
        if lstHeaders is None:
            lstHeaders = []
        if sSheet in [None, ""]:
            res.addErr("No sheet given", sys="MultiSheet Data", kat="Add Sheet")
            self.__result.addResult(res)
            return res
        if sSheet in self.__odicWb:
            res.addErr("Sheet already existing", sheet=sSheet, sys="MultiSheet Data", kat="Add Sheet")
            self.__result.addResult(res)
            return res
        if len(lstHeaders) == 0:
            res.addErr("No headers given for for Sheet", sheet=sSheet, sys="MultiSheet Data", kat="Add Sheet")
            self.__result.addResult(res)
            return res
        # if len(lstContent) == 0:
        #     res.addErr("No content given for for Sheet", sheet=sSheet, sys="MultiSheet Data", kat="Add Sheet")
        #     self.__result.addResult(res)
        #     return res
        if len(lstContent) != 0 and (len(lstHeaders)!= len(lstContent[0])):
            res.addErr("Numbers of columns not consistent", add="Headers: "+str(len(lstHeaders))+ ", Content:"+str(len(lstContent[0])), sheet=sSheet, sys="MultiSheet Data", kat="Add Sheet")
            self.__result.addResult(res)
            return res
        self.__odicWb[sSheet] = {"header": lstHeaders, "content": lstContent}
        self.__bEmpty = False
        return res

    def getSheet(self, sSheet=None):
        if sSheet is None:
            sSheet = self.getOpenSheetName()
        if sSheet is None:
            return None, None
        return self.__odicWb[sSheet]["header"], self.__odicWb[sSheet]["content"]

    def getStat(self):
        dicStat = {}
        for sSheet, odicData in self.__odicWb.items():
            dicStat[sSheet] = len(odicData["content"])
        return dicStat

    def getRows(self, sSheetIn=None):
        iRows = 0
        for sSheet, odicData in self.__odicWb.items():
                if (sSheet == sSheetIn) or (sSheetIn is None):
                    iRows += len(odicData["content"])
        return iRows

    def getNoOfSheets(self):
        return len(self.__odicWb.keys())

    def getSheetNames(self):
        return list(self.__odicWb.keys())

    def getColNames(self, sSheet):
        try:
            return self.__odicWb[sSheet]["header"]
        except:
            return None

class Package:
    """
    Kleines Datenpaket
    """

    def __init__(self, sName=None):
        self.__sName = sName
        self.__data = OrderedDict()

    def __str__(self):
        sOut = "* DataPackage" +(": "+str(self.__sName) if self.__sName is not None else "") +"\n"
        sOut += "stored data: " +str(len(self.__data)) +"\n"
        for iIdx, (k, v) in enumerate(self.__data.items()):
            sOut += str(iIdx)+ " " +str(k) +" : " +str(v) +"\n"
        return sOut

    def __createKey(self, lstKey):
        if type(lstKey) is not list:
            key = [lstKey]
        return "["+PrettyText.lst2str(lstKey,"][")+"]"

    def add(self, key, val):
        sKey = self.__createKey(key)
        self.__data[sKey] = val

    def get(self, key):
        sKey = self.__createKey(key)
        return self.__data.get(sKey)




class Result:
    """
    Result Klasse

    Hilft beim Verwalten von Ergebnissen eines Methodenaufrufes

    """
    def __init__(self, sName=None, **kwargs):

        """
        Args:
            sName (str): Name
            **kwargs:
                name (str): Name der Aktion, welche das Resultat ausgelöst hat
                err (list/str): Fehlermeldungen
                warn (list/str): Warnungen
                info (list/str): Informationen
                data (dict): Daten
        """

        self.__sName = sName

        self.__dicErr = {}
        self.__dicWarn = {}
        self.__dicInfo = {}

        self.__dicData = {}

        # Name
        if "name" in kwargs:
            self.__sName = kwargs['name']
        # Sowohl Strings als auch Listen werden den Error/Warnungs/Informations-Dictionaries hinzugefuegt
        if "err" in kwargs:
            self.addErr(kwargs['err'])
        if "warn" in kwargs:
            self.addWarn(kwargs['warn'])
        if "info" in kwargs:
            self.addInfo(kwargs['info'])

        # Daten müssen in einem Dictionary liegen um übergeben werden zu können
        if "data" in kwargs:
            self.addDataDict(kwargs['data'])

    def __str__(self):
        sInfo = "* Name: " + self.getName() + "\n"
        if not self.isAnything():
            sInfo += "Nothing special occurred ..."
        else:
            sInfo += self.getErrWarnInfoString()
            sInfo += self.getDataString()
        return sInfo

    # -------------------
    # Info
    # -------------------

    # --- Text
    def __getMsg(self, dicIn, **kwargs):

        # Parameter
        bShort = kwargs["short"] if "short" in kwargs else False

        # Welcher MessageTyp ist gemeint (Auswertung anhand des Dictionaries)
        if dicIn is self.__dicErr:
            sDicType = "Errors"
        elif dicIn is self.__dicWarn:
            sDicType = "Warnings"
        elif dicIn is self.__dicInfo:
            sDicType = "Information"
        else:
            return None

        """
        {2x}[1s2t]
        """
        # Ausdruck
        iSysNo = self.__getSysNo(dicIn)
        iKatNo = self.__getKatNo(dicIn)
        iKatNoUniq = self.__getKatNo(dicIn, uniq=True)
        iMsgNo = self.__getMsgNo(dicIn)
        iMsgNoUniq = self.__getMsgNo(dicIn, uniq=True)
        iMsgAddNo = self.__getMsgAddNo(dicIn)
        # * Typ
        sOut = ""
        if not bShort:
            sOut += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
        if not bShort:
            sOut += sDicType + ": " + str(iMsgAddNo) +"x (" + str(iSysNo) + "s" + str(iKatNo) + "c" + str(iMsgNo) + "t)\n"
        else:
            sOut += sDicType.upper() +" ("+str(iMsgAddNo) +"x)\n"
        if not bShort:
            sOut += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"
        for sSys, odicKat in dicIn.items():
            # * Sys
            iKatNo = self.__getKatNo(dicIn, sys=sSys)
            iMsgNoUniq = self.__getMsgNo(dicIn, sys=sSys, uniq=True)
            iMsgAddNo = self.__getMsgAddNo(dicIn, sys=sSys)
            if not bShort:
                sOut += ">>> System: " + str(sSys).title() + " (" + str(iMsgAddNo) + "x) ("+str(iKatNo) + "c" + str(iMsgNoUniq) + "t) <<<\n\n"
            else:
                sOut += "--- " + str(sSys).title() +"\n"
            for sKat, odicTitle in odicKat.items():
                # * Kat
                iMsgNo = self.__getMsgNo(dicIn, sys=sSys, kat=sKat)
                iMsgNoUniq = self.__getMsgNo(dicIn, sys=sSys, kat=sKat, uniq=True)
                iMsgAddNo = self.__getMsgAddNo(dicIn, sys=sSys, kat=sKat)
                if not bShort:
                    sOut += "--- Class: " + str(sKat).title() + " (" + str(iMsgAddNo) +"x) ("+ str(iMsgNo) + "t) ---\n\n"
                else:
                    sOut += "> " + str(sKat).title() +"\n"
                for sMsg, lstMsgExtra in odicTitle.items():
                    # * Msg
                    iMsgAddNo = self.__getMsgAddNo(dicIn, sys=sSys, kat=sKat, msg=sMsg)
                    if not bShort:
                        sOut += "* Type: " + str(sMsg).title() + " (" + str(iMsgAddNo) + "x)\n"
                    else:
                        sOut += "- " + str(sMsg).title() + "\n"
                    # Listen der Konstellationen
                    iExtraNo = 0

                    lstSheet = []
                    dicSheetRow = {}
                    lstSheetCol = []
                    dicSheetColRow = {}

                    lstAdd = []
                    lstAddRow = []
                    lstAddSheet = []
                    dicAddSheetRow = {}
                    lstAddSheetCol = []
                    dicAddSheetColRow = {}
                    # * Add
                    for tpMsgExtra in lstMsgExtra:
                        """
                        Msg
                        - (-)               
                        - Sheet             
                        - Sheet Row        
                        - Sheet Col     
                        - Sheet Col Row
                        - Add               
                        - Add Sheet             
                        - Add Sheet Row        
                        - Add Sheet Col     
                        - Add Sheet Col Row
                        """
                        # * Extra
                        iRow = tpMsgExtra[0]
                        sSheet = tpMsgExtra[1]
                        sCol = tpMsgExtra[2]
                        sAdd = tpMsgExtra[3]

                        # Aufteilen in Konstellationen
                        # Test vorweg
                        if sSheet is None:
                            # Kein Sheet -> Column und Row sind wertlos
                            sCol = None
                            iRow = None
                        # Sheet
                        if (sSheet is not None) and (sCol is None) and (iRow is None) and (sAdd is None):
                            lstSheet.append(sSheet)
                        # Sheet Row
                        elif (sSheet is not None) and (sCol is None) and (iRow is not None) and (sAdd is None):
                            if sSheet not in dicSheetRow:
                                dicSheetRow[sSheet] = []
                            dicSheetRow[sSheet].append(iRow)
                        # Sheet Col
                        elif (sSheet is not None) and (sCol is not None) and (iRow is None) and (sAdd is None):
                            sKey = sSheet + "." + sCol
                            lstSheetCol.append(sKey)
                        # Sheet Col Row
                        elif (sSheet is not None) and (sCol is not None) and (iRow is not None) and (sAdd is None):
                            sKey = sSheet + "." + sCol
                            if sKey not in dicSheetColRow:
                                dicSheetColRow[sKey] = []
                            dicSheetColRow[sKey].append(iRow)
                        # Add
                        elif (sSheet is None) and (sCol is None) and (iRow is None) and (sAdd is not None):
                            sKey = "[" + sAdd + "]"
                            lstAdd.append(sKey)
                        # Sheet Add
                        elif (sSheet is not None) and (sCol is None) and (iRow is None) and (sAdd is not None):
                            sKey = sSheet + " [" + sAdd + "]"
                            lstAddSheet.append(sKey)
                        # Sheet Add Row
                        elif (sSheet is not None) and (sCol is None) and (iRow is not None) and (sAdd is not None):
                            sKey = sSheet + " [" + sAdd + "]"
                            if sKey not in dicAddSheetRow:
                                dicAddSheetRow[sKey] = []
                            dicAddSheetRow[sKey].append(iRow)
                        # Sheet Col Add
                        elif (sSheet is not None) and (sCol is not None) and (iRow is None) and (sAdd is not None):
                            sKey = sSheet + "." + sCol + " [" + sAdd + "]"
                            lstAddSheetCol.append(sKey)
                        # Sheet Col Add Row
                        elif (sSheet is not None) and (sCol is not None) and (iRow is not None) and (sAdd is not None):
                            sKey = sSheet + "." + sCol + " [" + sAdd + "]"
                            if sKey not in dicAddSheetColRow:
                                dicAddSheetColRow[sKey] = []
                            dicAddSheetColRow[sKey].append(iRow)
                        # Nichts
                        else:
                            iExtraNo += 1

                    # Ausdruck der Konstellationen
                    # (-)
                    if iExtraNo > 0:
                        sTmp = "[" + sAdd + "] " if sAdd is not None else ""
                        sMulti = ("("+str(iExtraNo) + "x) ") if iExtraNo > 1 else ""
                        sOut += sTmp + sMulti + "\n"
                    # Sheet
                    if len(lstSheet) > 0:
                        cntSheet = Counter(lstSheet)
                        for sSheet, iNo in cntSheet.items():
                            sMulti = ("("+str(iNo) + "x) ") if iNo > 1 else ""
                            sOut += sSheet + " " + sMulti + "\n"
                    # Sheet Row
                    if len(dicSheetRow) > 0:
                        for sSheet, lstRows in dicSheetRow.items():
                            sMulti = ("("+str(len(lstRows)) + "x) ") if len(lstRows) > 1 else ""
                            sRows = "#" + PrettyText.lst2str(sorted(lstRows), ",") + " "
                            sOut += sSheet + " " + sMulti + sRows + "\n"
                    # Sheet Col
                    if len(lstSheetCol) > 0:
                        cntSheetCol = Counter(lstSheetCol)
                        for sSheetCol, iNo in cntSheetCol.items():
                            sMulti = ("("+str(iNo) + "x) ") if iNo > 1 else ""
                            sOut += sSheetCol + " " + sMulti + "\n"
                    # Sheet Col Row
                    if len(dicSheetColRow) > 0:
                        for sSheetCol, lstRows in dicSheetColRow.items():
                            sMulti = ("("+str(len(lstRows)) + "x) ") if len(lstRows) > 1 else ""
                            sRows = "#" + PrettyText.lst2str(sorted(lstRows), ",") + " "
                            sOut += sSheetCol + " " + sMulti + sRows + "\n"
                    # Add
                    if len(lstAdd) > 0:
                        cntAdd = Counter(lstAdd)
                        for sAdd, iNo in cntAdd.items():
                            sMulti = ("("+str(iNo) + "x) ") if iNo > 1 else ""
                            sOut += sAdd + " " + sMulti + "\n"
                    # Sheet Add
                    if len(lstAddSheet) > 0:
                        cntAddSheet = Counter(lstAddSheet)
                        for sAddSheet, iNo in cntAddSheet.items():
                            sMulti = ("("+str(iNo) + "x)") if iNo > 1 else ""
                            sOut += sAddSheet + " " + sMulti + "\n"
                    # Sheet Add Row
                    if len(dicAddSheetRow) > 0:
                        for sSheet, lstRows in dicAddSheetRow.items():
                            sMulti = ("("+str(len(lstRows)) + "x) ") if len(lstRows) > 1 else ""
                            sRows = "#" + PrettyText.lst2str(sorted(lstRows), ",") + " "
                            sOut += sSheet + " " + sMulti + sRows + "\n"
                    # Sheet Col Add
                    if len(lstAddSheetCol) > 0:
                        cntAddSheetCol = Counter(lstAddSheetCol)
                        for sAddSheetCol, iNo in cntAddSheetCol.items():
                            sMulti = ("("+str(iNo) + "x) ") if iNo > 1 else ""
                            sOut += sAddSheetCol + " " + sMulti + "\n"
                    # Sheet Col Add Row
                    if len(dicAddSheetColRow) > 0:
                        for sAddSheetCol, lstRows in dicAddSheetColRow.items():
                            sMulti = ("("+str(len(lstRows)) + "x) ") if len(lstRows) > 1 else ""
                            sRows = "#" + PrettyText.lst2str(sorted(lstRows), ",")
                            sOut += sAddSheetCol + " " + sMulti + sRows + "\n"
                    if not bShort:
                        sOut += "\n"

        return sOut

    # String
    def getErr(self, **kwargs):
        return self.__getMsg(self.__dicErr, **kwargs)

    def getWarn(self, **kwargs):
        return self.__getMsg(self.__dicWarn, **kwargs)

    def getInfo(self, **kwargs):
        return self.__getMsg(self.__dicInfo, **kwargs)

    # Number
    def getErrNo(self):
        return self.getErrMsgAddNo()

    def getWarnNo(self):
        return self.getWarnMsgAddNo()

    def getInfoNo(self):
        return self.getInfoMsgAddNo()

    # List of Strings (Type: Add)
    def __getMsgAddList(self, dic, **kwargs):
        bShort = kwargs["short"] if "short" in kwargs else False
        lstAdd = []
        for sSys in self.__getSysLst(dic):
            for sKat in self.__getKatLst(dic, sSys):
                for sMsg in self.__getMsgLst(dic, sSys, sKat):
                    for tpAdd in self.__getMsgAddLst(dic,sSys, sKat, sMsg):
                        iRow = tpAdd[0]
                        sSheet = tpAdd[1]
                        sCol = tpAdd[2]
                        sAdd = tpAdd[3]
                        sAddSingle = ""
                        if bShort:
                            # Short
                            if sMsg is not None:
                                sAddSingle += sMsg
                            if sAdd is not None:
                                if sAddSingle != "":
                                    sAddSingle += ": "
                                sAddSingle += sAdd
                            if sAddSingle != "":
                                lstAdd.append(sAddSingle)
                        else:
                            # Full
                            sAddSingle += str(sSys)+"/"+str(sKat)+"/"+str(sMsg)
                            sAddSingle_plus = ""
                            if sSheet is not None:
                                sAddSingle_plus += str(sSheet)
                            if sCol is not None:
                                if sAddSingle_plus != "":
                                    sAddSingle_plus += "."
                                sAddSingle_plus+= str(sCol)
                            if iRow is not None:
                                if sAddSingle_plus != "":
                                    sAddSingle_plus = "#" + str(iRow)+" "+sAddSingle_plus
                                else:
                                    sAddSingle_plus = "#"+str(iRow)
                            if sAdd is not None:
                                if sAddSingle_plus != "":
                                    sAddSingle_plus += " "
                                sAddSingle_plus += str(sAdd)
                            if sAddSingle_plus != "":
                                sAddSingle += ":["+sAddSingle_plus+"]"
                            if sAddSingle != "":
                                lstAdd.append(sAddSingle)
        return lstAdd

    def getErrLst(self, **kwargs):
        return self.__getMsgAddList(self.__dicErr, **kwargs)

    def getWarnLst(self, **kwargs):
        return self.__getMsgAddList(self.__dicWarn, **kwargs)

    def getInfoLst(self, **kwargs):
        return self.__getMsgAddList(self.__dicInfo, **kwargs)

    # --- Report
    def getErrWarnInfoStringHeader(self):
        sInfo = ""
        if self.isAnyMessage():
            sInfo = ""
            sInfo += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            if self.isErr():
                sInfo += "Errors: " + str(self.getErrMsgAddNo()) + "x ("+str(self.getErrSysNo())+"s"+str(self.getErrKatNo())+"c"+str(self.getErrMsgNo())+"t)\n"
            if self.isWarn():
                sInfo += "Warnings: " + str(self.getWarnMsgAddNo()) + "x ("+str(self.getWarnSysNo())+"s"+str(self.getWarnKatNo())+"c"+str(self.getWarnMsgNo())+"t)\n"
            if self.isInfo():
                sInfo += "Information: " + str(self.getInfoMsgAddNo()) + "x ("+str(self.getInfoSysNo())+"s"+str(self.getInfoKatNo())+"c"+str(self.getInfoMsgNo())+"t)\n"
            sInfo += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            sInfo += "\n"
        return sInfo

    def getErrWarnInfoStringFooter(self):
        sInfo = ""
        if self.isAnyMessage():
            sInfo += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            sInfo += "\n"
        return sInfo

    def getErrWarnInfoString(self, **kwargs):
        sInfo = ""
        if self.isAnyMessage():
            sInfo += (self.getErr(**kwargs)) if self.isErr() else ""
            sInfo += (self.getWarn(**kwargs)) if self.isWarn() else ""
            sInfo += (self.getInfo(**kwargs)) if self.isInfo() else ""
            sInfo += "\n"
        return sInfo

    def getDataString(self):
        sInfo = ""
        if self.isData():
            for k, v in self.__dicData.items():
                sInfo += str(k) + " : " + str(v) + "\n"
        if sInfo != "":
            sOut =  ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n"
            sOut += "Data (" + str(len(self.__dicData)) + ")\n"
            sOut += ">>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<\n\n"
            sInfo = sOut + sInfo
        return sInfo

    # -------------------
    # Status
    # -------------------

    def isOk(self):
        return not self.isErr()

    def isErr(self):
        return self.__dicErr != {}

    def isWarn(self):
        return self.__dicWarn != {}

    def isInfo(self):
        return self.__dicInfo != {}

    def isData(self):
        return self.__dicData != {}

    def isAnything(self):
        if self.isErr() or self.isWarn() or self.isInfo() or self.isData():
            return True
        else:
            return False

    def isAnyMessage(self):
        if self.isErr() or self.isWarn() or self.isInfo():
            return True
        else:
            return False

    # -------------------
    # Raw
    # -------------------

    def getErrDic(self):
        return self.__dicErr.copy()

    def getWarnDic(self):
        return self.__dicWarn.copy()

    def getInfoDic(self):
        return self.__dicInfo.copy()

    # -------------------
    # Result
    # -------------------

    def setName(self, sName):
        self.__sName = sName

    def getName(self):
        return str(self.__sName)

    def addResult(self, res, **kwargs):
        """
        Fügt einer Resultklasse die Inhalte einer anderen hinzu
        Achtung: Values bereits vorhandener Keys werden mit neuen Values überschrieben

        Args:
            res (tools.coding.Result): Result
        Kwargs:
            msg_only: Alles ohne Daten
            err: Error
            warn: Warning
            info: Info
            data: Daten
            sys: Override-System
            kat: Override-Kategorie
            sheet: Override-Sheet
            col: Override-Col
            row: Override-Row
        Returns:
            res (tools.coding.Result): Result
        """

        bMsgOnly = kwargs["msg_only"] if "msg_only" in kwargs else False
        bErr = kwargs["err"] if "err" in kwargs else False
        bWarn = kwargs["warn"] if "warn" in kwargs else False
        bInfo = kwargs["info"] if "info" in kwargs else False
        bData = kwargs["data"] if "data" in kwargs else False

        sSysOverride = kwargs["sys"] if "sys" in kwargs else None
        sKatOverride = kwargs["kat"] if "kat" in kwargs else None
        sSheetOverride = kwargs["sheet"] if "sheet" in kwargs else None
        sColOverride = kwargs["col"] if "col" in kwargs else None
        sMsgOverride = kwargs["msg"] if "msg" in kwargs else None
        sAddOverride = kwargs["add"] if "add" in kwargs else None
        iRowOverride = kwargs["row"] if "row" in kwargs else None

        if not bErr and not bWarn and not bInfo and not bData:
            bErr = bWarn = bInfo = bData = True
        if bMsgOnly:
            bErr = bWarn = bInfo = True
            bData = False
        lstWork = [(bErr, res.getErrDic(), self.__dicErr),
                   (bWarn, res.getWarnDic(), self.__dicWarn),
                   (bInfo, res.getInfoDic(), self.__dicInfo)]
        for bOk, dicIn, dicSelf in lstWork:
            if bOk:
                for sSys, odicKat in dicIn.items():
                    for sKat, odicMsg in odicKat.items():
                        for sMsg, lstMsgExtra in odicMsg.items():
                            for tpMsgExtra in lstMsgExtra:
                                iRow = tpMsgExtra[0]
                                sSheet = tpMsgExtra[1]
                                sCol = tpMsgExtra[2]
                                sAdd = tpMsgExtra[3]
                                # Override
                                if sSysOverride is not None:
                                    sSys = sSysOverride
                                if sKatOverride is not None:
                                    sKat = sKatOverride
                                if sSheetOverride is not None:
                                    sSheet = sSheetOverride
                                if sColOverride is not None:
                                    sCol = sColOverride
                                if sMsgOverride is not None:
                                    sMsg = sMsgOverride
                                if sAddOverride is not None:
                                    sAdd = sAddOverride
                                if iRowOverride is not None:
                                    iRow = iRowOverride
                                self.__addMsg(dicSelf, sMsg, sys=sSys, kat=sKat, sheet=sSheet, col=sCol, add=sAdd, row=iRow)
        if bData:
            self.addDataDict(res.getData())

    def addResultData(self, res, lstKeys=None):
        if lstKeys is None:
            # Alle Daten kopieren
            self.addDataDict(res.getData())
        else:
            # Nur aus Keyliste
            for key in lstKeys:
                val = res.getData(key)
                if val is not None:
                    if key in self.__dicData:
                        print("Internal message [add data to result object]: key exists -> old data is overridden: key=", str(key), ", val=",str(val))
                    else:
                        self.__dicData[key] = val

    # -------------------
    # System
    # -------------------

    # Number
    def __getSysNo(self, dic):
        return len(dic.keys())

    def getErrSysNo(self):
        return self.__getSysNo(self.__dicErr)

    def getWarnSysNo(self):
        return self.__getSysNo(self.__dicWarn)

    def getInfoSysNo(self):
        return self.__getSysNo(self.__dicInfo)

    # List
    def __getSysLst(self, dic):
        return [sSys for sSys, odicKat in dic.items()]

    def getErrSysLst(self):
        return self.__getSysLst(self.__dicErr)

    def getWarnSysLst(self):
        return self.__getSysLst(self.__dicWarn)

    def getInfoSysLst(self):
        return self.__getSysLst(self.__dicInfo)

    # -------------------
    # Kategorie (Class)
    # -------------------

    # Number
    def __getKatNo(self, dic, **kwargs):
        bUniq = kwargs["uniq"] if "uniq" in kwargs else False
        sSysIn = kwargs["sys"] if "sys" in kwargs else None
        lstKat = []
        for sSys, odicKat in dic.items():
            if ((sSys == sSysIn) or (sSysIn is None)):
                lstKat += list(odicKat.keys())
        if bUniq:
            iLen = len(set(lstKat))
        else:
            iLen = len(lstKat)
        return iLen

    def getErrKatNo(self, **kwargs):
        return self.__getKatNo(self.__dicErr, **kwargs)

    def getWarnKatNo(self, **kwargs):
        return self.__getKatNo(self.__dicWarn, **kwargs)

    def getInfoKatNo(self, **kwargs):
        return self.__getKatNo(self.__dicInfo, **kwargs)

    # List
    def __getKatLst(self, dic, sSysIn):
        try:
            return list(dic[sSysIn].keys())
        except:
            return []

    def getErrKatLst(self, sSys):
        return self.__getKatLst(self.__dicErr, sSys)

    def getWarnKatLst(self, sSys):
        return self.__getKatLst(self.__dicWarn, sSys)

    def getInfoKatLst(self, sSys):
        return self.__getKatLst(self.__dicInfo, sSys)

    # -------------------
    # Message (Type)
    # -------------------

    # --- Add
    def __addMsg(self, dicIn, sMsg, **kwargs):
        """
        Füge eine oder mehrere Nachrichten dem entsprechenen
        Dictionary hinzu

        Args:
            dicIn: Welcher Ditcionary soll benutzt werden?
            sMsg (str): Meldung

            **kwargs:
                sys: System
                kat: Kategorie
                col: Column-Name
                sheet: Sheet-Name
                row: Row
        """

        # Parameter
        sSys = kwargs['sys'] if "sys" in kwargs else None
        sKat = kwargs['kat'] if "kat" in kwargs else None
        iRow = kwargs['row'] if "row" in kwargs else None
        sSheet = kwargs['sheet'] if "sheet" in kwargs else None
        sCol = kwargs['col'] if "col" in kwargs else None
        sHint = kwargs['hint'] if "hint" in kwargs else None
        sAdd = kwargs['add'] if "add" in kwargs else None

        # Parameter
        if sSys in ["", None]:
            sSys = "No System"
        if sKat in ["", None]:
            sKat = "No Category"
        try:
            iRow = int(iRow)
        except:
            iRow = None
        if sSheet == "":
            sSheet = None
        if sCol == "":
            sCol = None
        if sMsg == "":
            sMsg = None
        if sAdd == "":
            sAdd = None
        if sHint == "":
            sHint = None

        # Msg
        if sMsg is None:
            print("Internal Error: Result Message Text not set")
            return
        # Sys
        if sSys not in dicIn:
            dicIn[sSys] = OrderedDict()
        # Kat
        if sKat not in dicIn[sSys]:
            dicIn[sSys][sKat] = OrderedDict()
        # Msg
        if sMsg is not None:
            sMsg = sMsg.replace("\n", " ")
        if sMsg not in dicIn[sSys][sKat]:
            dicIn[sSys][sKat][sMsg] = []
        # Add
        if sAdd is not None:
            sAdd = sAdd.replace("\n", " ")

        # Save
        dicIn[sSys][sKat][sMsg].append((iRow, sSheet, sCol, sAdd))

    def addErr(self, dat, **kwargs):
        self.__addMsg(self.__dicErr, dat, **kwargs)

    def addWarn(self, dat, **kwargs):
        self.__addMsg(self.__dicWarn, dat, **kwargs)

    def addInfo(self, dat, **kwargs):
        self.__addMsg(self.__dicInfo, dat, **kwargs)

    # Number (Msg)
    def __getMsgNo(self, dic, **kwargs):
        bUniq = kwargs["uniq"] if "uniq" in kwargs else False
        sSysIn = kwargs["sys"] if "sys" in kwargs else None
        sKatIn = kwargs["kat"] if "kat" in kwargs else None
        lstMsg = []
        for sSys, odicKat in dic.items():
            if ((sSys == sSysIn) or (sSysIn is None)):
                for sKat, odicMsg in odicKat.items():
                    if ((sKat == sKatIn) or (sKatIn is None)):
                        lstMsg += list(odicMsg.keys())
        if bUniq:
            iLen = len(set(lstMsg))
        else:
            iLen = len(lstMsg)
        return iLen

    def getErrMsgNo(self, **kwargs):
        return self.__getMsgNo(self.__dicErr, **kwargs)

    def getWarnMsgNo(self, **kwargs):
        return self.__getMsgNo(self.__dicWarn, **kwargs)

    def getInfoMsgNo(self, **kwargs):
        return self.__getMsgNo(self.__dicInfo, **kwargs)

    # List (Msg)
    def __getMsgLst(self, dic, sSys, sKat):
        try:
            return list(dic[sSys][sKat].keys())
        except:
            return []

    def getErrMsgLst(self, sSys, sKat):
        return self.__getMsgLst(self.__dicErr, sSys, sKat)

    def getWarnMsgLst(self, sSys, sKat):
        return self.__getMsgLst(self.__dicWarn, sSys, sKat)

    def getInfoMsgLst(self, sSys, sKat):
        return self.__getMsgLst(self.__dicInfo, sSys, sKat)

    # -------------------
    # Additional Info to Message (Type)
    # -------------------

    # --- Get
    # Number (Msg Add)
    def __getMsgAddNo(self, dic, **kwargs):
        sSysIn = kwargs["sys"] if "sys" in kwargs else None
        sKatIn = kwargs["kat"] if "kat" in kwargs else None
        sMsgIn = kwargs["msg"] if "msg" in kwargs else None
        iLen = 0
        for sSys, odicKat in dic.items():
            if ((sSys == sSysIn) or (sSysIn is None)):
                for sKat, odicMsg in odicKat.items():
                    if ((sKat == sKatIn) or (sKatIn is None)):
                        for sMsg, lstAdd in odicMsg.items():
                            if ((sMsg == sMsgIn) or (sMsgIn is None)):
                                iLen += len(lstAdd)
        return iLen

    def getErrMsgAddNo(self, **kwargs):
        return self.__getMsgAddNo(self.__dicErr, **kwargs)

    def getWarnMsgAddNo(self, **kwargs):
        return self.__getMsgAddNo(self.__dicWarn, **kwargs)

    def getInfoMsgAddNo(self, **kwargs):
        return self.__getMsgAddNo(self.__dicInfo, **kwargs)

    # List (Msg Add)
    def __getMsgAddLst(self, dic, sSys, sKat, sMsg):
        try:
            return list(dic[sSys][sKat][sMsg])
        except:
            return []

    def getErrMsgAddLst(self, sSys, sKat, sMsg):
        return self.__getMsgAddLst(self.__dicErr, sSys, sKat, sMsg)

    def getWarnMsgAddLst(self, sSys, sKat, sMsg):
        return self.__getMsgAddLst(self.__dicWarn, sSys, sKat, sMsg)

    def getInfoMsgAddLst(self, sSys, sKat, sMsg):
        return self.__getMsgAddLst(self.__dicInfo, sSys, sKat, sMsg)

    # -------------------
    # Daten
    # -------------------

    def getDataNumber(self):
        return len(self.__dicData)

    def getData(self, key=None):
        if key is None:
            return self.__dicData
        elif key in self.__dicData:
            return self.__dicData[key]
        else:
            return None

    def addData(self, key, val, **kwargs):
        bUpdate = kwargs["update"] if "update" in kwargs else False
        if bUpdate:
            if key not in self.__dicData:
                print("Internal message [update data in result object]: key does not exists -> update means create: key=" + str(key))
            self.__dicData[key] = val
        else:
            if key in self.__dicData:
                print("Internal message [add data to result object]: key exists -> old data is overridden: key=", str(key))
            self.__dicData[key] = val

    def addDataDict(self, dat):
        if type(dat) is dict:
            self.__dicData.update(dat)  # Achtung: Values bereits vorhandener Keys werden mit neuen Values überschrieben
        else:
            print("Internal message [add data to result object]: Result data is not a dictionary and therefore not added to Result-Object: " + str(dat))



