"""
Modul mit Hilfsklassen die beim Umgang mit CSV-Dateien helfen
"""

# coding=utf-8

import os, csv, sys, time
from tools.parallel import Stoppable
from tools.printing import SmartPrinter
from tools.coding import Result

class CsvTools(Stoppable):

    def __init__(self):
        super().__init__()

    """
    KLASSE MIT FUNKTIONALITÄTEN ZUM UMGANG MIT CSV-DATEIEN
    """

    def extractNewData(self, smtprnt, sHead, sIn1, sIn2, sOut):
        # Initialisierung
        """
        Alle Zeilen aus Datei in2 die sich nicht in Datei in1
        befinden, werden in Datei out geschrieben.
        Als Vergleichsfeld werden die Inhalte der Spalte mit der
        Überschrift head benutzt.
        Beide Dateien in1 und in2 müssen somit eine Spalte mit dem
        Name head besitzen.

        Args:
            smtprnt (tools.printtools.Smartprinter) : Smartprinter
            sIn1 (str): Dateiname CSV-Datei in1
            sIn2 (str): Dateiname CSV-Datei in2
            sOut (str): Dateiname CSV-Datei out
            sHead (str): Header anhand dessen Unterschiedliche Zeilen ausgewählt werden

        Returns:
            (tools.coding.Result): Vorgang war fehlerfrei?, Informationen
        """

        if smtprnt is None:
            smtprnt = SmartPrinter()
        pn = smtprnt.normal
        pv = smtprnt.verbose
        pd = smtprnt.debug
        status = smtprnt.status
        workflow = smtprnt.workflow
        progress = smtprnt.progress

        result = Result(name="Exctract new data from a csv file")

        iIndexHeadIn1 = None
        iDatensaetzeIn1 = 0
        lstIn1Header = list()
        iIndexHeadIn2 = None
        iDatensaetzeIn2 = 0
        lstIn2Header = list()

        progress("on-abort")
        workflow(top="Extracting new data from csv file", sub="Working", fmt=False)
        pn("* Extraction of new data sets from a CSV file")

        # Datei <in1> <in2> existieren?
        if not os.path.isfile(sIn1):
            sErr = "File '" + sIn1 + "' does not exist."
            pn(sErr)
            pn("Abort ...")
            result.addErr(sErr, kat="Extract new data", sys="Toolbox")
            return result
        if not os.path.isfile(sIn2):
            sErr = "File '" + sIn1 + "' does not exist."
            pn(sErr)
            result.addErr(sErr, kat="Extract new data", sys="Toolbox")
            return result

        # Hole die Header aus Datei1 und Datei2
        try:
            with open(sIn1, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstIn1Header = csvReader.__next__()
            with open(sIn2, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstIn2Header = csvReader.__next__()
        except Exception as ex:
            sErr = ("Fehler beim Auslesen der Header: "+str(ex))
            pn(sErr)
            result.addErr(sErr, kat="Extract new data", sys="Toolbox")
            return result

        # Hole Indices der Vergleichsheader
        try:
            iIndexHeadIn1 = lstIn1Header.index(sHead)
            iIndexHeadIn2 = lstIn2Header.index(sHead)
        except Exception as ex:
            print("Error: "+str(ex))

        pn("Settings:")
        pn("Name of column to compare: " + sHead)
        pn("File <in1>: " + sIn1)
        pn("Number of columns: "+str(len(lstIn1Header)))
        pn("Index of column to compare: "+str(iIndexHeadIn1))
        pn("File <in2>: "+ str(sIn2))
        pn("Number of columns: " + str(len(lstIn2Header)))
        pn("Index of column to compare: " + str(iIndexHeadIn2))
        pn("File <out>: " + str(sOut))

        bHeadInIn1In2 = sHead in lstIn1Header and sHead in lstIn2Header
        bHeaderGleich = lstIn1Header == lstIn2Header
        bHeaderAnzahlGleich = len(lstIn1Header) == len(lstIn2Header)

        pn("\nTests:")
        pn("Order of headers in file <in1> and <in2> are the same: " + ("yes" if bHeaderGleich else "no"))
        pn("Number of der headers in file <in1> and <in2> are equal: " + ("yes" if bHeaderAnzahlGleich else "no"))
        pn("Column to compare in file <in1> and <in2> found: " + ("yes" if bHeadInIn1In2 else "no"))

        # Test: Vergleichsheader
        if not bHeadInIn1In2:
            sErr = "Name of column to compare \'"+sHead+"\' not found in both files"
            pn(sErr)
            pn("Abort")
            result.addErr(sErr, kat="Extract new data", sys="Toolbox")
            return result

        # Mache Daten Uniq
        pn()
        pn("Loading data to compare from file: "+sIn1)
        setVergleich = set()

        try:
            # Lesen <in1>
            iZeilenGesamtIn1 = sum(1 for line in open(sIn1, newline='', encoding='utf-8'))
            with open(sIn1, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstHead = csvReader.__next__()
                for lstZeile in csvReader:
                    if self.isAbort():
                        result.addInfo("Aborted by user ...", kat="Extract new data", sys="Toolbox")
                        return result
                    setVergleich.add(lstZeile[iIndexHeadIn1])
                    progress(iDatensaetzeIn1, iZeilenGesamtIn1, 100, proz=True, cli=True)
                    status("Loading row: " + str(iDatensaetzeIn1) + "/" + str(iZeilenGesamtIn1))
                    iDatensaetzeIn1 += 1
            pn("Number of rows in data to compare:" + str(len(setVergleich)))
            pn()

            pn("Extract new data from file: "+sIn2)
            # Lesen <in2>
            lstHead = list()
            lstContent = list()
            iZeilenGesamtIn2 = sum(1 for line in open(sIn1, newline='', encoding='utf-8'))
            with open(sIn2, newline='', encoding='utf-8') as csvfile:
                csvReader = csv.reader(csvfile, delimiter=',', quotechar='"')
                lstHead = csvReader.__next__()
                iZeile = 2
                for lstZeile in csvReader:
                    if self.isAbort():
                        result.addInfo("Aborted by user ...", kat="Extract new data", sys="Toolbox")
                        return result
                    sHeadInhalt = lstZeile[iIndexHeadIn2]
                    if sHeadInhalt in setVergleich:
                        pass
                    else:
                        pv("Row " + str(
                            iZeile) + " from file <in2> is not found in file <in1> and will be appended to file <out>:" + sHead + "=" + sHeadInhalt)
                        lstContent.append(lstZeile)
                    if len(lstContent) != 0:
                        sStatus = "[new="+str(len(lstContent))+"]"
                    else:
                        sStatus = ""
                    progress(iDatensaetzeIn2, iZeilenGesamtIn2, 100, proz=True, status=sStatus, cli=True)
                    status("Comparing row: " + str(iDatensaetzeIn2) + "/" + str(iZeilenGesamtIn2)+ " "+sStatus)
                    iDatensaetzeIn2 += 1
            # Schreiben <out>
            pn()
            pn("Writing new data to file: " + sOut)

            # Datei <out> existiert?
            if os.path.isfile(sOut):
                pn("File '" + sOut + "' exists and will be overwritten.")

            with open(sOut, 'w', encoding='utf-8') as csvWriter:
                writer = csv.writer(csvWriter, delimiter=',', quotechar='"', quoting=csv.QUOTE_NONNUMERIC)
                writer.writerow(lstHead)
                writer.writerows(lstContent)

            # Info
            sInfo = "Number of extracted rows: "+str(len(lstContent))+ "\n"
            sInfo += "Rows in file <in1> : " + str(iDatensaetzeIn1) +"\n"
            sInfo += "Rows in file <in2> : " + str(iDatensaetzeIn2) +"\n"
            sInfo += "Rows in file <out>  : " + str(len(lstContent))
            pn()
            pn("done")
            pn(sInfo)
        except Exception as ex:
            result.addErr("Unexpected Error", add=str(ex), kat="Extract new data", sys="Toolbox")

        result.addInfo(sInfo, kat="Extract new data", sys="Toolbox")

        workflow()
        progress("off")
        return result


