"""
Modul mit verschiedensten Hilfsklassen für Datei IO
"""

# coding=utf-8

import os, datetime, traceback
from tools.coding import Result
from tools.text import PrettyText

class FileTools:
    """
    KLASSE MIT IO-HELFERN
    """

    @staticmethod
    def prepareSaveFile(sFile):
        """
        Überprüft ob eine Datei bereits exisitert und erzeugt evtl. ein Backup bei bestehen
        """

        result = Result(name="Prepare file to save")

        # File Daten
        sF = FileTools.getFileWithHomeExpanded(sFile)
        sFpath = FileTools.getFilePath(sF)
        sFname = FileTools.getFileName(sF)
        sFtype = FileTools.getFileType(sF)

        # File Backup?
        if FileTools.isFileExisting(sF):
            try:
                if sFpath != "":
                    sFpath = sFpath+"/"
                sFbak = sFpath+sFname+"_bak_"+PrettyText.createTimestamp()+"."+sFtype
                os.rename(sF,sFbak)
                result.addData("bak", sFbak)
            except:
                result.addErr("Error during file backup: "+str(traceback.print_exc()))
        return result

    @staticmethod
    def isFileExisting(sFile):
        return os.path.isfile(sFile) if sFile is not None else None

    @staticmethod
    def getFileWithHomeExpanded(sFile):
        return os.path.expanduser(sFile) if sFile is not None else None

    @staticmethod
    def getFileWithAbsPath(sFile):
        return os.path.abspath(sFile) if sFile is not None else None

    @staticmethod
    def getFileWithoutPath(sFile):
        return os.path.basename(sFile) if sFile is not None else None

    @staticmethod
    def getFilePath(sFile):
        return os.path.dirname(sFile) if sFile is not None else None

    @staticmethod
    def getFileName(sFile):
        if sFile is None:
            return None
        try:
            sBase = os.path.basename(sFile)
            sName = os.path.splitext(sBase)[0]
            return sName
        except:
            return None

    @staticmethod
    def getFileWithoutType(sFile):
        if sFile is None:
            return None
        try:
            sFileWithoutType = os.path.splitext(sFile)[0]
            return sFileWithoutType
        except:
            return None

    @staticmethod
    def getFileType(sFile):
        if sFile is None:
            return None
        sPfad = os.path.splitext(sFile)[1]
        return sPfad[1:]

    @staticmethod
    def createTimestamp():
        sTimeStamp = str(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        return sTimeStamp

    @staticmethod
    def isPathExiting(sPath):
        return os.path.exists(os.path.expanduser(os.path.normpath(sPath))) if sPath is not None else None

    @staticmethod
    def isPathFile(sPath):
        return os.path.isfile(os.path.expanduser(os.path.normpath(sPath))) if sPath is not None else None

    @staticmethod
    def isPathDir(sPath):
        return os.path.isdir(os.path.expanduser(os.path.normpath(sPath))) if sPath is not None else None

    @staticmethod
    def cutAtPathTail(sPath):
        if sPath is None:
            return None
        try:
            sPath = os.path.expanduser(os.path.normpath(sPath))
            lstTmp = os.path.split(sPath)
            sLast = lstTmp[1]
            sFront = lstTmp[0]
            return sFront, sLast
        except:
            print("Uuups")
            return None, None