"""
Modul mit Hilfsklassen zum Multithreading
"""

# coding=utf-8
import psycopg2, time, csv
from PyQt5.QtCore import QThread, pyqtSignal
from tools.coding import Result

class Stoppable:
    """
    Generische Klasse: Abbrechbarkeit eines Objektes
    """

    def __init__(self):
        self.__bAbort = False
        self.__lstChilds = []

    def addAbortChild(self, child):
        self.__lstChilds.append(child)

    def getAbortChilds(self):
        return self.__lstChilds

    def abort(self, bBool=True):
        self.__bAbort = bBool
        for child in self.__lstChilds:
            child.abort(bBool)

    def isAbort(self):
        return self.__bAbort


class Worker(QThread):
    """
    Generische Klasse: Thread Container
    """

    signalResult = pyqtSignal(Result)

    def __init__(self):
        super().__init__()
        self.__dicThreadInit = None  # Initialisierung des Objektes fuer Threadstart
        self.__bAbort = False

    def run(self):
        # Hier kommt der eigentliche Arbeitscode
        result = Result(info="This is a dummy worker ... Nothing to to ...")  # Dummy Antwort
        self.sendResult(result)

    def abort(self, bAbort=True):
        self.__bAbort = bAbort

    def isAbort(self):
        return self.__bAbort

    def connectFinish(self, finish):
        if finish is not None:
            self.signalResult.connect(finish)

    def disconnectFinish(self):
        try:
            self.signalResult.disconnect()
        except Exception:
            pass

    def init(self, **kwargs):
        self.__dicThreadInit = kwargs

    def getInit(self, key=None):
        if key is None:
            return self.__dicThreadInit
        elif key in self.__dicThreadInit:
            return self.__dicThreadInit[key]
        else:
            return None

    def sendResult(self, result):
        self.signalResult.emit(result)


