"""
Modul mit Hilfsklassen im Kontext Bildschirmausgaben
"""

# coding=utf-8
import sys, time, datetime, random, re

# Smartprinter
class SmartPrinter:


    """
    DIESE KLASSE ABSTRAHIERT DAS AUSDRUCKEN VON TEXTEN VOM KONTEXT (CLI/GUI)

    - Der Klasse wird der gewünschte Geschätzigkeitsgrad mitgeteilt.
    - Texte werden dann abstrakt via die Methoden normal(),verbose(),debug() ausgegeben.
    - Je nach eingestelltem Geschätzigkeitsgrad wird der Text dann tatsächlich ausgedruckt.
    - Geschätzigkeitsgrad Normal: normal() wird gedruckt.
    - Geschätzigkeitsgrad Verbose: normal(),verbose() wird gedruckt.
    - Geschätzigkeitsgrad Debug: normal(),verbose(),debug() wird gedruckt.
    - Weiterhin wird der Klasse eine Referenz zum Model übergeben.
    - Existiert ein Observer im Model befindet sich der Smartprinter im GUI-Modus und Texte
      auf der GUI.
    - Der Observer kennt dann das entsprechende GUI-Element auf dem der Text durch ihn
      ausgedruckt wird.
    - Existiert kein Observer befindet sich der Smartprinter im CLI-Modus und der Text
      wird auf der Konsole ausgedruckt.
    - Falls im Model das entsprechende Flag gesetzt ist, wird jede Textausgaber in eine Log-Datei
      geschrieben.
    """

    def __init__(self, **kwargs):
        """
        Kwargs:
            verbose (bool): Verbose an/aus
            debug (bool): Debug an/aus
            model (Model): Model
        """

        # Instanzvariablen
        # Default
        self.__bVerbose = False
        self.__bDebug = False
        self.__model = None
        self.__inputAnswer = None
        self.__bNoProgressInCli = False  # Progressbar in CLI nicht anzeigen, z.B. wenn man loggen will
        self.__bLastCmdProgressCli = False  # Tweak um zu bemerken, wenn ein Porgressbar im CLI-Modus läuft und mittendrin eine andere Ausgabe (-> Return) kommt

        # Werte aus Konstruktorparameterliste verteilen
        if "verbose" in kwargs:
            self.__bVerbose = kwargs['verbose']
        if "debug" in kwargs:
            self.__bDebug = kwargs['debug']
        if "model" in kwargs:
            self.__model = kwargs['model']

    def isGui(self):
        return self.__model is not None and self.__model.thread_ui_getObserver() is not None

    def inputAsk__comboMulti(self, sTitle, odicChoices):
        if self.isGui():
            # Gui
            # Gib Gui den Befehl eine Combobox zu öffnen
            self.__model.thread_ui_getObserver().inputComboMulti(sTitle, odicChoices)
            return
        else:
            bDebug = False
            # Cli
            dicUserChoice = {}
            self.__printConsole("* "+sTitle+" ("+str(len(odicChoices))+"x)", fmt=False)
            # Loop: Alle Auswahlen
            for iNr, (sName, lstChoices) in enumerate(odicChoices.items()):
                # Eine Auswahl
                if not bDebug:
                    self.__printConsole("-> "+str(sName)+" ("+str(iNr+1)+"/"+str(len(odicChoices))+")", fmt=False)
                    iFunde = 0
                    iIdxChoose_user = None
                    for i, sText in enumerate(lstChoices):
                        self.__printConsole(str(i) + ": " + sText)
                        iFunde += 1
                    bEingabeOk = False
                    while not bEingabeOk:
                        ein = input("Choose (0-" + str(iFunde - 1) + "): ")
                        try:
                            if int(ein) >= 0 and int(ein) < iFunde:
                                iIdxChoose_user = int(ein)
                                bEingabeOk = True
                        except:
                            pass
                    # Auswahl speichern
                    dicUserChoice[sName] = iIdxChoose_user
                else:
                    # Debug: Random Choice
                    dicUserChoice = {}
                    for sSn, lstTaxa in odicChoices.items():
                        iIdxChoose_user = random.randint(0, len(lstTaxa) - 1)
                        dicUserChoice[sSn] = iIdxChoose_user

            self.setInputAnswer(dicUserChoice)

    def setInputAnswer(self, data):
        self.__inputAnswer = data

    def getInputAnswer(self):
        data = self.__inputAnswer
        if self.__inputAnswer is not None:  # Wenn was angekommen ist wieder löschen
            self.__inputAnswer = None
        return data

    def setCliNoPorgress(self, bNoProgress=True):
        self.__bNoProgressInCli = bNoProgress

    def isCliNoPorgress(self):
        return self.__bNoProgressInCli

    def setVerbosity(self, verbose, debug):
        self.__bVerbose = verbose
        self.__bDebug = debug

    def isVerbose(self):
        return True if self.__bVerbose or self.__model is not None and self.__model.conf_isVerbose() else False

    def isDebug(self):
        return True if self.__bDebug or self.__model is not None and self.__model.conf_isDebug() else False

    def normal(self, text="", appendix="\n", **kwargs):
        #bFmt = kwargs["fmt"] if "fmt" in kwargs else True
        if self.isGui():
            self.__printGui(text, appendix, **kwargs)  # GUI
        else:
            self.__printConsole(text, appendix, **kwargs)  # CLI

    def verbose(self, text="", appendix="\n", **kwargs):
        bFmt = kwargs["fmt"] if "fmt" in kwargs else True
        if (self.isGui()) and (self.__model.conf_isVerbose() is True or self.__model.conf_isDebug() is True):
            self.__printGui(text, appendix, fmt=bFmt)  # GUI
        elif self.__model is not None and (self.__model.conf_isVerbose() is True or self.__model.conf_isDebug() is True):
            self.__printConsole(text, appendix, fmt=bFmt)  # CLI
        elif self.__bVerbose is True or self.__bDebug is True:
            self.__printConsole(text, appendix, fmt=bFmt)  # CLI

    def debug(self, text="", appendix="\n", **kwargs):
        bFmt = kwargs["fmt"] if "fmt" in kwargs else True
        if (self.isGui()) and self.__model.conf_isDebug() is True:
            self.__printGui(text, appendix, fmt=bFmt)  # GUI
        elif self.__model is not None and self.__model.conf_isDebug() is True:
            self.__printConsole(text, appendix, fmt=bFmt)  # CLI
        elif self.__bDebug is True:
            self.__printConsole(text, appendix, fmt=bFmt)  # CLI

    def __printConsole(self, text="", appendix="\n", **kwargs):
        bFmt = kwargs["fmt"] if "fmt" in kwargs else True
        if bFmt and (text not in [None, ""]) and (text[0] == "*"):
            text = text.title()
        if self.__bLastCmdProgressCli:
            self.__bLastCmdProgressCli = False
            ERASE_LINE = '\x1b[2K'
            CURSOR_BACK = '\r'
            sys.stdout.write(CURSOR_BACK)
            sys.stdout.write(ERASE_LINE)
            sys.stdout.flush()
        print(text, end=appendix)
        self.__logToFile(text)

    def __printGui(self, text="", appendix="\n", **kwargs):
        bFmt = kwargs["fmt"] if "fmt" in kwargs else True
        if bFmt and (text not in [None, ""]) and (text[0] == "*"):
            text = text.title()
        if self.isGui():
            self.__model.thread_ui_getObserver().printConsole(text, appendix)
            self.__logToFile(text)

    # Logging in Datei
    def __logToFile(self, text):
        if self.__model is not None and self.__model.conf_isLogging():
            sTimestamp= datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
            sTimestamp= datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
            with open(self.__model.conf_getDateiLogging(), 'a') as log:
                log.write(sTimestamp+" : "+text+"\n")

    def status(self, sText="", **kwargs):
        """
        Status setzen
        Kwargs:
            fmt (bool): Soll der Text zum Title gemacht werden?
        """
        bFmt = kwargs["fmt"] if ("fmt" in kwargs and kwargs["fmt"] in (True, False)) else True
        if bFmt:
            sText = sText.title()
        if self.__model is not None and self.__model.thread_ui_getObserver() is not None:
            self.__model.thread_ui_getObserver().setStatus(sText)

    def workflow(self, **kwargs):
        """
        WorkflowMultisheetTesting setzen
        Kwargs:
            fmt (bool): Soll der Text zum Title gemacht werden?
        """
        bFmt = kwargs["fmt"] if ("fmt" in kwargs and kwargs["fmt"] in (True,False)) else True
        sTop = kwargs["top"] if "top" in kwargs else None
        sSub = kwargs["sub"] if "sub" in kwargs else None
        tpSteps = kwargs["steps"] if "steps" in kwargs else None
        try:
            iStep = tpSteps[0]
            iStepMax = tpSteps[1]
        except:
            iStep = 0
            iStepMax = 0
        if bFmt:
            if sSub is not None:
                sSub = sSub.title()
            if sTop is not None:
                sTop = sTop.title()
        if self.isGui():
            self.__model.thread_ui_getObserver().setWorkflow(sTop, sSub, iStep, iStepMax)

    def progress(self, iWertAktuell="on-pulse", iWertMax=100, iSchritte=100, **kwargs):
        """
        Zeige Fortschritt (im Normalmodus)

        Args:
            iWertAktuell (int/str): Aktueller Wert
                                    Schlüsselwort: on((-pulse-)abort), off, pulse)
                                    pulse bringt progressbar nur zum pulsieren (workflow wird nicht beeinflusst)
            iWertMax (int): Maximalwert der Fortschrittskala
            iSchritte (int): Schrittanzahl in die die gesamte Skala durchlaufen werden soll

        Kwargs:
            proz (bool): Soll die Prozentzahl angezeigt werden
            abs (bool): Sollen neben der Prozentzahl auch die absoluten Schritte angezeigt werden?
            text (str): Text, der dem Fortschrittsbalken im GUI-Modus und CLI-Modus hinzugefügt werden soll (GUI: direkt in den Balken, CLI: Vor den Balken)
            time (str): Zeitangabe hinter dem Text
            status (str): Zusatztext, der dem Fortschrittsbalken im CLI-Modus angehaengt werden soll (entspricht ungefähr dem Statusbar)
            cli (bool): Soll der Fortschritt auch im CLI-Modus angezeigt werden?
            err (str): Errors
            warn (str): Warnings
            info (str): Information
            sysmon (bool): System Monitor updaten
            fmt (bool): Formatierung
        """

        sInfo = kwargs['text'] if "text" in kwargs else " " # Muss ein Zeichen sein um den Text zu leeren -> "" macht dies nicht
        sTime_left = kwargs['time'] if "time" in kwargs else None
        sStatus = kwargs['status'] if "status" in kwargs else ""
        bProz = kwargs['proz'] if "proz" in kwargs and kwargs['proz'] in (True,False) else False
        bFmt = kwargs['fmt'] if "fmt" in kwargs and kwargs['fmt'] in (True,False) else True
        bAbs= kwargs['abs'] if "abs" in kwargs and kwargs['abs'] in (True,False) else False
        bCli= kwargs['cli'] if "cli" in kwargs and kwargs['cli'] in (True,False) else False
        bMon= kwargs['mon'] if "mon" in kwargs and kwargs['mon'] in (True,False) else True
        iErr = kwargs['err'] if "err" in kwargs else None
        iWarn = kwargs['warn'] if "warn" in kwargs else None
        iInfo = kwargs['info'] if "info" in kwargs else None

        iMaxAusreisser = 0  # Aussreisser nach oben
        iMinAusreisser = 0  # Aussreisser nach unten

        # Text
        sInfo = re.sub(r'[^a-zA-Z0-9_ ]+', '', sInfo) # ASCII only
        sInfo = sInfo[:30] + (sInfo[30:] and '… ') # Kürzen
        if bFmt:
            sInfo = sInfo.title() # Capitals

        # Spezialfälle abfangen
        if type(iWertAktuell) == str and iWertAktuell not in ["on","on-abort", "on-pulse", "on-pulse-abort", "off", "pulse"]:
            iWertAktuell = "on-pulse"

        if bMon:
            if self.isGui():
                self.__model.thread_ui_getObserver().updateSysmon()

        if iErr is not None or iWarn is not None or iInfo is not None:
            if self.isGui():
                # GUI
                self.__model.thread_ui_getObserver().setProgressErrWarnInfo(iErr, iWarn, iInfo)
            else:
                # CLI
                lstEWI = []
                if iErr not in (0, None):
                    lstEWI.append("e"+str(iErr))
                if iWarn not in (0, None):
                    lstEWI.append("w"+str(iWarn))
                if iInfo not in (0, None):
                    lstEWI.append("i"+str(iInfo))
                if len(lstEWI) > 0:
                    sEWI = "["+" ".join(lstEWI)+"] "
                else:
                    sEWI = ""
                sStatus = sEWI+sStatus

        if iWertAktuell == "pulse":
            if self.isGui():
                self.__model.thread_ui_getObserver().setProgressbarWert(-1) # Pulse
            return

        if iWertAktuell == "on":
            if self.isGui():
                self.__model.thread_ui_getObserver().showProgress(abort=False, pulse=False)  # Progressbar anschalten
            return
        if iWertAktuell == "on-abort":
            if self.isGui():
                self.__model.thread_ui_getObserver().showProgress(abort=True, pulse=False)  # Progressbar anschalten
            return
        elif iWertAktuell == "off":
            if self.isGui():
                self.__model.thread_ui_getObserver().hideProgress()  # Progressbar ausschalten
            return
        elif iWertAktuell == "on-pulse":
            if self.isGui():
                self.__model.thread_ui_getObserver().showProgress(abort=False, pulse=True)  # Progressbar pulse
            return
        elif iWertAktuell == "on-pulse-abort":
            if self.isGui():
                self.__model.thread_ui_getObserver().showProgress(abort=True, pulse=True)  # Progressbar pulse
            return

        # Ausreißerwerte einfangen
        if iWertAktuell > iWertMax:
            iMaxAusreisser = iWertAktuell - iWertMax
            iWertAktuell = iWertMax
        if iWertAktuell < 0:
            iMinAusreisser = iWertAktuell * -1
            iWertAktuell = 0

        sProgress = self.__getProgress(iWertAktuell, iWertMax, iSchritte)
        if sProgress is not None:

            # Infotext Absolute Werte
            sAbs = " (" + str(iWertAktuell) + "/" + str(iWertMax) + ")"
            # Out of Range Max vermerken
            if iMaxAusreisser != 0:
                sAbs = " (" + str(iWertAktuell) + "+" + str(iMaxAusreisser) + "!/" + str(iWertMax) + ")"
            elif iMinAusreisser != 0:
                sAbs = " (0-" + str(iMinAusreisser) + "!/" + str(iWertMax) + ")"

            if self.isGui():
                # Gui
                sFortschrittsText = sInfo
                if bProz:
                    sFortschrittsText += " " + str(sProgress) + "%"
                if bAbs:
                    sFortschrittsText += sAbs
                if sTime_left is not None:
                    sFortschrittsText += " " + sTime_left
                self.__model.thread_ui_getObserver().setProgressbarTextAnzeigen(True)  # Prozentwert anzeigen
                self.__model.thread_ui_getObserver().setProgressbarMinMax(1, iWertMax)  # Range
                self.__model.thread_ui_getObserver().setProgressbarWert(iWertAktuell)  # Aktueller Wert
                self.__model.thread_ui_getObserver().setProgressbarText(sFortschrittsText)  # Text

            else:
                # Cli
                if bCli and not self.__bNoProgressInCli:
                    if not self.__bLastCmdProgressCli:  # War der letzte CLI-Print ein Progress-Ausdruck
                        self.__bLastCmdProgressCli = True
                    sInfo = "Progress" if sInfo == "" else sInfo
                    sStatusText = (" " + sStatus) if sStatus != "" else ""
                    iBarLength = 60
                    fProzent = float(iWertAktuell) / iWertMax
                    sPfeil = '-' * int(round(fProzent * iBarLength) - 1) + '>'
                    sSpaces = ' ' * (iBarLength - len(sPfeil))
                    sys.stdout.write("\r" + sInfo + " [{0}] {1}%".format(sPfeil + sSpaces, int(round(fProzent * 100))) + sStatusText + sAbs)
                    sys.stdout.flush()

    def __getProgress(self, iWertAktuell, iWertMax, iSchritte):
        """
        Ermittle Fortschritt in Prozent (Schrittweise)

        Args:
            iWertAktuell (int): Aktueller Wert (Achtung muss kontinuierlich erhöht werden sonst versagt modulo Erkennung)
            iWertMax (int): Maximalwert der Fortschrittskala
            iSchritte (int): Schrittanzahl in die die gesamte Skala durchlaufen werden soll

        Returns:
            int: Fortschritt in Prozent (0-100%)
        """
        try:
            # iSchritte teilt Gesamtfortschritt in Einzelteile auf (1 ~ einem 100%-Schritt, 100 ~ 100 1%-Schritten)
            if iSchritte > 100:
                # Mehr als 100 Teile gibt es nicht (1%-Schritte)
                iSchritte = 100
            fStepsZeilen = iWertMax / iSchritte
            if (iWertAktuell % fStepsZeilen) < 1:
                # Nächster Schritt erreicht -> String zurückgeben
                iProzent = int(iWertAktuell / fStepsZeilen / iSchritte * 100)
                return iProzent
        except:
            pass
        # Nächster Schritt noch nicht erreicht (oder Fehler)
        return None

    def batch(self, iNr=None, iMax=None, sFile=None):
        if (iNr, iMax, sFile) == (None,None,None):
            iNr = -1
            iMax = -1
            sFile = "end"
        if self.isGui() and self.__model is not None and self.__model.thread_ui_getObserver() is not None:
            self.__model.thread_ui_getObserver().setBatch(iNr, iMax, sFile)

    def updateMyworms(self):
        if self.isGui() and self.__model is not None and self.__model.thread_ui_getObserver() is not None:
            self.__model.thread_ui_getObserver().updateMyworms()

    def updateSysMon(self):
        if self.isGui() and self.__model is not None and self.__model.thread_ui_getObserver() is not None:
            self.__model.thread_ui_getObserver().updateSysmon()

    def updateIngestTable(self):
        if self.isGui() and self.__model is not None and self.__model.thread_ui_getObserver() is not None:
            self.__model.thread_ui_getObserver().updateIngestTable()