"""
Taxa
"""

# coding=utf-8

import openpyxl, traceback, pickle, zipfile
from difflib import SequenceMatcher
from collections import Counter, OrderedDict
import urllib, os, sys, csv, getpass, platform, datetime, time, copy, random, psycopg2, psycopg2.extras
import openpyxl.styles.numbers
import warnings
import requests
from psycopg2 import sql
from tools.parallel import Stoppable
from tools.ui import Colors
from tools.coding import Result, DataTools, ProgressString, WorkingObject
from tools.text import PrettyText
from tools.printing import SmartPrinter
from tools.excel import ExcelTools
from tools.fileio import FileTools
from shutil import copyfile

class Taxa(Stoppable):
    """
    TaxaList
    """

    def __init__(self, model):
        super().__init__()
        self.__model = model
        self.__smartprinter = model.ui_getSmartprinter()
        self.__myworms = model.subsystem_getMyWorms()
        self.__dicTaxa = {}

    # ---

    def __str__(self):
        sOut = "* Taxa ("+str(self.getTaxa_nr())+"x)\n"
        for iAid, t in self.__dicTaxa.items():
            sOut += t.signature() + "\n"
        return sOut

    # ---

    def __chooseTaxonName(t):
        pass
        # """
        # Hole denn Namen aus einem Taxon (Name -> Aname -> Vname -> PLATZHALTER)
        # return:
        #     (str) Name, (str) Comment, (bool ok): Name, Kommentar, Ok - wenn FALSE dann wurde etwas anderes als der Scientifiname gewählt
        # """
        # if t.getScientificName() is not None:
        #     return t.getScientificName(), "used ScientificName of found taxon", True
        # elif t.getScientificName_accepted() is not None:
        #     return t.getScientificName_accepted(), "found taxon: missing ScientificName -> used AcceptedScientificName", False
        # elif t.getScientificName_valid() is not None:
        #     return t.getScientificName_valid(), "found taxon: missing ScientificName and AcceptedScientificName -> used ValidScientificName", False
        # else:
        #     return "PLACEHOLDER", "found taxon has no names -> used placeholder", False

    def __checkGaps_Taxon(t):
        pass
        # if t.hasProblems_status():
        #     setAllTaxaWithNameGaps.add(t)
        #     res.addWarn("Taxa with name gaps used", add=str(t.signature()), row=iRow, sys="TaxaReport",
        #                 kat="Checking Taxa")
        # if t.hasProblems_status():
        #     setAllTaxaWithStatusProblems.add(t)
        #     res.addWarn("Taxa with status problems used", add=str(t.signature()), row=iRow, sys="TaxaReport",
        #                 kat="Checking Taxa")
        # if t.isUnaccepted():
        #     setAllUnacceptedTaxa.add(t)
        #     res.addWarn("Unaccepted taxa used", add=str(t.signature()), row=iRow, sys="TaxaReport", kat="Checking Taxa")
        # if t.isLimbo():
        #     setAllLimboTaxa.add(t)
        #     res.addWarn("Limbo taxa used", add=str(t.signature()), row=iRow, sys="TaxaReport", kat="Checking Taxa")

    # ---

    def getTaxa_nr(self):
        return len(self.__dicTaxa)

    def getTaxa(self, **kwargs):
        sSort = kwargs["sort"] if "sort" in kwargs else None
        lstTaxa = list(self.__dicTaxa.values())
        if sSort == "sn":
            lstTaxa.sort(key=lambda t: str(t.getScientificName()))
        elif sSort == "aid":
            lstTaxa.sort(key=lambda t: t.getAphiaId())
        return lstTaxa

    def getTaxa_aid(self):
        return list(self.__dicTaxa.keys())

    def getTaxon(self, iAid):
        if iAid in self.__dicTaxa:
            return self.__dicTaxa[iAid]
        else:
            return None

    def addTaxon(self, tax):
        try:
            iAid = tax.getAphiaId()
        except:
            return False
        if iAid is None:
            return False
        elif iAid in self.__dicTaxa:
            return False
        else:
            self.__dicTaxa[iAid] = tax
            return True

    def getStatistic(self, **kwargs):
        bHtml = kwargs["html"] if "html" in kwargs else False
        bAll = kwargs["all"] if "all" in kwargs else False
        bAccepted = kwargs["accepted"] if "accepted" in kwargs else False

        sStat = ""
        sColor_awi_blue_light = Colors.getColor("awi_blue_light", html=True)

        # Daten
        iInfo = 0
        iWarn = 0
        iErr = 0
        lstTaxa_all = self.getTaxa()
        lstTaxa_err = []
        lstTaxa_problems_names = []
        lstTaxa_problems_status = []
        lstTaxa_problems_aphiaid = []
        lstTaxa_accepted = []
        lstTaxa_redirected_accepted = []
        lstTaxa_redirected_unaccepted = []
        lstTaxa_unaccepted = []
        lstTaxa_limbo = []
        iTaxa_redirected = 0
        iTaxa_direct = 0
        iTaxa_problems = 0
        iTaxa_nomarine = 0
        lstStats = [
            ("All Taxa", lstTaxa_all),
            ("Taxa accepted (not redirected)", lstTaxa_accepted),
            ("Unaccepted taxa (not redirected)", lstTaxa_unaccepted),
            ("Taxa in Limbo (not redirected)", lstTaxa_limbo),
            ("Uaccepted taxa redirected to accepted taxa", lstTaxa_redirected_accepted),
            ("Uaccepted taxa redirected to unaccepted taxa", lstTaxa_redirected_unaccepted),
            ("Taxa with name problems", lstTaxa_problems_names),
            ("Taxa with status problems", lstTaxa_problems_status),
            ("Taxa with AphiaId problems", lstTaxa_problems_aphiaid),
            ("Taxa with other errors", lstTaxa_err),
        ]

        # Check
        for t in lstTaxa_all:
            if not t.isMarine() and (t.isAccepted() or t.isRedirected_2_accepted()):
                iTaxa_nomarine += 1
            if t.hasProblems():
                # *** Problem Taxa
                iTaxa_problems += 1
                # Name Problems
                if t.hasProblems_names():
                    lstTaxa_problems_names.append(t)
                    iWarn += 1
                # Status Problems
                if t.hasProblems_status():
                    lstTaxa_problems_status.append(t)
                    iWarn += 1
                # Aphia
                if t.hasProblems_aphia():
                    lstTaxa_problems_aphiaid.append(t)
                    iWarn += 1
            else:
                # *** Normale Taxa
                # Nicht/Redirected
                if t.isRedirected():
                    iTaxa_redirected += 1
                else:
                    iTaxa_direct += 1
                # Akzeptiert
                if t.isAccepted():
                    lstTaxa_accepted.append(t)
                # Unakzeptiert -> Akzeptiert
                elif t.isUnaccepted() and t.isRedirected_2_accepted():
                    lstTaxa_redirected_accepted.append(t)
                    iWarn += 1
                # Unakzeptiert
                elif t.isUnaccepted() and not t.isRedirected():
                    lstTaxa_unaccepted.append(t)
                    iInfo += 1
                # Unakzeptiert -> Unakzeptiert
                elif t.isUnaccepted() and t.isRedirected_2_unaccepted():
                    lstTaxa_redirected_unaccepted.append(t)
                    iInfo += 1
                # Limbo
                elif t.isLimbo():
                    lstTaxa_limbo.append(t)
                    iInfo += 1
                elif t.isErr():
                    lstTaxa_err.append(t)
                    iWarn += 1

        # *** Statistic
        # --- Overview
        def proz(iNr, iAll):
            if bHtml:
                return PrettyText.html_bar(iNr, iAll, color=sColor_awi_blue_light)
            else:
                sProz = ""
                try:
                    iProz = iNr/iAll*100
                    if iProz < 1:
                        sProz = str('{:3.2f}'.format(iProz))+"%"
                    else:
                        sProz = str(int(iProz))+"%"
                    return sProz
                except:
                    pass
                return sProz

        iAll = len(self.getTaxa())
        lstOverview = []
        if iAll >0:
            lstOverview.append(["all", "all",iAll, proz(iAll,iAll)])
        if iTaxa_direct > 0:
            lstOverview.append(["direct", "all", iTaxa_direct, proz(iTaxa_direct, iAll)])
        if len(lstTaxa_accepted) >0:
            lstOverview.append(["", "accepted", len(lstTaxa_accepted), proz(len(lstTaxa_accepted),iAll)])
        if len(lstTaxa_unaccepted) >0:
            lstOverview.append(["", "unaccepted", len(lstTaxa_unaccepted), proz(len(lstTaxa_unaccepted),iAll)])
        if len(lstTaxa_limbo) > 0:
            lstOverview.append(["", "Limbo", len(lstTaxa_limbo), proz(len(lstTaxa_limbo),iAll)])
        if iTaxa_redirected > 0:
            lstOverview.append(["redirected", "all", iTaxa_redirected , proz(iTaxa_redirected,iAll)])
        if len(lstTaxa_redirected_accepted) > 0:
            lstOverview.append(["", "accepted", len(lstTaxa_redirected_accepted), proz(len(lstTaxa_redirected_accepted),iAll)])
        if len(lstTaxa_redirected_unaccepted) > 0:
            lstOverview.append(["", "unaccepted", len(lstTaxa_redirected_unaccepted), proz(len(lstTaxa_redirected_unaccepted),iAll)])
        if iTaxa_nomarine > 0:
            lstOverview.append(["not marine", "accepted", iTaxa_nomarine, proz(iTaxa_nomarine, iAll)])
        if iTaxa_problems > 0:
            lstOverview.append(["problems", "all", iTaxa_problems, proz(iTaxa_problems,iAll)])
        if len(lstTaxa_problems_names) > 0:
            lstOverview.append(["", "names", len(lstTaxa_problems_names), proz(len(lstTaxa_problems_names),iAll)])
        if len(lstTaxa_problems_aphiaid) > 0:
            lstOverview.append(["", "AphiaId", len(lstTaxa_problems_aphiaid), proz(len(lstTaxa_problems_aphiaid),iAll)])
        if len(lstTaxa_problems_status) > 0:
            lstOverview.append(["", "names", len(lstTaxa_problems_status), proz(len(lstTaxa_problems_status),iAll)])
        if len(lstOverview) > 0:
            lstOverview.insert(0, [ "Class", "Type", "Count", "Ratio"])
            sStat += "--- Overview\n"
            sStat += PrettyText.table(lstOverview, header=True, html=bHtml) +"\n"
        # --- Details
        sStat += "--- Taxa\n"
        iNr_break = 3
        for sType, lst in lstStats:
            if len(lst) > 0:
                if lst is lstTaxa_accepted and not bAccepted:
                    continue
                if lst is lstTaxa_all and not bAll:
                    continue
                lstTab = []
                for t in lst:
                    lstTab.append([t.signature(html=bHtml)])
                sStat +=  "* "+sType+ " ("+str(len(lst))+"x)\n"
                sStat += PrettyText.table(lstTab, header=False, html=bHtml, preview=5) + "\n"
        if iWarn+iInfo+iErr > 0:
            sStat = "*** Taxa Statistics\n" + sStat
        return sStat