"""
Modul mit Hilfsklassen, die allgemein beim Lesen und Schreiben von Dateien helfen
"""

# coding=utf-8

import traceback, re, os, datetime
from collections import OrderedDict
from tools.printing import SmartPrinter
from cchardet import UniversalDetector
from tools.ui import Colors, ColorGradient

# from chardet.universaldetector import UniversalDetector

class TextEncoding:
    """
    KLASSE MIT FUNKTIONALITÄTEN ZUM UMGANG MIT TEXTENCODINGS
    """

    def checkout(sFile, smtprnt=SmartPrinter()):
        """
        Ermittelt die Encodingmethode einer Datei.
        Es wird das im Vergleich zum Standardmodul chardet
        deutlich schnellere Modul cchardet (in C geschrieben) benutzt.

        Args:
            sFile (str) : Dateipfad der zu überprüfenden Datei
            smtprnt (tools.printtools.Smartprinter) : Smartprinter

        Returns:
            (str,float): Kodierungsart, Wahrscheinlichkeit der korrekten Erkennung
        """
        try:
            detector = UniversalDetector()
            f = open(sFile, "rb")
            for line in f:
                detector.feed(line)
                if detector.done:
                    break
            detector.close()
            return ((detector.result["encoding"], detector.result["confidence"]))
        except Exception as ex:
            smtprnt.normal("* Error while checking the type of text encoding: " + str(ex))
            return ((None, None))

class PrettyText:
    """
    KLASSE MIT FUNKTIONALITÄTEN ZUM AUFHÜBSCHEN VON TEXT
    """

    def html_format(sTxt, **kwargs):
        bHtml = kwargs["html"] if "html" in kwargs and kwargs["html"] in [True, False] else True
        bBold = kwargs["bold"] if "bold" in kwargs and kwargs["bold"] in [True, False] else False
        bItalic = kwargs["italic"] if "italic" in kwargs and kwargs["italic"] in [True, False] else False
        if not bHtml:
            return sTxt
        if sTxt is None:
            return None
        sFormat_start = ""
        sFormat_end = ""
        if bBold:
            sFormat_start += "<b>"
            sFormat_end += "</b>"
        if bItalic:
            sFormat_start += "<i>"
            sFormat_end += "</i>"
        return sFormat_start + sTxt + sFormat_end

    def html_link(sUrl, **kwargs):
        bHtml = kwargs["html"] if "html" in kwargs and kwargs["html"] in [True, False] else True
        sTxt = kwargs["text"] if "text" in kwargs else None
        if not bHtml:
            return sUrl
        if sUrl is None:
            return None
        if sTxt is None:
            return "<a href=\""+sUrl+"\">"+sUrl+"</a>"
        else:
            return "<a href=\""+sUrl+"\">"+sTxt+"</a>"

    def html_special_info_fileSaved(sFile, **kwargs):
        """
        kwargs:
            name (str)
            place (place)
        return:
            (str)
        """
        sName = kwargs.get("name")
        sPlace = kwargs.get("place")
        if sName is None: sName = "File"
        if sPlace is None: sPlace = "here"

        sFile_name =  os.path.basename(sFile)
        sFile_path = os.path.dirname(sFile)

        sInfo = '<html><head/><body><p>'+str(sName)+'<br><b>' + str(sFile_name) + \
                '</b><br>was saved '
        if sFile_path not in [None, ""] :
            sInfo += '<a href="file:///' + str(sFile_path) + '"><span style=" text-decoration: underline; color:#0000ff;">'+str(sPlace)+'</span></a></p>'
        sInfo += '</body></html>'
        return sInfo

    def html_txt2html(sTxt, **kwargs):
        bHeading = kwargs["heading"] if "heading" in kwargs and kwargs["heading"] in [True, False] else True
        sTitle = kwargs["title"] if "title" in kwargs else ""
        sStdFont = kwargs["stdfont"] if "stdfont" in kwargs else "sans-serif"

        sColor_awi_blue_dark = "#173E6A"
        sColor_awi_blue_mid = "#425a89"
        sColor_awi_blue_light = "#4FAAE2"

        sColor_blue = "#488AC3"
        sColor_red = "#B22222"
        sColor_green = "#6B8E23"
        sColor_yellow = "#EDA766"
        sColor_blue = sColor_awi_blue_light

        # Loop Text
        sOut_html_body = ""
        urls = re.findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', sTxt)
        for sLine in sTxt.split("\n"):

            if bHeading:
                # *** Heading (***, ---, *, \n -> h1 h2 h3 p)
                if len(sLine.split("*** ")) == 2:
                    sOut_html_body += '<p><hr align="left" width="75%" style="border-top: dotted 1px;"><h2 style="color:' + sColor_awi_blue_dark + ';">' + sLine.split("***")[1] + '</h1><hr align="left" width="75%" style="border-top: dotted 1px;">'
                elif len(sLine.split("--- ")) == 2:
                    sOut_html_body += '<p><h2 style="color:'+sColor_awi_blue_dark+';">' + sLine.split("---")[1] + '</h2>'
                elif len(sLine.split("* ")) == 2:
                    sOut_html_body += '<p><h3 style="color:'+sColor_awi_blue_mid+';">' + sLine.split("*")[1] + '</h3>'
                else:
                    sOut_html_body += sLine + "<p>"

        # Start
        #Alter Doctype: <!doctype html>
        #Spezielle Sprache: <html>
        sOut_html_start = \
"""
<!doctype html>
<html>
<head>
<meta content="text/html">
<meta charset="utf-8">
<title>"""+sTitle+"""</title>
<style>

.center {
  margin: 20em auto;
  width: 400px;
}

body{
  font-family: """+sStdFont+""";
  font-weight:normal;
  font-size:13px;
}

/*{
  font-family: Arial, Helvetica, sans-serif;
  border-collapse: collapse;
  width: 100%;
}*/

td, th {
  border: 0px solid #ddd;
  padding: 8px;
}

tr:nth-child(odd) {background-color: #ddd;}

tr:hover {background-color: #c1c1c1;}

th {
  padding-top: 12px;
  padding-bottom: 12px;
  text-align: left;
  background-color: #4CAF50;
  color: white;
}

.tooltip {
    display:inline-block;
    position:relative;
    text-align:center;
}

.tooltip .right {
    min-width:200px; 
    top:50%;
    left:100%;
    margin-left:20px;
    transform:translate(0, -50%);
    padding:10px 20px;
    color:#444444;
    /*background-color:#EEEEEE;*/
    background-color:#FFFFF7;
    font-weight:normal;
    font-size:13px;
    border-radius:8px;
    position:absolute;
    z-index:99999999;
    box-sizing:border-box;
    box-shadow:0 1px 8px rgba(0,0,0,0.5);
    visibility:hidden; opacity:0; transition:opacity 0.8s;
}

.tooltip:hover .right {
    visibility:visible; opacity:1;
}

.tooltip .right i {
    position:absolute;
    top:50%;
    right:100%;
    margin-top:-10px;
    width:10px;
    height:20px;
    overflow:hidden;
}

.tooltip .right i::after {
    content:'';
    position:absolute;
    width:10px;
    height:10px;
    left:0;
    top:50%;
    transform:translate(50%,-50%) rotate(-45deg);
    /*background-color:#EEEEEE;*/
    background-color:#FFFFF7;
    box-shadow:0 1px 8px rgba(0,0,0,0.5);
}

</style>
</head>
<body>
"""

        # End
        sOut_html_end = \
            """
            </body>
            </html>
            """

        # Zusammenbauen
        sOut = sOut_html_start + sOut_html_body + sOut_html_end

        return sOut

    def html_bar(iNr, iAll, **kwargs):
        """
        HTML Bar
        """
        sColor = kwargs["color"] if "color" in kwargs  else Colors.getColor("awi_blue_light", html=True)
        bAbs = kwargs["abs"] if "abs" in kwargs else False
        sColorGrad = kwargs["colorgrad"] if ("colorgrad" in kwargs and (kwargs["colorgrad"] in ["pos","neg"])) else None
        try:
            iProz = iNr / iAll * 100
            if sColorGrad == "pos":
                grad = ColorGradient("#ff9494", "#89c496", 100)
                sColor = grad.getHtml(int(iProz))
            elif sColorGrad == "neg":
                grad = ColorGradient("#89c496", "#ff9494", 100)
                sColor = grad.getHtml(int(iProz))
            if iProz < 1:
                sProz = str('{:3.2f}'.format(iProz)) + "%"
            else:
                sProz = str(int(iProz)) + "%"
            sBar = '<div style="font: sans-serif; ; background-color: ' + sColor + '; text-align: left; padding: 3px; margin: 0px; color: black; width: 0' + str(int(iProz)) + 'px;">' + sProz + '</div>'
            if bAbs:
                sBar = sBar + '<div style="font: sans-serif; font-size: xx-small;">'+str(iNr) + "x</div>"
            return sBar
        except:
            return ""

    def barchart(lstData, **kwargs):
        """
        Args:
            lstData: BarName, BarLength
        Kwargs:
            head (Bool: Erstes Element in Daten ist Header
            maxbarsize (int): Max Länge der Balken
        Returns:
            str: Bar Chart
        """
        bHtml = kwargs["html"] if "html" in kwargs and kwargs["html"] in [True,False] else True
        bHeader = kwargs["header"] if "header" in kwargs and kwargs["header"] in [True,False] else True
        iMaxBarSize = kwargs["maxbarsize"] if "maxbarsize" in kwargs else 25

        sOutHeader = ""
        sOut = ""
        try:
            if bHeader:
                sHead1, sHead2 = lstData[0]
                del lstData[0]
            iMaxVal = max(count for _, count in lstData)
            iIncrement = iMaxVal / iMaxBarSize
            iLongestLabelLength = max(len(label) for label, _ in lstData)
            iLongestValueStrLength = max(len(str(val)) for _, val in lstData)
            iBarMax_len = 0
            for iNr, (sLabel, iCount)  in enumerate(lstData):
                iBarChunks, iRemainder = divmod(int(iCount * 8 / iIncrement), 8)
                sBar_unicode = '\u2589'
                #sBar_unicode = '█'
                iBar = sBar_unicode * iBarChunks
                # Funktioniert leider alles nicht unter Windows
                #if iRemainder > 0:
                #    iBar += chr(ord(sBar_unicode) + (8 - iRemainder))
                #iBar = iBar or '\u258f'  # If the bar is empty, add a left one-eighth block
                sBar = f'{sLabel.rjust(iLongestLabelLength)} | {str(iCount).rjust(iLongestValueStrLength)} {iBar}'
                if len(sBar) > iBarMax_len:
                    iBarMax_len = len(sBar)
                sOut += sBar + "\n"

            if bHeader:
                if len(sHead1) > iLongestLabelLength:
                    iLongestLabelLength = len(sHead1)
                sOutHeader += f'{sHead1.rjust(iLongestLabelLength)} | {sHead2}' + "\n"
                #sOutHeader += (iLongestLabelLength + int(iLongestValueStrLength/8) + 5 + iMaxBarSize) * "-"
                sOutHeader += iBarMax_len * "-"
                sOutHeader += "\n"
            sOut = sOutHeader + sOut
        except Exception as ex:
            print("barchart error: ", str(ex))
        return sOut

    @staticmethod
    def table(lst, bHeader=True, bHtml=False, **kwargs):
        """
        Liste als Tabelle konvertieren

        Args:
            lst (list): Liste mit Zeilen (listen) mit Spalten (str)
            bHeader (bool): Erste Zeile is ein Header
            bHtml (bool): HTML Tabelle
        Kwargs:
            header (bool)
            html (bool)
            preview (int): in HTML nur Vorschau von x Zeilen
        Returns:
            (str): Tabelle
        """
        if "header" in kwargs:
            bHeader = kwargs["header"]
        if "html" in kwargs:
            bHtml = kwargs["html"]
        iPreview = kwargs["preview"] if "preview" in kwargs else None
        if type(lst) is set:
            lst = list(lst)
        if bHtml:
            # HTML
            if iPreview is None:
                # No Preview
                return PrettyText.__table(lst, header=bHeader, html=True)
            else:
                # Preview
                iLenTable = len(lst)
                iPreview_table = iPreview
                if bHeader:
                    iLenTable -= 1
                    iPreview_table += 1
                if iPreview > 0 and iPreview < iLenTable :
                    sTable_html = ""
                    sTable_html += PrettyText.__table(lst[0: iPreview_table], header=bHeader, html=True)+"\n"
                    sTable_html += '<details><summary style="font-size: smaller; color:#000; text-decoration: none;">click here to see remaining ' + str(iLenTable - iPreview) + ' rows ...</summary>\n'
                    lst_rest = lst.copy()
                    iPop = 0
                    if bHeader:
                        iPop = 1
                    for i in range(iPreview):
                        lst_rest.pop(iPop)
                    sTable_html += PrettyText.__table(lst_rest, header=bHeader, html=True) +"\n"
                    sTable_html += "</details>\n"
                    return sTable_html
                else:
                    return PrettyText.__table(lst, header=bHeader, html=True)+"\n"
        else:
            # ASCII
            return PrettyText.__table(lst, header=bHeader, html=False)

    @staticmethod
    def __table(lst, **kwargs):
        bHeader = kwargs["header"] if "header" in kwargs else False
        bHtml = kwargs["html"] if "html" in kwargs else False
        iPreview = kwargs["preview"] if "preview" in kwargs else None

        # Alle Daten zu String konvertieren
        lstTmp = []
        for row in lst:
            rowTmp = [str(item) if item is not None else "" for item in row]
            lstTmp.append(rowTmp)
        lst = lstTmp
        if bHtml:
            sTable = '<table><tbody>'
        else:
            sTable = ""
        MaxRow = [max(map(len, col)) for col in zip(*lst)]
        for nr, row in enumerate(lst):
            if bHtml:
                # HTML
                if bHeader and nr == 0:
                    # Head
                    sTable += "<tr>"
                    for col in row:
                        sTable += "<td><b>"+str(col)+"</b></td>"
                    sTable += "</tr>"
                else:
                    # Inhalt
                    sTable += "<tr>"
                    for col in row:
                        sTable += "<td>" + str(col) + "</td>"
                    sTable += "</tr>"
            else:
                # Text
                # Inhalt
                sTable += " | ".join((val.ljust(Max) for val, Max in zip(row, MaxRow))) + "\n"
                # Headertrenner
                if bHeader and nr == 0:
                    sTable += "-|-".join((("-" * Max).ljust(Max) for val, Max in zip(row, MaxRow))) + "\n"
        if bHtml:
            sTable += "</tbody></table>"
        else:
            sTable = sTable[:-1]
        return sTable

    @staticmethod
    def lst2str(input, trenner):
        # Liste/Dict to String
        if input is None:
            return "-"
        if isinstance(input, str):
            return input
        lst = list()

        # Dict/Odict -> List
        if type(input) is dict or type(input) is OrderedDict:
            lst = [str(k) +":" + str(v) for k,v in input.items()]
        else:
            # List/Set -> Jedes Elemet wird zu String
            lst = ["None" if e is None else str(e) for e in input]  # Alles wird String

        return trenner.join(lst)

    @staticmethod
    def lst2strC(lst):
        # Liste to String (Trennung durch Komma)
        return PrettyText.lst2str(lst, ", ")

    @staticmethod
    def lst2strR(lst):
        # Liste to String (Trennung durch Komma)
        return PrettyText.lst2str(lst, "\n")

    @staticmethod
    def dic2str(dic, sTrenner=", "):
        lstOut = [str(k)+"["+str(v)+"]" for k, v in dic.items()]
        return PrettyText.lst2str(lstOut, sTrenner)

    @staticmethod
    def geoJson2str(geoJson):
        # GeoJson Objekt to String
        if geoJson is None:
            return
        try:
            lstFeatures = geoJson["features"] if "features" in geoJson else []
            if lstFeatures is None:
                lstFeatures = []
            sOut = "* GeoJson features\n"
            for dicFeature in lstFeatures:
                iId = dicFeature["id"]
                dicGeometry = dicFeature["geometry"]
                dicProperties = dicFeature["properties"]
                sOut += "--- id: " + str(iId) + "\n"
                for k, v in dicGeometry.items():
                    sOut += str(k) + ": " + str(v) + "\n"
                for k, v in dicProperties.items():
                    sOut += str(k) + ": " + str(v) + "\n"
            sOut += "Total no. of features: " + str(len(lstFeatures))
        except Exception as ex:
            sOut += "* Internal error: " + traceback.print_exc()
        return sOut

    @staticmethod
    def clean_wizard(sTxt, **kwargs):
        """
        Clean Text
        Artgs:
            sTxt:
        Kwargs:
            no_comments (bool)
            no_whitespaces_front_back (bool)
            max_one_empty_row (bool)
            max_one_space_between_words (bool)
            all_in_one_line (bool)
        """
        bNoComments = kwargs["no_comments"] if ("no_comments" in kwargs and kwargs["no_comments"] in [True, False]) else False
        bNoWhitespacesFrontEnd = kwargs["no_whitespaces_front_back"] if ("no_whitespaces_front_back" in kwargs and kwargs["no_whitespaces_front_back"] in [True, False]) else False
        bMaxOneEmptyRow = kwargs["max_one_empty_row"] if ("max_one_empty_row" in kwargs and kwargs["max_one_empty_row"] in [True, False]) else False
        bMaxOneSpaceBetweenWords = kwargs["max_one_space_between_words"] if ("max_one_space_between_words" in kwargs and kwargs["max_one_space_between_words"] in [True, False]) else False
        bAllInOneLine = kwargs["all_in_one_line"] if ("all_in_one_line" in kwargs and kwargs["all_in_one_line"] in [True, False]) else False
        if len(kwargs) == 0 or bAllInOneLine:
            bNoComments = True
            bNoWhitespacesFrontEnd = True
            bMaxOneEmptyRow = True
            bMaxOneSpaceBetweenWords = True
        if sTxt is None:
            return None
        sText_return = ""
        for sRow in sTxt.split("\n"):
            if bNoWhitespacesFrontEnd:
                sRow = sRow.strip()
            if bNoComments:
                if not (sRow != "" and sRow[0] == "#"):
                    sText_return += sRow + "\n"
            else:
                sText_return += sRow + "\n"
        if bMaxOneSpaceBetweenWords:
            sText_return = re.sub(r' +', ' ', sText_return)
        if bMaxOneEmptyRow:
            sText_return = re.sub(r'\n\n\n+', '\n\n', sText_return)
        if bAllInOneLine:
            sText_return = re.sub(r'\n+', ' ', sText_return)
        return sText_return

    @staticmethod
    def clean_2filename(sTxt):
        if sTxt is None:
            return None
        sTmp = re.sub(r'[ \.\-"]', "_", sTxt)
        sTmp = re.sub(r'[#:?)(&%$§\'"]', "", sTmp)
        sTmp = sTmp.lower()
        return sTmp

    @staticmethod
    def clean_2ascii_long(sText):
        try:
            return re.sub(r'[^a-zA-Z0-9,(){}\[\]:,.=?%§!/\-<> ]', "", sText)
        except:
            return sText

    @staticmethod
    def clean_ret2space(sText):
        try:
            sText = PrettyText.clean_2ascii_long(sText)
            sText = re.sub(r'[\n]', " ", sText)
            return sText
        except:
            return sText

    @staticmethod
    def duration(iSecs):
        try:
            iStd = int(iSecs // 3600)
            iMin = int((iSecs % 3600) // 60)
            iSek = int(iSecs % 60)
            iMsek = int((iSecs % 60) * 1000)
            sDuration = ""
            if iStd > 0:
                sDuration += str(iStd) + "h"
            if iMin > 0:
                sDuration += str(iMin) + "m"
            if iSek > 0:
                sDuration += str(iSek) + "s"
            else:
                sDuration += str(iMsek) +"ms"
            return sDuration
        except:
            return ""

    @staticmethod
    def geo_reformat(fValue, **kwargs):
        """
        Georefernz Reformatierung
        Max. Vorkomma: 3 Ziffern
        Max .Nachkomma: 5 Ziffern
        Args:
            fValue (float)
        kwargs:
            precision (int): Zu beachtenede Nachkommastellen
            dummy (bool): Fuelle fehlenden Nachkommastellen auf
            dummy_brace (bool): Klammern um Nachkommastellen, die aufgefüllt wurden
            dummy_char (str): Dummy Buchstaben
        return:
            (str): Value
        """
        # Parameters
        iNachkomma_dataset = kwargs.get("precision")
        bDummy = kwargs.get("dummy")
        bDummy_brace = kwargs.get("dummy_brace")
        sDummy_char = kwargs.get("dummy_char")

        # Daten
        iNachkomma_max = 5  # Alles andere wäre
        iVorkomma_max = 3  # Es sollte ja nur bis 180/90 gehen

        # Parameter checken
        if iNachkomma_dataset is None:
            iNachkomma_dataset = 0
        elif iNachkomma_dataset > iNachkomma_max:
            iNachkomma_dataset = iNachkomma_max
        elif iNachkomma_dataset < 0:
            iNachkomma_dataset = 0
        if bDummy is None:
            bDummy = True
        if sDummy_char is None:
            sDummy_char = "-"
        if bDummy_brace is None:
            bDummy_brace = False

        # Ist Wert wirklich ein Float?
        try:
            float(fValue)
        except:
            return str(fValue)

        # Aktuelle Anzahl der NAchkommastellen
        iNachkomma_aktuell = 0
        try:
            sValue = str(fValue).strip()
            iNachkomma_aktuell = len(sValue.split(".")[1])
        except:
            pass

        # Bringe in die gewünschte Form
        sValue_formated = str(fValue)
        try:
            if type(fValue) is int:
                fValue = float(fValue)
            iZeichen_gesamt = iVorkomma_max + iNachkomma_dataset  # Vorkomma (3) Nachkomma (5)
            iZeichen_gesamt += 1  # Vorzeichen (1)
            if iNachkomma_dataset > 0:
                iZeichen_gesamt += 1  # Komma (1) -> Wenn Nachkomma == 0 -> Wird das Komma nachher hinzugefuegt ... Eieiei ...
            sValue_formated = f'{fValue:{iZeichen_gesamt}.{iNachkomma_dataset}f}'  # TODO: Hier wird leider gerundet (das kann man nicht ausschalten)
            if iNachkomma_dataset == 0:
                sValue_formated += "."
        except:
            pass
        if bDummy:
            iNachkomma_dummy = iNachkomma_max - iNachkomma_dataset
            if iNachkomma_dummy > 0:
                sValue_formated += ("[" if bDummy_brace else "") + (sDummy_char * iNachkomma_dummy) + (
                    "]" if bDummy_brace else "")
        return sValue_formated

    @staticmethod
    # Versioniere einen Namen, der nicht in Referenzliste stehen darf
    def versionize(sName, lstRef):
        if sName is None:
            sName = "no_name"
        sName_versionized = sName
        if sName not in lstRef:
            return sName
        bLoop = True
        i = 0
        while bLoop:
            i += 1
            sName_versionized = str(sName + "_" + str(i))
            if sName_versionized not in lstRef:
                bLoop = False
            if i > 100:
                sName_versionized = sName + "_" + PrettyText.createTimestamp()
                bLoop = False
        return sName_versionized

    @staticmethod
    def createTimestamp():
        sTimeStamp = str(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        return sTimeStamp